@@ -13,7 +13,7 @@ Q_LOGGING_CATEGORY(lcTableModel, "qt.qml.tablemodel")
 
 /*!
     \qmltype TableModel
-//!    \instantiates QQmlTableModel
+//!    \nativetype QQmlTableModel
     \inqmlmodule Qt.labs.qmlmodels
     \brief Encapsulates a simple table model.
     \since 5.14
@@ -415,7 +415,8 @@ void QQmlTableModel::doInsert(int rowIndex, const QVariant &row)
 
     // Adding rowAsVariant.toList() will add each invidual variant in the list,
     // which is definitely not what we want.
-    const QVariant rowAsVariant = row.value<QJSValue>().toVariant();
+    const QVariant rowAsVariant = row.userType() == QMetaType::QVariantMap ? row : row.value<QJSValue>().toVariant();
+
     mRows.insert(rowIndex, rowAsVariant);
     ++mRowCount;
 
@@ -955,9 +956,11 @@ bool QQmlTableModel::validateNewRow(const char *functionName, const QVariant &ro
         return true;
     }
 
+    const bool isVariantMap = (row.userType() == QMetaType::QVariantMap);
+
     // Don't require each row to be a QJSValue when setting all rows,
     // as they won't be; they'll be QVariantMap.
-    if (operation != SetRowsOperation && !validateRowType(functionName, row))
+    if (operation != SetRowsOperation && (!isVariantMap && !validateRowType(functionName, row)))
         return false;
 
     if (operation == OtherOperation) {
@@ -974,7 +977,7 @@ bool QQmlTableModel::validateNewRow(const char *functionName, const QVariant &ro
         }
     }
 
-    const QVariant rowAsVariant = operation == SetRowsOperation
+    const QVariant rowAsVariant = operation == SetRowsOperation || isVariantMap
         ? row : row.value<QJSValue>().toVariant();
     if (rowAsVariant.userType() != QMetaType::QVariantMap) {
         qmlWarning(this) << functionName << ": row manipulation functions "
@@ -1050,6 +1053,12 @@ bool QQmlTableModel::validateRowIndex(const char *functionName, const char *argu
     return true;
 }
 
+Qt::ItemFlags QQmlTableModel::flags(const QModelIndex &index) const
+{
+    Q_UNUSED(index)
+    return Qt::ItemIsSelectable | Qt::ItemIsEnabled | Qt::ItemIsEditable;
+}
+
 QT_END_NAMESPACE
 
 #include "moc_qqmltablemodel_p.cpp"
