@@ -30,7 +30,7 @@ QT_BEGIN_NAMESPACE
 /*!
     \qmltype Menu
     \inherits QtObject
-//! \instantiates QQuickLabsPlatformMenu
+//! \nativetype QQuickLabsPlatformMenu
     \inqmlmodule Qt.labs.platform
     \since 5.8
     \brief A native menu.
@@ -91,43 +91,13 @@ QT_BEGIN_NAMESPACE
     }
     \endqml
 
-    \section2 Dynamically Generating Menu Items
+    \section2 Dynamically generating menu items
 
-    It is possible to dynamically generate menu items. One of the easiest ways
-    to do so is with \l Instantiator. For example, to implement a
-    "Recent Files" submenu, where the items are based on a list of files stored
-    in settings, the following code could be used:
+    You can dynamically generate menu items with \l Instantiator. The following
+    code shows how you can implement a "Recent Files" submenu, where the items
+    come from a list of files stored in settings:
 
-    \qml
-    Menu {
-        title: qsTr("File")
-
-        Menu {
-            id: recentFilesSubMenu
-            title: qsTr("Recent Files")
-            enabled: recentFilesInstantiator.count > 0
-
-            Instantiator {
-                id: recentFilesInstantiator
-                model: settings.recentFiles
-                delegate: MenuItem {
-                    text: settings.displayableFilePath(modelData)
-                    onTriggered: loadFile(modelData)
-                }
-
-                onObjectAdded: recentFilesSubMenu.insertItem(index, object)
-                onObjectRemoved: recentFilesSubMenu.removeItem(object)
-            }
-
-            MenuSeparator {}
-
-            MenuItem {
-                text: qsTr("Clear Recent Files")
-                onTriggered: settings.clearRecentFiles()
-            }
-        }
-    }
-    \endqml
+    \snippet qtlabsplatform-menu-instantiator.qml menu
 
     \section2 Availability
 
@@ -351,6 +321,7 @@ void QQuickLabsPlatformMenu::setParentMenu(QQuickLabsPlatformMenu *menu)
     emit parentMenuChanged();
 }
 
+#if QT_CONFIG(systemtrayicon)
 /*!
     \readonly
     \qmlproperty SystemTrayIcon Qt.labs.platform::Menu::systemTrayIcon
@@ -372,6 +343,7 @@ void QQuickLabsPlatformMenu::setSystemTrayIcon(QQuickLabsPlatformSystemTrayIcon
     destroy();
     emit systemTrayIconChanged();
 }
+#endif
 
 /*!
     \readonly
@@ -678,7 +650,7 @@ void QQuickLabsPlatformMenu::clear()
 
     Opens the menu at the specified \a target item, optionally aligned to a menu \a item.
 */
-void QQuickLabsPlatformMenu::open(QQmlV4Function *args)
+void QQuickLabsPlatformMenu::open(QQmlV4FunctionPtr args)
 {
     if (!m_handle)
         return;
@@ -723,7 +695,7 @@ void QQuickLabsPlatformMenu::open(QQmlV4Function *args)
 #endif
     }
     m_handle->showPopup(window,
-                        QHighDpi::toNativePixels(targetRect, window),
+                        QHighDpi::toNativeLocalPosition(targetRect, window),
                         menuItem ? menuItem->handle() : nullptr);
 }
 
