@@ -16,6 +16,7 @@
 #include <private/qqmlbuiltinfunctions_p.h>
 #include <private/qqmldebugconnector_p.h>
 #include <private/qv4qobjectwrapper_p.h>
+#include <private/qv4qmetaobjectwrapper_p.h>
 #include <private/qv4stackframe_p.h>
 #include <private/qv4module_p.h>
 #include <private/qv4symbol_p.h>
@@ -345,11 +346,8 @@ QJSEngine::QJSEngine()
 */
 
 QJSEngine::QJSEngine(QObject *parent)
-    : QObject(*new QJSEnginePrivate, parent)
-    , m_v4Engine(new QV4::ExecutionEngine(this))
+    : QJSEngine(*new QJSEnginePrivate, parent)
 {
-    checkForApplicationInstance();
-
     QJSEnginePrivate::addToDebugServer(this);
 }
 
@@ -372,6 +370,7 @@ QJSEngine::QJSEngine(QJSEnginePrivate &dd, QObject *parent)
 */
 QJSEngine::~QJSEngine()
 {
+    m_v4Engine->inShutdown = true;
     QJSEnginePrivate::removeFromDebugServer(this);
     delete m_v4Engine;
 }
@@ -391,7 +390,11 @@ QJSEngine::~QJSEngine()
     when the QJSEngine decides that it's wise to do so (i.e. when a certain number of new objects
     have been created). However, you can call this function to explicitly request that garbage
     collection should be performed as soon as possible.
-*/
+
+
+    \sa {Garbage Collection}
+    \sa {Qt::}{gc()}
+ */
 void QJSEngine::collectGarbage()
 {
     m_v4Engine->memoryManager->runGC();
@@ -533,7 +536,7 @@ QJSValue QJSEngine::evaluate(const QString& program, const QString& fileName, in
         result = v4->catchException(&trace);
         if (exceptionStackTrace) {
             for (auto &&frame: trace)
-                exceptionStackTrace->push_back(QString::fromLatin1("%1:%2:%3:%4").arg(
+                exceptionStackTrace->push_back(QLatin1StringView("%1:%2:%3:%4").arg(
                                           frame.function,
                                           QString::number(qAbs(frame.line)),
                                           QString::number(frame.column),
@@ -576,22 +579,25 @@ QJSValue QJSEngine::importModule(const QString &fileName)
     if (m_v4Engine->hasException)
         return QJSValuePrivate::fromReturnedValue(m_v4Engine->catchException());
 
+    // If there is neither a native nor a compiled module, we should have seen an exception
+    Q_ASSERT(module);
+
     QV4::Scope scope(m_v4Engine);
-    if (const auto compiled = module.compiled) {
-        QV4::Scoped<QV4::Module> moduleNamespace(scope, compiled->instantiate(m_v4Engine));
-        if (m_v4Engine->hasException)
-            return QJSValuePrivate::fromReturnedValue(m_v4Engine->catchException());
-        compiled->evaluate();
-        if (!m_v4Engine->isInterrupted.loadRelaxed())
-            return QJSValuePrivate::fromReturnedValue(moduleNamespace->asReturnedValue());
+    QV4::ScopedValue value(scope, module->value());
+    if (!value->isEmpty())
+        return QJSValuePrivate::fromReturnedValue(value->asReturnedValue());
+
+    QV4::Scoped<QV4::Module> moduleNamespace(scope, module->instantiate());
+    if (m_v4Engine->hasException)
+        return QJSValuePrivate::fromReturnedValue(m_v4Engine->catchException());
+
+    module->evaluate();
+    if (m_v4Engine->isInterrupted.loadRelaxed()) {
         return QJSValuePrivate::fromReturnedValue(
-                    m_v4Engine->newErrorObject(QStringLiteral("Interrupted"))->asReturnedValue());
+                m_v4Engine->newErrorObject(QStringLiteral("Interrupted"))->asReturnedValue());
     }
 
-    // If there is neither a native nor a compiled module, we should have seen an exception
-    Q_ASSERT(module.native);
-
-    return QJSValuePrivate::fromReturnedValue(module.native->asReturnedValue());
+    return QJSValuePrivate::fromReturnedValue(moduleNamespace->asReturnedValue());
 }
 
 /*!
@@ -928,7 +934,8 @@ bool QJSEngine::convertV2(const QJSValue &value, QMetaType metaType, void *ptr)
         return convertString(*string, metaType, ptr);
 
     // Does not need scoping since QJSValue still holds on to the value.
-    return QV4::ExecutionEngine::metaTypeFromJS(QJSValuePrivate::asReturnedValue(&value), metaType, ptr);
+    return QV4::ExecutionEngine::metaTypeFromJS(QV4::Value::fromReturnedValue(QJSValuePrivate::asReturnedValue(&value)),
+                                                metaType, ptr);
 }
 
 bool QJSEngine::convertVariant(const QVariant &value, QMetaType metaType, void *ptr)
@@ -960,6 +967,11 @@ QString QJSEngine::convertDateTimeToString(const QDateTime &dateTime)
     return QV4::DateObject::dateTimeToString(dateTime, handle());
 }
 
+double QJSEngine::convertDateTimeToNumber(const QDateTime &dateTime)
+{
+    return QV4::DateObject::dateTimeToNumber(dateTime);
+}
+
 QDate QJSEngine::convertDateTimeToDate(const QDateTime &dateTime)
 {
     return QV4::DateObject::dateTimeToDate(dateTime);
@@ -1023,7 +1035,7 @@ QDate QJSEngine::convertDateTimeToDate(const QDateTime &dateTime)
     conversions between JavaScript-equivalent types that are not
     performed by qvariant_cast by default.
 
-    \sa coerceValue(), fromScriptValue(), qvariant_cast()
+    \sa coerceValue(), fromScriptValue(), {QVariant::}{qvariant_cast()}
 */
 
 /*! \fn template <typename From, typename To> T QJSEngine::coerceValue(const From &from)
@@ -1035,7 +1047,7 @@ QDate QJSEngine::convertDateTimeToDate(const QDateTime &dateTime)
     performed by qvariant_cast by default. This method is a generalization of
     all the other conversion methods in this class.
 
-    \sa fromVariant(), qvariant_cast(), fromScriptValue(), toScriptValue()
+    \sa fromVariant(), {QVariant::}{qvariant_cast()}, fromScriptValue(), toScriptValue()
 */
 
 /*!
@@ -1155,7 +1167,8 @@ void QJSEngine::throwError(QJSValue::ErrorType errorType, const QString &message
 */
 void QJSEngine::throwError(const QJSValue &error)
 {
-    m_v4Engine->throwError(QJSValuePrivate::asReturnedValue(&error));
+    // safe, QJSValue holds a persistent reference
+    m_v4Engine->throwError(QV4::Value::fromReturnedValue(QJSValuePrivate::asReturnedValue(&error)));
 }
 
 /*!
