@@ -4,7 +4,6 @@
 #include <QtCore/qstring.h>
 #include <QtCore/qvarlengtharray.h>
 #include <QtCore/qdatetime.h>
-#include "qjsengine.h"
 #include "qjsvalue.h"
 #include "qjsprimitivevalue.h"
 #include "qjsmanagedvalue.h"
@@ -20,7 +19,9 @@
 #include <private/qv4mm_p.h>
 #include <private/qv4jscall_p.h>
 #include <private/qv4qobjectwrapper_p.h>
+#include <private/qv4qmetaobjectwrapper_p.h>
 #include <private/qv4urlobject_p.h>
+#include <private/qqmlbuiltins_p.h>
 
 /*!
   \since 5.0
@@ -164,10 +165,11 @@
 /*!
     \enum QJSValue::ObjectConversionBehavior
 
-    This enum is used to specify how JavaScript objects without an equivalent
+    This enum is used to specify how JavaScript objects and symbols without an equivalent
     native Qt type should be treated when converting to QVariant.
 
     \value ConvertJSObjects A best-effort, possibly lossy, conversion is attempted.
+           Symbols are converted to QString.
 
     \value RetainJSObjects The value is retained as QJSValue wrapped in QVariant.
 */
@@ -179,61 +181,59 @@ using namespace QV4;
 /*!
   Constructs a new QJSValue with a boolean \a value.
 */
-QJSValue::QJSValue(bool value) : d(QV4::Encode(value))
+QJSValue::QJSValue(bool value) : d(QJSValuePrivate::encode(value))
 {
 }
 
 /*!
   Constructs a new QJSValue with a number \a value.
 */
-QJSValue::QJSValue(int value) : d(QV4::Encode(value))
+QJSValue::QJSValue(int value) : d(QJSValuePrivate::encode(value))
 {
 }
 
 /*!
   Constructs a new QJSValue with a number \a value.
 */
-QJSValue::QJSValue(uint value) : d(QV4::Encode(value))
+QJSValue::QJSValue(uint value) : d(QJSValuePrivate::encode(value))
 {
 }
 
 /*!
   Constructs a new QJSValue with a number \a value.
 */
-QJSValue::QJSValue(double value) : d(QV4::Encode(value))
+QJSValue::QJSValue(double value) : d(QJSValuePrivate::encode(value))
 {
 }
 
 /*!
   Constructs a new QJSValue with a string \a value.
 */
-QJSValue::QJSValue(const QString& value)
+QJSValue::QJSValue(const QString &value) : d(QJSValuePrivate::encode(value))
 {
-    QJSValuePrivate::setString(this, QString(value));
 }
 
 /*!
   Constructs a new QJSValue with a special \a value.
 */
-QJSValue::QJSValue(SpecialValue value) : d(value == NullValue ? QV4::Encode::null() : 0)
+QJSValue::QJSValue(SpecialValue value)
+    : d(value == NullValue ? QJSValuePrivate::encodeNull() : QJSValuePrivate::encodeUndefined())
 {
 }
 
 /*!
   Constructs a new QJSValue with a string \a value.
 */
-QJSValue::QJSValue(const QLatin1String &value)
+QJSValue::QJSValue(const QLatin1String &value) : d(QJSValuePrivate::encode(value))
 {
-    QJSValuePrivate::setString(this, QString(value));
 }
 
 /*!
   Constructs a new QJSValue with a string \a value.
 */
 #ifndef QT_NO_CAST_FROM_ASCII
-QJSValue::QJSValue(const char *value)
+QJSValue::QJSValue(const char *value) : d(QJSValuePrivate::encode(QString::fromUtf8(value)))
 {
-    QJSValuePrivate::setString(this, QString(QString::fromUtf8(value)));
 }
 #endif
 
@@ -244,12 +244,24 @@ QJSValue::QJSValue(const char *value)
   true), then only a reference to the underlying object is copied into
   the new script value (i.e., the object itself is not copied).
 */
-QJSValue::QJSValue(const QJSValue &other) : d(0)
+QJSValue::QJSValue(const QJSValue &other) : d(other.d)
 {
-    if (const QString *string = QJSValuePrivate::asQString(&other))
-        QJSValuePrivate::setString(this, *string);
-    else
-        QJSValuePrivate::setValue(this, QJSValuePrivate::asReturnedValue(&other));
+    switch (QJSValuePrivate::tag(d)) {
+    case QJSValuePrivate::Kind::Undefined:
+    case QJSValuePrivate::Kind::Null:
+    case QJSValuePrivate::Kind::IntValue:
+    case QJSValuePrivate::Kind::BoolValue:
+        return;
+    case QJSValuePrivate::Kind::DoublePtr:
+        d = QJSValuePrivate::encode(*QJSValuePrivate::doublePtr(d));
+        return;
+    case QJSValuePrivate::Kind::QV4ValuePtr:
+        d = QJSValuePrivate::encode(*QJSValuePrivate::qv4ValuePtr(d));
+        return;
+    case QJSValuePrivate::Kind::QStringPtr:
+        d = QJSValuePrivate::encode(*QJSValuePrivate::qStringPtr(d));
+        break;
+    }
 }
 
 /*!
@@ -280,7 +292,7 @@ QJSValue::~QJSValue()
 */
 bool QJSValue::isBool() const
 {
-    return QV4::Value::fromReturnedValue(d).isBoolean();
+    return QJSValuePrivate::tag(d) == QJSValuePrivate::Kind::BoolValue;
 }
 
 /*!
@@ -291,7 +303,15 @@ bool QJSValue::isBool() const
 */
 bool QJSValue::isNumber() const
 {
-    return QV4::Value::fromReturnedValue(d).isNumber();
+    switch (QJSValuePrivate::tag(d)) {
+    case QJSValuePrivate::Kind::IntValue:
+    case QJSValuePrivate::Kind::DoublePtr:
+        return true;
+    default:
+        break;
+    }
+
+    return false;
 }
 
 /*!
@@ -300,7 +320,7 @@ bool QJSValue::isNumber() const
 */
 bool QJSValue::isNull() const
 {
-    return QV4::Value::fromReturnedValue(d).isNull();
+    return QJSValuePrivate::tag(d) == QJSValuePrivate::Kind::Null;
 }
 
 /*!
@@ -311,11 +331,17 @@ bool QJSValue::isNull() const
 */
 bool QJSValue::isString() const
 {
-    if (QJSValuePrivate::asQString(this))
+    switch (QJSValuePrivate::tag(d)) {
+    case QJSValuePrivate::Kind::QStringPtr:
         return true;
+    case QJSValuePrivate::Kind::QV4ValuePtr: {
+        return QJSValuePrivate::qv4ValuePtr(d)->isString();
+    }
+    default:
+        break;
+    }
 
-    // String is managed
-    return QV4::Value::fromReturnedValue(QJSValuePrivate::asReturnedValue(this)).isString();
+    return false;
 }
 
 /*!
@@ -324,14 +350,16 @@ bool QJSValue::isString() const
 */
 bool QJSValue::isUndefined() const
 {
-    if (QJSValuePrivate::asQString(this))
-        return false;
-    QV4::Value v = QV4::Value::fromReturnedValue(QJSValuePrivate::asReturnedValue(this));
-    if (v.isUndefined())
+    switch (QJSValuePrivate::tag(d)) {
+    case QJSValuePrivate::Kind::Undefined:
         return true;
-    if (!v.isManaged())
-        return false;
-    return v.managed() == nullptr;
+    case QJSValuePrivate::Kind::QV4ValuePtr:
+        return QJSValuePrivate::qv4ValuePtr(d)->isUndefined();
+    default:
+        break;
+    }
+
+    return false;
 }
 
 /*!
@@ -346,8 +374,11 @@ bool QJSValue::isError() const
 }
 
 /*!
-  Returns true if this QJSValue is an object of the URL class;
+  Returns true if this QJSValue is an object of the URL JavaScript class;
   otherwise returns false.
+
+  \note For a QJSValue that contains a QUrl, this function returns false.
+  However, \c{toVariant().value<QUrl>()} works in both cases.
 */
 bool QJSValue::isUrl() const
 {
@@ -382,8 +413,7 @@ QJSValue::ErrorType QJSValue::errorType() const
     case QV4::Heap::ErrorObject::URIError:
         return URIError;
     }
-    Q_UNREACHABLE();
-    return NoError;
+    Q_UNREACHABLE_RETURN(NoError);
 }
 
 /*!
@@ -422,16 +452,32 @@ bool QJSValue::isCallable() const
     return QJSValuePrivate::asManagedType<FunctionObject>(this);
 }
 
+#if QT_DEPRECATED_SINCE(6, 9)
 /*!
+  \deprecated [6.9]
   Returns true if this QJSValue is a variant value;
   otherwise returns false.
 
+  \warning This function is likely to give unexpected results.
+  A variant value is only constructed by the QJSEngine in a very
+  limited number of cases. This used to be different before Qt
+  5.14, where \l{QJSEngine::toScriptValue} would have created
+  them for more types instead of corresponding ECMAScript types.
+  You can get a valid \l QVariant via \l toVariant for many values
+  for which \c{isVariant} returns false.
+
   \sa toVariant()
 */
 bool QJSValue::isVariant() const
 {
-    return QJSValuePrivate::asManagedType<QV4::VariantObject>(this);
+    if (QJSValuePrivate::asManagedType<QV4::VariantObject>(this))
+        return true;
+    if (auto vt = QJSValuePrivate::asManagedType<QV4::QQmlValueTypeWrapper>(this))
+        if (vt->metaObject() == &QQmlVarForeign::staticMetaObject)
+            return true;
+    return false;
 }
+#endif
 
 /*!
   Returns the string value of this QJSValue, as defined in
@@ -609,8 +655,13 @@ QVariant QJSValue::toVariant(QJSValue::ObjectConversionBehavior behavior) const
 
     if (val.isString())
         return QVariant(val.toQString());
-    if (QV4::Managed *m = val.as<QV4::Managed>())
-        return m->engine()->toVariant(val, /*typeHint*/ QMetaType{}, behavior == RetainJSObjects);
+    if (val.as<QV4::Managed>()) {
+        if (behavior == RetainJSObjects)
+            return QV4::ExecutionEngine::toVariant(
+                    val, /*typeHint*/ QMetaType{}, /*createJSValueForObjectsAndSymbols=*/ true);
+        else
+            return QV4::ExecutionEngine::toVariantLossy(val);
+    }
 
     Q_ASSERT(false);
     return QVariant();
@@ -633,20 +684,7 @@ QJSPrimitiveValue QJSValue::toPrimitive() const
         return *string;
 
     const QV4::Value val = QV4::Value::fromReturnedValue(QJSValuePrivate::asReturnedValue(this));
-    if (val.isUndefined())
-        return QJSPrimitiveUndefined();
-    if (val.isNull())
-        return QJSPrimitiveNull();
-    if (val.isBoolean())
-        return val.toBoolean();
-    if (val.isInteger())
-        return val.integerValue();
-    if (val.isDouble())
-        return val.doubleValue();
-
-    bool ok;
-    const QString result = val.toQString(&ok);
-    return ok ? QJSPrimitiveValue(result) : QJSPrimitiveValue(QJSPrimitiveUndefined());
+    return QV4::ExecutionEngine::createPrimitive(&val);
 }
 
 /*!
@@ -871,7 +909,10 @@ QJSValue& QJSValue::operator=(const QJSValue& other)
     if (const QString *string = QJSValuePrivate::asQString(&other))
         QJSValuePrivate::setString(this, *string);
     else
-        QJSValuePrivate::setValue(this, QJSValuePrivate::asReturnedValue(&other));
+        // fomReturnedValue is safe, as the QJSValue still has a persistent reference
+        QJSValuePrivate::setValue(
+            this,
+            QV4::Value::fromReturnedValue(QJSValuePrivate::asReturnedValue(&other)));
 
     return *this;
 }
@@ -880,22 +921,22 @@ QJSValue::QJSValue(QJSPrimitiveValue &&value)
 {
     switch (value.type()) {
     case QJSPrimitiveValue::Undefined:
-        d = QV4::Encode::undefined();
+        d = QJSValuePrivate::encodeUndefined();
         return;
     case QJSPrimitiveValue::Null:
-        d = QV4::Encode::null();
+        d = QJSValuePrivate::encodeNull();
         return;
     case QJSPrimitiveValue::Boolean:
-        d = QV4::Encode(value.asBoolean());
+        d = QJSValuePrivate::encode(value.asBoolean());
         return;
     case QJSPrimitiveValue::Integer:
-        d = QV4::Encode(value.asInteger());
+        d = QJSValuePrivate::encode(value.asInteger());
         return;
     case QJSPrimitiveValue::Double:
-        d = QV4::Encode(value.asDouble());
+        d = QJSValuePrivate::encode(value.asDouble());
         return;
     case QJSPrimitiveValue::String:
-        QJSValuePrivate::setString(this, value.asString());
+        d = QJSValuePrivate::encode(value.asString());
         return;
     }
 
@@ -907,10 +948,11 @@ QJSValue::QJSValue(QJSManagedValue &&value)
     if (!value.d) {
         d = QV4::Encode::undefined();
     } else if (value.d->isManaged()) {
-        QJSValuePrivate::setRawValue(this, value.d);
+        // If it's managed, we can adopt the persistent value.
+        QJSValuePrivate::adoptPersistentValue(this, value.d);
         value.d = nullptr;
     } else {
-        d = value.d->asReturnedValue();
+        d = QJSValuePrivate::encode(*value.d);
         QV4::PersistentValueStorage::free(value.d);
         value.d = nullptr;
     }
@@ -958,17 +1000,19 @@ static bool js_equal(const QString &string, const QV4::Value &value)
 */
 bool QJSValue::equals(const QJSValue& other) const
 {
+    // QJSValue stores heap items in persistent values, which already ensures marking
+    // therefore, fromReturnedValue below is safe
     if (const QString *string = QJSValuePrivate::asQString(this)) {
         if (const QString *otherString = QJSValuePrivate::asQString(&other))
             return *string == *otherString;
-        return js_equal(*string, QJSValuePrivate::asReturnedValue(&other));
+        return js_equal(*string, Value::fromReturnedValue(QJSValuePrivate::asReturnedValue(&other)));
     }
 
     if (const QString *otherString = QJSValuePrivate::asQString(&other))
-        return js_equal(*otherString, QJSValuePrivate::asReturnedValue(this));
+        return js_equal(*otherString, Value::fromReturnedValue(QJSValuePrivate::asReturnedValue(this)));
 
-    return Runtime::CompareEqual::call(QJSValuePrivate::asReturnedValue(this),
-                                       QJSValuePrivate::asReturnedValue(&other));
+    return Runtime::CompareEqual::call(Value::fromReturnedValue(QJSValuePrivate::asReturnedValue(this)),
+                                       Value::fromReturnedValue(QJSValuePrivate::asReturnedValue(&other)));
 }
 
 /*!
@@ -1009,8 +1053,10 @@ bool QJSValue::strictlyEquals(const QJSValue& other) const
         return false;
     }
 
-    return RuntimeHelpers::strictEqual(QJSValuePrivate::asReturnedValue(this),
-                                       QJSValuePrivate::asReturnedValue(&other));
+    // QJSValue stores heap objects persistently, so we can be sure that they'll be marked
+    // thus we can safely use fromReturnedValue
+    return RuntimeHelpers::strictEqual(Value::fromReturnedValue(QJSValuePrivate::asReturnedValue(this)),
+                                       Value::fromReturnedValue(QJSValuePrivate::asReturnedValue(&other)));
 }
 
 /*!
