@@ -14,23 +14,28 @@
 // We mean it.
 //
 
-#include "qv4global_p.h"
-#include "qv4runtime_p.h"
 #include "qv4engine_p.h"
-#include "qv4context_p.h"
 #include "qv4object_p.h"
 #include "qv4internalclass_p.h"
 #include "qv4qmlcontext_p.h"
 #include <private/qqmltypewrapper_p.h>
+#include <private/qv4mm_p.h>
 
 QT_BEGIN_NAMESPACE
 
 namespace QV4 {
 
+namespace Heap {
+    struct QObjectMethod;
+}
+
+template <typename T, int PhantomTag>
+using HeapObjectWrapper = WriteBarrier::HeapObjectWrapper<T, PhantomTag>;
+
 // Note: We cannot hide the copy ctor and assignment operator of this class because it needs to
 //       be trivially copyable. But you should never ever copy it. There are refcounted members
 //       in there.
-struct Q_QML_PRIVATE_EXPORT Lookup {
+struct Q_QML_EXPORT Lookup {
     union {
         ReturnedValue (*getter)(Lookup *l, ExecutionEngine *engine, const Value &object);
         ReturnedValue (*globalGetter)(Lookup *l, ExecutionEngine *engine);
@@ -47,7 +52,7 @@ struct Q_QML_PRIVATE_EXPORT Lookup {
             quintptr unused2;
         } markDef;
         struct {
-            Heap::InternalClass *ic;
+            HeapObjectWrapper<Heap::InternalClass, 0> ic;
             quintptr unused;
             uint index;
             uint offset;
@@ -58,8 +63,8 @@ struct Q_QML_PRIVATE_EXPORT Lookup {
             const Value *data;
         } protoLookup;
         struct {
-            Heap::InternalClass *ic;
-            Heap::InternalClass *ic2;
+            HeapObjectWrapper<Heap::InternalClass, 1> ic;
+            HeapObjectWrapper<Heap::InternalClass, 2> ic2;
             uint offset;
             uint offset2;
         } objectLookupTwoClasses;
@@ -72,12 +77,12 @@ struct Q_QML_PRIVATE_EXPORT Lookup {
         struct {
             // Make sure the next two values are in sync with protoLookup
             quintptr protoId;
-            Heap::Object *proto;
+            HeapObjectWrapper<Heap::Object, 3> proto;
             const Value *data;
             quintptr type;
         } primitiveLookup;
         struct {
-            Heap::InternalClass *newClass;
+            HeapObjectWrapper<Heap::InternalClass, 4> newClass;
             quintptr protoId;
             uint offset;
             uint unused;
@@ -89,11 +94,17 @@ struct Q_QML_PRIVATE_EXPORT Lookup {
             uint unused;
         } indexedLookup;
         struct {
-            Heap::InternalClass *ic;
-            Heap::InternalClass *qmlTypeIc; // only used when lookup goes through QQmlTypeWrapper
+            HeapObjectWrapper<Heap::InternalClass, 5> ic;
+            HeapObjectWrapper<Heap::InternalClass, 6> qmlTypeIc; // only used when lookup goes through QQmlTypeWrapper
             const QQmlPropertyCache *propertyCache;
             const QQmlPropertyData *propertyData;
         } qobjectLookup;
+        struct {
+            HeapObjectWrapper<Heap::InternalClass, 7> ic;
+            HeapObjectWrapper<Heap::QObjectMethod, 8> method;
+            const QQmlPropertyCache *propertyCache;
+            const QQmlPropertyData *propertyData;
+        } qobjectMethodLookup;
         struct {
             quintptr isConstant; // This is a bool, encoded as 0 or 1. Both values are ignored by gc
             quintptr metaObject; // a (const QMetaObject* & 1) or nullptr
@@ -101,7 +112,7 @@ struct Q_QML_PRIVATE_EXPORT Lookup {
             int notifyIndex;
         } qobjectFallbackLookup;
         struct {
-            Heap::InternalClass *ic;
+            HeapObjectWrapper<Heap::InternalClass, 9> ic;
             quintptr metaObject; // a (const QMetaObject* & 1) or nullptr
             const QtPrivate::QMetaTypeInterface *metaType; // cannot use QMetaType; class must be trivial
             quint16 coreIndex;
@@ -114,7 +125,7 @@ struct Q_QML_PRIVATE_EXPORT Lookup {
             int scriptIndex;
         } qmlContextScriptLookup;
         struct {
-            Heap::Base *singletonObject;
+            HeapObjectWrapper<Heap::Base, 10> singletonObject;
             quintptr unused2;
             QV4::ReturnedValue singletonValue;
         } qmlContextSingletonLookup;
@@ -131,20 +142,24 @@ struct Q_QML_PRIVATE_EXPORT Lookup {
             ReturnedValue (*getterTrampoline)(Lookup *l, ExecutionEngine *engine);
         } qmlContextGlobalLookup;
         struct {
-            Heap::Base *qmlTypeWrapper;
+            HeapObjectWrapper<Heap::Base, 11> qmlTypeWrapper;
             quintptr unused2;
         } qmlTypeLookup;
         struct {
-            Heap::InternalClass *ic;
+            HeapObjectWrapper<Heap::InternalClass, 12> ic;
             quintptr unused;
             ReturnedValue encodedEnumValue;
+            const QtPrivate::QMetaTypeInterface *metaType;
         } qmlEnumValueLookup;
         struct {
-            Heap::InternalClass *ic;
-            Heap::Object *qmlScopedEnumWrapper;
+            HeapObjectWrapper<Heap::InternalClass, 13> ic;
+            HeapObjectWrapper<Heap::Object, 14> qmlScopedEnumWrapper;
         } qmlScopedEnumWrapperLookup;
     };
-    uint nameIndex;
+
+    uint nameIndex: 28; // Same number of bits we store in the compilation unit for name indices
+    uint forCall: 1;    // Whether we are looking up a value in order to call it right away
+    uint reserved: 3;
 
     ReturnedValue resolveGetter(ExecutionEngine *engine, const Object *object);
     ReturnedValue resolvePrimitiveGetter(ExecutionEngine *engine, const Value &object);
@@ -154,6 +169,7 @@ struct Q_QML_PRIVATE_EXPORT Lookup {
     static ReturnedValue getterGeneric(Lookup *l, ExecutionEngine *engine, const Value &object);
     static ReturnedValue getterTwoClasses(Lookup *l, ExecutionEngine *engine, const Value &object);
     static ReturnedValue getterFallback(Lookup *l, ExecutionEngine *engine, const Value &object);
+    static ReturnedValue getterFallbackAsVariant(Lookup *l, ExecutionEngine *engine, const Value &object);
 
     static ReturnedValue getter0MemberData(Lookup *l, ExecutionEngine *engine, const Value &object);
     static ReturnedValue getter0Inline(Lookup *l, ExecutionEngine *engine, const Value &object);
@@ -167,6 +183,8 @@ struct Q_QML_PRIVATE_EXPORT Lookup {
     static ReturnedValue getterProtoAccessorTwoClasses(Lookup *l, ExecutionEngine *engine, const Value &object);
     static ReturnedValue getterIndexed(Lookup *l, ExecutionEngine *engine, const Value &object);
     static ReturnedValue getterQObject(Lookup *l, ExecutionEngine *engine, const Value &object);
+    static ReturnedValue getterQObjectAsVariant(Lookup *l, ExecutionEngine *engine, const Value &object);
+    static ReturnedValue getterQObjectMethod(Lookup *l, ExecutionEngine *engine, const Value &object);
 
     static ReturnedValue primitiveGetterProto(Lookup *l, ExecutionEngine *engine, const Value &object);
     static ReturnedValue primitiveGetterAccessor(Lookup *l, ExecutionEngine *engine, const Value &object);
@@ -180,11 +198,13 @@ struct Q_QML_PRIVATE_EXPORT Lookup {
     static bool setterGeneric(Lookup *l, ExecutionEngine *engine, Value &object, const Value &value);
     Q_NEVER_INLINE static bool setterTwoClasses(Lookup *l, ExecutionEngine *engine, Value &object, const Value &value);
     static bool setterFallback(Lookup *l, ExecutionEngine *engine, Value &object, const Value &value);
+    static bool setterFallbackAsVariant(Lookup *l, ExecutionEngine *engine, Value &object, const Value &value);
     static bool setter0MemberData(Lookup *l, ExecutionEngine *engine, Value &object, const Value &value);
     static bool setter0Inline(Lookup *l, ExecutionEngine *engine, Value &object, const Value &value);
     static bool setter0setter0(Lookup *l, ExecutionEngine *engine, Value &object, const Value &value);
     static bool setterInsert(Lookup *l, ExecutionEngine *engine, Value &object, const Value &value);
     static bool setterQObject(Lookup *l, ExecutionEngine *engine, Value &object, const Value &value);
+    static bool setterQObjectAsVariant(Lookup *l, ExecutionEngine *engine, Value &object, const Value &value);
     static bool arrayLengthSetter(Lookup *l, ExecutionEngine *engine, Value &object, const Value &value);
 
     void markObjects(MarkStack *stack) {
@@ -194,19 +214,23 @@ struct Q_QML_PRIVATE_EXPORT Lookup {
             markDef.h2->mark(stack);
     }
 
-    void clear() {
-        memset(&markDef, 0, sizeof(markDef));
-    }
-
     void releasePropertyCache()
     {
         if (getter == getterQObject
                 || getter == QQmlTypeWrapper::lookupSingletonProperty
                 || setter == setterQObject
                 || qmlContextPropertyGetter == QQmlContextWrapper::lookupScopeObjectProperty
-                || qmlContextPropertyGetter == QQmlContextWrapper::lookupContextObjectProperty) {
+                || qmlContextPropertyGetter == QQmlContextWrapper::lookupContextObjectProperty
+                || getter == getterQObjectAsVariant
+                || setter == setterQObjectAsVariant) {
             if (const QQmlPropertyCache *pc = qobjectLookup.propertyCache)
                 pc->release();
+        } else if (getter == getterQObjectMethod
+                   || getter == QQmlTypeWrapper::lookupSingletonMethod
+                   || qmlContextPropertyGetter == QQmlContextWrapper::lookupScopeObjectMethod
+                   || qmlContextPropertyGetter == QQmlContextWrapper::lookupContextObjectMethod) {
+            if (const QQmlPropertyCache *pc = qobjectMethodLookup.propertyCache)
+                pc->release();
         }
     }
 };
@@ -230,8 +254,8 @@ inline void setupQObjectLookup(
         Lookup *lookup, const QQmlData *ddata, const QQmlPropertyData *propertyData,
         const Object *self)
 {
-    lookup->qobjectLookup.ic = self->internalClass();
     setupQObjectLookup(lookup, ddata, propertyData);
+    lookup->qobjectLookup.ic.set(self->engine(), self->internalClass());
 }
 
 
@@ -239,8 +263,32 @@ inline void setupQObjectLookup(
         Lookup *lookup, const QQmlData *ddata, const QQmlPropertyData *propertyData,
         const Object *self, const Object *qmlType)
 {
-    lookup->qobjectLookup.qmlTypeIc = qmlType->internalClass();
     setupQObjectLookup(lookup, ddata, propertyData, self);
+    lookup->qobjectLookup.qmlTypeIc.set(self->engine(), qmlType->internalClass());
+}
+
+// template parameter is an ugly trick to avoid pulling in the QObjectMethod header here
+template<typename QObjectMethod = Heap::QObjectMethod>
+inline void setupQObjectMethodLookup(
+        Lookup *lookup, const QQmlData *ddata, const QQmlPropertyData *propertyData,
+        const Object *self, QObjectMethod *method)
+{
+    lookup->releasePropertyCache();
+    Q_ASSERT(!ddata->propertyCache.isNull());
+    auto engine = self->engine();
+    lookup->qobjectMethodLookup.method.set(engine, method);
+    lookup->qobjectMethodLookup.ic.set(engine, self->internalClass());
+    lookup->qobjectMethodLookup.propertyCache = ddata->propertyCache.data();
+    lookup->qobjectMethodLookup.propertyCache->addref();
+    lookup->qobjectMethodLookup.propertyData = propertyData;
+}
+
+inline bool qualifiesForMethodLookup(const QQmlPropertyData *propertyData)
+{
+    return propertyData->isFunction()
+            && !propertyData->isSignalHandler() // TODO: Optimize SignalHandler, too
+            && !propertyData->isVMEFunction() // Handled by QObjectLookup
+            && !propertyData->isVarProperty();
 }
 
 }
