@@ -7,10 +7,7 @@
 #include "qv4argumentsobject_p.h"
 #include <private/qv4mm_p.h>
 #include "qv4scopedvalue_p.h"
-#include "qv4runtime_p.h"
 #include "qv4objectiterator_p.h"
-#include "qv4string_p.h"
-#include "qv4jscall_p.h"
 #include "qv4symbol_p.h"
 #include "qv4propertykey_p.h"
 
@@ -22,9 +19,9 @@ using namespace QV4;
 
 DEFINE_OBJECT_VTABLE(ObjectCtor);
 
-void Heap::ObjectCtor::init(QV4::ExecutionContext *scope)
+void Heap::ObjectCtor::init(QV4::ExecutionEngine *engine)
 {
-    Heap::FunctionObject::init(scope, QStringLiteral("Object"));
+    Heap::FunctionObject::init(engine, QStringLiteral("Object"));
 }
 
 ReturnedValue ObjectCtor::virtualCallAsConstructor(const FunctionObject *f, const Value *argv, int argc, const Value *newTarget)
