@@ -19,6 +19,7 @@
 #include <private/qv4module_p.h>
 #include <private/qv4objectproto_p.h>
 #include <private/qv4qobjectwrapper_p.h>
+#include <private/qv4stackframe_p.h>
 #include <private/qv4value_p.h>
 
 #include <QtCore/qloggingcategory.h>
@@ -103,6 +104,13 @@ bool performLookup(ScopedValue *result, bool *hasProperty, const Lookup &lookup)
     return false;
 }
 
+static QV4::QObjectWrapper::Flags getQmlPropertyFlags(const Lookup *l)
+{
+    return (l && l->forCall)
+            ? QV4::QObjectWrapper::CheckRevision
+            : (QV4::QObjectWrapper::CheckRevision | QV4::QObjectWrapper::AttachMethods);
+}
+
 ReturnedValue QQmlContextWrapper::getPropertyAndBase(const QQmlContextWrapper *resource, PropertyKey id, const Value *receiver, bool *hasProperty, Value *base, Lookup *lookup)
 {
     if (!id.isString())
@@ -189,9 +197,11 @@ ReturnedValue QQmlContextWrapper::getPropertyAndBase(const QQmlContextWrapper *r
             return result->asReturnedValue();
     }
 
-    if (context->imports() && name->startsWithUpper()) {
+    QQmlEnginePrivate *ep = QQmlEnginePrivate::get(v4->qmlEngine());
+    if (context->imports() && (name->startsWithUpper() || context->valueTypesAreAddressable())) {
         // Search for attached properties, enums and imported scripts
-        QQmlTypeNameCache::Result r = context->imports()->query<QQmlImport::AllowRecursion>(name);
+        QQmlTypeNameCache::Result r = context->imports()->query<QQmlImport::AllowRecursion>(
+                name, QQmlTypeLoader::get(ep));
 
         if (r.isValid()) {
             if (hasProperty)
@@ -202,7 +212,7 @@ ReturnedValue QQmlContextWrapper::getPropertyAndBase(const QQmlContextWrapper *r
                     lookup->qmlContextPropertyGetter = QQmlContextWrapper::lookupScript;
                     return lookup->qmlContextPropertyGetter(lookup, v4, base);
                 }
-                QV4::ScopedObject scripts(scope, context->importedScripts().valueRef());
+                QV4::ScopedObject scripts(scope, context->importedScripts());
                 if (scripts)
                     return scripts->get(r.scriptIndex);
                 return QV4::Encode::null();
@@ -213,10 +223,10 @@ ReturnedValue QQmlContextWrapper::getPropertyAndBase(const QQmlContextWrapper *r
                         QQmlEnginePrivate *e = QQmlEnginePrivate::get(v4->qmlEngine());
                         if (r.type.isQObjectSingleton() || r.type.isCompositeSingleton()) {
                             e->singletonInstance<QObject*>(r.type);
-                            lookup->qmlContextSingletonLookup.singletonObject =
+                            lookup->qmlContextSingletonLookup.singletonObject.set(v4,
                                         Value::fromReturnedValue(
                                             QQmlTypeWrapper::create(v4, nullptr, r.type)
-                                        ).heapObject();
+                                        ).heapObject());
                         } else {
                             QJSValue singleton = e->singletonInstance<QJSValue>(r.type);
 
@@ -225,7 +235,7 @@ ReturnedValue QQmlContextWrapper::getPropertyAndBase(const QQmlContextWrapper *r
                             Q_ASSERT(!QJSValuePrivate::asQString(&singleton));
 
                             if (QV4::Value *val = QJSValuePrivate::takeManagedValue(&singleton)) {
-                                lookup->qmlContextSingletonLookup.singletonObject = val->heapObject();
+                                lookup->qmlContextSingletonLookup.singletonObject.set(v4, val->heapObject());
                             } else {
                                 lookup->qmlContextSingletonLookup.singletonValue = QJSValuePrivate::asReturnedValue(&singleton);
                                 isValueSingleton = true;
@@ -241,7 +251,7 @@ ReturnedValue QQmlContextWrapper::getPropertyAndBase(const QQmlContextWrapper *r
                 result = QQmlTypeWrapper::create(v4, scopeObject, context->imports(), r.importNamespace);
             }
             if (lookup) {
-                lookup->qmlTypeLookup.qmlTypeWrapper = result->heapObject();
+                lookup->qmlTypeLookup.qmlTypeWrapper.set(v4, result->heapObject());
                 lookup->qmlContextPropertyGetter = QQmlContextWrapper::lookupType;
             }
             return result->asReturnedValue();
@@ -250,7 +260,6 @@ ReturnedValue QQmlContextWrapper::getPropertyAndBase(const QQmlContextWrapper *r
         // Fall through
     }
 
-    QQmlEnginePrivate *ep = QQmlEnginePrivate::get(v4->qmlEngine());
     Lookup * const originalLookup = lookup;
 
     decltype(lookup->qmlContextPropertyGetter) contextGetterFunction = QQmlContextWrapper::lookupContextObjectProperty;
@@ -261,29 +270,69 @@ ReturnedValue QQmlContextWrapper::getPropertyAndBase(const QQmlContextWrapper *r
         contextGetterFunction = QQmlContextWrapper::lookupScopeObjectProperty;
     }
 
+    QQmlRefPointer<QQmlContextData> outer = context;
     while (context) {
-        if (auto property = searchContextProperties(v4, context, name, hasProperty, base, lookup, originalLookup, ep))
-            return *property;
+        if (outer == context) {
+            if (auto property = searchContextProperties(
+                        v4, context, name, hasProperty, base, lookup, originalLookup, ep)) {
+                return *property;
+            }
+
+            outer = outer->parent();
+
+            if (const auto cu = context->typeCompilationUnit(); cu && cu->componentsAreBound()) {
+                // If components are bound in this CU, we can search the whole context hierarchy
+                // of the file. Bound components' contexts override their local properties.
+                // You also can't instantiate bound components outside of their creation
+                // context. Therefore this is safe.
+
+                for (;
+                     outer && outer->typeCompilationUnit() == cu;
+                     outer = outer->parent()) {
+                    if (auto property = searchContextProperties(
+                                v4, outer, name, hasProperty, base,
+                                nullptr, originalLookup, ep)) {
+                        return *property;
+                    }
+                }
+            }
+        }
 
         // Search scope object
         if (scopeObject) {
             bool hasProp = false;
 
             const QQmlPropertyData *propertyData = nullptr;
-            QV4::ScopedValue result(scope, QV4::QObjectWrapper::getQmlProperty(v4, context, scopeObject,
-                                                                               name, QV4::QObjectWrapper::CheckRevision, &hasProp, &propertyData));
+
+            QV4::ScopedObject wrapper(scope, QV4::QObjectWrapper::wrap(v4, scopeObject));
+            QV4::ScopedValue result(scope, QV4::QObjectWrapper::getQmlProperty(
+                                        v4, context, wrapper->d(), scopeObject, name,
+                                        getQmlPropertyFlags(lookup), &hasProp, &propertyData));
             if (hasProp) {
                 if (hasProperty)
                     *hasProperty = true;
                 if (base)
-                    *base = QV4::QObjectWrapper::wrap(v4, scopeObject);
+                    *base = wrapper;
 
                 if (lookup && propertyData) {
                     QQmlData *ddata = QQmlData::get(scopeObject, false);
                     if (ddata && ddata->propertyCache) {
-                        ScopedValue val(scope, base ? *base : Value::fromReturnedValue(QV4::QObjectWrapper::wrap(v4, scopeObject)));
-                        QV4::setupQObjectLookup(lookup, ddata, propertyData, val->objectValue());
-                        lookup->qmlContextPropertyGetter = QQmlContextWrapper::lookupScopeObjectProperty;
+                        ScopedValue val(
+                                    scope,
+                                    base ? *base : Value::fromReturnedValue(
+                                               QV4::QObjectWrapper::wrap(v4, scopeObject)));
+                        if (QObjectMethod *method = result->as<QObjectMethod>()) {
+                            QV4::setupQObjectMethodLookup(
+                                        lookup, ddata, propertyData, val->objectValue(),
+                                        method->d());
+                            lookup->qmlContextPropertyGetter
+                                    = QQmlContextWrapper::lookupScopeObjectMethod;
+                        } else {
+                            QV4::setupQObjectLookup(
+                                        lookup, ddata, propertyData, val->objectValue());
+                            lookup->qmlContextPropertyGetter
+                                    = QQmlContextWrapper::lookupScopeObjectProperty;
+                        }
                     }
                 }
 
@@ -297,25 +346,38 @@ ReturnedValue QQmlContextWrapper::getPropertyAndBase(const QQmlContextWrapper *r
         if (QObject *contextObject = context->contextObject()) {
             bool hasProp = false;
             const QQmlPropertyData *propertyData = nullptr;
-            result = QV4::QObjectWrapper::getQmlProperty(v4, context, contextObject,
-                                                         name, QV4::QObjectWrapper::CheckRevision,
-                                                         &hasProp, &propertyData);
+            QV4::ScopedObject wrapper(scope, QV4::QObjectWrapper::wrap(v4, contextObject));
+            result = QV4::QObjectWrapper::getQmlProperty(
+                        v4, context, wrapper->d(), contextObject, name, getQmlPropertyFlags(lookup),
+                        &hasProp, &propertyData);
             if (hasProp) {
                 if (hasProperty)
                     *hasProperty = true;
                 if (base)
-                    *base = QV4::QObjectWrapper::wrap(v4, contextObject);
+                    *base = wrapper;
 
                 if (propertyData) {
                     if (lookup) {
                         QQmlData *ddata = QQmlData::get(contextObject, false);
                         if (ddata && ddata->propertyCache
                                 && lookup->qmlContextPropertyGetter != contextGetterFunction) {
-                            ScopedValue val(scope, base ? *base
-                                                        : Value::fromReturnedValue(QV4::QObjectWrapper::wrap(v4, contextObject)));
-                            QV4::setupQObjectLookup(lookup, ddata, propertyData,
-                                                    val->objectValue());
-                            lookup->qmlContextPropertyGetter = contextGetterFunction;
+                            ScopedValue val(
+                                        scope,
+                                        base ? *base : Value::fromReturnedValue(
+                                                   QV4::QObjectWrapper::wrap(v4, contextObject)));
+                            if (QObjectMethod *method = result->as<QObjectMethod>()) {
+                                setupQObjectMethodLookup(
+                                            lookup, ddata, propertyData, val->objectValue(),
+                                            method->d());
+                                if (contextGetterFunction == lookupScopeObjectProperty)
+                                    lookup->qmlContextPropertyGetter = lookupScopeObjectMethod;
+                                else
+                                    lookup->qmlContextPropertyGetter = lookupContextObjectMethod;
+                            } else {
+                                setupQObjectLookup(
+                                            lookup, ddata, propertyData, val->objectValue());
+                                lookup->qmlContextPropertyGetter = contextGetterFunction;
+                            }
                         }
                     } else if (originalLookup) {
                         originalLookup->qmlContextPropertyGetter = lookupInParentContextHierarchy;
@@ -439,8 +501,8 @@ ReturnedValue QQmlContextWrapper::resolveQmlContextPropertyLookupGetter(Lookup *
 {
     Scope scope(engine);
     auto *func = engine->currentStackFrame->v4Function;
-    PropertyKey name =engine->identifierTable->asPropertyKey(
-                func->compilationUnit->runtimeStrings[l->nameIndex]);
+    ScopedPropertyKey name(scope, engine->identifierTable->asPropertyKey(
+                            func->compilationUnit->runtimeStrings[l->nameIndex]));
 
     // Special hack for bounded signal expressions, where the parameters of signals are injected
     // into the handler expression through the locals of the call context. So for onClicked: { ... }
@@ -454,7 +516,7 @@ ReturnedValue QQmlContextWrapper::resolveQmlContextPropertyLookupGetter(Lookup *
                     const auto location = func->sourceLocation();
                     qCWarning(lcQmlContext).nospace().noquote()
                             << location.sourceFile << ":" << location.line << ":" << location.column
-                            << " Parameter \"" << name.toQString() << "\" is not declared."
+                            << " Parameter \"" << name->toQString() << "\" is not declared."
                             << " Injection of parameters into signal handlers is deprecated."
                             << " Use JavaScript functions with formal parameters instead.";
 
@@ -478,8 +540,8 @@ ReturnedValue QQmlContextWrapper::resolveQmlContextPropertyLookupGetter(Lookup *
     Scoped<QmlContext> callingQmlContext(scope, engine->qmlContext());
     if (callingQmlContext) {
         Scoped<QQmlContextWrapper> qmlContextWrapper(scope, callingQmlContext->d()->qml());
-        result = QQmlContextWrapper::getPropertyAndBase(qmlContextWrapper, name, /*receiver*/nullptr, &hasProperty,
-                                                        base, l);
+        result = QQmlContextWrapper::getPropertyAndBase(
+                    qmlContextWrapper, name, /*receiver*/nullptr, &hasProperty, base, l);
     } else {
         // Code path typical to worker scripts, compiled with lookups but no qml context.
         result = l->resolveGlobalGetter(engine);
@@ -490,7 +552,7 @@ ReturnedValue QQmlContextWrapper::resolveQmlContextPropertyLookupGetter(Lookup *
         }
     }
     if (!hasProperty)
-        return engine->throwReferenceError(name.toQString());
+        return engine->throwReferenceError(name->toQString());
     return result->asReturnedValue();
 }
 
@@ -506,7 +568,7 @@ ReturnedValue QQmlContextWrapper::lookupScript(Lookup *l, ExecutionEngine *engin
     if (!context)
         return QV4::Encode::null();
 
-    QV4::ScopedObject scripts(scope, context->importedScripts().valueRef());
+    QV4::ScopedObject scripts(scope, context->importedScripts());
     if (!scripts)
         return QV4::Encode::null();
     return scripts->get(l->qmlContextScriptLookup.scriptIndex);
@@ -556,7 +618,24 @@ ReturnedValue QQmlContextWrapper::lookupIdObjectInParentContext(
     return QQmlContextWrapper::resolveQmlContextPropertyLookupGetter(l, engine, base);
 }
 
-ReturnedValue QQmlContextWrapper::lookupScopeObjectProperty(Lookup *l, ExecutionEngine *engine, Value *base)
+static ReturnedValue revertObjectPropertyLookup(Lookup *l, ExecutionEngine *engine, Value *base)
+{
+    l->qobjectLookup.propertyCache->release();
+    l->qobjectLookup.propertyCache = nullptr;
+    l->qmlContextPropertyGetter = QQmlContextWrapper::resolveQmlContextPropertyLookupGetter;
+    return QQmlContextWrapper::resolveQmlContextPropertyLookupGetter(l, engine, base);
+}
+
+static ReturnedValue revertObjectMethodLookup(Lookup *l, ExecutionEngine *engine, Value *base)
+{
+    l->qobjectMethodLookup.propertyCache->release();
+    l->qobjectMethodLookup.propertyCache = nullptr;
+    l->qmlContextPropertyGetter = QQmlContextWrapper::resolveQmlContextPropertyLookupGetter;
+    return QQmlContextWrapper::resolveQmlContextPropertyLookupGetter(l, engine, base);
+}
+
+template<typename Call>
+ReturnedValue callWithScopeObject(ExecutionEngine *engine, Value *base, Call c)
 {
     Scope scope(engine);
     Scoped<QmlContext> qmlContext(scope, engine->qmlContext());
@@ -570,22 +649,40 @@ ReturnedValue QQmlContextWrapper::lookupScopeObjectProperty(Lookup *l, Execution
     if (QQmlData::wasDeleted(scopeObject))
         return QV4::Encode::undefined();
 
-    const auto revertLookup = [l, engine, base]() {
-        l->qobjectLookup.propertyCache->release();
-        l->qobjectLookup.propertyCache = nullptr;
-        l->qmlContextPropertyGetter = QQmlContextWrapper::resolveQmlContextPropertyLookupGetter;
-        return QQmlContextWrapper::resolveQmlContextPropertyLookupGetter(l, engine, base);
-    };
-
     ScopedValue obj(scope, QV4::QObjectWrapper::wrap(engine, scopeObject));
 
     if (base)
         *base = obj;
 
-    return QObjectWrapper::lookupGetterImpl(l, engine, obj, /*useOriginalProperty*/ true, revertLookup);
+    return c(obj);
 }
 
-ReturnedValue QQmlContextWrapper::lookupContextObjectProperty(Lookup *l, ExecutionEngine *engine, Value *base)
+ReturnedValue QQmlContextWrapper::lookupScopeObjectProperty(Lookup *l, ExecutionEngine *engine, Value *base)
+{
+    return callWithScopeObject(engine, base, [l, engine, base](const Value &obj) {
+        const QObjectWrapper::Flags flags = l->forCall
+                ? QObjectWrapper::NoFlag
+                : QObjectWrapper::AttachMethods;
+        return QObjectWrapper::lookupPropertyGetterImpl(l, engine, obj, flags, [l, engine, base]() {
+            return revertObjectPropertyLookup(l, engine, base);
+        });
+    });
+}
+
+ReturnedValue QQmlContextWrapper::lookupScopeObjectMethod(Lookup *l, ExecutionEngine *engine, Value *base)
+{
+    return callWithScopeObject(engine, base, [l, engine, base](const Value &obj) {
+        const QObjectWrapper::Flags flags = l->forCall
+                ? QObjectWrapper::NoFlag
+                : QObjectWrapper::AttachMethods;
+        return QObjectWrapper::lookupMethodGetterImpl(l, engine, obj, flags, [l, engine, base]() {
+            return revertObjectMethodLookup(l, engine, base);
+        });
+    });
+}
+
+template<typename Call>
+ReturnedValue callWithContextObject(ExecutionEngine *engine, Value *base, Call c)
 {
     Scope scope(engine);
     Scoped<QmlContext> qmlContext(scope, engine->qmlContext());
@@ -603,19 +700,38 @@ ReturnedValue QQmlContextWrapper::lookupContextObjectProperty(Lookup *l, Executi
     if (QQmlData::wasDeleted(contextObject))
         return QV4::Encode::undefined();
 
-    const auto revertLookup = [l, engine, base]() {
-        l->qobjectLookup.propertyCache->release();
-        l->qobjectLookup.propertyCache = nullptr;
-        l->qmlContextPropertyGetter = QQmlContextWrapper::resolveQmlContextPropertyLookupGetter;
-        return QQmlContextWrapper::resolveQmlContextPropertyLookupGetter(l, engine, base);
-    };
-
     ScopedValue obj(scope, QV4::QObjectWrapper::wrap(engine, contextObject));
 
     if (base)
         *base = obj;
 
-    return QObjectWrapper::lookupGetterImpl(l, engine, obj, /*useOriginalProperty*/ true, revertLookup);
+    return c(obj);
+}
+
+ReturnedValue QQmlContextWrapper::lookupContextObjectProperty(
+        Lookup *l, ExecutionEngine *engine, Value *base)
+{
+    return callWithContextObject(engine, base, [l, engine, base](const Value &obj) {
+        const QObjectWrapper::Flags flags = l->forCall
+                ? QObjectWrapper::NoFlag
+                : QObjectWrapper::AttachMethods;
+        return QObjectWrapper::lookupPropertyGetterImpl(l, engine, obj, flags, [l, engine, base]() {
+            return revertObjectPropertyLookup(l, engine, base);
+        });
+    });
+}
+
+ReturnedValue QQmlContextWrapper::lookupContextObjectMethod(
+        Lookup *l, ExecutionEngine *engine, Value *base)
+{
+    return callWithContextObject(engine, base, [l, engine, base](const Value &obj) {
+        const QObjectWrapper::Flags flags = l->forCall
+                ? QObjectWrapper::NoFlag
+                : QObjectWrapper::AttachMethods;
+        return QObjectWrapper::lookupMethodGetterImpl(l, engine, obj, flags, [l, engine, base]() {
+            return revertObjectMethodLookup(l, engine, base);
+        });
+    });
 }
 
 ReturnedValue QQmlContextWrapper::lookupScopeFallbackProperty(Lookup *l, ExecutionEngine *engine, Value *base)
@@ -663,12 +779,13 @@ ReturnedValue QQmlContextWrapper::lookupInParentContextHierarchy(Lookup *l, Exec
         // Search context object
         if (QObject *contextObject = context->contextObject()) {
             bool hasProp = false;
+            QV4::ScopedObject wrapper(scope, QV4::QObjectWrapper::wrap(engine, contextObject));
             result = QV4::QObjectWrapper::getQmlProperty(
-                        engine, context, contextObject, name,
-                        QV4::QObjectWrapper::CheckRevision, &hasProp);
+                        engine, context, wrapper->d(), contextObject, name, getQmlPropertyFlags(l),
+                        &hasProp);
             if (hasProp) {
                 if (base)
-                    *base = QV4::QObjectWrapper::wrap(engine, contextObject);
+                    *base = wrapper;
 
                 return result->asReturnedValue();
             }
@@ -698,7 +815,7 @@ ReturnedValue QQmlContextWrapper::lookupType(Lookup *l, ExecutionEngine *engine,
 
     Heap::Base *heapObject = l->qmlTypeLookup.qmlTypeWrapper;
     if (static_cast<Heap::QQmlTypeWrapper *>(heapObject)->object != scopeObject) {
-        l->qmlTypeLookup.qmlTypeWrapper = nullptr;
+        l->qmlTypeLookup.qmlTypeWrapper.clear();
         l->qmlContextPropertyGetter = QQmlContextWrapper::resolveQmlContextPropertyLookupGetter;
         return QQmlContextWrapper::resolveQmlContextPropertyLookupGetter(l, engine, base);
     }
