@@ -19,6 +19,8 @@
 #include <private/qqmlpropertybinding_p.h>
 #include <private/qqmlbinding_p.h>
 
+QT_BEGIN_NAMESPACE
+
 // Fully inline so that subsequent prop.isBindable check might get ellided.
 
 /*!
@@ -62,6 +64,28 @@ public:
         return binding;
     }
 
+    /*!
+        \overload
+
+        \a object must be non-null
+     */
+    static QQmlAnyBinding ofProperty(QObject *object, QQmlPropertyIndex index)
+    {
+        QQmlAnyBinding binding;
+        Q_ASSERT(object);
+        auto coreIndex = index.coreIndex();
+        // we don't support bindable properties on value types so far
+        if (!index.hasValueTypeIndex()
+            && QQmlData::ensurePropertyCache(object)->property(coreIndex)->isBindable()) {
+            auto metaProp = object->metaObject()->property(coreIndex);
+            QUntypedBindable bindable = metaProp.bindable(object);
+            binding = bindable.binding();
+        } else {
+            binding = QQmlPropertyPrivate::binding(object, index);
+        }
+        return binding;
+    }
+
     /*!
         Removes the binding from the property \a prop, and returns it as a
         QQmlAnyBinding if there was any. Otherwise returns a null
@@ -300,6 +324,28 @@ public:
         return d.asT1();
     }
 
+    /*!
+        \internal
+        Reevaluates the binding. If the binding was disabled,
+        it gets enabled.
+    */
+    void refresh()
+    {
+        if (d.isNull())
+            return;
+        if (d.isT1()) {
+            auto binding = static_cast<QQmlBinding *>(d.asT1());
+            binding->setEnabledFlag(true);
+            binding->refresh();
+        } else {
+            auto bindingPriv = d.asT2();
+            PendingBindingObserverList bindingObservers;
+            bindingPriv->evaluateRecursive(bindingObservers);
+            bindingPriv->notifyNonRecursive(bindingObservers);
+        }
+
+    }
+
     /*!
         \internal
         Stores \a binding and keeps a reference to it.
@@ -413,8 +459,7 @@ private:
                 delete  qqmlptr;
         } else if (d.isT2()) {
             QPropertyBindingPrivate *priv = d.asT2();
-            priv->ref--;
-            if (!priv->ref)
+            if (!priv->deref())
                 QPropertyBindingPrivate::destroyAndFreeMemory(priv);
         }
         d = static_cast<QQmlAbstractBinding *>(nullptr);
@@ -422,5 +467,7 @@ private:
     QBiPointer<QQmlAbstractBinding, QPropertyBindingPrivate> d;
 };
 
+QT_END_NAMESPACE
+
 
 #endif // QQMLANYBINDINGPTR_P_H
