@@ -24,18 +24,21 @@
 QT_BEGIN_NAMESPACE
 
 class QFileSelector;
-class Q_QML_PRIVATE_EXPORT QQmlApplicationEnginePrivate : public QQmlEnginePrivate
+class Q_QML_EXPORT QQmlApplicationEnginePrivate : public QQmlEnginePrivate
 {
     Q_DECLARE_PUBLIC(QQmlApplicationEngine)
 public:
     QQmlApplicationEnginePrivate(QQmlEngine *e);
     ~QQmlApplicationEnginePrivate();
+    void ensureInitialized();
     void init();
     void cleanUp();
 
     void startLoad(const QUrl &url, const QByteArray &data = QByteArray(), bool dataFlag = false);
+    void startLoad(QAnyStringView uri, QAnyStringView type);
     void _q_loadTranslations();
     void finishLoad(QQmlComponent *component);
+    void ensureLoadingFinishes(QQmlComponent *component);
     QList<QObject *> objects;
     QVariantMap initialProperties;
     QStringList extraFileSelectors;
