@@ -3,64 +3,46 @@
 
 #include "qqmlengine_p.h"
 #include "qqmlengine.h"
-#include "qqmlcomponentattached_p.h"
 
-#include "qqmlcontext_p.h"
-#include "qqml.h"
-#include "qqmlcontext.h"
-#include "qqmlexpression.h"
-#include "qqmlcomponent.h"
-#include "qqmlvme_p.h"
-#include "qqmlstringconverters_p.h"
-#include "qqmlscriptstring.h"
-#include "qqmlglobal_p.h"
-#include "qqmlcomponent_p.h"
-#include "qqmlextensioninterface.h"
-#include "qqmllist_p.h"
-#include "qqmltypenamecache_p.h"
-#include "qqmlnotifier_p.h"
-#include "qqmlincubator.h"
-#include "qqmlabstracturlinterceptor.h"
-#include "qqmlsourcecoordinate_p.h"
-#include <private/qqmldirparser_p.h>
+#include <private/qqmlabstractbinding_p.h>
 #include <private/qqmlboundsignal_p.h>
-#include <private/qqmljsdiagnosticmessage_p.h>
-#include <private/qqmltype_p_p.h>
+#include <private/qqmlcontext_p.h>
+#include <private/qqmlnotifier_p.h>
 #include <private/qqmlpluginimporter_p.h>
-#include <QtCore/qstandardpaths.h>
-#include <QtCore/qmetaobject.h>
-#include <QDebug>
+#include <private/qqmlprofiler_p.h>
+#include <private/qqmlscriptdata_p.h>
+#include <private/qqmlsourcecoordinate_p.h>
+#include <private/qqmltype_p.h>
+#include <private/qqmltypedata_p.h>
+#include <private/qqmlvmemetaobject_p.h>
+#include <private/qqmlcomponent_p.h>
+
+#include <private/qobject_p.h>
+#include <private/qthread_p.h>
+
+#include <QtQml/qqml.h>
+#include <QtQml/qqmlcomponent.h>
+#include <QtQml/qqmlcontext.h>
+#include <QtQml/qqmlincubator.h>
+#include <QtQml/qqmlscriptstring.h>
+
 #include <QtCore/qcoreapplication.h>
 #include <QtCore/qcryptographichash.h>
 #include <QtCore/qdir.h>
+#include <QtCore/qmetaobject.h>
 #include <QtCore/qmutex.h>
+#include <QtCore/qstandardpaths.h>
 #include <QtCore/qthread.h>
-#include <private/qthread_p.h>
-#include <private/qqmlscriptdata_p.h>
 
 #if QT_CONFIG(qml_network)
-#include "qqmlnetworkaccessmanagerfactory.h"
-#include <QNetworkAccessManager>
+#include <QtQml/qqmlnetworkaccessmanagerfactory.h>
+#include <QtNetwork/qnetworkaccessmanager.h>
 #endif
 
-#include <private/qobject_p.h>
-#include <private/qmetaobject_p.h>
-#if QT_CONFIG(qml_locale)
-#include <private/qqmllocale_p.h>
-#endif
-#include <private/qqmlbind_p.h>
-#include <private/qqmlconnections_p.h>
-#if QT_CONFIG(qml_animation)
-#include <private/qqmltimer_p.h>
-#endif
-#include <private/qqmlplatform_p.h>
-#include <private/qqmlloggingcategory_p.h>
-#include <private/qv4sequenceobject_p.h>
-
 #ifdef Q_OS_WIN // for %APPDATA%
 #  include <qt_windows.h>
 #  include <shlobj.h>
-#  include <qlibrary.h>
+#  include <QtCore/qlibrary.h>
 #  ifndef CSIDL_APPDATA
 #    define CSIDL_APPDATA           0x001a  // <username>\Application Data
 #  endif
@@ -68,9 +50,11 @@
 
 QT_BEGIN_NAMESPACE
 
+void qml_register_types_QML();
+
 /*!
   \qmltype QtObject
-    \instantiates QObject
+    \nativetype QObject
   \inqmlmodule QtQml
   \ingroup qml-utility-elements
   \brief A basic QML type.
@@ -214,27 +198,39 @@ QQmlEnginePrivate::~QQmlEnginePrivate()
 #if QT_CONFIG(qml_debug)
     delete profiler;
 #endif
+    qDeleteAll(cachedValueTypeInstances);
 }
 
 void QQmlPrivate::qdeclarativeelement_destructor(QObject *o)
 {
-    if (QQmlData *d = QQmlData::get(o)) {
+    QObjectPrivate *p = QObjectPrivate::get(o);
+    if (QQmlData *d = QQmlData::get(p)) {
+        const auto invalidate = [](QQmlContextData *c) {c->invalidate();};
         if (d->ownContext) {
-            for (QQmlRefPointer<QQmlContextData> lc = d->ownContext->linkedContext(); lc;
-                 lc = lc->linkedContext()) {
-                lc->invalidate();
-                if (lc->contextObject() == o)
-                    lc->setContextObject(nullptr);
-            }
-            d->ownContext->invalidate();
-            if (d->ownContext->contextObject() == o)
-                d->ownContext->setContextObject(nullptr);
+            d->ownContext->deepClearContextObject(o, invalidate, invalidate);
             d->ownContext.reset();
             d->context = nullptr;
+            Q_ASSERT(!d->outerContext || d->outerContext->contextObject() != o);
+        } else if (d->outerContext && d->outerContext->contextObject() == o) {
+            d->outerContext->deepClearContextObject(o, invalidate, invalidate);
         }
 
-        if (d->outerContext && d->outerContext->contextObject() == o)
-            d->outerContext->setContextObject(nullptr);
+        if (d->hasVMEMetaObject || d->hasInterceptorMetaObject) {
+            // This is somewhat dangerous because another thread might concurrently
+            // try to resolve the dynamic metaobject. In practice this will then
+            // lead to either the code path that still returns the interceptor
+            // metaobject or the code path that returns the string casted one. Both
+            // is fine if you cannot actually touch the object itself. Since the
+            // other thread is obviously not synchronized to this one, it can't.
+            //
+            // In particular we do this when delivering the frameSwapped() signal
+            // in QQuickWindow. The handler for frameSwapped() is written in a way
+            // that is thread safe as long as QQuickWindow's dtor hasn't finished.
+            // QQuickWindow's dtor does synchronize with the render thread, but it
+            // runs _after_ qdeclarativeelement_destructor.
+            static_cast<QQmlInterceptorMetaObject *>(p->metaObject)->invalidate();
+            d->hasVMEMetaObject = d->hasInterceptorMetaObject = false;
+        }
 
         // Mark this object as in the process of deletion to
         // prevent it resolving in bindings
@@ -242,14 +238,11 @@ void QQmlPrivate::qdeclarativeelement_destructor(QObject *o)
     }
 }
 
-QQmlData::QQmlData()
-    : ownMemory(true), indestructible(true), explicitIndestructibleSet(false),
+QQmlData::QQmlData(Ownership ownership)
+    : ownMemory(ownership == OwnsMemory), indestructible(true), explicitIndestructibleSet(false),
       hasTaintedV4Object(false), isQueuedForDeletion(false), rootObjectInCreation(false),
-      hasInterceptorMetaObject(false), hasVMEMetaObject(false), hasConstWrapper(false),
-      bindingBitsArraySize(InlineBindingArraySize), notifyList(nullptr),
-      bindings(nullptr), signalHandlers(nullptr), nextContextObject(nullptr), prevContextObject(nullptr),
-      lineNumber(0), columnNumber(0), jsEngineId(0),
-      guards(nullptr), extendedData(nullptr)
+      hasInterceptorMetaObject(false), hasVMEMetaObject(false), hasConstWrapper(false), dummy(0),
+      bindingBitsArraySize(InlineBindingArraySize)
 {
     memset(bindingBitsValue, 0, sizeof(bindingBitsValue));
     init();
@@ -299,7 +292,10 @@ void QQmlData::signalEmitted(QAbstractDeclarativeData *, QObject *object, int in
     // QQmlEngine to emit signals from a different thread.  These signals are then automatically
     // marshalled back onto the QObject's thread and handled by QML from there.  This is tested
     // by the qqmlecmascript::threadSignal() autotest.
-    if (!ddata->notifyList)
+
+    // Relaxed semantics here. If we're on a different thread we might schedule a useless event,
+    // but that should be rare.
+    if (!ddata->notifyList.loadRelaxed())
         return;
 
     auto objectThreadData = QObjectPrivate::get(object)->threadData.loadRelaxed();
@@ -312,12 +308,12 @@ void QQmlData::signalEmitted(QAbstractDeclarativeData *, QObject *object, int in
 
         auto ev = std::make_unique<QMetaCallEvent>(m.methodIndex(), 0, nullptr,
                                                    object, index,
-                                                   parameterTypes.count() + 1);
+                                                   parameterTypes.size() + 1);
 
         void **args = ev->args();
         QMetaType *types = ev->types();
 
-        for (int ii = 0; ii < parameterTypes.count(); ++ii) {
+        for (int ii = 0; ii < parameterTypes.size(); ++ii) {
             const QByteArray &typeName = parameterTypes.at(ii);
             if (typeName.endsWith('*'))
                 types[ii + 1] = QMetaType(QMetaType::VoidStar);
@@ -373,11 +369,15 @@ int QQmlData::endpointCount(int index)
 
 void QQmlData::markAsDeleted(QObject *o)
 {
-    QQmlData::setQueuedForDeletion(o);
-
-    QObjectPrivate *p = QObjectPrivate::get(o);
-    for (QList<QObject *>::const_iterator it = p->children.constBegin(), end = p->children.constEnd(); it != end; ++it) {
-        QQmlData::markAsDeleted(*it);
+    QVarLengthArray<QObject *> workStack;
+    workStack.push_back(o);
+    while (!workStack.isEmpty()) {
+        auto currentObject = workStack.last();
+        workStack.pop_back();
+        QQmlData::setQueuedForDeletion(currentObject);
+        auto currentObjectPriv = QObjectPrivate::get(currentObject);
+        for (QObject *child: std::as_const(currentObjectPriv->children))
+            workStack.push_back(child);
     }
 }
 
@@ -387,9 +387,7 @@ void QQmlData::setQueuedForDeletion(QObject *object)
         if (QQmlData *ddata = QQmlData::get(object)) {
             if (ddata->ownContext) {
                 Q_ASSERT(ddata->ownContext.data() == ddata->context);
-                ddata->context->emitDestruction();
-                if (ddata->ownContext->contextObject() == object)
-                    ddata->ownContext->setContextObject(nullptr);
+                ddata->ownContext->deepClearContextObject(object);
                 ddata->ownContext.reset();
                 ddata->context = nullptr;
             }
@@ -400,7 +398,7 @@ void QQmlData::setQueuedForDeletion(QObject *object)
             // possible to get the metaobject anymore.
             // Also, there is no point in evaluating bindings in order to set properties on
             // half-deleted objects.
-            ddata->disconnectNotifiers();
+            ddata->disconnectNotifiers(DeleteNotifyList::No);
         }
     }
 }
@@ -424,38 +422,57 @@ void QQmlData::flushPendingBinding(int coreIndex)
 QQmlData::DeferredData::DeferredData() = default;
 QQmlData::DeferredData::~DeferredData() = default;
 
+template<>
+int qmlRegisterType<void>(const char *uri, int versionMajor, int versionMinor, const char *qmlName)
+{
+    QQmlPrivate::RegisterType type = {
+        QQmlPrivate::RegisterType::CurrentVersion,
+        QMetaType(),
+        QMetaType(),
+        0, nullptr, nullptr,
+        QString(),
+        nullptr,
+        uri,
+        QTypeRevision::fromVersion(versionMajor, versionMinor),
+        qmlName,
+        nullptr,
+        nullptr,
+        nullptr,
+        -1,
+        -1,
+        -1,
+        nullptr,
+        nullptr,
+        nullptr,
+        QTypeRevision::zero(),
+        -1,
+        QQmlPrivate::ValueTypeCreationMethod::None,
+    };
+
+    return QQmlPrivate::qmlregister(QQmlPrivate::TypeRegistration, &type);
+}
+
 bool QQmlEnginePrivate::baseModulesUninitialized = true;
 void QQmlEnginePrivate::init()
 {
     Q_Q(QQmlEngine);
 
     if (baseModulesUninitialized) {
-
-        // required for the Compiler.
-        qmlRegisterType<QObject>("QML", 1, 0, "QtObject");
-        qmlRegisterType<QQmlComponent>("QML", 1, 0, "Component");
-        qmlRegisterAnonymousSequentialContainer<QList<QVariant>>("QML", 1);
-        qmlRegisterAnonymousSequentialContainer<QList<bool>>("QML", 1);
-        qmlRegisterAnonymousSequentialContainer<QList<int>>("QML", 1);
-        qmlRegisterAnonymousSequentialContainer<QList<float>>("QML", 1);
-        qmlRegisterAnonymousSequentialContainer<QList<double>>("QML", 1);
-        qmlRegisterAnonymousSequentialContainer<QList<QString>>("QML", 1);
-        qmlRegisterAnonymousSequentialContainer<QList<QUrl>>("QML", 1);
-        qmlRegisterAnonymousSequentialContainer<QList<QDateTime>>("QML", 1);
-        qmlRegisterAnonymousSequentialContainer<QList<QRegularExpression>>("QML", 1);
-        qmlRegisterAnonymousSequentialContainer<QList<QByteArray>>("QML", 1);
+        // Register builtins
+        qml_register_types_QML();
 
         // No need to specifically register those.
         static_assert(std::is_same_v<QStringList, QList<QString>>);
         static_assert(std::is_same_v<QVariantList, QList<QVariant>>);
 
-        qRegisterMetaType<QVariant>();
         qRegisterMetaType<QQmlScriptString>();
-        qRegisterMetaType<QJSValue>();
         qRegisterMetaType<QQmlComponent::Status>();
         qRegisterMetaType<QList<QObject*> >();
         qRegisterMetaType<QQmlBinding*>();
 
+        // Protect the module: We don't want any URL interceptor to mess with the builtins.
+        qmlProtectModule("QML", 1);
+
         QQmlData::init();
         baseModulesUninitialized = false;
     }
@@ -471,29 +488,16 @@ void QQmlEnginePrivate::init()
   \inmodule QtQml
   \brief The QQmlEngine class provides an environment for instantiating QML components.
 
-  Each QML component is instantiated in a QQmlContext.
-  QQmlContext's are essential for passing data to QML
-  components.  In QML, contexts are arranged hierarchically and this
-  hierarchy is managed by the QQmlEngine.
+  A QQmlEngine is used to manage \l{QQmlComponent}{components} and objects created from
+  them and execute their bindings and functions. QQmlEngine also inherits from
+  \l{QJSEngine} which allows seamless integration between your QML components and
+  JavaScript code.
 
-  Prior to creating any QML components, an application must have
-  created a QQmlEngine to gain access to a QML context.  The
-  following example shows how to create a simple Text item.
+  Each QML component is instantiated in a QQmlContext. In QML, contexts are arranged
+  hierarchically and this hierarchy is managed by the QQmlEngine. By default,
+  components are instantiated in the \l {QQmlEngine::rootContext()}{root context}.
 
-  \code
-  QQmlEngine engine;
-  QQmlComponent component(&engine);
-  component.setData("import QtQuick 2.0\nText { text: \"Hello world!\" }", QUrl());
-  QQuickItem *item = qobject_cast<QQuickItem *>(component.create());
-
-  //add item to view, etc
-  ...
-  \endcode
-
-  In this case, the Text item will be created in the engine's
-  \l {QQmlEngine::rootContext()}{root context}.
-
-  \sa QQmlComponent, QQmlContext, {QML Global Object}
+  \sa QQmlComponent, QQmlContext, {QML Global Object}, QQmlApplicationEngine
 */
 
 /*!
@@ -524,11 +528,12 @@ QQmlEngine::QQmlEngine(QQmlEnginePrivate &dd, QObject *parent)
   invalidated, but not destroyed (unless they are parented to the
   QQmlEngine object).
 
-  See QJSEngine docs for details on cleaning up the JS engine.
+  See ~QJSEngine() for details on cleaning up the JS engine.
 */
 QQmlEngine::~QQmlEngine()
 {
     Q_D(QQmlEngine);
+    handle()->inShutdown = true;
     QJSEnginePrivate::removeFromDebugServer(this);
 
     // Emit onDestruction signals for the root context before
@@ -570,14 +575,16 @@ QQmlEngine::~QQmlEngine()
 /*!
   Clears the engine's internal component cache.
 
-  This function causes the property metadata of all components previously
-  loaded by the engine to be destroyed.  All previously loaded components and
-  the property bindings for all extant objects created from those components will
-  cease to function.
+  This function causes the property metadata of most components previously
+  loaded by the engine to be destroyed. It does so by dropping unreferenced
+  components from the engine's component cache. It does not drop components that
+  are still referenced since that would almost certainly lead to crashes further
+  down the line.
 
-  This function returns the engine to a state where it does not contain any loaded
-  component data.  This may be useful in order to reload a smaller subset of the
-  previous component set, or to load a new version of a previously loaded component.
+  If no components are referenced, this function returns the engine to a state
+  where it does not contain any loaded component data. This may be useful in
+  order to reload a smaller subset of the previous component set, or to load a
+  new version of a previously loaded component.
 
   Once the component cache has been cleared, components must be loaded before
   any new objects can be created.
@@ -597,9 +604,20 @@ QQmlEngine::~QQmlEngine()
 void QQmlEngine::clearComponentCache()
 {
     Q_D(QQmlEngine);
+
+    // Contexts can hold on to CUs but live on the JS heap.
+    // Use a non-incremental GC run to get rid of those.
+    QV4::MemoryManager *mm = handle()->memoryManager;
+    auto oldLimit = mm->gcStateMachine->timeLimit;
+    mm->setGCTimeLimit(-1);
+    mm->runGC();
+    mm->gcStateMachine->timeLimit = std::move(oldLimit);
+
+    handle()->trimCompilationUnits();
     d->typeLoader.lock();
     d->typeLoader.clearCache();
     d->typeLoader.unlock();
+    QQmlMetaType::freeUnusedTypesAndCaches();
 }
 
 /*!
@@ -617,6 +635,7 @@ void QQmlEngine::clearComponentCache()
 void QQmlEngine::trimComponentCache()
 {
     Q_D(QQmlEngine);
+    handle()->trimCompilationUnits();
     d->typeLoader.trimCache();
 }
 
@@ -914,6 +933,45 @@ void QQmlEngine::setOutputWarningsToStandardError(bool enabled)
     d->outputWarningsToMsgLog = enabled;
 }
 
+
+/*!
+  \since 6.6
+  If this method is called inside of a function that is part of
+  a binding in QML, the binding will be treated as a translation binding.
+
+  \code
+  class I18nAwareClass : public QObject {
+
+    //...
+
+     QString text() const
+     {
+          if (auto engine = qmlEngine(this))
+              engine->markCurrentFunctionAsTranslationBinding();
+          return tr("Hello, world!");
+     }
+  };
+  \endcode
+
+  \note This function is mostly useful if you wish to provide your
+  own alternative to the qsTr function. To ensure that properties
+  exposed from C++ classes are updated on language changes, it is
+  instead recommended to react to \c LanguageChange events. That
+  is a more general mechanism which also works when the class is
+  used in a non-QML context, and has slightly less overhead. However,
+  using \c markCurrentFunctionAsTranslationBinding can be acceptable
+  when the class is already closely tied to the QML engine.
+  For more details, see \l {Prepare for Dynamic Language Changes}
+
+  \sa QQmlEngine::retranslate
+*/
+void QQmlEngine::markCurrentFunctionAsTranslationBinding()
+{
+    Q_D(QQmlEngine);
+    if (auto propertyCapture = d->propertyCapture)
+        propertyCapture->captureTranslation();
+}
+
 /*!
   \internal
 
@@ -986,6 +1044,48 @@ QJSValue QQmlEngine::singletonInstance<QJSValue>(int qmlTypeId)
     return d->singletonInstance<QJSValue>(type);
 }
 
+
+/*!
+  \fn template<typename T> T QQmlEngine::singletonInstance(QAnyStringView uri, QAnyStringView typeName)
+
+  \overload
+  Returns the instance of a singleton type named \a typeName from the module specified by \a uri.
+
+  This method can be used as an alternative to calling qmlTypeId followed by the id based overload of
+  singletonInstance. This is convenient when one only needs to do a one time setup of a
+  singleton; if repeated access to the singleton is required, caching its typeId will allow
+  faster subsequent access via the
+  \l {QQmlEngine::singletonInstance(int qmlTypeId)}{type-id based overload}.
+
+  The template argument \e T may be either QJSValue or a pointer to a QObject-derived
+  type and depends on how the singleton was registered. If no instance of \e T has been
+  created yet, it is created now. If \a typeName does not represent a valid singleton
+  type, either a default constructed QJSValue or a \c nullptr is returned.
+
+  \snippet code/src_qml_qqmlengine.cpp 5
+
+  \sa QML_SINGLETON, qmlRegisterSingletonType(), qmlTypeId()
+  \since 6.5
+*/
+template<>
+QJSValue QQmlEngine::singletonInstance<QJSValue>(QAnyStringView uri, QAnyStringView typeName)
+{
+    Q_D(QQmlEngine);
+
+    auto loadHelper = QQml::makeRefPointer<LoadHelper>(&d->typeLoader, uri);
+
+    auto [moduleStatus, type] = loadHelper->resolveType(typeName);
+
+    if (moduleStatus == LoadHelper::ResolveTypeResult::NoSuchModule)
+        return {};
+    if (!type.isValid())
+        return {};
+    if (!type.isSingleton())
+        return {};
+
+    return d->singletonInstance<QJSValue>(type);
+}
+
 /*!
   Refreshes all binding expressions that use strings marked for translation.
 
@@ -1002,13 +1102,14 @@ void QQmlEngine::retranslate()
 }
 
 /*!
-  Returns the QQmlContext for the \a object, or 0 if no
+  Returns the QQmlContext for the \a object, or nullptr if no
   context has been set.
 
-  When the QQmlEngine instantiates a QObject, the context is
-  set automatically.
+  When the QQmlEngine instantiates a QObject, an internal context is assigned
+  to it automatically. Such internal contexts are read-only. You cannot set
+  context properties on them.
 
-  \sa qmlContext(), qmlEngine()
+  \sa qmlContext(), qmlEngine(), QQmlContext::setContextProperty()
   */
 QQmlContext *QQmlEngine::contextForObject(const QObject *object)
 {
@@ -1130,19 +1231,21 @@ void QQmlData::NotifyList::layout()
 
 void QQmlData::deferData(
         int objectIndex, const QQmlRefPointer<QV4::ExecutableCompilationUnit> &compilationUnit,
-        const QQmlRefPointer<QQmlContextData> &context)
+        const QQmlRefPointer<QQmlContextData> &context, const QString &inlineComponentName)
 {
     QQmlData::DeferredData *deferData = new QQmlData::DeferredData;
     deferData->deferredIdx = objectIndex;
     deferData->compilationUnit = compilationUnit;
     deferData->context = context;
+    deferData->inlineComponentName = inlineComponentName;
 
     const QV4::CompiledData::Object *compiledObject = compilationUnit->objectAt(objectIndex);
-    const QV4::BindingPropertyData &propertyData = compilationUnit->bindingPropertyDataPerObject.at(objectIndex);
+    const QV4::CompiledData::BindingPropertyData *propertyData
+            = compilationUnit->bindingPropertyDataPerObjectAt(objectIndex);
 
     const QV4::CompiledData::Binding *binding = compiledObject->bindingTable();
     for (quint32 i = 0; i < compiledObject->nBindings; ++i, ++binding) {
-        const QQmlPropertyData *property = propertyData.at(i);
+        const QQmlPropertyData *property = propertyData->at(i);
         if (binding->hasFlag(QV4::CompiledData::Binding::IsDeferredBinding))
             deferData->bindings.insert(property ? property->coreIndex() : -1, binding);
     }
@@ -1166,49 +1269,73 @@ void QQmlData::releaseDeferredData()
 
 void QQmlData::addNotify(int index, QQmlNotifierEndpoint *endpoint)
 {
-    if (!notifyList) {
-        notifyList = (NotifyList *)malloc(sizeof(NotifyList));
-        notifyList->connectionMask = 0;
-        notifyList->maximumTodoIndex = 0;
-        notifyList->notifiesSize = 0;
-        notifyList->todo = nullptr;
-        notifyList->notifies = nullptr;
+    // Can only happen on "home" thread. We apply relaxed semantics when loading the atomics.
+
+    NotifyList *list = notifyList.loadRelaxed();
+
+    if (!list) {
+        list = new NotifyList;
+        // We don't really care when this change takes effect on other threads. The notifyList can
+        // only become non-null once in the life time of a QQmlData. It becomes null again when the
+        // underlying QObject is deleted. At that point any interaction with the QQmlData is UB
+        // anyway. So, for all intents and purposese, the list becomes non-null once and then stays
+        // non-null "forever". We can apply relaxed semantics.
+        notifyList.storeRelaxed(list);
     }
 
     Q_ASSERT(!endpoint->isConnected());
 
     index = qMin(index, 0xFFFF - 1);
-    notifyList->connectionMask |= (1ULL << quint64(index % 64));
 
-    if (index < notifyList->notifiesSize) {
+    // Likewise, we don't really care _when_ the change in the connectionMask is propagated to other
+    // threads. Cross-thread event ordering is inherently nondeterministic. Therefore, when querying
+    // the conenctionMask in the presence of concurrent modification, any result is correct.
+    list->connectionMask.storeRelaxed(
+            list->connectionMask.loadRelaxed() | (1ULL << quint64(index % 64)));
 
-        endpoint->next = notifyList->notifies[index];
+    if (index < list->notifiesSize) {
+        endpoint->next = list->notifies[index];
         if (endpoint->next) endpoint->next->prev = &endpoint->next;
-        endpoint->prev = &notifyList->notifies[index];
-        notifyList->notifies[index] = endpoint;
-
+        endpoint->prev = &list->notifies[index];
+        list->notifies[index] = endpoint;
     } else {
-        notifyList->maximumTodoIndex = qMax(int(notifyList->maximumTodoIndex), index);
+        list->maximumTodoIndex = qMax(int(list->maximumTodoIndex), index);
 
-        endpoint->next = notifyList->todo;
+        endpoint->next = list->todo;
         if (endpoint->next) endpoint->next->prev = &endpoint->next;
-        endpoint->prev = &notifyList->todo;
-        notifyList->todo = endpoint;
+        endpoint->prev = &list->todo;
+        list->todo = endpoint;
     }
 }
 
-void QQmlData::disconnectNotifiers()
+void QQmlData::disconnectNotifiers(QQmlData::DeleteNotifyList doDelete)
 {
-    if (notifyList) {
-        while (notifyList->todo)
-            notifyList->todo->disconnect();
-        for (int ii = 0; ii < notifyList->notifiesSize; ++ii) {
-            while (QQmlNotifierEndpoint *ep = notifyList->notifies[ii])
+    // Can only happen on "home" thread. We apply relaxed semantics when loading  the atomics.
+    if (NotifyList *list = notifyList.loadRelaxed()) {
+        while (QQmlNotifierEndpoint *todo = list->todo)
+            todo->disconnect();
+        for (int ii = 0; ii < list->notifiesSize; ++ii) {
+            while (QQmlNotifierEndpoint *ep = list->notifies[ii])
                 ep->disconnect();
         }
-        free(notifyList->notifies);
-        free(notifyList);
-        notifyList = nullptr;
+        free(list->notifies);
+
+        if (doDelete == DeleteNotifyList::Yes) {
+            // We can only get here from QQmlData::destroyed(), and that can only come from the
+            // the QObject dtor. If you're still sending signals at that point you have UB already
+            // without any threads. Therefore, it's enough to apply relaxed semantics.
+            notifyList.storeRelaxed(nullptr);
+            delete list;
+        } else {
+            // We can use relaxed semantics here. The worst thing that can happen is that some
+            // signal is falsely reported as connected. Signal connectedness across threads
+            // is not quite deterministic anyway.
+            list->connectionMask.storeRelaxed(0);
+            list->maximumTodoIndex = 0;
+            list->notifiesSize = 0;
+            list->notifies = nullptr;
+
+        }
     }
 }
 
@@ -1293,7 +1420,7 @@ void QQmlData::destroyed(QObject *object)
             guard->objectDestroyed(guard);
     }
 
-    disconnectNotifiers();
+    disconnectNotifiers(DeleteNotifyList::Yes);
 
     if (extendedData)
         delete extendedData;
@@ -1334,7 +1461,7 @@ QQmlData *QQmlData::createQQmlData(QObjectPrivate *priv)
 {
     Q_ASSERT(priv);
     Q_ASSERT(!priv->isDeletingChildren);
-    priv->declarativeData = new QQmlData;
+    priv->declarativeData = new QQmlData(OwnsMemory);
     return static_cast<QQmlData *>(priv->declarativeData);
 }
 
@@ -1391,7 +1518,7 @@ static void dumpwarning(const QQmlError &error)
 
 static void dumpwarning(const QList<QQmlError> &errors)
 {
-    for (int ii = 0; ii < errors.count(); ++ii)
+    for (int ii = 0; ii < errors.size(); ++ii)
         dumpwarning(errors.at(ii));
 }
 
@@ -1489,7 +1616,8 @@ void QQmlEnginePrivate::cleanupScarceResources()
 
   The newly added \a path will be first in the importPathList().
 
-  \sa setImportPathList(), {QML Modules}
+  \b {See also} \l setImportPathList(), \l {QML Modules},
+    and \l [QtQml] {QML Import Path}
 */
 void QQmlEngine::addImportPath(const QString& path)
 {
@@ -1507,9 +1635,8 @@ void QQmlEngine::addImportPath(const QString& path)
   provided by that module. A \c qmldir file is required for defining the
   type version mapping and possibly QML extensions plugins.
 
-  By default, the list contains the directory of the application executable,
-  paths specified in the \c QML_IMPORT_PATH environment variable,
-  and the builtin \c QmlImportsPath from QLibraryInfo.
+  By default, this list contains the paths mentioned in
+  \l {QML Import Path}.
 
   \sa addImportPath(), setImportPathList()
 */
@@ -1523,9 +1650,11 @@ QStringList QQmlEngine::importPathList() const
   Sets \a paths as the list of directories where the engine searches for
   installed modules in a URL-based directory structure.
 
-  By default, the list contains the directory of the application executable,
-  paths specified in the \c QML_IMPORT_PATH environment variable,
-  and the builtin \c QmlImportsPath from QLibraryInfo.
+  By default, this list contains the paths mentioned in
+  \l {QML Import Path}.
+
+  \warning Calling setImportPathList does not preserve the default
+  import paths.
 
   \sa importPathList(), addImportPath()
   */
@@ -1629,10 +1758,20 @@ bool QQmlEngine::importPlugin(const QString &filePath, const QString &uri, QList
 
   \sa {Qt Quick Local Storage QML Types}
 */
+
+/*!
+  \fn void QQmlEngine::offlineStoragePathChanged()
+  This signal is emitted when \l offlineStoragePath changes.
+  \since 6.5
+*/
+
 void QQmlEngine::setOfflineStoragePath(const QString& dir)
 {
     Q_D(QQmlEngine);
+    if (dir == d->offlineStoragePath)
+        return;
     d->offlineStoragePath = dir;
+    Q_EMIT offlineStoragePathChanged();
 }
 
 QString QQmlEngine::offlineStoragePath() const
@@ -1642,10 +1781,12 @@ QString QQmlEngine::offlineStoragePath() const
     if (d->offlineStoragePath.isEmpty()) {
         QString dataLocation = QStandardPaths::writableLocation(QStandardPaths::AppDataLocation);
         QQmlEnginePrivate *e = const_cast<QQmlEnginePrivate *>(d);
-        if (!dataLocation.isEmpty())
+        if (!dataLocation.isEmpty()) {
             e->offlineStoragePath = dataLocation.replace(QLatin1Char('/'), QDir::separator())
                                   + QDir::separator() + QLatin1String("QML")
                                   + QDir::separator() + QLatin1String("OfflineStorage");
+            Q_EMIT e->q_func()->offlineStoragePathChanged();
+        }
     }
 
     return d->offlineStoragePath;
@@ -1677,14 +1818,13 @@ QJSValue QQmlEnginePrivate::singletonInstance<QJSValue>(const QQmlType &type)
 {
     Q_Q(QQmlEngine);
 
-    QJSValue value = singletonInstances.value(type);
-    if (!value.isUndefined()) {
-        return value;
-    }
-
-    QQmlType::SingletonInstanceInfo *siinfo = type.singletonInstanceInfo();
+    QQmlType::SingletonInstanceInfo::ConstPtr siinfo = type.singletonInstanceInfo();
     Q_ASSERT(siinfo != nullptr);
 
+    QJSValue value = singletonInstances.value(siinfo);
+    if (!value.isUndefined())
+        return value;
+
     if (siinfo->scriptCallback) {
         value = siinfo->scriptCallback(q, q);
         if (value.isQObject()) {
@@ -1693,7 +1833,7 @@ QJSValue QQmlEnginePrivate::singletonInstance<QJSValue>(const QQmlType &type)
             // should behave identically to QML singleton types.
             q->setContextForObject(o, new QQmlContext(q->rootContext(), q));
         }
-        singletonInstances.convertAndInsert(v4engine(), type, &value);
+        singletonInstances.convertAndInsert(v4engine(), siinfo, &value);
 
     } else if (siinfo->qobjectCallback) {
         QObject *o = siinfo->qobjectCallback(q, q);
@@ -1722,7 +1862,7 @@ QJSValue QQmlEnginePrivate::singletonInstance<QJSValue>(const QQmlType &type)
         }
 
         value = q->newQObject(o);
-        singletonInstances.convertAndInsert(v4engine(), type, &value);
+        singletonInstances.convertAndInsert(v4engine(), siinfo, &value);
     } else if (!siinfo->url.isEmpty()) {
         QQmlComponent component(q, siinfo->url, QQmlComponent::PreferSynchronous);
         if (component.isError()) {
@@ -1732,8 +1872,26 @@ QJSValue QQmlEnginePrivate::singletonInstance<QJSValue>(const QQmlType &type)
             return QJSValue(QJSValue::UndefinedValue);
         }
         QObject *o = component.beginCreate(q->rootContext());
+        auto *compPriv = QQmlComponentPrivate::get(&component);
+        if (compPriv->state.hasUnsetRequiredProperties()) {
+            /* We would only get the errors from the component after (complete)Create.
+                We can't call create, as we need to convertAndInsert before completeCreate (otherwise
+                tst_qqmllanguage::compositeSingletonCircular fails).
+                On the other hand, we don't want to call cnovertAndInsert if we have an error
+                So create the unset required component errors manually.
+            */
+            delete o;
+            const auto requiredProperties = compPriv->state.requiredProperties();
+            QList<QQmlError> errors (requiredProperties->size());
+            for (const auto &reqProp: *requiredProperties)
+                errors.push_back(QQmlComponentPrivate::unsetRequiredPropertyToQQmlError(reqProp));
+            warning(errors);
+            v4engine()->throwError(QLatin1String("Due to the preceding error(s), Singleton \"%1\" could not be loaded.").arg(QString::fromUtf8(type.typeName())));
+            return QJSValue(QJSValue::UndefinedValue);
+        }
+
         value = q->newQObject(o);
-        singletonInstances.convertAndInsert(v4engine(), type, &value);
+        singletonInstances.convertAndInsert(v4engine(), siinfo, &value);
         component.completeCreate();
     }
 
@@ -1765,7 +1923,7 @@ void QQmlEnginePrivate::executeRuntimeFunction(const QV4::ExecutableCompilationU
                                                int argc, void **args, QMetaType *types)
 {
     Q_ASSERT(unit);
-    Q_ASSERT((functionIndex >= 0) && (functionIndex < unit->runtimeFunctions.length()));
+    Q_ASSERT((functionIndex >= 0) && (functionIndex < unit->runtimeFunctions.size()));
     Q_ASSERT(thisObject);
 
     QQmlData *ddata = QQmlData::get(thisObject);
@@ -1796,7 +1954,7 @@ void QQmlEnginePrivate::executeRuntimeFunction(const QV4::ExecutableCompilationU
         // different version of ExecutionEngine::callInContext() that returns a
         // QV4::ReturnedValue with no arguments since they are not needed by the
         // outer function anyhow
-        QV4::ScopedFunctionObject result(scope,
+        QV4::Scoped<QV4::JavaScriptFunctionObject> result(scope,
             v4->callInContext(function, thisObject, callContext, 0, nullptr));
         Q_ASSERT(result->function());
         Q_ASSERT(result->function()->compilationUnit == function->compilationUnit);
@@ -1811,12 +1969,20 @@ void QQmlEnginePrivate::executeRuntimeFunction(const QV4::ExecutableCompilationU
 
 QV4::ExecutableCompilationUnit *QQmlEnginePrivate::compilationUnitFromUrl(const QUrl &url)
 {
+    QV4::ExecutionEngine *v4 = v4engine();
+    if (auto unit = v4->compilationUnitForUrl(url)) {
+        if (!unit->runtimeStrings)
+            unit->populate();
+        return unit.data();
+    }
+
     auto unit = typeLoader.getType(url)->compilationUnit();
     if (!unit)
         return nullptr;
-    if (!unit->engine)
-        unit->linkToEngine(v4engine());
-    return unit;
+
+    auto executable = v4->executableCompilationUnit(std::move(unit));
+    executable->populate();
+    return executable.data();
 }
 
 QQmlRefPointer<QQmlContextData>
@@ -1829,21 +1995,21 @@ QQmlEnginePrivate::createInternalContext(const QQmlRefPointer<QV4::ExecutableCom
     QQmlRefPointer<QQmlContextData> context;
     context = QQmlContextData::createRefCounted(parentContext);
     context->setInternal(true);
-    context->setImports(unit->typeNameCache);
+    context->setImports(unit->typeNameCache());
     context->initFromTypeCompilationUnit(unit, subComponentIndex);
 
-    if (isComponentRoot && unit->dependentScripts.count()) {
+    const auto *dependentScripts = unit->dependentScriptsPtr();
+    const qsizetype dependentScriptsSize = dependentScripts->size();
+    if (isComponentRoot && dependentScriptsSize) {
         QV4::ExecutionEngine *v4 = v4engine();
         Q_ASSERT(v4);
         QV4::Scope scope(v4);
 
-        QV4::ScopedObject scripts(scope, v4->newArrayObject(unit->dependentScripts.count()));
-        context->setImportedScripts(QV4::PersistentValue(v4, scripts.asReturnedValue()));
+        QV4::ScopedObject scripts(scope, v4->newArrayObject(dependentScriptsSize));
+        context->setImportedScripts(v4, scripts);
         QV4::ScopedValue v(scope);
-        for (int i = 0; i < unit->dependentScripts.count(); ++i) {
-            QQmlRefPointer<QQmlScriptData> s = unit->dependentScripts.at(i);
-            scripts->put(i, (v = s->scriptValueForContext(context)));
-        }
+        for (qsizetype i = 0; i < dependentScriptsSize; ++i)
+            scripts->put(i, (v = dependentScripts->at(i)->scriptValueForContext(context)));
     }
 
     return context;
@@ -1885,7 +2051,7 @@ static inline QString shellNormalizeFileName(const QString &name)
 
 bool QQml_isFileCaseCorrect(const QString &fileName, int lengthIn /* = -1 */)
 {
-#if defined(Q_OS_MAC) || defined(Q_OS_WIN)
+#if defined(Q_OS_DARWIN) || defined(Q_OS_WIN)
     QFileInfo info(fileName);
     const QString absolute = info.absoluteFilePath();
 
@@ -1958,8 +2124,51 @@ bool QQml_isFileCaseCorrect(const QString &fileName, int lengthIn /* = -1 */)
 
 void hasJsOwnershipIndicator(QQmlGuardImpl *) {};
 
+LoadHelper::LoadHelper(QQmlTypeLoader *loader, QAnyStringView uri)
+    : QQmlTypeLoader::Blob({}, QQmlDataBlob::QmlFile, loader)
+    , m_uri(uri.toString())
+
+{
+    auto import = std::make_shared<PendingImport>();
+    import->uri = m_uri;
+    QList<QQmlError> errorList;
+    if (!Blob::addImport(import, &errorList)) {
+        qCDebug(lcQmlImport) << "LoadHelper: Errors loading " << m_uri << errorList;
+        m_uri.clear(); // reset m_uri to remember the failure
+    }
+}
+
+LoadHelper::ResolveTypeResult LoadHelper::resolveType(QAnyStringView typeName)
+{
+    QQmlType type;
+    if (!couldFindModule())
+        return {ResolveTypeResult::NoSuchModule, type};
+    QQmlTypeModule *module = QQmlMetaType::typeModule(m_uri, QTypeRevision{});
+    if (module) {
+        type = module->type(typeName.toString(), {});
+        if (type.isValid())
+            return {ResolveTypeResult::ModuleFound, type};
+    }
+    // The module exists (see check above), but there is no QQmlTypeModule
+    // ==> pure QML module, attempt resolveType
+    QTypeRevision versionReturn;
+    QList<QQmlError> errors;
+    QQmlImportNamespace *ns_return = nullptr;
+    m_importCache->resolveType(
+            typeLoader(), typeName.toString(), &type, &versionReturn, &ns_return, &errors);
+    return {ResolveTypeResult::ModuleFound, type};
+}
+
+bool LoadHelper::couldFindModule() const
+{
+    if (m_uri.isEmpty())
+        return false;
+    for (const auto &import: std::as_const(m_unresolvedImports))
+        if (import->priority == 0) // compare QQmlTypeData::allDependenciesDone
+            return false;
+    return true;
+}
+
 QT_END_NAMESPACE
 
-#include "moc_qqmlengine_p.cpp"
-
 #include "moc_qqmlengine.cpp"
