@@ -25,6 +25,8 @@
 #include <private/qv4generatorobject_p.h>
 #include <private/qv4resolvedtypereference_p.h>
 #include <private/qqmlpropertybinding_p.h>
+#include <private/qqmlanybinding_p.h>
+#include <QtQml/private/qqmlvme_p.h>
 
 #include <QScopedValueRollback>
 
@@ -36,60 +38,78 @@ Q_LOGGING_CATEGORY(lcQmlDefaultMethod, "qt.qml.defaultmethod")
 
 QT_USE_NAMESPACE
 
+Q_TRACE_PREFIX(qtqml,
+"namespace QV4 {" \
+"struct ExecutionEngine;" \
+"class ExecutableCompilationUnit;" \
+"namespace CompiledData {" \
+"struct Object;" \
+"}}" \
+"class QQmlEngine;"
+)
+
+Q_TRACE_POINT(qtqml, QQmlObjectCreator_createInstance_entry, const QV4::ExecutableCompilationUnit *compilationUnit, const QV4::CompiledData::Object *object, const QUrl &url)
+Q_TRACE_POINT(qtqml, QQmlObjectCreator_createInstance_exit, const QString &typeName)
+
 QQmlObjectCreator::QQmlObjectCreator(
-        QQmlRefPointer<QQmlContextData> parentContext,
+        const QQmlRefPointer<QQmlContextData> &parentContext,
         const QQmlRefPointer<QV4::ExecutableCompilationUnit> &compilationUnit,
         const QQmlRefPointer<QQmlContextData> &creationContext,
+        const QString &inlineComponentName,
         QQmlIncubatorPrivate *incubator)
     : phase(Startup)
+    , m_inlineComponentName(inlineComponentName)
     , compilationUnit(compilationUnit)
-    , propertyCaches(&compilationUnit->propertyCaches)
+    , propertyCaches(compilationUnit->propertyCachesPtr())
     , sharedState(new QQmlObjectCreatorSharedState, QQmlRefPointer<QQmlObjectCreatorSharedState>::Adopt)
     , topLevelCreator(true)
     , isContextObject(true)
     , incubator(incubator)
 {
-    init(std::move(parentContext));
+    init(parentContext);
 
     sharedState->componentAttached = nullptr;
-    sharedState->allCreatedBindings.allocate(compilationUnit->totalBindingsCount());
-    sharedState->allParserStatusCallbacks.allocate(compilationUnit->totalParserStatusCount());
-    sharedState->allCreatedObjects.allocate(compilationUnit->totalObjectCount());
-    sharedState->allJavaScriptObjects = nullptr;
+    sharedState->allCreatedBindings.allocate(compilationUnit->totalBindingsCount(inlineComponentName));
+    sharedState->allParserStatusCallbacks.allocate(compilationUnit->totalParserStatusCount(inlineComponentName));
+    sharedState->allCreatedObjects.allocate(compilationUnit->totalObjectCount(inlineComponentName));
+    sharedState->allJavaScriptObjects = ObjectInCreationGCAnchorList();
     sharedState->creationContext = creationContext;
     sharedState->rootContext.reset();
     sharedState->hadTopLevelRequiredProperties = false;
 
     if (auto profiler = QQmlEnginePrivate::get(engine)->profiler) {
         Q_QML_PROFILE_IF_ENABLED(QQmlProfilerDefinitions::ProfileCreating, profiler,
-                sharedState->profiler.init(profiler, compilationUnit->totalParserStatusCount()));
+                sharedState->profiler.init(profiler, compilationUnit->totalParserStatusCount(inlineComponentName)));
     } else {
         Q_UNUSED(profiler);
     }
 }
 
-QQmlObjectCreator::QQmlObjectCreator(QQmlRefPointer<QQmlContextData> parentContext,
-        const QQmlRefPointer<QV4::ExecutableCompilationUnit> &compilationUnit,
-        QQmlObjectCreatorSharedState *inheritedSharedState, bool isContextObject)
+QQmlObjectCreator::QQmlObjectCreator(const QQmlRefPointer<QQmlContextData> &parentContext,
+                                     const QQmlRefPointer<QV4::ExecutableCompilationUnit> &compilationUnit, const QString inlineComponentName,
+                                     QQmlObjectCreatorSharedState *inheritedSharedState, bool isContextObject)
     : phase(Startup)
+    , m_inlineComponentName(inlineComponentName)
     , compilationUnit(compilationUnit)
-    , propertyCaches(&compilationUnit->propertyCaches)
+    , propertyCaches(compilationUnit->propertyCachesPtr())
     , sharedState(inheritedSharedState)
     , topLevelCreator(false)
     , isContextObject(isContextObject)
     , incubator(nullptr)
 {
-    init(std::move(parentContext));
+    init(parentContext);
 }
 
-void QQmlObjectCreator::init(QQmlRefPointer<QQmlContextData> providedParentContext)
+void QQmlObjectCreator::init(const QQmlRefPointer<QQmlContextData> &providedParentContext)
 {
-    parentContext = std::move(providedParentContext);
+    parentContext = providedParentContext;
     engine = parentContext->engine();
     v4 = engine->handle();
 
-    if (compilationUnit && !compilationUnit->engine)
-        compilationUnit->linkToEngine(v4);
+    Q_ASSERT(compilationUnit);
+    Q_ASSERT(compilationUnit->engine == v4);
+    if (!compilationUnit->runtimeStrings)
+        compilationUnit->populate();
 
     qmlUnit = compilationUnit->unitData();
     _qobject = nullptr;
@@ -139,7 +159,7 @@ QObject *QQmlObjectCreator::create(int subComponentIndex, QObject *parent, QQmlI
     } else {
         Q_ASSERT(subComponentIndex >= 0);
         if (flags & CreationFlags::InlineComponent) {
-            if (compilationUnit->unitData()->flags & QV4::CompiledData::Unit::ComponentsBound
+            if (compilationUnit->componentsAreBound()
                     && compilationUnit != parentContext->typeCompilationUnit()) {
                 recordError({}, tr("Cannot instantiate bound inline component in different file"));
                 phase = ObjectsCreated;
@@ -149,7 +169,7 @@ QObject *QQmlObjectCreator::create(int subComponentIndex, QObject *parent, QQmlI
             isComponentRoot = true;
         } else {
             Q_ASSERT(flags & CreationFlags::NormalObject);
-            if (compilationUnit->unitData()->flags & QV4::CompiledData::Unit::ComponentsBound
+            if (compilationUnit->componentsAreBound()
                     && sharedState->creationContext != parentContext) {
                 recordError({}, tr("Cannot instantiate bound component "
                                    "outside its creation context"));
@@ -172,13 +192,14 @@ QObject *QQmlObjectCreator::create(int subComponentIndex, QObject *parent, QQmlI
 
     QV4::Scope scope(v4);
 
-    Q_ASSERT(sharedState->allJavaScriptObjects || topLevelCreator);
+    Q_ASSERT(sharedState->allJavaScriptObjects.canTrack() || topLevelCreator);
     if (topLevelCreator)
-        sharedState->allJavaScriptObjects = scope.alloc(compilationUnit->totalObjectCount());
+        sharedState->allJavaScriptObjects = ObjectInCreationGCAnchorList(scope, compilationUnit->totalObjectCount(m_inlineComponentName));
 
     if (!isComponentRoot && sharedState->creationContext) {
         // otherwise QQmlEnginePrivate::createInternalContext() handles it
-        context->setImportedScripts(sharedState->creationContext->importedScripts());
+        QV4::ScopedValue scripts(scope, sharedState->creationContext->importedScripts());
+        context->setImportedScripts(v4, scripts);
     }
 
     QObject *instance = createInstance(objectToCreate, parent, /*isContextObject*/true);
@@ -189,7 +210,7 @@ QObject *QQmlObjectCreator::create(int subComponentIndex, QObject *parent, QQmlI
     }
 
     if (topLevelCreator)
-        sharedState->allJavaScriptObjects = nullptr;
+        sharedState->allJavaScriptObjects = ObjectInCreationGCAnchorList();
 
     phase = CreatingObjectsPhase2;
 
@@ -218,10 +239,11 @@ void QQmlObjectCreator::beginPopulateDeferred(const QQmlRefPointer<QQmlContextDa
     sharedState->rootContext = newContext;
 
     Q_ASSERT(topLevelCreator);
-    Q_ASSERT(!sharedState->allJavaScriptObjects);
+    Q_ASSERT(!sharedState->allJavaScriptObjects.canTrack());
 
+    // FIXME (QTBUG-122956): allocating from the short lived scope does not make any sense
     QV4::Scope valueScope(v4);
-    sharedState->allJavaScriptObjects = valueScope.alloc(compilationUnit->totalObjectCount());
+    sharedState->allJavaScriptObjects = ObjectInCreationGCAnchorList(valueScope, compilationUnit->totalObjectCount(m_inlineComponentName));
 }
 
 void QQmlObjectCreator::populateDeferred(QObject *instance, int deferredIndex,
@@ -275,6 +297,16 @@ void QQmlObjectCreator::populateDeferredBinding(const QQmlProperty &qmlProperty,
     }
 }
 
+void QQmlObjectCreator::populateDeferredInstance(
+        QObject *outerObject, int deferredIndex, int index, QObject *instance,
+        QObject *bindingTarget, const QQmlPropertyData *valueTypeProperty,
+        const QV4::CompiledData::Binding *binding)
+{
+    doPopulateDeferred(outerObject, deferredIndex, [&]() {
+        populateInstance(index, instance, bindingTarget, valueTypeProperty, binding);
+    });
+}
+
 void QQmlObjectCreator::finalizePopulateDeferred()
 {
     phase = ObjectsCreated;
@@ -288,8 +320,11 @@ void QQmlObjectCreator::setPropertyValue(const QQmlPropertyData *property, const
     QMetaType propertyType = property->propType();
 
     if (property->isEnum()) {
-        if (binding->hasFlag(QV4::CompiledData::Binding::IsResolvedEnum)) {
-            propertyType = QMetaType::fromType<int>();
+        if (binding->hasFlag(QV4::CompiledData::Binding::IsResolvedEnum) ||
+                // TODO: For historical reasons you can assign any number to an enum property alias
+                //       This can be fixed with an opt-out mechanism, for example a pragma.
+                (property->isAlias() && binding->isNumberBinding())) {
+            propertyType = property->propType().underlyingType();
         } else {
             // ### This should be resolved earlier at compile time and the binding value should be changed accordingly.
             QVariant value = compilationUnit->bindingValueAsString(binding);
@@ -411,6 +446,49 @@ void QQmlObjectCreator::setPropertyValue(const QQmlPropertyData *property, const
         break;
     }
     break;
+    case QMetaType::SChar: {
+        assertType(QV4::CompiledData::Binding::Type_Number);
+        double d = compilationUnit->bindingValueAsNumber(binding);
+        qint8 value = qint8(d);
+        property->writeProperty(_qobject, &value, propertyWriteFlags);
+        break;
+    }
+    case QMetaType::UChar: {
+        assertType(QV4::CompiledData::Binding::Type_Number);
+        double d = compilationUnit->bindingValueAsNumber(binding);
+        quint8 value = quint8(d);
+        property->writeProperty(_qobject, &value, propertyWriteFlags);
+        break;
+    }
+    case QMetaType::Short: {
+        assertType(QV4::CompiledData::Binding::Type_Number);
+        double d = compilationUnit->bindingValueAsNumber(binding);
+        qint16 value = qint16(d);
+        property->writeProperty(_qobject, &value, propertyWriteFlags);
+        break;
+    }
+    case QMetaType::UShort: {
+        assertType(QV4::CompiledData::Binding::Type_Number);
+        double d = compilationUnit->bindingValueAsNumber(binding);
+        quint16 value = quint16(d);
+        property->writeProperty(_qobject, &value, propertyWriteFlags);
+        break;
+    }
+    case QMetaType::LongLong: {
+        assertType(QV4::CompiledData::Binding::Type_Number);
+        double d = compilationUnit->bindingValueAsNumber(binding);
+        qint64 value = qint64(d);
+        property->writeProperty(_qobject, &value, propertyWriteFlags);
+        break;
+    }
+    case QMetaType::ULongLong: {
+        assertType(QV4::CompiledData::Binding::Type_Number);
+        double d = compilationUnit->bindingValueAsNumber(binding);
+        quint64 value = quint64(d);
+        property->writeProperty(_qobject, &value, propertyWriteFlags);
+        break;
+    }
+    break;
     case QMetaType::Float: {
         assertType(QV4::CompiledData::Binding::Type_Number);
         float value = float(compilationUnit->bindingValueAsNumber(binding));
@@ -424,9 +502,9 @@ void QQmlObjectCreator::setPropertyValue(const QQmlPropertyData *property, const
     }
     break;
     case QMetaType::QColor: {
-        QVariant data;
-        if (QQml_valueTypeProvider()->createValueType(
-                    propertyType, compilationUnit->bindingValueAsString(binding), data)) {
+        QVariant data = QQmlValueTypeProvider::createValueType(
+                    compilationUnit->bindingValueAsString(binding), propertyType);
+        if (data.isValid()) {
             property->writeProperty(_qobject, data.data(), propertyWriteFlags);
         }
     }
@@ -507,11 +585,9 @@ void QQmlObjectCreator::setPropertyValue(const QQmlPropertyData *property, const
     case QMetaType::QVector3D:
     case QMetaType::QVector4D:
     case QMetaType::QQuaternion: {
-        QVariant result;
-        bool ok = QQml_valueTypeProvider()->createValueType(
-                    propertyType, compilationUnit->bindingValueAsString(binding), result);
-        assertOrNull(ok);
-        Q_UNUSED(ok);
+        QVariant result = QQmlValueTypeProvider::createValueType(
+                    compilationUnit->bindingValueAsString(binding), propertyType);
+        assertOrNull(result.isValid());
         property->writeProperty(_qobject, result.data(), propertyWriteFlags);
         break;
     }
@@ -575,6 +651,36 @@ void QQmlObjectCreator::setPropertyValue(const QQmlPropertyData *property, const
             }
             property->writeProperty(_qobject, &value, propertyWriteFlags);
             break;
+        } else {
+            QVariant source;
+            switch (binding->type()) {
+            case QV4::CompiledData::Binding::Type_Boolean:
+                source = binding->valueAsBoolean();
+                break;
+            case QV4::CompiledData::Binding::Type_Number: {
+                const double n = compilationUnit->bindingValueAsNumber(binding);
+                if (double(int(n)) == n)
+                    source = int(n);
+                else
+                    source = n;
+                break;
+            }
+            case QV4::CompiledData::Binding::Type_Null:
+                source = QVariant::fromValue<std::nullptr_t>(nullptr);
+                break;
+            case QV4::CompiledData::Binding::Type_Invalid:
+                break;
+            default:
+                source = compilationUnit->bindingValueAsString(binding);
+                break;
+            }
+
+            QVariant target = QQmlValueTypeProvider::createValueType(
+                    source, propertyType, engine->handle());
+            if (target.isValid()) {
+                property->writeProperty(_qobject, target.data(), propertyWriteFlags);
+                break;
+            }
         }
 
         // string converters are not exposed, so ending up here indicates an error
@@ -603,10 +709,11 @@ void QQmlObjectCreator::setupBindings(BindingSetupFlags mode)
     QQmlListProperty<void> savedList;
     qSwap(_currentList, savedList);
 
-    const QV4::BindingPropertyData &propertyData = compilationUnit->bindingPropertyDataPerObject.at(_compiledObjectIndex);
+    const QV4::CompiledData::BindingPropertyData *propertyData
+            = compilationUnit->bindingPropertyDataPerObjectAt(_compiledObjectIndex);
 
     if (_compiledObject->idNameIndex) {
-        const QQmlPropertyData *idProperty = propertyData.last();
+        const QQmlPropertyData *idProperty = propertyData->last();
         Q_ASSERT(!idProperty || !idProperty->isValid() || idProperty->name(_qobject) == QLatin1String("id"));
         if (idProperty && idProperty->isValid() && idProperty->isWritable() && idProperty->propType().id() == QMetaType::QString) {
             QV4::CompiledData::Binding idBinding;
@@ -653,7 +760,7 @@ void QQmlObjectCreator::setupBindings(BindingSetupFlags mode)
 
     const QV4::CompiledData::Binding *binding = _compiledObject->bindingTable();
     for (quint32 i = 0; i < _compiledObject->nBindings; ++i, ++binding) {
-        const QQmlPropertyData *const property = propertyData.at(i);
+        const QQmlPropertyData *const property = propertyData->at(i);
         if (property) {
             const QQmlPropertyData *targetProperty = property;
             if (targetProperty->isAlias()) {
@@ -663,8 +770,9 @@ void QQmlObjectCreator::setupBindings(BindingSetupFlags mode)
                 QQmlData *data = QQmlData::get(targetObject);
                 Q_ASSERT(data && data->propertyCache);
                 targetProperty = data->propertyCache->property(targetIndex.coreIndex());
+                sharedState->requiredProperties.remove({targetObject, targetProperty});
             }
-            sharedState->requiredProperties.remove(targetProperty);
+            sharedState->requiredProperties.remove({_bindingTarget, property});
         }
 
 
@@ -733,16 +841,17 @@ bool QQmlObjectCreator::setPropertyBinding(const QQmlPropertyData *bindingProper
         QV4::ResolvedTypeReference *tr = resolvedType(binding->propertyNameIndex);
         Q_ASSERT(tr);
         QQmlType attachedType = tr->type();
+        QQmlEnginePrivate *enginePrivate = QQmlEnginePrivate::get(engine);
         if (!attachedType.isValid()) {
             QQmlTypeNameCache::Result res = context->imports()->query(
-                        stringAt(binding->propertyNameIndex));
+                    stringAt(binding->propertyNameIndex), QQmlTypeLoader::get(enginePrivate));
             if (res.isValid())
                 attachedType = res.type;
             else
                 return false;
         }
         QObject *qmlObject = qmlAttachedPropertiesObject(
-                _qobject, attachedType.attachedPropertiesFunction(QQmlEnginePrivate::get(engine)));
+                _qobject, attachedType.attachedPropertiesFunction(enginePrivate));
         if (!qmlObject) {
             recordError(binding->location,
                         QStringLiteral("Could not create attached properties object '%1'")
@@ -817,8 +926,18 @@ bool QQmlObjectCreator::setPropertyBinding(const QQmlPropertyData *bindingProper
                 void *argv[1] = { &groupObject };
                 QMetaObject::metacall(_qobject, QMetaObject::ReadProperty, bindingProperty->coreIndex(), argv);
                 if (!groupObject) {
-                    recordError(binding->location, tr("Cannot set properties on %1 as it is null").arg(stringAt(binding->propertyNameIndex)));
-                    return false;
+                    QQmlPropertyIndex index(bindingProperty->coreIndex());
+                    auto anyBinding = QQmlAnyBinding::ofProperty(_qobject, index);
+                    if (anyBinding) {
+                        // if there is a binding, try to force-evaluate it now
+                        // this might instantiate a necessary part of a grouped property
+                        anyBinding.refresh();
+                        QMetaObject::metacall(_qobject, QMetaObject::ReadProperty, bindingProperty->coreIndex(), argv);
+                    }
+                    if (!groupObject) {
+                        recordError(binding->location, tr("Cannot set properties on %1 as it is null").arg(stringAt(binding->propertyNameIndex)));
+                        return false;
+                    }
                 }
 
                 bindingTarget = groupObject;
@@ -846,14 +965,13 @@ bool QQmlObjectCreator::setPropertyBinding(const QQmlPropertyData *bindingProper
             && !(bindingFlags & QV4::CompiledData::Binding::IsPropertyObserver)
             && !_valueTypeProperty;
 
-    if (_ddata->hasBindingBit(bindingProperty->coreIndex()) && allowedToRemoveBinding) {
-        QQmlPropertyPrivate::removeBinding(_bindingTarget, QQmlPropertyIndex(bindingProperty->coreIndex()));
-    } else if (bindingProperty->isBindable() && allowedToRemoveBinding) {
-        QList<DeferredQPropertyBinding> &pendingBindings = sharedState.data()->allQPropertyBindings;
-        auto it = std::remove_if(pendingBindings.begin(), pendingBindings.end(), [&](const DeferredQPropertyBinding &deferred) {
-            return deferred.properyIndex == bindingProperty->coreIndex() && deferred.target == _bindingTarget;
-        });
-        pendingBindings.erase(it, pendingBindings.end());
+    if (allowedToRemoveBinding) {
+        if (bindingProperty->isBindable()) {
+            removePendingBinding(_bindingTarget, bindingProperty->coreIndex());
+        } else {
+            QQmlPropertyPrivate::removeBinding(
+                    _bindingTarget, QQmlPropertyIndex(bindingProperty->coreIndex()));
+        }
     }
 
     if (bindingType == QV4::CompiledData::Binding::Type_Script || binding->isTranslationBinding()) {
@@ -899,6 +1017,9 @@ bool QQmlObjectCreator::setPropertyBinding(const QQmlPropertyData *bindingProper
                 qmlBinding = QQmlPropertyBinding::create(bindingProperty, runtimeFunction, _scopeObject, context, currentQmlContext(), _bindingTarget, index);
             }
             sharedState.data()->allQPropertyBindings.push_back(DeferredQPropertyBinding {_bindingTarget, bindingProperty->coreIndex(), qmlBinding });
+
+            QQmlData *data = QQmlData::get(_bindingTarget, true);
+            data->setBindingBit(_bindingTarget, bindingProperty->coreIndex());
         } else {
             // When writing bindings to grouped properties implemented as value types,
             // such as point.x: { someExpression; }, then the binding is installed on
@@ -1024,9 +1145,9 @@ bool QQmlObjectCreator::setPropertyBinding(const QQmlPropertyData *bindingProper
                 recordError(binding->valueLocation, tr("Cannot assign object type %1 with no default method").arg(QString::fromLatin1(createdSubObject->metaObject()->className())));
                 return false;
             }
-            qCWarning(lcQmlDefaultMethod) << "Assigning an object to a signal handler is deprecated."
-                                             "Instead, create the object, give it an id, and call the desired slot from the signal handler."
-                                             ;
+            qCWarning(lcQmlDefaultMethod) << "Assigning an object to a signal handler is deprecated. "
+                                             "Instead, create the object, give it an id, and call the desired slot "
+                                             "from the signal handler. The object is:" << createdSubObject;
 
             QMetaMethod signalMethod = _qobject->metaObject()->method(bindingProperty->coreIndex());
             if (!QMetaObject::checkConnectArgs(signalMethod, method)) {
@@ -1182,21 +1303,12 @@ QObject *QQmlObjectCreator::createInstance(int index, QObject *parent, bool isCo
         if (type.isValid() && !type.isInlineComponentType()) {
             typeName = type.qmlTypeName();
 
-            void *ddataMemory = nullptr;
-            instance = type.create(&ddataMemory, sizeof(QQmlData));
+            instance = type.createWithQQmlData();
             if (!instance) {
                 recordError(obj->location, tr("Unable to create object of type %1").arg(stringAt(obj->inheritedTypeNameIndex)));
                 return nullptr;
             }
 
-            {
-                QQmlData *ddata = new (ddataMemory) QQmlData;
-                ddata->ownMemory = false;
-                QObjectPrivate* p = QObjectPrivate::get(instance);
-                Q_ASSERT(!p->isDeletingChildren);
-                p->declarativeData = ddata;
-            }
-
             const int finalizerCast = type.finalizerCast();
             if (finalizerCast != -1) {
                 auto hook = reinterpret_cast<QQmlFinalizerHook *>(reinterpret_cast<char *>(instance) + finalizerCast);
@@ -1216,7 +1328,7 @@ QObject *QQmlObjectCreator::createInstance(int index, QObject *parent, bool isCo
 
             sharedState->allCreatedObjects.push(instance);
         } else {
-            const auto compilationUnit = typeRef->compilationUnit();
+            auto compilationUnit = typeRef->compilationUnit();
             Q_ASSERT(compilationUnit);
             typeName = compilationUnit->fileName();
             // compilation unit is shared between root type and its inline component types
@@ -1227,19 +1339,31 @@ QObject *QQmlObjectCreator::createInstance(int index, QObject *parent, bool isCo
             }
 
             if (!type.isInlineComponentType()) {
-                QQmlObjectCreator subCreator(context, compilationUnit, sharedState.data(),
-                                             isContextObject);
+                QQmlObjectCreator subCreator(
+                        context, engine->handle()->executableCompilationUnit(
+                                         std::move(compilationUnit)),
+                        QString(), // not an inline component
+                        sharedState.data(), isContextObject);
                 instance = subCreator.create();
                 if (!instance) {
                     errors += subCreator.errors;
                     return nullptr;
                 }
             } else {
-                int subObjectId = type.inlineComponentId();
-                QScopedValueRollback<int> rollback {compilationUnit->icRoot, subObjectId};
-                QQmlObjectCreator subCreator(context, compilationUnit, sharedState.data(),
-                                             isContextObject);
-                instance = subCreator.create(subObjectId, nullptr, nullptr, CreationFlags::InlineComponent);
+                const QString inlineComponentName = type.elementName();
+
+                const int inlineComponentId
+                        = compilationUnit->inlineComponentId(inlineComponentName);
+                QQmlObjectCreator subCreator(
+                        context,
+                        engine->handle()->executableCompilationUnit(
+                                QQmlRefPointer<QV4::CompiledData::CompilationUnit>(
+                                        compilationUnit)),
+                        inlineComponentName,
+                        sharedState.data(),
+                        isContextObject);
+                instance = subCreator.create(
+                        inlineComponentId, nullptr, nullptr, CreationFlags::InlineComponent);
                 if (!instance) {
                     errors += subCreator.errors;
                     return nullptr;
@@ -1292,7 +1416,7 @@ QObject *QQmlObjectCreator::createInstance(int index, QObject *parent, bool isCo
 
     if (customParser && obj->hasFlag(QV4::CompiledData::Object::HasCustomParserBindings)) {
         customParser->engine = QQmlEnginePrivate::get(engine);
-        customParser->imports = compilationUnit->typeNameCache.data();
+        customParser->imports = compilationUnit->typeNameCache().data();
 
         QList<const QV4::CompiledData::Binding *> bindings;
         const QV4::CompiledData::Object *obj = compilationUnit->objectAt(index);
@@ -1320,9 +1444,8 @@ QObject *QQmlObjectCreator::createInstance(int index, QObject *parent, bool isCo
     QObject *scopeObject = instance;
     qSwap(_scopeObject, scopeObject);
 
-    Q_ASSERT(sharedState->allJavaScriptObjects);
-    *sharedState->allJavaScriptObjects = QV4::QObjectWrapper::wrap(v4, instance);
-    ++sharedState->allJavaScriptObjects;
+    Q_ASSERT(sharedState->allJavaScriptObjects.canTrack());
+    sharedState->allJavaScriptObjects.trackObject(v4, instance);
 
     QV4::Scope valueScope(v4);
     QV4::QmlContext *qmlContext = static_cast<QV4::QmlContext *>(valueScope.alloc());
@@ -1404,20 +1527,38 @@ bool QQmlObjectCreator::finalize(QQmlInstantiationInterrupt &interrupt)
 
     while (!sharedState->allQPropertyBindings.isEmpty()) {
         auto& [target, index, qmlBinding] = sharedState->allQPropertyBindings.first();
+
+        QQmlData *data = QQmlData::get(target);
+        if (!data || !data->hasBindingBit(index)) {
+            // The target property has been overwritten since we stashed the binding.
+            sharedState->allQPropertyBindings.pop_front();
+            continue;
+        }
+
         QUntypedBindable bindable;
         void *argv[] = { &bindable };
         // allow interception
         target->metaObject()->metacall(target, QMetaObject::BindableProperty, index, argv);
-        bindable.setBinding(qmlBinding);
+        const bool success = bindable.setBinding(qmlBinding);
+
+        const auto bindingPrivateRefCount = QPropertyBindingPrivate::get(qmlBinding)->refCount();
+
+        // Only pop_front after setting the binding as the bindings are refcounted.
         sharedState->allQPropertyBindings.pop_front();
-        if (auto priv = QPropertyBindingPrivate::get(qmlBinding); priv->hasCustomVTable()) {
-            auto qmlBindingPriv = static_cast<QQmlPropertyBinding *>(priv);
-            auto jsExpression = qmlBindingPriv->jsExpression();
-            const bool canRemove = !qmlBinding.error().hasError() && !qmlBindingPriv->hasDependencies()
-                    && !jsExpression->hasUnresolvedNames();
-            if (canRemove)
-                bindable.takeBinding();
+
+        // If the binding was actually not set, it's deleted now.
+        if (success && bindingPrivateRefCount > 1) {
+            if (auto priv = QPropertyBindingPrivate::get(qmlBinding); priv->hasCustomVTable()) {
+                auto qmlBindingPriv = static_cast<QQmlPropertyBinding *>(priv);
+                auto jsExpression = qmlBindingPriv->jsExpression();
+                const bool canRemove = !qmlBinding.error().hasError()
+                        && !qmlBindingPriv->hasDependencies()
+                        && !jsExpression->hasUnresolvedNames();
+                if (canRemove)
+                    bindable.takeBinding();
+            }
         }
+
         if (watcher.hasRecursed() || interrupt.shouldInterrupt())
             return false;
     }
@@ -1522,7 +1663,7 @@ bool QQmlObjectCreator::populateInstance(int index, QObject *instance, QObject *
 
     _ddata->compilationUnit = compilationUnit;
     if (_compiledObject->hasFlag(QV4::CompiledData::Object::HasDeferredBindings))
-        _ddata->deferData(_compiledObjectIndex, compilationUnit, context);
+        _ddata->deferData(_compiledObjectIndex, compilationUnit, context, m_inlineComponentName);
 
     const qsizetype oldRequiredPropertiesCount = sharedState->requiredProperties.size();
     QSet<QString> postHocRequired;
@@ -1541,7 +1682,7 @@ bool QQmlObjectCreator::populateInstance(int index, QObject *instance, QObject *
             postHocRequired.erase(postHocIt);
         if (isContextObject)
             sharedState->hadTopLevelRequiredProperties = true;
-        sharedState->requiredProperties.insert(propertyData,
+        sharedState->requiredProperties.insert({_qobject, propertyData},
                                                RequiredPropertyInfo {compilationUnit->stringAt(property->nameIndex), compilationUnit->finalUrl(), property->location, {}});
 
     }
@@ -1601,7 +1742,7 @@ bool QQmlObjectCreator::populateInstance(int index, QObject *instance, QObject *
         if (isContextObject)
             sharedState->hadTopLevelRequiredProperties = true;
         sharedState->requiredProperties.insert(
-                propertyData,
+                {_qobject, propertyData},
                 RequiredPropertyInfo {
                         name, compilationUnit->finalUrl(), _compiledObject->location, {} });
     }
@@ -1632,7 +1773,7 @@ bool QQmlObjectCreator::populateInstance(int index, QObject *instance, QObject *
             if (isContextObject)
                 sharedState->hadTopLevelRequiredProperties = true;
             sharedState->requiredProperties.insert(
-                    propertyData,
+                    {_qobject, propertyData},
                     RequiredPropertyInfo {
                             name, compilationUnit->finalUrl(), _compiledObject->location, {} });
         }
@@ -1665,7 +1806,7 @@ bool QQmlObjectCreator::populateInstance(int index, QObject *instance, QObject *
         const QQmlPropertyData *const targetProperty = targetDData->propertyCache->property(coreIndex);
         if (!targetProperty)
             continue;
-        auto it = sharedState->requiredProperties.find(targetProperty);
+        auto it = sharedState->requiredProperties.find({target, targetProperty});
         if (it != sharedState->requiredProperties.end())
             it->aliasesToRequired.push_back(
                     AliasToRequiredInfo {
@@ -1705,3 +1846,15 @@ QQmlObjectCreatorRecursionWatcher::QQmlObjectCreatorRecursionWatcher(QQmlObjectC
     , watcher(creator->sharedState.data())
 {
 }
+
+void ObjectInCreationGCAnchorList::trackObject(QV4::ExecutionEngine *engine, QObject *instance)
+{
+    *allJavaScriptObjects = QV4::QObjectWrapper::wrap(engine, instance);
+    // we have to handle the case where the gc is already running, but the scope is discarded
+    // before the collector runs again. In that case, rescanning won't help us. Thus, mark the
+    // object.
+    QV4::WriteBarrier::markCustom(engine, [this](QV4::MarkStack *ms) {
+        allJavaScriptObjects->heapObject()->mark(ms);
+    });
+    ++allJavaScriptObjects;
+}
