@@ -17,47 +17,44 @@
 QT_BEGIN_NAMESPACE
 
 QQmlTypePrivate::QQmlTypePrivate(QQmlType::RegistrationType type)
-    : regType(type), iid(nullptr), revision(QTypeRevision::zero()),
-    containsRevisionedAttributes(false), baseMetaObject(nullptr),
-    index(-1), isSetup(false), isEnumFromCacheSetup(false), isEnumFromBaseSetup(false),
-    haveSuperType(false)
+    : regType(type)
 {
     switch (type) {
     case QQmlType::CppType:
-        extraData.cd = new QQmlCppTypeData;
-        extraData.cd->allocationSize = 0;
-        extraData.cd->newFunc = nullptr;
-        extraData.cd->createValueTypeFunc = nullptr;
-        extraData.cd->parserStatusCast = -1;
-        extraData.cd->extFunc = nullptr;
-        extraData.cd->extMetaObject = nullptr;
-        extraData.cd->customParser = nullptr;
-        extraData.cd->attachedPropertiesFunc = nullptr;
-        extraData.cd->attachedPropertiesType = nullptr;
-        extraData.cd->propertyValueSourceCast = -1;
-        extraData.cd->propertyValueInterceptorCast = -1;
-        extraData.cd->finalizerCast = -1;
-        extraData.cd->registerEnumClassesUnscoped = true;
-        extraData.cd->registerEnumsFromRelatedTypes = true;
+        extraData.cppTypeData = new QQmlCppTypeData;
+        extraData.cppTypeData->allocationSize = 0;
+        extraData.cppTypeData->newFunc = nullptr;
+        extraData.cppTypeData->createValueTypeFunc = nullptr;
+        extraData.cppTypeData->parserStatusCast = -1;
+        extraData.cppTypeData->extFunc = nullptr;
+        extraData.cppTypeData->extMetaObject = nullptr;
+        extraData.cppTypeData->customParser = nullptr;
+        extraData.cppTypeData->attachedPropertiesFunc = nullptr;
+        extraData.cppTypeData->attachedPropertiesType = nullptr;
+        extraData.cppTypeData->propertyValueSourceCast = -1;
+        extraData.cppTypeData->propertyValueInterceptorCast = -1;
+        extraData.cppTypeData->finalizerCast = -1;
+        extraData.cppTypeData->registerEnumClassesUnscoped = true;
+        extraData.cppTypeData->registerEnumsFromRelatedTypes = true;
         break;
     case QQmlType::SingletonType:
     case QQmlType::CompositeSingletonType:
-        extraData.sd = new QQmlSingletonTypeData;
-        extraData.sd->singletonInstanceInfo = nullptr;
-        extraData.sd->extFunc = nullptr;
-        extraData.sd->extMetaObject = nullptr;
+        extraData.singletonTypeData = new QQmlSingletonTypeData;
+        extraData.singletonTypeData->singletonInstanceInfo = nullptr;
+        extraData.singletonTypeData->extFunc = nullptr;
+        extraData.singletonTypeData->extMetaObject = nullptr;
         break;
     case QQmlType::InterfaceType:
-        extraData.cd = nullptr;
+        extraData.interfaceTypeData = nullptr;
         break;
     case QQmlType::CompositeType:
-        extraData.fd = new QQmlCompositeTypeData;
+        new (&extraData.compositeTypeData) QUrl();
         break;
     case QQmlType::InlineComponentType:
-        extraData.id = new QQmlInlineTypeData;
+        new (&extraData.inlineComponentTypeData) QUrl();
         break;
     case QQmlType::SequentialContainerType:
-        extraData.ld = new QQmlSequenceTypeData;
+        new (&extraData.sequentialContainerTypeData) QMetaSequence();
         break;
     default: qFatal("QQmlTypePrivate Internal Error.");
     }
@@ -65,27 +62,32 @@ QQmlTypePrivate::QQmlTypePrivate(QQmlType::RegistrationType type)
 
 QQmlTypePrivate::~QQmlTypePrivate()
 {
-    qDeleteAll(scopedEnums);
-    for (const auto &metaObject : metaObjects)
-        free(metaObject.metaObject);
+    delete enums.fetchAndStoreAcquire(nullptr);
+    delete proxyMetaObjects.fetchAndStoreAcquire(nullptr);
+
+    if (const QtPrivate::QMetaTypeInterface *iface = typeId.iface()) {
+        if (iface->metaObjectFn == &dynamicQmlMetaObject)
+            QQmlMetaType::unregisterInternalCompositeType(typeId, listId);
+    }
+
     switch (regType) {
     case QQmlType::CppType:
-        delete extraData.cd->customParser;
-        delete extraData.cd;
+        delete extraData.cppTypeData->customParser;
+        delete extraData.cppTypeData;
         break;
     case QQmlType::SingletonType:
     case QQmlType::CompositeSingletonType:
-        delete extraData.sd->singletonInstanceInfo;
-        delete extraData.sd;
+        extraData.singletonTypeData->singletonInstanceInfo.reset();
+        delete extraData.singletonTypeData;
         break;
     case QQmlType::CompositeType:
-        delete extraData.fd;
+        extraData.compositeTypeData.~QUrl();
         break;
     case QQmlType::InlineComponentType:
-        delete  extraData.id;
+        extraData.inlineComponentTypeData.~QUrl();
         break;
     case QQmlType::SequentialContainerType:
-        delete extraData.ld;
+        extraData.sequentialContainerTypeData.~QMetaSequence();
         break;
     default: //Also InterfaceType, because it has no extra data
         break;
@@ -144,22 +146,22 @@ QQmlType QQmlTypePrivate::resolveCompositeBaseType(QQmlEnginePrivate *engine) co
     QQmlRefPointer<QQmlTypeData> td(engine->typeLoader.getType(sourceUrl()));
     if (td.isNull() || !td->isComplete())
         return QQmlType();
-    QV4::ExecutableCompilationUnit *compilationUnit = td->compilationUnit();
+    QV4::CompiledData::CompilationUnit *compilationUnit = td->compilationUnit();
     const QMetaObject *mo = compilationUnit->rootPropertyCache()->firstCppMetaObject();
     return QQmlMetaType::qmlType(mo);
 }
 
 QQmlPropertyCache::ConstPtr QQmlTypePrivate::compositePropertyCache(
-        QQmlEnginePrivate *engine) const
+        QQmlTypeLoader *typeLoader) const
 {
     // similar logic to resolveCompositeBaseType
     Q_ASSERT(isComposite());
-    if (!engine)
+    if (!typeLoader)
         return nullptr;
-    QQmlRefPointer<QQmlTypeData> td(engine->typeLoader.getType(sourceUrl()));
+    QQmlRefPointer<QQmlTypeData> td(typeLoader->getType(sourceUrl()));
     if (td.isNull() || !td->isComplete())
         return nullptr;
-    QV4::ExecutableCompilationUnit *compilationUnit = td->compilationUnit();
+    QV4::CompiledData::CompilationUnit *compilationUnit = td->compilationUnit();
     return compilationUnit->rootPropertyCache();
 }
 
@@ -168,21 +170,29 @@ static bool isPropertyRevisioned(const QMetaObject *mo, int index)
     return mo->property(index).revision();
 }
 
-void QQmlTypePrivate::init() const
+const QQmlTypePrivate::ProxyMetaObjects *QQmlTypePrivate::init() const
 {
-    if (isSetup.loadAcquire())
-        return;
+    if (const ProxyMetaObjects *result = proxyMetaObjects.loadRelaxed())
+        return result;
 
-    QMutexLocker lock(QQmlMetaType::typeRegistrationLock());
-    if (isSetup.loadAcquire())
-        return;
+    ProxyMetaObjects *proxies = new ProxyMetaObjects;
+    auto finalize = [this, proxies]() -> const ProxyMetaObjects *{
+        const ProxyMetaObjects *concurrentModification;
+        if (proxyMetaObjects.testAndSetOrdered(nullptr, proxies, concurrentModification))
+            return proxies;
+
+        delete proxies;
+        return concurrentModification;
+    };
 
     const QMetaObject *mo = baseMetaObject;
     if (!mo) {
         // version 0 singleton type without metaobject information
-        return;
+        return finalize();
     }
 
+    QList<QQmlProxyMetaObject::ProxyData> metaObjects;
+
     auto setupExtendedMetaObject = [&](const QMetaObject *extMetaObject,
                                        QObject *(*extFunc)(QObject *)) {
         if (!extMetaObject)
@@ -200,9 +210,9 @@ void QQmlTypePrivate::init() const
     };
 
     if (regType == QQmlType::SingletonType)
-        setupExtendedMetaObject(extraData.sd->extMetaObject, extraData.sd->extFunc);
+        setupExtendedMetaObject(extraData.singletonTypeData->extMetaObject, extraData.singletonTypeData->extFunc);
     else if (regType == QQmlType::CppType)
-        setupExtendedMetaObject(extraData.cd->extMetaObject, extraData.cd->extFunc);
+        setupExtendedMetaObject(extraData.cppTypeData->extMetaObject, extraData.cppTypeData->extFunc);
 
     metaObjects.append(QQmlMetaType::proxyData(
             mo, baseMetaObject, metaObjects.isEmpty() ? nullptr
@@ -215,6 +225,8 @@ void QQmlTypePrivate::init() const
                 metaObjects.at(ii).metaObject->methodOffset();
     }
 
+    bool containsRevisionedAttributes = false;
+
     // Check for revisioned details
     {
         const QMetaObject *mo = nullptr;
@@ -234,46 +246,54 @@ void QQmlTypePrivate::init() const
         }
     }
 
-    isSetup.storeRelease(true);
-    lock.unlock();
+    proxies->data = std::move(metaObjects);
+    proxies->containsRevisionedAttributes = containsRevisionedAttributes;
+
+    return finalize();
 }
 
-void QQmlTypePrivate::initEnums(QQmlEnginePrivate *engine) const
+const QQmlTypePrivate::Enums *QQmlTypePrivate::initEnums(QQmlTypeLoader *typeLoader) const
 {
-    QQmlPropertyCache::ConstPtr cache = (!isEnumFromCacheSetup.loadAcquire() && isComposite())
-            ? compositePropertyCache(engine)
-            : QQmlPropertyCache::ConstPtr();
+    if (const Enums *result = enums.loadRelaxed())
+        return result;
+
+    QQmlPropertyCache::ConstPtr cache;
+    if (isComposite()) {
+        cache = compositePropertyCache(typeLoader);
+        if (!cache)
+            return nullptr; // Composite type not ready, yet.
+    }
+
+    Enums *newEnums = new Enums;
 
     // beware: It could be a singleton type without metaobject
-    const QMetaObject *metaObject = !isEnumFromBaseSetup.loadAcquire()
-            ? baseMetaObject
-            : nullptr;
 
-    if (!cache && !metaObject)
-        return;
+    if (cache)
+        insertEnumsFromPropertyCache(newEnums, cache);
 
-    init(); // init() can add to the metaObjects list. Therefore, check metaObjects only below
-
-    QMutexLocker lock(QQmlMetaType::typeRegistrationLock());
-
-    if (cache) {
-        insertEnumsFromPropertyCache(cache);
-        isEnumFromCacheSetup.storeRelease(true);
+    if (baseMetaObject) {
+        // init() can add to the metaObjects list. Therefore, check proxies->data only below
+        const ProxyMetaObjects *proxies = init();
+        insertEnums(
+                newEnums,
+                proxies->data.isEmpty() ? baseMetaObject : proxies->data.constFirst().metaObject);
     }
 
-    if (metaObject) {
-        insertEnums(metaObjects.isEmpty() ? baseMetaObject : metaObjects.constFirst().metaObject);
-        isEnumFromBaseSetup.storeRelease(true);
-    }
+    const Enums *concurrentModification;
+    if (enums.testAndSetOrdered(nullptr, newEnums, concurrentModification))
+        return newEnums;
+
+    delete newEnums;
+    return concurrentModification;
 }
 
-void QQmlTypePrivate::insertEnums(const QMetaObject *metaObject) const
+void QQmlTypePrivate::insertEnums(Enums *enums, const QMetaObject *metaObject) const
 {
     // Add any enum values defined by 'related' classes
-    if (regType != QQmlType::CppType || extraData.cd->registerEnumsFromRelatedTypes) {
+    if (regType != QQmlType::CppType || extraData.cppTypeData->registerEnumsFromRelatedTypes) {
         if (const auto *related = metaObject->d.relatedMetaObjects) {
             while (const QMetaObject *relatedMetaObject = *related) {
-                insertEnums(relatedMetaObject);
+                insertEnums(enums, relatedMetaObject);
                 ++related;
             }
         }
@@ -282,6 +302,19 @@ void QQmlTypePrivate::insertEnums(const QMetaObject *metaObject) const
     QSet<QString> localEnums;
     const QMetaObject *localMetaObject = nullptr;
 
+    // ### TODO (QTBUG-123294): track this at instance creation time
+    auto shouldSingletonAlsoRegisterUnscoped = [&](){
+        Q_ASSERT(regType == QQmlType::SingletonType);
+        if (!baseMetaObject)
+            return true;
+        int idx = baseMetaObject->indexOfClassInfo("RegisterEnumClassesUnscoped");
+        if (idx == -1)
+            return true;
+        if (qstrcmp(baseMetaObject->classInfo(idx).value(), "false") == 0)
+            return false;
+        return true;
+    };
+
     // Add any enum values defined by this class, overwriting any inherited values
     for (int ii = 0; ii < metaObject->enumeratorCount(); ++ii) {
         QMetaEnum e = metaObject->enumerator(ii);
@@ -298,29 +331,33 @@ void QQmlTypePrivate::insertEnums(const QMetaObject *metaObject) const
             localEnums.clear();
             localMetaObject = e.enclosingMetaObject();
         }
+        const bool shouldRegisterUnscoped = !isScoped
+                || (regType == QQmlType::CppType && extraData.cppTypeData->registerEnumClassesUnscoped)
+                || (regType == QQmlType::SingletonType && shouldSingletonAlsoRegisterUnscoped())
+        ;
 
         for (int jj = 0; jj < e.keyCount(); ++jj) {
             const QString key = QString::fromUtf8(e.key(jj));
             const int value = e.value(jj);
-            if (!isScoped || (regType == QQmlType::CppType && extraData.cd->registerEnumClassesUnscoped)) {
+            if (shouldRegisterUnscoped) {
                 if (localEnums.contains(key)) {
-                    auto existingEntry = enums.find(key);
-                    if (existingEntry != enums.end() && existingEntry.value() != value) {
+                    auto existingEntry = enums->enums.find(key);
+                    if (existingEntry != enums->enums.end() && existingEntry.value() != value) {
                         qWarning("Previously registered enum will be overwritten due to name clash: %s.%s", metaObject->className(), key.toUtf8().constData());
                         createEnumConflictReport(metaObject, key);
                     }
                 } else {
                     localEnums.insert(key);
                 }
-                enums.insert(key, value);
+                enums->enums.insert(key, value);
             }
             if (isScoped)
                 scoped->insert(key, value);
         }
 
         if (isScoped) {
-            scopedEnums << scoped;
-            scopedEnumIndex.insert(QString::fromUtf8(e.name()), scopedEnums.size()-1);
+            enums->scopedEnums << scoped;
+            enums->scopedEnumIndex.insert(QString::fromUtf8(e.name()), enums->scopedEnums.size()-1);
         }
     }
 }
@@ -379,7 +416,7 @@ void QQmlTypePrivate::createEnumConflictReport(const QMetaObject *metaObject, co
 }
 
 void QQmlTypePrivate::insertEnumsFromPropertyCache(
-        const QQmlPropertyCache::ConstPtr &cache) const
+        Enums *enums, const QQmlPropertyCache::ConstPtr &cache) const
 {
     const QMetaObject *cppMetaObject = cache->firstCppMetaObject();
 
@@ -394,20 +431,14 @@ void QQmlTypePrivate::insertEnumsFromPropertyCache(
 
             for (int jj = 0; jj < enumData->values.size(); ++jj) {
                 const QQmlEnumValue &value = enumData->values.at(jj);
-                enums.insert(value.namedValue, value.value);
+                enums->enums.insert(value.namedValue, value.value);
                 scoped->insert(value.namedValue, value.value);
             }
-            scopedEnums << scoped;
-            scopedEnumIndex.insert(enumData->name, scopedEnums.size()-1);
+            enums->scopedEnums << scoped;
+            enums->scopedEnumIndex.insert(enumData->name, enums->scopedEnums.size()-1);
         }
     }
-    insertEnums(cppMetaObject);
-}
-
-void QQmlTypePrivate::setContainingType(QQmlType *containingType)
-{
-    Q_ASSERT(regType == QQmlType::InlineComponentType);
-    extraData.id->containingType = containingType->d.data();
+    insertEnums(enums, cppMetaObject);
 }
 
 void QQmlTypePrivate::setName(const QString &uri, const QString &element)
@@ -421,11 +452,9 @@ QByteArray QQmlType::typeName() const
 {
     if (d) {
         if (d->regType == SingletonType || d->regType == CompositeSingletonType)
-            return d->extraData.sd->singletonInstanceInfo->typeName.toUtf8();
+            return d->extraData.singletonTypeData->singletonInstanceInfo->typeName;
         else if (d->baseMetaObject)
             return d->baseMetaObject->className();
-        else if (d->regType == InlineComponentType)
-            return d->extraData.id->inlineComponentName.toUtf8();
     }
     return QByteArray();
 }
@@ -473,23 +502,39 @@ QObject *QQmlType::create(void **memory, size_t additionalMemory) const
     if (!d || !isCreatable())
         return nullptr;
 
-    d->init();
-
-    QObject *rv = (QObject *)operator new(d->extraData.cd->allocationSize + additionalMemory);
-    d->extraData.cd->newFunc(rv, d->extraData.cd->userdata);
+    QObject *rv = (QObject *)operator new(d->extraData.cppTypeData->allocationSize + additionalMemory);
+    d->extraData.cppTypeData->newFunc(rv, d->extraData.cppTypeData->userdata);
 
     createProxy(rv);
-    *memory = ((char *)rv) + d->extraData.cd->allocationSize;
+    *memory = ((char *)rv) + d->extraData.cppTypeData->allocationSize;
     return rv;
 }
 
-QQmlType::SingletonInstanceInfo *QQmlType::singletonInstanceInfo() const
+/*!
+    \internal
+    Like create, but also allocates memory behind the object, constructs a QQmlData there
+    and lets the objects declarativeData point to the newly created QQmlData.
+ */
+QObject *QQmlType::createWithQQmlData() const
+{
+    void *ddataMemory = nullptr;
+    auto instance = create(&ddataMemory, sizeof(QQmlData));
+    if (!instance)
+        return nullptr;
+    QObjectPrivate* p = QObjectPrivate::get(instance);
+    Q_ASSERT(!p->isDeletingChildren);
+    if (!p->declarativeData)
+        p->declarativeData = new (ddataMemory) QQmlData(QQmlData::DoesNotOwnMemory);
+    return instance;
+}
+
+QQmlType::SingletonInstanceInfo::ConstPtr QQmlType::singletonInstanceInfo() const
 {
     if (!d)
-        return nullptr;
+        return {};
     if (d->regType != SingletonType && d->regType != CompositeSingletonType)
-        return nullptr;
-    return d->extraData.sd->singletonInstanceInfo;
+        return {};
+    return d->extraData.singletonTypeData->singletonInstanceInfo;
 }
 
 QQmlCustomParser *QQmlType::customParser() const
@@ -498,33 +543,47 @@ QQmlCustomParser *QQmlType::customParser() const
         return nullptr;
     if (d->regType != CppType)
         return nullptr;
-    return d->extraData.cd->customParser;
+    return d->extraData.cppTypeData->customParser;
 }
 
 QQmlType::CreateValueTypeFunc QQmlType::createValueTypeFunction() const
 {
     if (!d || d->regType != CppType)
         return nullptr;
-    return d->extraData.cd->createValueTypeFunc;
+    return d->extraData.cppTypeData->createValueTypeFunc;
+}
+
+bool QQmlType::canConstructValueType() const
+{
+    if (!d || d->regType != CppType)
+        return false;
+    return d->extraData.cppTypeData->constructValueType;
+}
+
+bool QQmlType::canPopulateValueType() const
+{
+    if (!d || d->regType != CppType)
+        return false;
+    return d->extraData.cppTypeData->populateValueType;
 }
 
 QQmlType::CreateFunc QQmlType::createFunction() const
 {
     if (!d || d->regType != CppType)
         return nullptr;
-    return d->extraData.cd->newFunc;
+    return d->extraData.cppTypeData->newFunc;
 }
 
 QString QQmlType::noCreationReason() const
 {
     if (!d || d->regType != CppType)
         return QString();
-    return d->extraData.cd->noCreationReason;
+    return d->extraData.cppTypeData->noCreationReason;
 }
 
 bool QQmlType::isCreatable() const
 {
-    return d && d->regType == CppType && d->extraData.cd->newFunc;
+    return d && d->regType == CppType && d->extraData.cppTypeData->newFunc;
 }
 
 QQmlType::ExtensionFunc QQmlType::extensionFunction() const
@@ -534,9 +593,9 @@ QQmlType::ExtensionFunc QQmlType::extensionFunction() const
 
     switch (d->regType) {
     case CppType:
-        return d->extraData.cd->extFunc;
+        return d->extraData.cppTypeData->extFunc;
     case SingletonType:
-        return d->extraData.sd->extFunc;
+        return d->extraData.singletonTypeData->extFunc;
     default:
         return nullptr;
     }
@@ -549,9 +608,9 @@ const QMetaObject *QQmlType::extensionMetaObject() const
 
     switch (d->regType) {
     case CppType:
-        return d->extraData.cd->extMetaObject;
+        return d->extraData.cppTypeData->extMetaObject;
     case SingletonType:
-        return d->extraData.sd->extMetaObject;
+        return d->extraData.singletonTypeData->extMetaObject;
     default:
         return nullptr;
     }
@@ -559,11 +618,7 @@ const QMetaObject *QQmlType::extensionMetaObject() const
 
 bool QQmlType::isExtendedType() const
 {
-    if (!d)
-        return false;
-    d->init();
-
-    return !d->metaObjects.isEmpty();
+    return d && !d->init()->data.isEmpty();
 }
 
 bool QQmlType::isSingleton() const
@@ -592,12 +647,12 @@ bool QQmlType::isCompositeSingleton() const
 
 bool QQmlType::isQObjectSingleton() const
 {
-    return d && d->regType == SingletonType && d->extraData.sd->singletonInstanceInfo->qobjectCallback;
+    return d && d->regType == SingletonType && d->extraData.singletonTypeData->singletonInstanceInfo->qobjectCallback;
 }
 
 bool QQmlType::isQJSValueSingleton() const
 {
-    return d && d->regType == SingletonType && d->extraData.sd->singletonInstanceInfo->scriptCallback;
+    return d && d->regType == SingletonType && d->extraData.singletonTypeData->singletonInstanceInfo->scriptCallback;
 }
 
 bool QQmlType::isSequentialContainer() const
@@ -605,6 +660,11 @@ bool QQmlType::isSequentialContainer() const
     return d && d->regType == SequentialContainerType;
 }
 
+bool QQmlType::isValueType() const
+{
+    return d && d->isValueType();
+}
+
 QMetaType QQmlType::typeId() const
 {
     return d ? d->typeId : QMetaType{};
@@ -617,20 +677,18 @@ QMetaType QQmlType::qListTypeId() const
 
 QMetaSequence QQmlType::listMetaSequence() const
 {
-    return isSequentialContainer() ? *d->extraData.ld : QMetaSequence();
+    return isSequentialContainer() ? d->extraData.sequentialContainerTypeData : QMetaSequence();
 }
 
 const QMetaObject *QQmlType::metaObject() const
 {
-    if (!d)
-        return nullptr;
-    d->init();
-
-    if (d->metaObjects.isEmpty())
-        return d->baseMetaObject;
-    else
-        return d->metaObjects.constFirst().metaObject;
+    return d ? d->metaObject() : nullptr;
+}
 
+const QMetaObject *QQmlType::metaObjectForValueType() const
+{
+    Q_ASSERT(d);
+    return d->metaObjectForValueType();
 }
 
 const QMetaObject *QQmlType::baseMetaObject() const
@@ -640,11 +698,7 @@ const QMetaObject *QQmlType::baseMetaObject() const
 
 bool QQmlType::containsRevisionedAttributes() const
 {
-    if (!d)
-        return false;
-    d->init();
-
-    return d->containsRevisionedAttributes;
+    return d && d->init()->containsRevisionedAttributes;
 }
 
 QTypeRevision QQmlType::metaObjectRevision() const
@@ -655,14 +709,14 @@ QTypeRevision QQmlType::metaObjectRevision() const
 QQmlAttachedPropertiesFunc QQmlType::attachedPropertiesFunction(QQmlEnginePrivate *engine) const
 {
     if (const QQmlTypePrivate *base = d ? d->attachedPropertiesBase(engine) : nullptr)
-        return base->extraData.cd->attachedPropertiesFunc;
+        return base->extraData.cppTypeData->attachedPropertiesFunc;
     return nullptr;
 }
 
 const QMetaObject *QQmlType::attachedPropertiesType(QQmlEnginePrivate *engine) const
 {
     if (const QQmlTypePrivate *base = d ? d->attachedPropertiesBase(engine) : nullptr)
-        return base->extraData.cd->attachedPropertiesType;
+        return base->extraData.cppTypeData->attachedPropertiesType;
     return nullptr;
 }
 
@@ -670,35 +724,35 @@ int QQmlType::parserStatusCast() const
 {
     if (!d || d->regType != CppType)
         return -1;
-    return d->extraData.cd->parserStatusCast;
+    return d->extraData.cppTypeData->parserStatusCast;
 }
 
 int QQmlType::propertyValueSourceCast() const
 {
     if (!d || d->regType != CppType)
         return -1;
-    return d->extraData.cd->propertyValueSourceCast;
+    return d->extraData.cppTypeData->propertyValueSourceCast;
 }
 
 int QQmlType::propertyValueInterceptorCast() const
 {
     if (!d || d->regType != CppType)
         return -1;
-    return d->extraData.cd->propertyValueInterceptorCast;
+    return d->extraData.cppTypeData->propertyValueInterceptorCast;
 }
 
 int QQmlType::finalizerCast() const
 {
     if (!d || d->regType != CppType)
         return -1;
-    return d->extraData.cd->finalizerCast;
+    return d->extraData.cppTypeData->finalizerCast;
 }
 
 const char *QQmlType::interfaceIId() const
 {
     if (!d || d->regType != InterfaceType)
         return nullptr;
-    return d->iid;
+    return d->extraData.interfaceTypeData;
 }
 
 int QQmlType::index() const
@@ -710,200 +764,49 @@ bool QQmlType::isInlineComponentType() const {
     return d ? d->regType == QQmlType::InlineComponentType : false;
 }
 
-int QQmlType::inlineComponentId() const {
-    bool ok = false;
-    if (d->regType == QQmlType::RegistrationType::InlineComponentType) {
-        Q_ASSERT(d->extraData.id->objectId != -1);
-        return d->extraData.id->objectId;
-    }
-    int subObjectId = sourceUrl().fragment().toInt(&ok);
-    return ok ? subObjectId : -1;
-}
-
 QUrl QQmlType::sourceUrl() const
 {
-    auto url = d ? d->sourceUrl() : QUrl();
-    if (url.isValid() && d->regType == QQmlType::RegistrationType::InlineComponentType && d->extraData.id->objectId) {
-        Q_ASSERT(url.hasFragment());
-        url.setFragment(QString::number(inlineComponentId()));
-    }
-    return url;
+    return d ? d->sourceUrl() : QUrl();
 }
 
-int QQmlType::enumValue(QQmlEnginePrivate *engine, const QHashedStringRef &name, bool *ok) const
+int QQmlType::enumValue(QQmlTypeLoader *typeLoader, const QHashedStringRef &name, bool *ok) const
 {
-    Q_ASSERT(ok);
-    if (d) {
-        *ok = true;
-
-        d->initEnums(engine);
-
-        int *rv = d->enums.value(name);
-        if (rv)
-            return *rv;
-    }
-
-    *ok = false;
-    return -1;
+    return QQmlTypePrivate::enumValue(d, typeLoader, name, ok);
 }
 
-int QQmlType::enumValue(QQmlEnginePrivate *engine, const QHashedCStringRef &name, bool *ok) const
+int QQmlType::enumValue(QQmlTypeLoader *typeLoader, const QHashedCStringRef &name, bool *ok) const
 {
-    Q_ASSERT(ok);
-    if (d) {
-        *ok = true;
-
-        d->initEnums(engine);
-
-        int *rv = d->enums.value(name);
-        if (rv)
-            return *rv;
-    }
-
-    *ok = false;
-    return -1;
+    return QQmlTypePrivate::enumValue(d, typeLoader, name, ok);
 }
 
-int QQmlType::enumValue(QQmlEnginePrivate *engine, const QV4::String *name, bool *ok) const
+int QQmlType::enumValue(QQmlTypeLoader *typeLoader, const QV4::String *name, bool *ok) const
 {
-    Q_ASSERT(ok);
-    if (d) {
-        *ok = true;
-
-        d->initEnums(engine);
-
-        int *rv = d->enums.value(name);
-        if (rv)
-            return *rv;
-    }
-
-    *ok = false;
-    return -1;
+    return QQmlTypePrivate::enumValue(d, typeLoader, name, ok);
 }
 
-int QQmlType::scopedEnumIndex(QQmlEnginePrivate *engine, const QV4::String *name, bool *ok) const
+int QQmlType::scopedEnumIndex(QQmlTypeLoader *typeLoader, const QV4::String *name, bool *ok) const
 {
-    Q_ASSERT(ok);
-    if (d) {
-        *ok = true;
-
-        d->initEnums(engine);
-
-        int *rv = d->scopedEnumIndex.value(name);
-        if (rv)
-            return *rv;
-    }
-
-    *ok = false;
-    return -1;
+    return QQmlTypePrivate::scopedEnumIndex(d, typeLoader, name, ok);
 }
 
-int QQmlType::scopedEnumIndex(QQmlEnginePrivate *engine, const QString &name, bool *ok) const
+int QQmlType::scopedEnumIndex(QQmlTypeLoader *typeLoader, const QString &name, bool *ok) const
 {
-    Q_ASSERT(ok);
-    if (d) {
-        *ok = true;
-
-        d->initEnums(engine);
-
-        int *rv = d->scopedEnumIndex.value(name);
-        if (rv)
-            return *rv;
-    }
-
-    *ok = false;
-    return -1;
+    return QQmlTypePrivate::scopedEnumIndex(d, typeLoader, name, ok);
 }
 
-int QQmlType::scopedEnumValue(QQmlEnginePrivate *engine, int index, const QV4::String *name, bool *ok) const
+int QQmlType::scopedEnumValue(QQmlTypeLoader *typeLoader, int index, const QV4::String *name, bool *ok) const
 {
-    Q_UNUSED(engine);
-    Q_ASSERT(ok);
-    *ok = true;
-
-    if (d) {
-        Q_ASSERT(index > -1 && index < d->scopedEnums.size());
-        int *rv = d->scopedEnums.at(index)->value(name);
-        if (rv)
-            return *rv;
-    }
-
-    *ok = false;
-    return -1;
+    return QQmlTypePrivate::scopedEnumValue(d, typeLoader, index, name, ok);
 }
 
-int QQmlType::scopedEnumValue(QQmlEnginePrivate *engine, int index, const QString &name, bool *ok) const
+int QQmlType::scopedEnumValue(QQmlTypeLoader *typeLoader, int index, const QString &name, bool *ok) const
 {
-    Q_UNUSED(engine);
-    Q_ASSERT(ok);
-    *ok = true;
-
-    if (d) {
-        Q_ASSERT(index > -1 && index < d->scopedEnums.size());
-        int *rv = d->scopedEnums.at(index)->value(name);
-        if (rv)
-            return *rv;
-    }
-
-    *ok = false;
-    return -1;
+    return QQmlTypePrivate::scopedEnumValue(d, typeLoader, index, name, ok);
 }
 
-int QQmlType::scopedEnumValue(QQmlEnginePrivate *engine, const QByteArray &scopedEnumName, const QByteArray &name, bool *ok) const
+int QQmlType::scopedEnumValue(QQmlTypeLoader *typeLoader, const QHashedStringRef &scopedEnumName, const QHashedStringRef &name, bool *ok) const
 {
-    Q_ASSERT(ok);
-    if (d) {
-        *ok = true;
-
-        d->initEnums(engine);
-
-        int *rv = d->scopedEnumIndex.value(QHashedCStringRef(scopedEnumName.constData(), scopedEnumName.size()));
-        if (rv) {
-            int index = *rv;
-            Q_ASSERT(index > -1 && index < d->scopedEnums.size());
-            rv = d->scopedEnums.at(index)->value(QHashedCStringRef(name.constData(), name.size()));
-            if (rv)
-                return *rv;
-        }
-    }
-
-    *ok = false;
-    return -1;
-}
-
-int QQmlType::scopedEnumValue(QQmlEnginePrivate *engine, QStringView scopedEnumName, QStringView name, bool *ok) const
-{
-    Q_ASSERT(ok);
-    if (d) {
-        *ok = true;
-
-        d->initEnums(engine);
-
-        int *rv = d->scopedEnumIndex.value(QHashedStringRef(scopedEnumName));
-        if (rv) {
-            int index = *rv;
-            Q_ASSERT(index > -1 && index < d->scopedEnums.size());
-            rv = d->scopedEnums.at(index)->value(QHashedStringRef(name));
-            if (rv)
-                return *rv;
-        }
-    }
-
-    *ok = false;
-    return -1;
-}
-
-int QQmlType::inlineComponentObjectId() const
-{
-    if (!isInlineComponentType())
-        return -1;
-    return d->extraData.id->objectId;
-}
-
-void QQmlType::setInlineComponentObjectId(int id) const
-{
-    Q_ASSERT(d && d->regType == QQmlType::InlineComponentType);
-    d->extraData.id->objectId = id;
+    return QQmlTypePrivate::scopedEnumValue(d, typeLoader, scopedEnumName, name, ok);
 }
 
 void QQmlType::refHandle(const QQmlTypePrivate *priv)
@@ -925,72 +828,11 @@ int QQmlType::refCount(const QQmlTypePrivate *priv)
     return -1;
 }
 
-int QQmlType::lookupInlineComponentIdByName(const QString &name) const
-{
-    Q_ASSERT(d);
-    return d->namesToInlineComponentObjectIndex.value(name, -1);
-}
-
-QQmlType QQmlType::containingType() const
-{
-    Q_ASSERT(d && d->regType == QQmlType::RegistrationType::InlineComponentType);
-    auto ret = QQmlType {d->extraData.id->containingType};
-    Q_ASSERT(!ret.isInlineComponentType());
-    return ret;
-}
-
-QQmlType QQmlType::lookupInlineComponentById(int objectid) const
-{
-    Q_ASSERT(d);
-    return d->objectIdToICType.value(objectid, QQmlType(nullptr));
-}
-
-int QQmlType::generatePlaceHolderICId() const
-{
-    Q_ASSERT(d);
-    int id = -2;
-    for (auto it = d->objectIdToICType.keyBegin(); it != d->objectIdToICType.keyEnd(); ++it)
-        if (*it < id)
-                id = *it;
-    return id;
-}
-
-void QQmlType::associateInlineComponent(const QString &name, int objectID, const CompositeMetaTypeIds &metaTypeIds, QQmlType existingType)
-{
-    bool const reuseExistingType = existingType.isValid();
-    auto priv = reuseExistingType ? const_cast<QQmlTypePrivate *>(existingType.d.data()) : new QQmlTypePrivate { RegistrationType::InlineComponentType } ;
-    priv->setName( QString::fromUtf8(typeName()), name);
-    auto icUrl = QUrl(sourceUrl());
-    icUrl.setFragment(QString::number(objectID));
-    priv->extraData.id->url = icUrl;
-    priv->extraData.id->containingType = d.data();
-    priv->extraData.id->objectId = objectID;
-    priv->typeId = metaTypeIds.id;
-    priv->listId = metaTypeIds.listId;
-    d->namesToInlineComponentObjectIndex.insert(name, objectID);
-    QQmlType icType(priv);
-    d->objectIdToICType.insert(objectID, icType);
-    if (!reuseExistingType)
-        priv->release();
-}
-
-void QQmlType::setPendingResolutionName(const QString &name)
-{
-    Q_ASSERT(d && d->regType == QQmlType::RegistrationType::InlineComponentType);
-    Q_ASSERT(d->extraData.id->inlineComponentName == name|| d->extraData.id->inlineComponentName.isEmpty());
-    d->extraData.id->inlineComponentName = name;
-}
-
-QString QQmlType::pendingResolutionName() const
-{
-    Q_ASSERT(d && d->regType == QQmlType::RegistrationType::InlineComponentType);
-    return d->extraData.id->inlineComponentName;
-}
-
 void QQmlType::createProxy(QObject *instance) const
 {
-    if (!d->metaObjects.isEmpty())
-        (void)new QQmlProxyMetaObject(instance, &d->metaObjects);
+    const QQmlTypePrivate::ProxyMetaObjects *proxies = d->init();
+    if (!proxies->data.isEmpty())
+        (void)new QQmlProxyMetaObject(instance, &proxies->data);
 }
 
 QT_END_NAMESPACE
