@@ -3,114 +3,127 @@
 
 #include "qqmlvaluetypewrapper_p.h"
 
-#include <private/qqmlvaluetype_p.h>
 #include <private/qqmlbinding_p.h>
-#include <private/qqmlglobal_p.h>
 #include <private/qqmlbuiltinfunctions_p.h>
+#include <private/qqmlvaluetype_p.h>
 
+#include <private/qv4alloca_p.h>
+#include <private/qv4arraybuffer_p.h>
+#include <private/qv4dateobject_p.h>
 #include <private/qv4engine_p.h>
 #include <private/qv4functionobject_p.h>
-#include <private/qv4variantobject_p.h>
-#include <private/qv4alloca_p.h>
-#include <private/qv4stackframe_p.h>
-#include <private/qv4objectiterator_p.h>
-#include <private/qv4qobjectwrapper_p.h>
 #include <private/qv4identifiertable_p.h>
+#include <private/qv4jsonobject_p.h>
 #include <private/qv4lookup_p.h>
+#include <private/qv4qobjectwrapper_p.h>
+#include <private/qv4stackframe_p.h>
+#include <private/qv4variantobject_p.h>
+
+#include <QtCore/qline.h>
+#include <QtCore/qsize.h>
+#include <QtCore/qdatetime.h>
 #include <QtCore/qloggingcategory.h>
-#include <QtCore/QLine>
-#include <QtCore/QLineF>
-#include <QtCore/QSize>
-#include <QtCore/QSizeF>
+
+#if QT_CONFIG(regularexpression)
+#include <private/qv4regexpobject_p.h>
+#endif
 
 QT_BEGIN_NAMESPACE
 
-Q_DECLARE_LOGGING_CATEGORY(lcBindingRemoval)
+Q_DECLARE_LOGGING_CATEGORY(lcBuiltinsBindingRemoval)
 
 DEFINE_OBJECT_VTABLE(QV4::QQmlValueTypeWrapper);
-DEFINE_OBJECT_VTABLE(QV4::QQmlValueTypeReference);
 
-using namespace QV4;
+namespace QV4 {
+
+Heap::QQmlValueTypeWrapper *Heap::QQmlValueTypeWrapper::detached() const
+{
+    return internalClass->engine->memoryManager->allocate<QV4::QQmlValueTypeWrapper>(
+        m_gadgetPtr, QMetaType(m_metaType), m_metaObject, nullptr, -1, NoFlag);
+}
 
 void Heap::QQmlValueTypeWrapper::destroy()
 {
     if (m_gadgetPtr) {
-        m_valueType->metaType.destruct(m_gadgetPtr);
+        metaType().destruct(m_gadgetPtr);
         ::operator delete(m_gadgetPtr);
     }
-    Object::destroy();
-}
-
-void Heap::QQmlValueTypeWrapper::setData(const void *data) const
-{
-    if (auto *gadget = gadgetPtr())
-        valueType()->metaType.destruct(gadget);
-    if (!gadgetPtr())
-        setGadgetPtr(::operator new(valueType()->metaType.sizeOf()));
-    valueType()->metaType.construct(gadgetPtr(), data);
-}
-
-void Heap::QQmlValueTypeWrapper::setValue(const QVariant &value) const
-{
-    Q_ASSERT(valueType()->metaType.id() == value.userType());
-    setData(value.constData());
+    ReferenceObject::destroy();
 }
 
 QVariant Heap::QQmlValueTypeWrapper::toVariant() const
 {
     Q_ASSERT(gadgetPtr());
-    return QVariant(valueType()->metaType, gadgetPtr());
+    return QVariant(metaType(), gadgetPtr());
 }
 
-
-bool QQmlValueTypeReference::readReferenceValue() const
+bool Heap::QQmlValueTypeWrapper::setVariant(const QVariant &variant)
 {
-    if (!d()->object)
-        return false;
-    // A reference resource may be either a "true" reference (eg, to a QVector3D property)
-    // or a "variant" reference (eg, to a QVariant property which happens to contain a value-type).
-    QMetaProperty writebackProperty = d()->object->metaObject()->property(d()->property);
-    if (writebackProperty.userType() == QMetaType::QVariant) {
-        // variant-containing-value-type reference
-        QVariant variantReferenceValue;
+    Q_ASSERT(isVariant());
 
-        void *a[] = { &variantReferenceValue, nullptr };
-        QMetaObject::metacall(d()->object, QMetaObject::ReadProperty, d()->property, a);
-
-        const QMetaType variantReferenceType = variantReferenceValue.metaType();
-        if (variantReferenceType != type()) {
-            // This is a stale VariantReference.  That is, the variant has been
-            // overwritten with a different type in the meantime.
-            // We need to modify this reference to the updated value type, if
-            // possible, or return false if it is not a value type.
-            if (QQmlMetaType::isValueType(variantReferenceType)) {
-                const QMetaObject *mo = QQmlMetaType::metaObjectForValueType(variantReferenceType);
-                if (d()->gadgetPtr()) {
-                    d()->valueType()->metaType.destruct(d()->gadgetPtr());
-                    ::operator delete(d()->gadgetPtr());
-                }
-                d()->setGadgetPtr(nullptr);
-                d()->setMetaObject(mo);
-                d()->setValueType(QQmlMetaType::valueType(variantReferenceType));
-                if (!mo)
-                    return false;
-            } else {
-                return false;
+    const QMetaType variantReferenceType = variant.metaType();
+    if (variantReferenceType != metaType()) {
+        // This is a stale VariantReference.  That is, the variant has been
+        // overwritten with a different type in the meantime.
+        // We need to modify this reference to the updated value type, if
+        // possible, or return false if it is not a value type.
+        if (QQmlMetaType::isValueType(variantReferenceType)) {
+            const QMetaObject *mo = QQmlMetaType::metaObjectForValueType(variantReferenceType);
+            if (gadgetPtr()) {
+                metaType().destruct(gadgetPtr());
+                ::operator delete(gadgetPtr());
             }
+            setGadgetPtr(nullptr);
+            setMetaObject(mo);
+            setMetaType(variantReferenceType);
+            if (!mo)
+                return false;
+        } else {
+            return false;
         }
-        d()->setValue(variantReferenceValue);
-    } else {
-        if (!d()->gadgetPtr()) {
-            d()->setGadgetPtr(::operator new(d()->valueType()->metaType.sizeOf()));
-            d()->valueType()->metaType.construct(d()->gadgetPtr(), nullptr);
-        }
-        // value-type reference
-        void *args[] = { d()->gadgetPtr(), nullptr };
-        QMetaObject::metacall(d()->object, QMetaObject::ReadProperty, d()->property, args);
     }
+
+    setData(variant.constData());
     return true;
 }
 
+void *Heap::QQmlValueTypeWrapper::storagePointer()
+{
+    if (!gadgetPtr()) {
+        setGadgetPtr(::operator new(metaType().sizeOf()));
+        metaType().construct(gadgetPtr(), nullptr);
+    }
+    return gadgetPtr();
+}
+
+bool Heap::QQmlValueTypeWrapper::readReference()
+{
+    // If locations are enforced we only read once
+    return enforcesLocation() || QV4::ReferenceObject::readReference(this);
+}
+
+bool Heap::QQmlValueTypeWrapper::writeBack(int propertyIndex)
+{
+    return isAttachedToProperty() && QV4::ReferenceObject::writeBack(this, propertyIndex);
+}
+
+ReturnedValue QQmlValueTypeWrapper::create(
+        ExecutionEngine *engine, Heap::QQmlValueTypeWrapper *cloneFrom, Heap::Object *object)
+{
+    QV4::Scope scope(engine);
+    initProto(engine);
+
+    // Either we're enforcing the location, then we have to read right away.
+    // Or we don't then we lazy-load. In neither case we pass any data.
+    Scoped<QQmlValueTypeWrapper> r(scope, engine->memoryManager->allocate<QQmlValueTypeWrapper>(
+                                       nullptr, cloneFrom->metaType(), cloneFrom->metaObject(),
+                                       object, cloneFrom->property(), cloneFrom->flags()));
+    r->d()->setLocation(cloneFrom->function(), cloneFrom->statementIndex());
+    if (cloneFrom->enforcesLocation())
+        QV4::ReferenceObject::readReference(r->d());
+    return r->asReturnedValue();
+}
+
 void QQmlValueTypeWrapper::initProto(ExecutionEngine *v4)
 {
     if (v4->valueTypeWrapperPrototype()->d_unchecked())
@@ -122,31 +135,70 @@ void QQmlValueTypeWrapper::initProto(ExecutionEngine *v4)
     v4->jsObjects[QV4::ExecutionEngine::ValueTypeProto] = o->d();
 }
 
-ReturnedValue QQmlValueTypeWrapper::create(ExecutionEngine *engine, QObject *object, int property, const QMetaObject *metaObject, QMetaType type)
+int QQmlValueTypeWrapper::virtualMetacall(
+        Object *object, QMetaObject::Call call, int index, void **a)
+{
+    QQmlValueTypeWrapper *wrapper = object->as<QQmlValueTypeWrapper>();
+    Q_ASSERT(wrapper);
+
+    switch (call) {
+    case QMetaObject::InvokeMetaMethod:
+    case QMetaObject::ReadProperty:
+    case QMetaObject::BindableProperty:
+    case QMetaObject::CustomCall:
+        if (wrapper->d()->object())
+            wrapper->d()->readReference();
+        break;
+    default:
+        break;
+    }
+
+    const QMetaObject *mo = wrapper->d()->metaObject();
+    if (!mo->d.static_metacall)
+        return 0;
+
+    mo->d.static_metacall(static_cast<QObject *>(wrapper->d()->gadgetPtr()), call, index, a);
+
+    switch (call) {
+    case QMetaObject::ReadProperty:
+        break;
+    case QMetaObject::WriteProperty:
+    case QMetaObject::ResetProperty:
+        if (wrapper->d()->object())
+            wrapper->d()->writeBack(index);
+        break;
+    case QMetaObject::InvokeMetaMethod:
+    case QMetaObject::CustomCall:
+        if (wrapper->d()->object())
+            wrapper->d()->writeBack();
+        break;
+    default:
+        break;
+    }
+
+    return -1;
+}
+
+ReturnedValue QQmlValueTypeWrapper::create(
+        ExecutionEngine *engine, const void *data, const QMetaObject *metaObject, QMetaType type,
+        Heap::Object *object, int property, Heap::ReferenceObject::Flags flags)
 {
     Scope scope(engine);
     initProto(engine);
 
-    Scoped<QQmlValueTypeReference> r(scope, engine->memoryManager->allocate<QQmlValueTypeReference>());
-    r->d()->object = object;
-    r->d()->property = property;
-    r->d()->setMetaObject(metaObject);
-    auto valueType = QQmlMetaType::valueType(type);
-    if (!valueType) {
+    if (!type.isValid()) {
         return engine->throwTypeError(QLatin1String("Type %1 is not a value type")
                                       .arg(QString::fromUtf8(type.name())));
     }
-    r->d()->setValueType(valueType);
-    r->d()->setGadgetPtr(nullptr);
-    return r->asReturnedValue();
-}
 
-ReturnedValue QQmlValueTypeWrapper::create(
-        ExecutionEngine *engine, const QVariant &value, const QMetaObject *metaObject,
-        QMetaType type)
-{
-    Q_ASSERT(value.metaType() == QQmlMetaType::valueType(type)->metaType);
-    return create(engine, value.constData(), metaObject, type);
+    // If data is given explicitly, we assume it has just been read from the property
+    Scoped<QQmlValueTypeWrapper> r(scope, engine->memoryManager->allocate<QQmlValueTypeWrapper>(
+                                       data, type, metaObject, object, property, flags));
+    if (CppStackFrame *frame = engine->currentStackFrame)
+        r->d()->setLocation(frame->v4Function, frame->statementNumber());
+    if (!data && r->d()->enforcesLocation())
+        QV4::ReferenceObject::readReference(r->d());
+    return r->asReturnedValue();
 }
 
 ReturnedValue QQmlValueTypeWrapper::create(
@@ -155,33 +207,29 @@ ReturnedValue QQmlValueTypeWrapper::create(
     Scope scope(engine);
     initProto(engine);
 
-    Scoped<QQmlValueTypeWrapper> r(scope, engine->memoryManager->allocate<QQmlValueTypeWrapper>());
-    r->d()->setMetaObject(metaObject);
-    auto valueType = QQmlMetaType::valueType(type);
-    if (!valueType) {
+    if (!type.isValid()) {
         return engine->throwTypeError(QLatin1String("Type %1 is not a value type")
                                       .arg(QString::fromUtf8(type.name())));
     }
-    r->d()->setValueType(valueType);
-    r->d()->setGadgetPtr(nullptr);
-    r->d()->setData(data);
+
+    Scoped<QQmlValueTypeWrapper> r(
+                scope, engine->memoryManager->allocate<QQmlValueTypeWrapper>(
+                    data, type, metaObject, nullptr,  -1, Heap::ReferenceObject::NoFlag));
     return r->asReturnedValue();
 }
 
 QVariant QQmlValueTypeWrapper::toVariant() const
 {
-    if (const QQmlValueTypeReference *ref = as<const QQmlValueTypeReference>())
-        if (!ref->readReferenceValue())
-            return QVariant();
+    if (d()->isReference() && !readReferenceValue())
+        return QVariant();
     return d()->toVariant();
 }
 
 bool QQmlValueTypeWrapper::toGadget(void *data) const
 {
-    if (const QQmlValueTypeReference *ref = as<const QQmlValueTypeReference>())
-        if (!ref->readReferenceValue())
-            return false;
-    const QMetaType type = d()->valueType()->metaType;
+    if (d()->isReference() && !readReferenceValue())
+        return false;
+    const QMetaType type = d()->metaType();
     type.destruct(data);
     type.construct(data, d()->gadgetPtr());
     return true;
@@ -201,43 +249,164 @@ bool QQmlValueTypeWrapper::virtualIsEqualTo(Managed *m, Managed *other)
     return false;
 }
 
+bool QQmlValueTypeWrapper::virtualHasProperty(const Managed *m, PropertyKey id)
+{
+    if (!id.isString())
+        return Object::virtualHasProperty(m, id);
+    Q_ASSERT(m && m->as<QQmlValueTypeWrapper>());
+    auto wrapper = static_cast<const QQmlValueTypeWrapper *>(m);
+    if (auto mo = wrapper->d()->metaObject())
+        if (mo->indexOfProperty(id.toQString().toUtf8()) != -1)
+            return true;
+
+    /* we don't want to fallback to QObject::virtualHasProperty
+       as that would end up calling getOwnProperty which is wasteful,
+       as it calls our own virtualGetOwnProperty.
+       As we know that our own properties are only those found on the meta-object,
+       we can instead skip the call, and simply check whether the property exists
+       on the prototype.
+    */
+    Scope scope(m->engine());
+    ScopedObject o(scope, m);
+    o = o->getPrototypeOf();
+    if (o)
+        return o->hasProperty(id);
+
+    return false;
+}
+
+static Heap::ReferenceObject::Flags referenceFlags(const QMetaObject *metaObject, int index)
+{
+    return metaObject->property(index).isWritable()
+            ? (Heap::ReferenceObject::CanWriteBack | Heap::ReferenceObject::EnforcesLocation)
+            : Heap::ReferenceObject::EnforcesLocation;
+}
+
+static void doStaticReadCall(
+        const QMetaObject *metaObject, Heap::QQmlValueTypeWrapper *valueTypeWrapper,
+        int index, void **args)
+{
+    metaObject->d.static_metacall(
+                reinterpret_cast<QObject*>(
+                    valueTypeWrapper->gadgetPtr()), QMetaObject::ReadProperty, index, args);
+}
+
 static ReturnedValue getGadgetProperty(ExecutionEngine *engine,
                                        Heap::QQmlValueTypeWrapper *valueTypeWrapper,
                                        QMetaType metaType, quint16 coreIndex, bool isFunction, bool isEnum)
 {
     if (isFunction) {
         // calling a Q_INVOKABLE function of a value type
-        return QV4::QObjectMethod::create(engine->rootContext(), valueTypeWrapper, coreIndex);
-    }
-#define VALUE_TYPE_LOAD(metatype, cpptype, constructor) \
-    if (metaTypeId == metatype) { \
-        cpptype v; \
-        void *args[] = { &v, nullptr }; \
-        metaObject->d.static_metacall(reinterpret_cast<QObject*>(valueTypeWrapper->gadgetPtr()), \
-                                      QMetaObject::ReadProperty, index, args); \
-        return QV4::Encode(constructor(v)); \
+        return QV4::QObjectMethod::create(engine, valueTypeWrapper, coreIndex);
     }
+
     const QMetaObject *metaObject = valueTypeWrapper->metaObject();
     int index = coreIndex;
-    QQmlMetaObject::resolveGadgetMethodOrPropertyIndex(QMetaObject::ReadProperty, &metaObject, &index);
-    // These four types are the most common used by the value type wrappers
-    int metaTypeId = metaType.id();
-    VALUE_TYPE_LOAD(QMetaType::Double, double, double);
-    VALUE_TYPE_LOAD(QMetaType::Float, float, float);
-    VALUE_TYPE_LOAD(QMetaType::Int || isEnum, int, int);
-    VALUE_TYPE_LOAD(QMetaType::QString, QString, engine->newString);
-    VALUE_TYPE_LOAD(QMetaType::Bool, bool, bool);
-    QVariant v;
-    void *args[] = { nullptr, nullptr };
-    if (metaType == QMetaType::fromType<QVariant>()) {
-        args[0] = &v;
-    } else {
-        v = QVariant(metaType, static_cast<void *>(nullptr));
-        args[0] = v.data();
+
+    const auto wrapChar16 = [engine](char16_t c) {
+        return engine->newString(QChar(c));
+    };
+    const auto wrapQObject = [engine](QObject *object) {
+        return QObjectWrapper::wrap(engine, object);
+    };
+    const auto wrapJsonValue = [engine](const QJsonValue &value) {
+        return JsonObject::fromJsonValue(engine, value);
+    };
+    const auto wrapJsonObject = [engine](const QJsonObject &object) {
+        return JsonObject::fromJsonObject(engine, object);
+    };
+    const auto wrapJsonArray = [engine](const QJsonArray &array) {
+        return JsonObject::fromJsonArray(engine, array);
+    };
+
+    const auto wrapQDateTime = [&](const QDateTime &dateTime) {
+        return engine->newDateObject(
+                    dateTime, valueTypeWrapper, index, referenceFlags(metaObject, index));
+    };
+    const auto wrapQDate = [&](QDate date) {
+        return engine->newDateObject(
+                    date, valueTypeWrapper, index, referenceFlags(metaObject, index));
+    };
+    const auto wrapQTime = [&](QTime time) {
+        return engine->newDateObject(
+                    time, valueTypeWrapper, index, referenceFlags(metaObject, index));
+    };
+
+#define VALUE_TYPE_LOAD(metatype, cpptype, constructor) \
+    case metatype: { \
+        cpptype v; \
+        void *args[] = { &v, nullptr }; \
+        doStaticReadCall(metaObject, valueTypeWrapper, index, args); \
+        return QV4::Encode(constructor(v)); \
     }
-    metaObject->d.static_metacall(reinterpret_cast<QObject*>(valueTypeWrapper->gadgetPtr()), QMetaObject::ReadProperty,
-                                  index, args);
-    return engine->fromVariant(v);
+
+    QQmlMetaObject::resolveGadgetMethodOrPropertyIndex(
+                QMetaObject::ReadProperty, &metaObject, &index);
+
+    const int metaTypeId = isEnum
+            ? metaType.underlyingType().id()
+            : (metaType.flags() & QMetaType::PointerToQObject)
+              ? QMetaType::QObjectStar
+              : metaType.id();
+
+    switch (metaTypeId) {
+    case QMetaType::UnknownType:
+    case QMetaType::Void:
+        return Encode::undefined();
+    case QMetaType::Nullptr:
+    case QMetaType::VoidStar:
+        return Encode::null();
+    VALUE_TYPE_LOAD(QMetaType::Bool, bool, bool);
+    VALUE_TYPE_LOAD(QMetaType::Int, int, int);
+    VALUE_TYPE_LOAD(QMetaType::UInt, uint, uint);
+    VALUE_TYPE_LOAD(QMetaType::Long, long, double);
+    VALUE_TYPE_LOAD(QMetaType::ULong, ulong, double);
+    VALUE_TYPE_LOAD(QMetaType::LongLong, qlonglong, double);
+    VALUE_TYPE_LOAD(QMetaType::ULongLong, qulonglong, double);
+    VALUE_TYPE_LOAD(QMetaType::Double, double, double);
+    VALUE_TYPE_LOAD(QMetaType::QString, QString, engine->newString);
+    VALUE_TYPE_LOAD(QMetaType::QByteArray, QByteArray, engine->newArrayBuffer);
+    VALUE_TYPE_LOAD(QMetaType::Float, float, float);
+    VALUE_TYPE_LOAD(QMetaType::Short, short, int);
+    VALUE_TYPE_LOAD(QMetaType::UShort, unsigned short, int);
+    VALUE_TYPE_LOAD(QMetaType::Char, char, int);
+    VALUE_TYPE_LOAD(QMetaType::UChar, unsigned char, int);
+    VALUE_TYPE_LOAD(QMetaType::SChar, signed char, int);
+    VALUE_TYPE_LOAD(QMetaType::QChar, QChar, engine->newString);
+    VALUE_TYPE_LOAD(QMetaType::Char16, char16_t, wrapChar16);
+    VALUE_TYPE_LOAD(QMetaType::QDateTime, QDateTime, wrapQDateTime);
+    VALUE_TYPE_LOAD(QMetaType::QDate, QDate, wrapQDate);
+    VALUE_TYPE_LOAD(QMetaType::QTime, QTime, wrapQTime);
+#if QT_CONFIG(regularexpression)
+    VALUE_TYPE_LOAD(QMetaType::QRegularExpression, QRegularExpression, engine->newRegExpObject);
+#endif
+    VALUE_TYPE_LOAD(QMetaType::QObjectStar, QObject*, wrapQObject);
+    VALUE_TYPE_LOAD(QMetaType::QJsonValue, QJsonValue, wrapJsonValue);
+    VALUE_TYPE_LOAD(QMetaType::QJsonObject, QJsonObject, wrapJsonObject);
+    VALUE_TYPE_LOAD(QMetaType::QJsonArray, QJsonArray, wrapJsonArray);
+    case QMetaType::QPixmap:
+    case QMetaType::QImage: {
+        QVariant v(metaType);
+        void *args[] = { v.data(), nullptr };
+        doStaticReadCall(metaObject, valueTypeWrapper, index, args);
+        return Encode(engine->newVariantObject(metaType, v.data()));
+    }
+    case QMetaType::QVariant: {
+        QVariant v;
+        void *args[] = { &v, nullptr };
+        doStaticReadCall(metaObject, valueTypeWrapper, index, args);
+        return engine->fromVariant(
+                    v, valueTypeWrapper, index,
+                    referenceFlags(metaObject, index) | Heap::ReferenceObject::IsVariant);
+    }
+    default:
+        break;
+    }
+
+    QVariant v(metaType);
+    void *args[] = { v.data(), nullptr };
+    doStaticReadCall(metaObject, valueTypeWrapper, index, args);
+    return engine->fromVariant(v, valueTypeWrapper, index, referenceFlags(metaObject, index));
 #undef VALUE_TYPE_LOAD
 }
 
@@ -254,8 +423,7 @@ PropertyAttributes QQmlValueTypeWrapper::virtualGetOwnProperty(const Managed *m,
         if (!p)
             return Attr_Data; // Property exists, but we're not interested in the value
 
-        const QQmlValueTypeReference *ref = r->as<const QQmlValueTypeReference>();
-        if (!ref || ref->readReferenceValue()) {
+        if (!r->d()->isReference() || r->readReferenceValue()) {
             // Property exists, and we can retrieve it
             p->value = getGadgetProperty(
                         r->engine(), r->d(), result.propType(), result.coreIndex(),
@@ -282,10 +450,8 @@ struct QQmlValueTypeWrapperOwnPropertyKeyIterator : ObjectOwnPropertyKeyIterator
 PropertyKey QQmlValueTypeWrapperOwnPropertyKeyIterator::next(const Object *o, Property *pd, PropertyAttributes *attrs) {
     const QQmlValueTypeWrapper *that = static_cast<const QQmlValueTypeWrapper *>(o);
 
-    if (const QQmlValueTypeReference *ref = that->as<QQmlValueTypeReference>()) {
-        if (!ref->readReferenceValue())
-            return PropertyKey::invalid();
-    }
+    if (that->d()->isReference() && !that->readReferenceValue())
+        return PropertyKey::invalid();
 
     const QMetaObject *mo = that->d()->metaObject();
     // We don't return methods, ie. they are not visible when iterating
@@ -317,9 +483,8 @@ OwnPropertyKeyIterator *QQmlValueTypeWrapper::virtualOwnPropertyKeys(const Objec
 
 bool QQmlValueTypeWrapper::isEqual(const QVariant& value) const
 {
-    if (const QQmlValueTypeReference *ref = as<const QQmlValueTypeReference>())
-        if (!ref->readReferenceValue())
-            return false;
+    if (d()->isReference() && !readReferenceValue())
+        return false;
     int id1 = value.metaType().id();
     QVariant v = d()->toVariant();
     int id2 = v.metaType().id();
@@ -367,26 +532,26 @@ bool QQmlValueTypeWrapper::isEqual(const QVariant& value) const
 
 int QQmlValueTypeWrapper::typeId() const
 {
-    return d()->valueType()->metaType.id();
+    return d()->metaType().id();
 }
 
 QMetaType QQmlValueTypeWrapper::type() const
 {
-    return d()->valueType()->metaType;
+    return d()->metaType();
 }
 
 bool QQmlValueTypeWrapper::write(QObject *target, int propertyIndex) const
 {
     bool destructGadgetOnExit = false;
     Q_ALLOCA_DECLARE(void, gadget);
-    if (const QQmlValueTypeReference *ref = as<const QQmlValueTypeReference>()) {
+    if (d()->isReference()) {
         if (!d()->gadgetPtr()) {
-            Q_ALLOCA_ASSIGN(void, gadget, d()->valueType()->metaType.sizeOf());
+            Q_ALLOCA_ASSIGN(void, gadget, d()->metaType().sizeOf());
             d()->setGadgetPtr(gadget);
-            d()->valueType()->metaType.construct(d()->gadgetPtr(), nullptr);
+            d()->metaType().construct(d()->gadgetPtr(), nullptr);
             destructGadgetOnExit = true;
         }
-        if (!ref->readReferenceValue())
+        if (!readReferenceValue())
             return false;
     }
 
@@ -396,7 +561,7 @@ bool QQmlValueTypeWrapper::write(QObject *target, int propertyIndex) const
     QMetaObject::metacall(target, QMetaObject::WriteProperty, propertyIndex, a);
 
     if (destructGadgetOnExit) {
-        d()->valueType()->metaType.destruct(d()->gadgetPtr());
+        d()->metaType().destruct(d()->gadgetPtr());
         d()->setGadgetPtr(nullptr);
     }
     return true;
@@ -429,14 +594,13 @@ ReturnedValue QQmlValueTypeWrapper::method_toString(const FunctionObject *b, con
     if (!w)
         return b->engine()->throwTypeError();
 
-    if (const QQmlValueTypeReference *ref = w->as<QQmlValueTypeReference>())
-        if (!ref->readReferenceValue())
-            RETURN_UNDEFINED();
+    if (w->d()->isReference() && !w->readReferenceValue())
+        RETURN_UNDEFINED();
 
     QString result;
-    if (!QMetaType::convert(w->d()->valueType()->metaType, w->d()->gadgetPtr(),
+    if (!QMetaType::convert(w->d()->metaType(), w->d()->gadgetPtr(),
                             QMetaType(QMetaType::QString), &result)) {
-        result = QString::fromUtf8(w->d()->valueType()->metaType.name()) + QLatin1Char('(');
+        result = QString::fromUtf8(w->d()->metaType().name()) + QLatin1Char('(');
         const QMetaObject *mo = w->d()->metaObject();
         const int propCount = mo->propertyCount();
         for (int i = 0; i < propCount; ++i) {
@@ -466,16 +630,14 @@ ReturnedValue QQmlValueTypeWrapper::virtualResolveLookupGetter(const Object *obj
     ScopedString name(scope, id.asStringOrSymbol());
 
     // Note: readReferenceValue() can change the reference->type.
-    if (const QQmlValueTypeReference *reference = r->as<QQmlValueTypeReference>()) {
-        if (!reference->readReferenceValue())
-            return Value::undefinedValue().asReturnedValue();
-    }
+    if (r->d()->isReference() && !r->readReferenceValue())
+        return Value::undefinedValue().asReturnedValue();
 
     QQmlPropertyData result = r->dataForPropertyKey(id);
     if (!result.isValid())
         return QV4::Object::virtualResolveLookupGetter(object, engine, lookup);
 
-    lookup->qgadgetLookup.ic = r->internalClass();
+    lookup->qgadgetLookup.ic.set(engine, r->internalClass());
     // & 1 to tell the gc that this is not heap allocated; see markObjects in qv4lookup_p.h
     lookup->qgadgetLookup.metaObject = quintptr(r->d()->metaObject()) + 1;
     lookup->qgadgetLookup.metaType = result.propType().iface();
@@ -505,21 +667,47 @@ ReturnedValue QQmlValueTypeWrapper::lookupGetter(Lookup *lookup, ExecutionEngine
     if (valueTypeWrapper->metaObject() != reinterpret_cast<const QMetaObject *>(lookup->qgadgetLookup.metaObject - 1))
         return revertLookup();
 
-    if (lookup->qgadgetLookup.ic->vtable == QQmlValueTypeReference::staticVTable()) {
-        Scope scope(engine);
-        Scoped<QQmlValueTypeReference> referenceWrapper(scope, valueTypeWrapper);
-        referenceWrapper->readReferenceValue();
-    }
+    if (valueTypeWrapper->isReference() && !valueTypeWrapper->readReference())
+        return Encode::undefined();
 
-    return getGadgetProperty(engine, valueTypeWrapper, QMetaType(lookup->qgadgetLookup.metaType), lookup->qgadgetLookup.coreIndex, lookup->qgadgetLookup.isFunction, lookup->qgadgetLookup.isEnum);
+    return getGadgetProperty(
+                engine, valueTypeWrapper, QMetaType(lookup->qgadgetLookup.metaType),
+                lookup->qgadgetLookup.coreIndex, lookup->qgadgetLookup.isFunction,
+                lookup->qgadgetLookup.isEnum);
 }
 
 bool QQmlValueTypeWrapper::lookupSetter(
         Lookup *l, ExecutionEngine *engine, Value &object, const Value &value)
 {
+    if (&QQmlValueTypeWrapper::lookupSetter == &QV4::Lookup::setterFallback) {
+        // Certain compilers, e.g. MSVC, will "helpfully" deduplicate methods that are completely
+        // equal. As a result, the pointers are the same, which wreaks havoc on the logic that
+        // decides how to retrieve the property.
+        qFatal("Your C++ compiler is broken.");
+    }
+
+    // This setter marks the presence of a value type setter lookup.
+    // It falls back to the fallback lookup when run through the interpreter, but AOT-compiled
+    // code can get clever with it.
     return QV4::Lookup::setterFallback(l, engine, object, value);
 }
 
+bool QQmlValueTypeWrapper::lookupSetterAsVariant(
+        Lookup *l, ExecutionEngine *engine, Value &object, const Value &value)
+{
+    if (&QQmlValueTypeWrapper::lookupSetterAsVariant == &QQmlValueTypeWrapper::lookupSetter) {
+        // Certain compilers, e.g. MSVC, will "helpfully" deduplicate methods that are completely
+        // equal. As a result, the pointers are the same, which wreaks havoc on the logic that
+        // decides how to retrieve the property.
+        qFatal("Your C++ compiler is broken.");
+    }
+
+    // This setter marks the presence of a value type setter lookup with QVariant conversion.
+    // It falls back to the fallback lookup when run through the interpreter, but AOT-compiled
+    // code can get clever with it.
+    return lookupSetter(l, engine, object, value);
+}
+
 bool QQmlValueTypeWrapper::virtualResolveLookupSetter(Object *object, ExecutionEngine *engine, Lookup *lookup,
                                                       const Value &value)
 {
@@ -537,10 +725,8 @@ ReturnedValue QQmlValueTypeWrapper::virtualGet(const Managed *m, PropertyKey id,
     QV4::ExecutionEngine *v4 = r->engine();
 
     // Note: readReferenceValue() can change the reference->type.
-    if (const QQmlValueTypeReference *reference = r->as<QQmlValueTypeReference>()) {
-        if (!reference->readReferenceValue())
-            return Value::undefinedValue().asReturnedValue();
-    }
+    if (r->d()->isReference() && !r->readReferenceValue())
+        return Value::undefinedValue().asReturnedValue();
 
     QQmlPropertyData result = r->dataForPropertyKey(id);
     if (!result.isValid())
@@ -564,17 +750,11 @@ bool QQmlValueTypeWrapper::virtualPut(Managed *m, PropertyKey id, const Value &v
         return false;
 
     Scoped<QQmlValueTypeWrapper> r(scope, static_cast<QQmlValueTypeWrapper *>(m));
-    Scoped<QQmlValueTypeReference> reference(scope, m->d());
-
-    QMetaType writeBackPropertyType;
-
-    if (reference) {
-        QMetaProperty writebackProperty = reference->d()->object->metaObject()->property(reference->d()->property);
-
-        if (!writebackProperty.isWritable() || !reference->readReferenceValue())
+    Heap::Object *heapObject = nullptr;
+    if (r->d()->isReference()) {
+        heapObject = r->d()->object();
+        if (!r->readReferenceValue() || !r->d()->canWriteBack())
             return false;
-
-        writeBackPropertyType = writebackProperty.metaType();
     }
 
     const QMetaObject *metaObject = r->d()->metaObject();
@@ -582,10 +762,18 @@ bool QQmlValueTypeWrapper::virtualPut(Managed *m, PropertyKey id, const Value &v
     if (!pd.isValid())
         return false;
 
-    if (reference) {
+    if (heapObject) {
+        QObject *referenceObject = nullptr;
         QV4::ScopedFunctionObject f(scope, value);
-        const QV4QPointer<QObject> &referenceObject = reference->d()->object;
-        const int referencePropertyIndex = reference->d()->property;
+        const int referencePropertyIndex = r->d()->property();
+        QV4::Scoped<QV4::QObjectWrapper> o(scope, heapObject);
+        if (o) {
+            referenceObject = o->object();
+        } else {
+            QV4::Scoped<QV4::QQmlTypeWrapper> t(scope, heapObject);
+            if (t)
+                referenceObject = t->object();
+        }
 
         if (f) {
             if (!f->isBinding()) {
@@ -596,6 +784,17 @@ bool QQmlValueTypeWrapper::virtualPut(Managed *m, PropertyKey id, const Value &v
                 return false;
             }
 
+            if (!referenceObject) {
+                QString error = QStringLiteral("Cannot create binding on nested value type property");
+                ScopedString e(scope, v4->newString(error));
+                v4->throwError(e);
+                return false;
+            }
+
+            const QMetaProperty writebackProperty
+                    = referenceObject->metaObject()->property(referencePropertyIndex);
+            const QMetaType writeBackPropertyType = writebackProperty.metaType();
+
             QQmlRefPointer<QQmlContextData> context = v4->callingQmlContext();
 
             QQmlPropertyData cacheData;
@@ -605,7 +804,7 @@ bool QQmlValueTypeWrapper::virtualPut(Managed *m, PropertyKey id, const Value &v
 
             QV4::Scoped<QQmlBindingFunction> bindingFunction(scope, (const Value &)f);
 
-            QV4::ScopedFunctionObject f(scope, bindingFunction->bindingFunction());
+            QV4::Scoped<JavaScriptFunctionObject> f(scope, bindingFunction->bindingFunction());
             QV4::ScopedContext ctx(scope, f->scope());
             QQmlBinding *newBinding = QQmlBinding::create(&cacheData, f->function(), referenceObject, context, ctx);
             newBinding->setSourceLocation(bindingFunction->currentLocation());
@@ -615,13 +814,13 @@ bool QQmlValueTypeWrapper::virtualPut(Managed *m, PropertyKey id, const Value &v
             newBinding->setTarget(referenceObject, cacheData, &pd);
             QQmlPropertyPrivate::setBinding(newBinding);
             return true;
-        } else {
-            if (Q_UNLIKELY(lcBindingRemoval().isInfoEnabled())) {
+        } else if (referenceObject) {
+            if (Q_UNLIKELY(lcBuiltinsBindingRemoval().isInfoEnabled())) {
                 if (auto binding = QQmlPropertyPrivate::binding(referenceObject, QQmlPropertyIndex(referencePropertyIndex, pd.coreIndex()))) {
                     Q_ASSERT(binding->kind() == QQmlAbstractBinding::QmlBinding);
                     const auto qmlBinding = static_cast<const QQmlBinding*>(binding);
                     const auto stackFrame = v4->currentStackFrame;
-                    qCInfo(lcBindingRemoval,
+                    qCInfo(lcBuiltinsBindingRemoval,
                            "Overwriting binding on %s::%s which was initially bound at %s by setting \"%s\" at %s:%d",
                            referenceObject->metaObject()->className(), referenceObject->metaObject()->property(referencePropertyIndex).name(),
                            qPrintable(qmlBinding->expressionIdentifier()),
@@ -635,19 +834,35 @@ bool QQmlValueTypeWrapper::virtualPut(Managed *m, PropertyKey id, const Value &v
 
     QMetaProperty property = metaObject->property(pd.coreIndex());
     Q_ASSERT(property.isValid());
+    if (value.isUndefined() && pd.isResettable()) {
+        property.resetOnGadget(reinterpret_cast<QObject *>(r->d()->gadgetPtr()));
+        if (heapObject)
+            r->d()->writeBack(pd.coreIndex());
+        return true;
+    }
 
-    QVariant v = v4->toVariant(value, property.metaType());
+    QVariant v = QV4::ExecutionEngine::toVariant(value, property.metaType());
 
     if (property.isEnumType() && (QMetaType::Type)v.userType() == QMetaType::Double)
         v = v.toInt();
 
     void *gadget = r->d()->gadgetPtr();
-    property.writeOnGadget(gadget, v);
+    const QMetaType type = v.metaType();
+    if (!property.writeOnGadget(gadget, std::move(v))) {
+        const QString error = QLatin1String("Cannot assign ") +
+                QLatin1String(type.name()) +
+                QLatin1String(" to ") +
+                QLatin1String(property.metaType().name());
+        scope.engine->throwError(error);
+        return true;
+    }
 
-    if (reference)
-        reference->d()->writeBack();
+    if (heapObject)
+        r->d()->writeBack(pd.coreIndex());
 
     return true;
 }
 
+} // namespace QV4
+
 QT_END_NAMESPACE
