@@ -6,10 +6,12 @@
 #include <private/qqmlirbuilder_p.h>
 #include <private/qqmljsbasicblocks_p.h>
 #include <private/qqmljscodegenerator_p.h>
+#include <private/qqmljscompilerstats_p.h>
 #include <private/qqmljsfunctioninitializer_p.h>
 #include <private/qqmljsimportvisitor_p.h>
 #include <private/qqmljslexer_p.h>
 #include <private/qqmljsloadergenerator_p.h>
+#include <private/qqmljsoptimizations_p.h>
 #include <private/qqmljsparser_p.h>
 #include <private/qqmljsshadowcheck_p.h>
 #include <private/qqmljsstoragegeneralizer_p.h>
@@ -19,6 +21,8 @@
 #include <QtCore/qfileinfo.h>
 #include <QtCore/qloggingcategory.h>
 
+#include <QtQml/private/qqmlsignalnames_p.h>
+
 #include <limits>
 
 QT_BEGIN_NAMESPACE
@@ -123,9 +127,7 @@ static bool checkArgumentsObjectUseInSignalHandlers(const QmlIR::Document &doc,
             if (binding->type() != QV4::CompiledData::Binding::Type_Script)
                 continue;
             const QString propName =  doc.stringAt(binding->propertyNameIndex);
-            if (!propName.startsWith(QLatin1String("on"))
-                || propName.length() < 3
-                || !propName.at(2).isUpper())
+            if (!QQmlSignalNames::isHandlerName(propName))
                 continue;
             auto compiledFunction = doc.jsModule.functions.value(object->runtimeFunctionIndices.at(binding->value.compiledScriptIndex));
             if (!compiledFunction)
@@ -221,7 +223,7 @@ bool qCompileQmlFile(QmlIR::Document &irDocument, const QString &inputFileName,
     QQmlJSAotFunctionMap aotFunctionsByIndex;
 
     {
-        QmlIR::JSCodeGen v4CodeGen(&irDocument, *illegalNames(), interface);
+        QmlIR::JSCodeGen v4CodeGen(&irDocument, *illegalNames(), interface, storeSourceLocation);
 
         if (aotCompiler)
             aotCompiler->setDocument(&v4CodeGen, &irDocument);
@@ -231,7 +233,7 @@ bool qCompileQmlFile(QmlIR::Document &irDocument, const QString &inputFileName,
             if (object->functionsAndExpressions->count == 0 && object->bindingCount() == 0)
                 continue;
 
-            if (!v4CodeGen.generateRuntimeFunctions(object, storeSourceLocation)) {
+            if (!v4CodeGen.generateRuntimeFunctions(object)) {
                 Q_ASSERT(v4CodeGen.hasError());
                 error->appendDiagnostic(inputFileName, v4CodeGen.error());
                 return false;
@@ -265,11 +267,11 @@ bool qCompileQmlFile(QmlIR::Document &irDocument, const QString &inputFileName,
 
             // AOT-compile bindings and functions in the same order as above so that the runtime
             // class indices match
+            auto contextMap = v4CodeGen.module()->contextMap;
             std::sort(bindingsAndFunctions.begin(), bindingsAndFunctions.end());
             std::for_each(bindingsAndFunctions.begin(), bindingsAndFunctions.end(),
                           [&](const BindingOrFunction &bindingOrFunction) {
                 std::variant<QQmlJSAotFunction, QQmlJS::DiagnosticMessage> result;
-                auto *module = v4CodeGen.module();
                 if (const auto *binding = bindingOrFunction.binding()) {
                     switch (binding->type()) {
                     case QmlIR::Binding::Type_AttachedProperty:
@@ -289,27 +291,53 @@ bool qCompileQmlFile(QmlIR::Document &irDocument, const QString &inputFileName,
                         break;
                     }
 
-                    Q_ASSERT(quint32(functionsToCompile.length()) > binding->value.compiledScriptIndex);
-                    auto *node = functionsToCompile[binding->value.compiledScriptIndex].parentNode;
-                    Q_ASSERT(node);
-                    Q_ASSERT(module->contextMap.contains(node));
-                    QV4::Compiler::Context *context = module->contextMap[node];
+                    Q_ASSERT(quint32(functionsToCompile.size()) > binding->value.compiledScriptIndex);
+                    const auto &functionToCompile
+                            = functionsToCompile[binding->value.compiledScriptIndex];
+                    auto *parentNode = functionToCompile.parentNode;
+                    Q_ASSERT(parentNode);
+                    Q_ASSERT(contextMap.contains(parentNode));
+                    QV4::Compiler::Context *context = contextMap.take(parentNode);
                     Q_ASSERT(context);
 
+                    auto *node = functionToCompile.node;
+                    Q_ASSERT(node);
+
+                    if (context->returnsClosure) {
+                        QQmlJS::AST::Node *inner
+                                = QQmlJS::AST::cast<QQmlJS::AST::ExpressionStatement *>(
+                                    node)->expression;
+                        Q_ASSERT(inner);
+                        QV4::Compiler::Context *innerContext = contextMap.take(inner);
+                        Q_ASSERT(innerContext);
+                        qCDebug(lcAotCompiler) << "Compiling signal handler for"
+                                               << irDocument.stringAt(binding->propertyNameIndex);
+                        std::variant<QQmlJSAotFunction, QQmlJS::DiagnosticMessage> innerResult
+                                = aotCompiler->compileBinding(innerContext, *binding, inner);
+
+                        if (auto *error = std::get_if<QQmlJS::DiagnosticMessage>(&innerResult)) {
+                            qCDebug(lcAotCompiler) << "Compilation failed:"
+                                                   << diagnosticErrorMessage(inputFileName, *error);
+                        } else if (auto *func = std::get_if<QQmlJSAotFunction>(&innerResult)) {
+                            qCDebug(lcAotCompiler) << "Generated code:" << func->code;
+                            aotFunctionsByIndex[innerContext->functionIndex] = *func;
+                        }
+                    }
+
                     qCDebug(lcAotCompiler) << "Compiling binding for property"
                                            << irDocument.stringAt(binding->propertyNameIndex);
-                    result = aotCompiler->compileBinding(context, *binding);
+                    result = aotCompiler->compileBinding(context, *binding, node);
                 } else if (const auto *function = bindingOrFunction.function()) {
-                    Q_ASSERT(quint32(functionsToCompile.length()) > function->index);
+                    Q_ASSERT(quint32(functionsToCompile.size()) > function->index);
                     auto *node = functionsToCompile[function->index].node;
                     Q_ASSERT(node);
-                    Q_ASSERT(module->contextMap.contains(node));
-                    QV4::Compiler::Context *context = module->contextMap[node];
+                    Q_ASSERT(contextMap.contains(node));
+                    QV4::Compiler::Context *context = contextMap.take(node);
                     Q_ASSERT(context);
 
-                    qCDebug(lcAotCompiler) << "Compiling function"
-                                           << irDocument.stringAt(function->nameIndex);
-                    result = aotCompiler->compileFunction(context, *function);
+                    const QString functionName = irDocument.stringAt(function->nameIndex);
+                    qCDebug(lcAotCompiler) << "Compiling function" << functionName;
+                    result = aotCompiler->compileFunction(context, functionName, node);
                 } else {
                     Q_UNREACHABLE();
                 }
@@ -337,8 +365,8 @@ bool qCompileQmlFile(QmlIR::Document &irDocument, const QString &inputFileName,
         const quint32 saveFlags
                 = QV4::CompiledData::Unit::StaticData
                 | QV4::CompiledData::Unit::PendingTypeCompilation;
-        QV4::CompiledData::SaveableUnitPointer saveable(irDocument.javaScriptCompilationUnit.data,
-                                                        saveFlags);
+        QV4::CompiledData::SaveableUnitPointer saveable(
+                irDocument.javaScriptCompilationUnit->unitData(), saveFlags);
         if (!saveFunction(saveable, aotFunctionsByIndex, &error->message))
             return false;
     }
@@ -347,7 +375,7 @@ bool qCompileQmlFile(QmlIR::Document &irDocument, const QString &inputFileName,
 
 bool qCompileJSFile(const QString &inputFileName, const QString &inputFileUrl, QQmlJSSaveFunction saveFunction, QQmlJSCompileError *error)
 {
-    QV4::CompiledData::CompilationUnit unit;
+    QQmlRefPointer<QV4::CompiledData::CompilationUnit> unit;
 
     QString sourceCode;
     {
@@ -371,7 +399,7 @@ bool qCompileJSFile(const QString &inputFileName, const QString &inputFileUrl, Q
         unit = QV4::Compiler::Codegen::compileModule(/*debugMode*/false, url, sourceCode,
                                                      QDateTime(), &diagnostics);
         error->appendDiagnostics(inputFileName, diagnostics);
-        if (!unit.unitData())
+        if (!unit || !unit->unitData())
             return false;
     } else {
         QmlIR::Document irDocument(/*debugMode*/false);
@@ -429,32 +457,14 @@ bool qCompileJSFile(const QString &inputFileName, const QString &inputFileUrl, Q
     }
 
     QQmlJSAotFunctionMap empty;
-    return saveFunction(QV4::CompiledData::SaveableUnitPointer(unit.data), empty, &error->message);
+    return saveFunction(
+            QV4::CompiledData::SaveableUnitPointer(unit->unitData()), empty, &error->message);
 }
 
-static const char *wrapCallCode = R"(
-template <typename Binding>
-void wrapCall(const QQmlPrivate::AOTCompiledContext *aotContext, void *dataPtr, void **argumentsPtr, Binding &&binding)
-{
-    using return_type = std::invoke_result_t<Binding, const QQmlPrivate::AOTCompiledContext *, void **>;
-    if constexpr (std::is_same_v<return_type, void>) {
-       Q_UNUSED(dataPtr)
-       binding(aotContext, argumentsPtr);
-    } else {
-        if (dataPtr) {
-           new (dataPtr) return_type(binding(aotContext, argumentsPtr));
-        } else {
-           binding(aotContext, argumentsPtr);
-        }
-    }
-}
-)";
-
 static const char *funcHeaderCode = R"(
-    [](const QQmlPrivate::AOTCompiledContext *aotContext, void *dataPtr, void **argumentsPtr) {
-        wrapCall(aotContext, dataPtr, argumentsPtr, [](const QQmlPrivate::AOTCompiledContext *aotContext, void **argumentsPtr) {
+    [](const QQmlPrivate::AOTCompiledContext *aotContext, void **argv) {
 Q_UNUSED(aotContext)
-Q_UNUSED(argumentsPtr)
+Q_UNUSED(argv)
 )";
 
 bool qSaveQmlJSUnitAsCpp(const QString &inputFileName, const QString &outputFileName, const QV4::CompiledData::SaveableUnitPointer &unit, const QQmlJSAotFunctionMap &aotFunctions, QString *errorString)
@@ -551,13 +561,12 @@ bool qSaveQmlJSUnitAsCpp(const QString &inputFileName, const QString &outputFile
     if (aotFunctions.size() <= 1) {
         // FileScopeCodeIndex is always there, but it may be the only one.
         writeStr("extern const QQmlPrivate::AOTCompiledFunction aotBuiltFunctions[];\n"
-                 "extern const QQmlPrivate::AOTCompiledFunction aotBuiltFunctions[] = { { 0, QMetaType::fromType<void>(), {}, nullptr } };");
+                 "extern const QQmlPrivate::AOTCompiledFunction aotBuiltFunctions[] = { { 0, 0, nullptr, nullptr } };\n");
     } else {
-        writeStr(wrapCallCode);
         writeStr("extern const QQmlPrivate::AOTCompiledFunction aotBuiltFunctions[];\n"
                  "extern const QQmlPrivate::AOTCompiledFunction aotBuiltFunctions[] = {\n");
 
-        QString footer = QStringLiteral("});}\n");
+        QString footer = QStringLiteral("}\n");
 
         for (QQmlJSAotFunctionMap::ConstIterator func = aotFunctions.constBegin(),
              end = aotFunctions.constEnd();
@@ -566,25 +575,18 @@ bool qSaveQmlJSUnitAsCpp(const QString &inputFileName, const QString &outputFile
             if (func.key() == FileScopeCodeIndex)
                 continue;
 
-            QString function = QString::fromUtf8(funcHeaderCode) + func.value().code + footer;
+            const QString function = QString::fromUtf8(funcHeaderCode) + func.value().code + footer;
 
-            QString argumentTypes = func.value().argumentTypes.join(
-                        QStringLiteral(">(), QMetaType::fromType<"));
-            if (!argumentTypes.isEmpty()) {
-                argumentTypes = QStringLiteral("QMetaType::fromType<")
-                        + argumentTypes + QStringLiteral(">()");
-            }
-
-            writeStr(QStringLiteral("{ %1, QMetaType::fromType<%2>(), { %3 }, %4 },")
+            writeStr(QStringLiteral("{ %1, %2, [](QV4::ExecutableCompilationUnit *unit, "
+                                    "QMetaType *argTypes) {\n%3}, %4 },")
                      .arg(func.key())
-                     .arg(func.value().returnType)
-                     .arg(argumentTypes)
-                     .arg(function)
+                     .arg(func->numArguments)
+                     .arg(func->signature, function)
                      .toUtf8().constData());
         }
 
         // Conclude the list with a nullptr
-        writeStr("{ 0, QMetaType::fromType<void>(), {}, nullptr }");
+        writeStr("{ 0, 0, nullptr, nullptr }");
         writeStr("};\n");
     }
 
@@ -621,10 +623,10 @@ void QQmlJSAotCompiler::setDocument(
     Q_UNUSED(codegen);
     m_document = irDocument;
     const QFileInfo resourcePathInfo(m_resourcePath);
-    m_logger->setFileName(resourcePathInfo.fileName());
+    if (m_logger->filePath().isEmpty())
+        m_logger->setFilePath(resourcePathInfo.fileName());
     m_logger->setCode(irDocument->code);
     m_unitGenerator = &irDocument->jsGenerator;
-    m_entireSourceCodeLines = irDocument->code.split(u'\n');
     QQmlJSScope::Ptr target = QQmlJSScope::create();
     QQmlJSImportVisitor visitor(target, m_importer, m_logger,
                                 resourcePathInfo.canonicalPath() + u'/',
@@ -652,7 +654,7 @@ QQmlJS::DiagnosticMessage QQmlJSAotCompiler::diagnose(
 {
     if (isStrict(m_document)
             && (type == QtWarningMsg || type == QtCriticalMsg || type == QtFatalMsg)
-            && m_logger->isCategoryFatal(Log_Compiler)) {
+            && m_logger->isCategoryFatal(qmlCompiler)) {
         qFatal("%s:%d: (strict mode) %s",
                qPrintable(QFileInfo(m_resourcePath).fileName()),
                location.startLine, qPrintable(message));
@@ -660,7 +662,7 @@ QQmlJS::DiagnosticMessage QQmlJSAotCompiler::diagnose(
 
     // TODO: this is a special place that explicitly sets the severity through
     // logger's private function
-    m_logger->log(message, Log_Compiler, location, type);
+    m_logger->log(message, qmlCompiler, location, type);
 
     return QQmlJS::DiagnosticMessage {
         message,
@@ -670,13 +672,17 @@ QQmlJS::DiagnosticMessage QQmlJSAotCompiler::diagnose(
 }
 
 std::variant<QQmlJSAotFunction, QQmlJS::DiagnosticMessage> QQmlJSAotCompiler::compileBinding(
-        const QV4::Compiler::Context *context, const QmlIR::Binding &irBinding)
+        const QV4::Compiler::Context *context, const QmlIR::Binding &irBinding,
+        QQmlJS::AST::Node *astNode)
 {
-    QQmlJSFunctionInitializer initializer(&m_typeResolver, m_currentObject, m_currentScope);
+    QQmlJSFunctionInitializer initializer(
+                &m_typeResolver, m_currentObject->location, m_currentScope->location);
     QQmlJS::DiagnosticMessage error;
     const QString name = m_document->stringAt(irBinding.propertyNameIndex);
-    QQmlJSCompilePass::Function function = initializer.run(context, name, irBinding, &error);
-    const QQmlJSAotFunction aotFunction = doCompile(context, &function, &error);
+    QQmlJSCompilePass::Function function = initializer.run(
+                context, name, astNode, irBinding, &error);
+    const QQmlJSAotFunction aotFunction = doCompileAndRecordAotStats(
+            context, &function, &error, name, astNode->firstSourceLocation());
 
     if (error.isValid()) {
         // If it's a signal and the function just returns a closure, it's harmless.
@@ -694,13 +700,14 @@ std::variant<QQmlJSAotFunction, QQmlJS::DiagnosticMessage> QQmlJSAotCompiler::co
 }
 
 std::variant<QQmlJSAotFunction, QQmlJS::DiagnosticMessage> QQmlJSAotCompiler::compileFunction(
-        const QV4::Compiler::Context *context, const QmlIR::Function &irFunction)
+        const QV4::Compiler::Context *context, const QString &name, QQmlJS::AST::Node *astNode)
 {
-    QQmlJSFunctionInitializer initializer(&m_typeResolver, m_currentObject, m_currentScope);
+    QQmlJSFunctionInitializer initializer(
+                &m_typeResolver, m_currentObject->location, m_currentScope->location);
     QQmlJS::DiagnosticMessage error;
-    const QString name = m_document->stringAt(irFunction.nameIndex);
-    QQmlJSCompilePass::Function function = initializer.run(context, name, irFunction, &error);
-    const QQmlJSAotFunction aotFunction = doCompile(context, &function, &error);
+    QQmlJSCompilePass::Function function = initializer.run(context, name, astNode, &error);
+    const QQmlJSAotFunction aotFunction = doCompileAndRecordAotStats(
+            context, &function, &error, name, astNode->firstSourceLocation());
 
     if (error.isValid())
         return diagnose(error.message, QtWarningMsg, error.loc);
@@ -720,8 +727,10 @@ QQmlJSAotFunction QQmlJSAotCompiler::globalCode() const
         u"QtQml/qqmlcomponent.h"_s,
         u"QtQml/qqmlcontext.h"_s,
         u"QtQml/qqmlengine.h"_s,
+        u"QtQml/qqmllist.h"_s,
 
         u"QtCore/qdatetime.h"_s,
+        u"QtCore/qtimezone.h"_s,
         u"QtCore/qobject.h"_s,
         u"QtCore/qstring.h"_s,
         u"QtCore/qstringlist.h"_s,
@@ -733,7 +742,6 @@ QQmlJSAotFunction QQmlJSAotCompiler::globalCode() const
     return global;
 }
 
-
 QQmlJSAotFunction QQmlJSAotCompiler::doCompile(
         const QV4::Compiler::Context *context, QQmlJSCompilePass::Function *function,
         QQmlJS::DiagnosticMessage *error)
@@ -744,31 +752,61 @@ QQmlJSAotFunction QQmlJSAotCompiler::doCompile(
         return QQmlJSAotFunction();
     };
 
-    QQmlJSTypePropagator propagator(m_unitGenerator, &m_typeResolver, m_logger);
-    auto typePropagationResult = propagator.run(function, error);
     if (error->isValid())
         return compileError();
 
-    QQmlJSBasicBlocks basicBlocks(m_unitGenerator, &m_typeResolver, m_logger);
-    typePropagationResult = basicBlocks.run(function, typePropagationResult);
+    bool basicBlocksValidationFailed = false;
+    QQmlJSBasicBlocks basicBlocks(context, m_unitGenerator, &m_typeResolver, m_logger);
+    auto passResult = basicBlocks.run(function, m_flags, basicBlocksValidationFailed);
+    auto &[blocks, annotations] = passResult;
 
-    QQmlJSShadowCheck shadowCheck(m_unitGenerator, &m_typeResolver, m_logger);
-    shadowCheck.run(&typePropagationResult, function, error);
+    QQmlJSTypePropagator propagator(m_unitGenerator, &m_typeResolver, m_logger, blocks, annotations);
+    passResult = propagator.run(function, error);
+    if (error->isValid())
+        return compileError();
+
+    QQmlJSShadowCheck shadowCheck(m_unitGenerator, &m_typeResolver, m_logger, blocks, annotations);
+    passResult = shadowCheck.run(function, error);
+    if (error->isValid())
+        return compileError();
+
+    QQmlJSOptimizations optimizer(m_unitGenerator, &m_typeResolver, m_logger, blocks, annotations,
+                                  basicBlocks.objectAndArrayDefinitions());
+    passResult = optimizer.run(function, error);
     if (error->isValid())
         return compileError();
 
     // Generalize all arguments, registers, and the return type.
-    QQmlJSStorageGeneralizer generalizer(
-                m_unitGenerator, &m_typeResolver, m_logger);
-    typePropagationResult = generalizer.run(typePropagationResult, function, error);
+    QQmlJSStorageGeneralizer generalizer(m_unitGenerator, &m_typeResolver, m_logger, blocks, annotations);
+    passResult = generalizer.run(function, error);
     if (error->isValid())
         return compileError();
 
-    QQmlJSCodeGenerator codegen(
-                context, m_unitGenerator, &m_typeResolver, m_logger,
-                m_entireSourceCodeLines);
-    QQmlJSAotFunction result = codegen.run(function, &typePropagationResult, error);
+    QQmlJSCodeGenerator codegen(context, m_unitGenerator, &m_typeResolver, m_logger, blocks, annotations);
+    QQmlJSAotFunction result = codegen.run(function, error, basicBlocksValidationFailed);
     return error->isValid() ? compileError() : result;
 }
 
+QQmlJSAotFunction QQmlJSAotCompiler::doCompileAndRecordAotStats(
+        const QV4::Compiler::Context *context, QQmlJSCompilePass::Function *function,
+        QQmlJS::DiagnosticMessage *error, const QString &name, QQmlJS::SourceLocation location)
+{
+    auto t1 = std::chrono::high_resolution_clock::now();
+    auto &&result = doCompile(context, function, error);
+    auto t2 = std::chrono::high_resolution_clock::now();
+
+    if (QQmlJS::QQmlJSAotCompilerStats::recordAotStats()) {
+        QQmlJS::AotStatsEntry entry;
+        entry.codegenDuration = std::chrono::duration_cast<std::chrono::microseconds>(t2 - t1);
+        entry.functionName = name;
+        entry.errorMessage = error->message;
+        entry.line = location.startLine;
+        entry.column = location.startColumn;
+        entry.codegenSuccessful = !error->isValid();
+        QQmlJS::QQmlJSAotCompilerStats::addEntry(function->qmlScope->filePath(), entry);
+    }
+
+    return result;
+}
+
 QT_END_NAMESPACE
