@@ -6,6 +6,7 @@
 #include "qqmljstypereader_p.h"
 #include "qqmljsimportvisitor_p.h"
 #include "qqmljslogger_p.h"
+#include "qqmljsutils_p.h"
 
 #include <QtQml/private/qqmlimportresolver_p.h>
 
@@ -16,8 +17,24 @@ QT_BEGIN_NAMESPACE
 
 using namespace Qt::StringLiterals;
 
-static const QLatin1String SlashQmldir             = QLatin1String("/qmldir");
-static const QLatin1String SlashPluginsDotQmltypes = QLatin1String("/plugins.qmltypes");
+static const QLatin1String SlashQmldir        = QLatin1String("/qmldir");
+static const QLatin1String PluginsDotQmltypes = QLatin1String("plugins.qmltypes");
+
+
+QQmlJS::Import::Import(QString prefix, QString name, QTypeRevision version, bool isFile,
+                       bool isDependency) :
+      m_prefix(std::move(prefix)),
+      m_name(std::move(name)),
+      m_version(version),
+      m_isFile(isFile),
+      m_isDependency(isDependency)
+{
+}
+
+bool QQmlJS::Import::isValid() const
+{
+    return !m_name.isEmpty();
+}
 
 static const QString prefixedName(const QString &prefix, const QString &name)
 {
@@ -25,22 +42,29 @@ static const QString prefixedName(const QString &prefix, const QString &name)
     return prefix.isEmpty() ? name : (prefix  + QLatin1Char('.') + name);
 }
 
-static QQmlDirParser createQmldirParserForFile(const QString &filename)
+QQmlDirParser QQmlJSImporter::createQmldirParserForFile(const QString &filename, Import *import)
 {
+    Q_ASSERT(import);
     QFile f(filename);
-    f.open(QFile::ReadOnly);
     QQmlDirParser parser;
-    parser.parse(QString::fromUtf8(f.readAll()));
+    if (f.open(QFile::ReadOnly)) {
+        parser.parse(QString::fromUtf8(f.readAll()));
+    } else {
+        import->warnings.append({
+            QStringLiteral("Could not open qmldir file: ") + filename,
+            QtWarningMsg,
+            QQmlJS::SourceLocation()
+        });
+    }
+
     return parser;
 }
 
-void QQmlJSImporter::readQmltypes(
-        const QString &filename, QList<QQmlJSExportedScope> *objects,
-        QList<QQmlDirParser::Import> *dependencies)
+void QQmlJSImporter::readQmltypes(const QString &filename, Import *result)
 {
     const QFileInfo fileInfo(filename);
     if (!fileInfo.exists()) {
-        m_warnings.append({
+        result->warnings.append({
                               QStringLiteral("QML types file does not exist: ") + filename,
                               QtWarningMsg,
                               QQmlJS::SourceLocation()
@@ -49,7 +73,7 @@ void QQmlJSImporter::readQmltypes(
     }
 
     if (fileInfo.isDir()) {
-        m_warnings.append({
+        result->warnings.append({
                               QStringLiteral("QML types file cannot be a directory: ") + filename,
                               QtWarningMsg,
                               QQmlJS::SourceLocation()
@@ -59,7 +83,7 @@ void QQmlJSImporter::readQmltypes(
 
     QFile file(filename);
     if (!file.open(QFile::ReadOnly)) {
-        m_warnings.append({
+        result->warnings.append({
                               QStringLiteral("QML types file cannot be opened: ") + filename,
                               QtWarningMsg,
                               QQmlJS::SourceLocation()
@@ -69,18 +93,18 @@ void QQmlJSImporter::readQmltypes(
 
     QQmlJSTypeDescriptionReader reader { filename, QString::fromUtf8(file.readAll()) };
     QStringList dependencyStrings;
-    auto succ = reader(objects, &dependencyStrings);
+    auto succ = reader(&result->objects, &dependencyStrings);
     if (!succ)
-        m_warnings.append({ reader.errorMessage(), QtCriticalMsg, QQmlJS::SourceLocation() });
+        result->warnings.append({ reader.errorMessage(), QtCriticalMsg, QQmlJS::SourceLocation() });
 
     const QString warningMessage = reader.warningMessage();
     if (!warningMessage.isEmpty())
-        m_warnings.append({ warningMessage, QtWarningMsg, QQmlJS::SourceLocation() });
+        result->warnings.append({ warningMessage, QtWarningMsg, QQmlJS::SourceLocation() });
 
     if (dependencyStrings.isEmpty())
         return;
 
-    m_warnings.append({
+    result->warnings.append({
                           QStringLiteral("Found deprecated dependency specifications in %1."
                                          "Specify dependencies in qmldir and use qmltyperegistrar "
                                          "to generate qmltypes files without dependencies.")
@@ -92,15 +116,16 @@ void QQmlJSImporter::readQmltypes(
     for (const QString &dependency : std::as_const(dependencyStrings)) {
         const auto blank = dependency.indexOf(u' ');
         if (blank < 0) {
-            dependencies->append(QQmlDirParser::Import(dependency, {},
-                                                       QQmlDirParser::Import::Default));
+            result->dependencies.append(
+                    QQmlDirParser::Import(dependency, {}, QQmlDirParser::Import::Default));
             continue;
         }
 
         const QString module = dependency.left(blank);
         const QString versionString = dependency.mid(blank + 1).trimmed();
         if (versionString == QStringLiteral("auto")) {
-            dependencies->append(QQmlDirParser::Import(module, {}, QQmlDirParser::Import::Auto));
+            result->dependencies.append(
+                    QQmlDirParser::Import(module, {}, QQmlDirParser::Import::Auto));
             continue;
         }
 
@@ -111,8 +136,8 @@ void QQmlJSImporter::readQmltypes(
                 : QTypeRevision::fromVersion(versionString.left(dot).toUShort(),
                                              versionString.mid(dot + 1).toUShort());
 
-        dependencies->append(QQmlDirParser::Import(module, version,
-                                                   QQmlDirParser::Import::Default));
+        result->dependencies.append(
+                QQmlDirParser::Import(module, version, QQmlDirParser::Import::Default));
     }
 }
 
@@ -129,24 +154,79 @@ static bool isComposite(const QQmlJSScope::ConstPtr &scope)
     return scope.factory() || scope->isComposite();
 }
 
+static QStringList aliases(const QQmlJSScope::ConstPtr &scope)
+{
+    return isComposite(scope)
+        ? QStringList()
+        : scope->aliases();
+}
+
 QQmlJSImporter::QQmlJSImporter(const QStringList &importPaths, QQmlJSResourceFileMapper *mapper,
-                               bool useOptionalImports)
+                               QQmlJSImporterFlags flags)
     : m_importPaths(importPaths),
       m_mapper(mapper),
-      m_useOptionalImports(useOptionalImports),
-      m_createImportVisitor([](const QQmlJSScope::Ptr &target, QQmlJSImporter *importer,
-                               QQmlJSLogger *logger, const QString &implicitImportDirectory,
-                               const QStringList &qmldirFiles) {
-          return new QQmlJSImportVisitor(target, importer, logger, implicitImportDirectory,
-                                         qmldirFiles);
+      m_flags(flags),
+      m_importVisitor([](QQmlJS::AST::Node *rootNode, QQmlJSImporter *self,
+                         const ImportVisitorPrerequisites &p) {
+          auto visitor = std::unique_ptr<QQmlJS::AST::BaseVisitor>(new QQmlJSImportVisitor(
+                  p.m_target, self, p.m_logger, p.m_implicitImportDirectory, p.m_qmldirFiles));
+          QQmlJS::AST::Node::accept(rootNode, visitor.get());
       })
 {
 }
 
-QQmlJSImporter::Import QQmlJSImporter::readQmldir(const QString &path)
+static QString resolvePreferredPath(
+        const QString &qmldirPath, const QString &prefer, QQmlJSResourceFileMapper *mapper)
+{
+    if (prefer.isEmpty())
+        return qmldirPath;
+
+    if (!prefer.endsWith(u'/')) {
+        qWarning() << "Ignoring invalid prefer path" << prefer << "(has to end with slash)";
+        return qmldirPath;
+    }
+
+    if (prefer.startsWith(u':')) {
+        // Resource path: Resolve via resource file mapper if possible.
+        if (!mapper)
+            return qmldirPath;
+
+        Q_ASSERT(prefer.endsWith(u'/'));
+        const auto entry = mapper->entry(
+                QQmlJSResourceFileMapper::resourceFileFilter(prefer.mid(1) + SlashQmldir.mid(1)));
+
+        // This can be empty if the .qrc files does not belong to this module.
+        // In that case we trust the given qmldir file.
+        return entry.filePath.endsWith(SlashQmldir)
+                ? entry.filePath
+                : qmldirPath;
+    }
+
+    // Host file system path. This should be rare. We don't generate it.
+    const QFileInfo f(prefer + SlashQmldir);
+    const QString canonical = f.canonicalFilePath();
+    if (canonical.isEmpty()) {
+        qWarning() << "No qmldir at" << prefer;
+        return qmldirPath;
+    }
+    return canonical;
+}
+
+QQmlJSImporter::Import QQmlJSImporter::readQmldir(const QString &modulePath)
 {
     Import result;
-    auto reader = createQmldirParserForFile(path + SlashQmldir);
+    const QString moduleQmldirPath = modulePath + SlashQmldir;
+    auto reader = createQmldirParserForFile(moduleQmldirPath, &result);
+
+    const QString resolvedQmldirPath
+            = resolvePreferredPath(moduleQmldirPath, reader.preferredPath(), m_mapper);
+    if (resolvedQmldirPath != moduleQmldirPath)
+        reader = createQmldirParserForFile(resolvedQmldirPath, &result);
+
+    // Leave the trailing slash
+    Q_ASSERT(resolvedQmldirPath.endsWith(SlashQmldir));
+    QStringView resolvedPath = QStringView(resolvedQmldirPath).chopped(SlashQmldir.size() - 1);
+
     result.name = reader.typeNamespace();
 
     result.isStaticModule = reader.isStaticModule();
@@ -157,31 +237,32 @@ QQmlJSImporter::Import QQmlJSImporter::readQmldir(const QString &path)
     const auto typeInfos = reader.typeInfos();
     for (const auto &typeInfo : typeInfos) {
         const QString typeInfoPath = QFileInfo(typeInfo).isRelative()
-                ? path + u'/' + typeInfo : typeInfo;
-        readQmltypes(typeInfoPath, &result.objects, &result.dependencies);
+                ? resolvedPath + typeInfo
+                : typeInfo;
+        readQmltypes(typeInfoPath, &result);
     }
 
     if (typeInfos.isEmpty() && !reader.plugins().isEmpty()) {
-        const QString defaultTypeInfoPath = path + SlashPluginsDotQmltypes;
+        const QString defaultTypeInfoPath = resolvedPath + PluginsDotQmltypes;
         if (QFile::exists(defaultTypeInfoPath)) {
-            m_warnings.append({
+            result.warnings.append({
                                   QStringLiteral("typeinfo not declared in qmldir file: ")
                                     + defaultTypeInfoPath,
                                   QtWarningMsg,
                                   QQmlJS::SourceLocation()
                               });
-            readQmltypes(defaultTypeInfoPath, &result.objects, &result.dependencies);
+            readQmltypes(defaultTypeInfoPath, &result);
         }
     }
 
     QHash<QString, QQmlJSExportedScope> qmlComponents;
     const auto components = reader.components();
     for (auto it = components.begin(), end = components.end(); it != end; ++it) {
-        const QString filePath = path + QLatin1Char('/') + it->fileName;
+        const QString filePath = resolvedPath + it->fileName;
         if (!QFile::exists(filePath)) {
-            m_warnings.append({
+            result.warnings.append({
                                   it->fileName + QStringLiteral(" is listed as component in ")
-                                        + path + SlashQmldir
+                                        + resolvedQmldirPath
                                         + QStringLiteral(" but does not exist.\n"),
                                   QtWarningMsg,
                                   QQmlJS::SourceLocation()
@@ -208,7 +289,7 @@ QQmlJSImporter::Import QQmlJSImporter::readQmldir(const QString &path)
 
     const auto scripts = reader.scripts();
     for (const auto &script : scripts) {
-        const QString filePath = path + QLatin1Char('/') + script.fileName;
+        const QString filePath = resolvedPath + script.fileName;
         auto mo = result.scripts.find(script.fileName);
         if (mo == result.scripts.end())
             mo = result.scripts.insert(script.fileName, { localFile2ScopeTree(filePath), {} });
@@ -272,10 +353,9 @@ QQmlJSImporter::Import QQmlJSImporter::readDirectory(const QString &directory)
     return import;
 }
 
-void QQmlJSImporter::importDependencies(const QQmlJSImporter::Import &import,
-                                        QQmlJSImporter::AvailableTypes *types,
-                                        const QString &prefix, QTypeRevision version,
-                                        bool isDependency)
+void QQmlJSImporter::importDependencies(
+        const QQmlJSImporter::Import &import, QQmlJSImporter::AvailableTypes *types,
+        const QString &prefix, QTypeRevision version, bool isDependency)
 {
     // Import the dependencies with an invalid prefix. The prefix will never be matched by actual
     // QML code but the C++ types will be visible.
@@ -286,7 +366,7 @@ void QQmlJSImporter::importDependencies(const QQmlJSImporter::Import &import,
     for (auto const &import : std::as_const(import.imports)) {
         if (import.flags & QQmlDirParser::Import::Optional) {
             hasOptionalImports = true;
-            if (!m_useOptionalImports) {
+            if (!useOptionalImports()) {
                 continue;
             }
 
@@ -299,16 +379,17 @@ void QQmlJSImporter::importDependencies(const QQmlJSImporter::Import &import,
                      isDependency);
     }
 
-    if (hasOptionalImports && !m_useOptionalImports) {
-        m_warnings.append(
-                { u"%1 uses optional imports which are not supported. Some types might not be found."_s
-                          .arg(import.name),
-                  QtCriticalMsg, QQmlJS::SourceLocation() });
+    if (hasOptionalImports && !useOptionalImports()) {
+        types->warnings.append({
+            u"%1 uses optional imports which are not supported. Some types might not be found."_s
+                                          .arg(import.name),
+            QtCriticalMsg, QQmlJS::SourceLocation()
+        });
     }
 }
 
 static bool isVersionAllowed(const QQmlJSScope::Export &exportEntry,
-                             const QQmlJSScope::Import &importDescription)
+                             const QQmlJS::Import &importDescription)
 {
     const QTypeRevision importVersion = importDescription.version();
     const QTypeRevision exportVersion = exportEntry.version();
@@ -320,9 +401,9 @@ static bool isVersionAllowed(const QQmlJSScope::Export &exportEntry,
             || exportVersion.minorVersion() <= importVersion.minorVersion();
 }
 
-void QQmlJSImporter::processImport(const QQmlJSScope::Import &importDescription,
-                                   const QQmlJSImporter::Import &import,
-                                   QQmlJSImporter::AvailableTypes *types)
+void QQmlJSImporter::processImport(
+        const QQmlJS::Import &importDescription, const QQmlJSImporter::Import &import,
+        QQmlJSImporter::AvailableTypes *types)
 {
     // In the list of QML types we prefix unresolvable QML names with $anonymous$, and C++
     // names with $internal$. This is to avoid clashes between them.
@@ -333,6 +414,12 @@ void QQmlJSImporter::processImport(const QQmlJSScope::Import &importDescription,
     const QString modulePrefix = QStringLiteral("$module$");
     QHash<QString, QList<QQmlJSScope::Export>> seenExports;
 
+    const auto insertAliases = [&](const QQmlJSScope::ConstPtr &scope, QTypeRevision revision) {
+        const QStringList cppAliases = aliases(scope);
+        for (const QString &alias : cppAliases)
+            types->cppNames.setType(alias, { scope, revision });
+    };
+
     const auto insertExports = [&](const QQmlJSExportedScope &val, const QString &cppName) {
         QQmlJSScope::Export bestExport;
 
@@ -377,7 +464,7 @@ void QQmlJSImporter::processImport(const QQmlJSScope::Import &importDescription,
                 case LowerVersion:
                     break;
                 case SameVersion: {
-                    m_warnings.append({
+                    types->warnings.append({
                         QStringLiteral("Ambiguous type detected. "
                                        "%1 %2.%3 is defined multiple times.")
                             .arg(qmlName)
@@ -405,6 +492,8 @@ void QQmlJSImporter::processImport(const QQmlJSScope::Import &importDescription,
                 : QTypeRevision::zero();
         types->cppNames.setType(cppName, { val.scope, bestRevision });
 
+        insertAliases(val.scope, bestRevision);
+
         const QTypeRevision bestVersion = bestExport.isValid()
                 ? bestExport.version()
                 : QTypeRevision::zero();
@@ -415,16 +504,17 @@ void QQmlJSImporter::processImport(const QQmlJSScope::Import &importDescription,
     if (!importDescription.prefix().isEmpty())
         types->qmlNames.setType(importDescription.prefix(), {});
 
-    // Add a marker to show that this module has been imported
-    if (!importDescription.isDependency())
+    if (!importDescription.isDependency()) {
+        // Add a marker to show that this module has been imported
         types->qmlNames.setType(prefixedName(modulePrefix, importDescription.name()), {});
 
-    if (!importDescription.isDependency()) {
         if (import.isStaticModule)
             types->staticModules << import.name;
 
         if (import.isSystemModule)
             types->hasSystemModule = true;
+
+        types->warnings.append(import.warnings);
     }
 
     for (auto it = import.scripts.begin(); it != import.scripts.end(); ++it) {
@@ -444,6 +534,7 @@ void QQmlJSImporter::processImport(const QQmlJSScope::Import &importDescription,
             types->qmlNames.setType(
                         prefixedName(internalPrefix, cppName), { val.scope, QTypeRevision() });
             types->cppNames.setType(cppName, { val.scope, QTypeRevision() });
+            insertAliases(val.scope, QTypeRevision());
         } else {
             insertExports(val, cppName);
         }
@@ -494,7 +585,7 @@ void QQmlJSImporter::processImport(const QQmlJSScope::Import &importDescription,
             // ... except that old qmltypes files might specify composite types with C++ names.
             // Warn about those.
             if (val.scope->isComposite()) {
-                m_warnings.append({
+                types->warnings.append({
                     QStringLiteral("Found incomplete composite type %1. Do not use qmlplugindump.")
                                       .arg(val.scope->internalName()),
                     QtWarningMsg,
@@ -512,7 +603,8 @@ void QQmlJSImporter::processImport(const QQmlJSScope::Import &importDescription,
  */
 QQmlJSImporter::ImportedTypes QQmlJSImporter::importBuiltins()
 {
-    return builtinImportHelper().qmlNames;
+    auto builtins = builtinImportHelper();
+    return ImportedTypes(std::move(builtins.qmlNames), std::move(builtins.warnings));
 }
 
 
@@ -521,43 +613,39 @@ QQmlJSImporter::AvailableTypes QQmlJSImporter::builtinImportHelper()
     if (m_builtins)
         return *m_builtins;
 
-    AvailableTypes builtins(ImportedTypes(ImportedTypes::INTERNAL, {}, {}));
+    AvailableTypes builtins(QQmlJS::ContextualTypes(QQmlJS::ContextualTypes::INTERNAL, {}, {}));
 
     Import result;
     result.name = QStringLiteral("QML");
 
-    QStringList qmltypesFiles = { QStringLiteral("builtins.qmltypes"),
-                                  QStringLiteral("jsroot.qmltypes") };
-    const auto importBuiltins = [&](const QStringList &imports) {
+    const auto importBuiltins = [&](const QString &qmltypesFile, const QStringList &imports) {
         for (auto const &dir : imports) {
-            QDirIterator it { dir, qmltypesFiles, QDir::NoFilter };
-            while (it.hasNext() && !qmltypesFiles.isEmpty()) {
-                readQmltypes(it.next(), &result.objects, &result.dependencies);
-                qmltypesFiles.removeOne(it.fileName());
-            }
+            const QDir importDir(dir);
+            if (!importDir.exists(qmltypesFile))
+                continue;
+
+            readQmltypes(importDir.filePath(qmltypesFile), &result);
             setQualifiedNamesOn(result);
             importDependencies(result, &builtins);
-
-            if (qmltypesFiles.isEmpty())
-                return;
+            return true;
         }
+
+        return false;
     };
 
-    importBuiltins(m_importPaths);
-    if (!qmltypesFiles.isEmpty()) {
-        const QString pathsString =
-                m_importPaths.isEmpty() ? u"<empty>"_s : m_importPaths.join(u"\n\t");
-        m_warnings.append({ QStringLiteral("Failed to find the following builtins: %1 (so will use "
-                                           "qrc). Import paths used:\n\t%2")
-                                    .arg(qmltypesFiles.join(u", "), pathsString),
-                            QtWarningMsg, QQmlJS::SourceLocation() });
-        importBuiltins({ u":/qt-project.org/qml/builtins"_s }); // use qrc as a "last resort"
+    {
+        // If the same name (such as "Qt") appears in the JS root and in the builtins,
+        // we want the builtins to override the JS root. Therefore, process jsroot first.
+        const QStringList builtinsPath{ u":/qt-project.org/qml/builtins"_s };
+        for (const QString qmltypesFile : { "jsroot.qmltypes"_L1, "builtins.qmltypes"_L1 }) {
+            if (!importBuiltins(qmltypesFile, builtinsPath))
+                qFatal() << u"Failed to find the following builtin:" << qmltypesFile;
+        }
     }
-    Q_ASSERT(qmltypesFiles.isEmpty()); // since qrc must cover it in all the bad cases
 
     // Process them together since there they have interdependencies that wouldn't get resolved
     // otherwise
-    const QQmlJSScope::Import builtinImport(
+    const QQmlJS::Import builtinImport(
                 QString(), QStringLiteral("QML"), QTypeRevision::fromVersion(1, 0), false, true);
 
     QQmlJSScope::ConstPtr intType;
@@ -578,10 +666,10 @@ QQmlJSImporter::AvailableTypes QQmlJSImporter::builtinImportHelper()
     Q_ASSERT(intType);
     Q_ASSERT(arrayType);
 
-    m_builtins = AvailableTypes(
-                ImportedTypes(ImportedTypes::INTERNAL, builtins.cppNames.types(), arrayType));
-    m_builtins->qmlNames
-            = ImportedTypes(ImportedTypes::QML, builtins.qmlNames.types(), arrayType);
+    m_builtins = AvailableTypes(QQmlJS::ContextualTypes(
+            QQmlJS::ContextualTypes::INTERNAL, builtins.cppNames.types(), arrayType));
+    m_builtins->qmlNames = QQmlJS::ContextualTypes(
+            QQmlJS::ContextualTypes::QML, builtins.qmlNames.types(), arrayType);
 
     processImport(builtinImport, result, &(*m_builtins));
 
@@ -591,23 +679,25 @@ QQmlJSImporter::AvailableTypes QQmlJSImporter::builtinImportHelper()
 /*!
  * Imports types from the specified \a qmltypesFiles.
  */
-void QQmlJSImporter::importQmldirs(const QStringList &qmldirFiles)
+QList<QQmlJS::DiagnosticMessage> QQmlJSImporter::importQmldirs(const QStringList &qmldirFiles)
 {
+    QList<QQmlJS::DiagnosticMessage> warnings;
     for (const auto &file : qmldirFiles) {
         Import result;
         QString qmldirName;
         if (file.endsWith(SlashQmldir)) {
             result = readQmldir(file.chopped(SlashQmldir.size()));
+            setQualifiedNamesOn(result);
             qmldirName = file;
         } else {
-            m_warnings.append({
+            warnings.append({
                 QStringLiteral("Argument %1 to -i option is not a qmldir file. Assuming qmltypes.")
                                   .arg(file),
                 QtWarningMsg,
                 QQmlJS::SourceLocation()
             });
 
-            readQmltypes(file, &result.objects, &result.dependencies);
+            readQmltypes(file, &result);
 
             // Append _FAKE_QMLDIR to our made up qmldir name so that if it ever gets used somewhere
             // else except for cache lookups, it will blow up due to a missing file instead of
@@ -615,6 +705,7 @@ void QQmlJSImporter::importQmldirs(const QStringList &qmldirFiles)
             qmldirName = file + QStringLiteral("_FAKE_QMLDIR");
         }
 
+        warnings.append(result.warnings);
         m_seenQmldirFiles.insert(qmldirName, result);
 
         for (const auto &object : std::as_const(result.objects)) {
@@ -625,6 +716,8 @@ void QQmlJSImporter::importQmldirs(const QStringList &qmldirFiles)
             }
         }
     }
+
+    return warnings;
 }
 
 QQmlJSImporter::ImportedTypes QQmlJSImporter::importModule(const QString &module,
@@ -635,11 +728,12 @@ QQmlJSImporter::ImportedTypes QQmlJSImporter::importModule(const QString &module
     const AvailableTypes builtins = builtinImportHelper();
     AvailableTypes result(builtins.cppNames);
     if (!importHelper(module, &result, prefix, version)) {
-        m_warnings.append({
-                              QStringLiteral("Failed to import %1. Are your import paths set up properly?").arg(module),
-                              QtWarningMsg,
-                              QQmlJS::SourceLocation()
-                          });
+        result.warnings.append({
+            QStringLiteral("Failed to import %1. Are your import paths set up properly?")
+                                         .arg(module),
+            QtWarningMsg,
+            QQmlJS::SourceLocation()
+        });
     }
 
     // If we imported a system module add all builtin QML types
@@ -653,12 +747,13 @@ QQmlJSImporter::ImportedTypes QQmlJSImporter::importModule(const QString &module
     if (staticModuleList)
         *staticModuleList << result.staticModules;
 
-    return result.qmlNames;
+    return ImportedTypes(std::move(result.qmlNames), std::move(result.warnings));
 }
 
 QQmlJSImporter::ImportedTypes QQmlJSImporter::builtinInternalNames()
 {
-    return builtinImportHelper().cppNames;
+    auto builtins = builtinImportHelper();
+    return ImportedTypes(std::move(builtins.cppNames), std::move(builtins.warnings));
 }
 
 bool QQmlJSImporter::importHelper(const QString &module, AvailableTypes *types,
@@ -671,7 +766,7 @@ bool QQmlJSImporter::importHelper(const QString &module, AvailableTypes *types,
     if (isDependency)
         Q_ASSERT(prefix.isEmpty());
 
-    const QQmlJSScope::Import cacheKey(prefix, moduleCacheName, version, isFile, isDependency);
+    const QQmlJS::Import cacheKey(prefix, moduleCacheName, version, isFile, isDependency);
 
     auto getTypesFromCache = [&]() -> bool {
         if (!m_cachedImportTypes.contains(cacheKey))
@@ -684,8 +779,10 @@ bool QQmlJSImporter::importHelper(const QString &module, AvailableTypes *types,
         types->hasSystemModule |= cacheEntry->hasSystemModule;
 
         // No need to import qml names for dependencies
-        if (!isDependency)
+        if (!isDependency) {
+            types->warnings.append(cacheEntry->warnings);
             types->qmlNames.addTypes(cacheEntry->qmlNames);
+        }
 
         return true;
     };
@@ -699,8 +796,8 @@ bool QQmlJSImporter::importHelper(const QString &module, AvailableTypes *types,
         return true;
 
     auto cacheTypes = QSharedPointer<QQmlJSImporter::AvailableTypes>(
-                new QQmlJSImporter::AvailableTypes(
-                    ImportedTypes(ImportedTypes::INTERNAL, {}, types->cppNames.arrayType())));
+            new QQmlJSImporter::AvailableTypes(QQmlJS::ContextualTypes(
+                    QQmlJS::ContextualTypes::INTERNAL, {}, types->cppNames.arrayType())));
     m_cachedImportTypes[cacheKey] = cacheTypes;
 
     const QPair<QString, QTypeRevision> importId { module, version };
@@ -795,15 +892,19 @@ bool QQmlJSImporter::importHelper(const QString &module, AvailableTypes *types,
 
 QQmlJSScope::Ptr QQmlJSImporter::localFile2ScopeTree(const QString &filePath)
 {
-    const auto seen = m_importedFiles.find(filePath);
+    const QString sourceFolderFile = preferQmlFilesFromSourceFolder()
+            ? QQmlJSUtils::qmlSourcePathFromBuildPath(m_mapper, filePath)
+            : filePath;
+
+    const auto seen = m_importedFiles.find(sourceFolderFile);
     if (seen != m_importedFiles.end())
         return *seen;
 
-    return *m_importedFiles.insert(filePath, {
-                                       QQmlJSScope::create(),
-                                       QSharedPointer<QDeferredFactory<QQmlJSScope>>(
-                                            new QDeferredFactory<QQmlJSScope>(this, filePath))
-                                   });
+    return *m_importedFiles.insert(
+            sourceFolderFile,
+            { QQmlJSScope::create(),
+              QSharedPointer<QDeferredFactory<QQmlJSScope>>(new QDeferredFactory<QQmlJSScope>(
+                      this, sourceFolderFile)) });
 }
 
 QQmlJSScope::Ptr QQmlJSImporter::importFile(const QString &file)
@@ -815,11 +916,10 @@ QQmlJSImporter::ImportedTypes QQmlJSImporter::importDirectory(
         const QString &directory, const QString &prefix)
 {
     const AvailableTypes builtins = builtinImportHelper();
-    QQmlJSImporter::AvailableTypes types(
-                ImportedTypes(
-                    ImportedTypes::INTERNAL, {}, builtins.cppNames.arrayType()));
+    QQmlJSImporter::AvailableTypes types(QQmlJS::ContextualTypes(
+            QQmlJS::ContextualTypes::INTERNAL, {}, builtins.cppNames.arrayType()));
     importHelper(directory, &types, prefix, QTypeRevision(), false, true);
-    return types.qmlNames;
+    return ImportedTypes(std::move(types.qmlNames), std::move(types.warnings));
 }
 
 void QQmlJSImporter::setImportPaths(const QStringList &importPaths)
@@ -851,27 +951,18 @@ void QQmlJSImporter::setQualifiedNamesOn(const Import &import)
     for (auto &object : import.objects) {
         if (object.exports.isEmpty())
             continue;
-        const QString qualifiedName = QQmlJSScope::qualifiedNameFrom(
-                    import.name, object.exports.first().type(),
-                    object.exports.first().revision(),
-                    object.exports.last().revision());
         if (auto *factory = object.scope.factory()) {
-            factory->setQualifiedName(qualifiedName);
             factory->setModuleName(import.name);
         } else {
-            object.scope->setQualifiedName(qualifiedName);
-            object.scope->setModuleName(import.name);
+            object.scope->setOwnModuleName(import.name);
         }
     }
 }
 
-std::unique_ptr<QQmlJSImportVisitor>
-QQmlJSImporter::makeImportVisitor(const QQmlJSScope::Ptr &target, QQmlJSImporter *importer,
-                                  QQmlJSLogger *logger, const QString &implicitImportDirectory,
-                                  const QStringList &qmldirFiles)
+void QQmlJSImporter::runImportVisitor(QQmlJS::AST::Node *rootNode,
+                                      const ImportVisitorPrerequisites &p)
 {
-    return std::unique_ptr<QQmlJSImportVisitor>(
-            m_createImportVisitor(target, importer, logger, implicitImportDirectory, qmldirFiles));
+    m_importVisitor(rootNode, this, p);
 }
 
 QT_END_NAMESPACE
