@@ -122,7 +122,6 @@ QObject *QQmlTableInstanceModel::object(int index, QQmlIncubator::IncubationMode
 {
     Q_ASSERT(m_delegate);
     Q_ASSERT(index >= 0 && index < m_adaptorModel.count());
-    Q_ASSERT(m_qmlContext && m_qmlContext->isValid());
 
     QQmlDelegateModelItem *modelItem = resolveModelItem(index);
     if (!modelItem)
@@ -165,6 +164,9 @@ QQmlInstanceModel::ReleaseFlags QQmlTableInstanceModel::release(QObject *object,
     Q_ASSERT(object);
     auto modelItem = qvariant_cast<QQmlDelegateModelItem *>(object->property(kModelItemTag));
     Q_ASSERT(modelItem);
+    // Ensure that the object was incubated by this QQmlTableInstanceModel
+    Q_ASSERT(m_modelItems.contains(modelItem->index));
+    Q_ASSERT(m_modelItems[modelItem->index]->object == object);
 
     if (!modelItem->releaseObject())
         return QQmlDelegateModel::Referenced;
@@ -214,6 +216,9 @@ void QQmlTableInstanceModel::dispose(QObject *object)
     // The item is not referenced by anyone
     Q_ASSERT(!modelItem->isObjectReferenced());
     Q_ASSERT(!modelItem->isReferenced());
+    // Ensure that the object was incubated by this QQmlTableInstanceModel
+    Q_ASSERT(m_modelItems.contains(modelItem->index));
+    Q_ASSERT(m_modelItems[modelItem->index]->object == object);
 
     m_modelItems.remove(modelItem->index);
 
@@ -284,7 +289,7 @@ void QQmlTableInstanceModel::incubateModelItem(QQmlDelegateModelItem *modelItem,
         const bool sync = (incubationMode == QQmlIncubator::Synchronous || incubationMode == QQmlIncubator::AsynchronousIfNested);
         if (sync && modelItem->incubationTask->incubationMode() == QQmlIncubator::Asynchronous)
             modelItem->incubationTask->forceCompletion();
-    } else {
+    } else if (m_qmlContext && m_qmlContext->isValid()) {
         modelItem->incubationTask = new QQmlTableInstanceModelIncubationTask(this, modelItem, incubationMode);
 
         QQmlContext *creationContext = modelItem->delegate->creationContext();
@@ -391,8 +396,8 @@ bool QQmlTableInstanceModel::setRequiredProperty(int index, const QString &name,
 
     bool wasInRequired = false;
     const auto task = QQmlIncubatorPrivate::get(modelItem->incubationTask);
-    RequiredProperties &props = task->requiredProperties();
-    if (props.empty())
+    RequiredProperties *props = task->requiredProperties();
+    if (props->empty())
         return false;
 
     QQmlProperty componentProp = QQmlComponentPrivate::removePropertyFromRequired(
@@ -430,11 +435,15 @@ void QQmlTableInstanceModel::setModel(const QVariant &model)
     // needs to stay in sync with the model. So we need to drain the pool
     // completely when the model changes.
     drainReusableItemsPool(0);
-    if (auto const aim = abstractItemModel())
+    if (auto const aim = abstractItemModel()) {
         disconnect(aim, &QAbstractItemModel::dataChanged, this, &QQmlTableInstanceModel::dataChangedCallback);
+        disconnect(aim, &QAbstractItemModel::modelAboutToBeReset, this, &QQmlTableInstanceModel::modelAboutToBeResetCallback);
+    }
     m_adaptorModel.setModel(model);
-    if (auto const aim = abstractItemModel())
+    if (auto const aim = abstractItemModel()) {
         connect(aim, &QAbstractItemModel::dataChanged, this, &QQmlTableInstanceModel::dataChangedCallback);
+        connect(aim, &QAbstractItemModel::modelAboutToBeReset, this, &QQmlTableInstanceModel::modelAboutToBeResetCallback);
+    }
 }
 
 void QQmlTableInstanceModel::dataChangedCallback(const QModelIndex &begin, const QModelIndex &end, const QVector<int> &roles)
@@ -452,6 +461,21 @@ void QQmlTableInstanceModel::dataChangedCallback(const QModelIndex &begin, const
     }
 }
 
+void QQmlTableInstanceModel::modelAboutToBeResetCallback()
+{
+    // When the model is reset, we can no longer rely on any of the data it has
+    // provided us so far. Normally it's enough for the view to recreate all the
+    // delegate items in that case, except if the model roles has changed as well
+    // (since those are cached by QQmlAdaptorModel / Accessors). For the latter case, we
+    // simply set the model once more in the delegate model to rebuild everything.
+    auto const aim = abstractItemModel();
+    auto oldRoleNames = aim->roleNames();
+    QObject::connect(aim, &QAbstractItemModel::modelReset, this, [this, aim, oldRoleNames](){
+        if (oldRoleNames != aim->roleNames())
+            setModel(model());
+    }, Qt::SingleShotConnection);
+}
+
 QQmlComponent *QQmlTableInstanceModel::delegate() const
 {
     return m_delegate;
@@ -486,7 +510,7 @@ void QQmlTableInstanceModelIncubationTask::setInitialState(QObject *object)
     modelItemToIncubate->object = object;
     emit tableInstanceModel->initItem(modelItemToIncubate->index, object);
 
-    if (!QQmlIncubatorPrivate::get(this)->requiredProperties().empty()) {
+    if (!QQmlIncubatorPrivate::get(this)->requiredProperties()->empty()) {
         modelItemToIncubate->object = nullptr;
         object->deleteLater();
     }
