@@ -441,6 +441,10 @@ QAccessible::State QAccessibleQuickItem::state() const
     if (role() == QAccessible::EditableText)
         if (auto ti = qobject_cast<QQuickTextInput *>(item()))
             state.passwordEdit = ti->echoMode() != QQuickTextInput::Normal;
+    if (!item()->isEnabled()) {
+        state.focusable = false;
+        state.disabled = true;
+    }
     return state;
 }
 
@@ -452,7 +456,7 @@ QAccessible::Role QAccessibleQuickItem::role() const
 
     QAccessible::Role role = QAccessible::NoRole;
     if (item())
-        role = QQuickItemPrivate::get(item())->accessibleRole();
+        role = QQuickItemPrivate::get(item())->effectiveAccessibleRole();
     if (role == QAccessible::NoRole) {
         if (qobject_cast<QQuickText*>(const_cast<QQuickItem *>(item())))
             role = QAccessible::StaticText;
@@ -600,6 +604,20 @@ QString QAccessibleQuickItem::text(QAccessible::Text textType) const
         if (!accessibleDecription.isNull())
             return accessibleDecription.toString();
         break;}
+    case QAccessible::Identifier: {
+        QVariant accessibleIdentifier = QQuickAccessibleAttached::property(object(), "id");
+        if (!accessibleIdentifier.isNull())
+            return accessibleIdentifier.toString();
+        auto quickItem = item();
+        if (quickItem->isComponentComplete()) {
+            QQmlContext *context = qmlContext(quickItem);
+            if (context) {
+                const auto objectId = context->nameForObject(quickItem);
+                if (!objectId.isEmpty())
+                    return objectId;
+            }
+        }
+        break;}
 #ifdef Q_ACCESSIBLE_QUICK_ITEM_ENABLE_DEBUG_DESCRIPTION
     case QAccessible::DebugDescription: {
         QString debugString;
@@ -618,6 +636,8 @@ QString QAccessibleQuickItem::text(QAccessible::Text textType) const
     // the following block handles item-specific behavior
     if (role() == QAccessible::EditableText) {
         if (textType == QAccessible::Value) {
+            if (auto textInput = qobject_cast<QQuickTextInput *>(item()))
+                return textInput->displayText();
             if (QTextDocument *doc = textDocument()) {
                 return doc->toPlainText();
             }
