@@ -1,41 +1,5 @@
-/****************************************************************************
-**
-** Copyright (C) 2018 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the QtQuick module of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:LGPL$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 3 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL3 included in the
-** packaging of this file. Please review the following information to
-** ensure the GNU Lesser General Public License version 3 requirements
-** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 2.0 or (at your option) the GNU General
-** Public license version 3 or any later version approved by the KDE Free
-** Qt Foundation. The licenses are as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-2.0.html and
-** https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2018 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
 #ifndef QQUICKTABLEVIEW_P_P_H
 #define QQUICKTABLEVIEW_P_P_H
@@ -54,7 +18,6 @@
 #include "qquicktableview_p.h"
 
 #include <QtCore/qtimer.h>
-#include <QtCore/private/qflatmap_p.h>
 #include <QtCore/qitemselectionmodel.h>
 #include <QtQmlModels/private/qqmltableinstancemodel_p.h>
 #include <QtQml/private/qqmlincubator_p.h>
@@ -63,7 +26,18 @@
 
 #include <QtQuick/private/qquickflickable_p_p.h>
 #include <QtQuick/private/qquickitemviewfxitem_p_p.h>
+#include <QtQuick/private/qquickanimation_p.h>
 #include <QtQuick/private/qquickselectable_p.h>
+#include <QtQuick/private/qquicksinglepointhandler_p.h>
+#include <QtQuick/private/qquickhoverhandler_p.h>
+#include <QtQuick/private/qquicktaphandler_p.h>
+
+#include <QtCore/private/qminimalflatset_p.h>
+
+#if QT_CONFIG(quick_draganddrop)
+#include <QtGui/qdrag.h>
+#include <QtQuick/private/qquickdroparea_p.h>
+#endif
 
 QT_BEGIN_NAMESPACE
 
@@ -71,33 +45,147 @@ Q_DECLARE_LOGGING_CATEGORY(lcTableViewDelegateLifecycle)
 
 static const qreal kDefaultRowHeight = 50;
 static const qreal kDefaultColumnWidth = 50;
+static const int kEdgeIndexNotSet = -2;
+static const int kEdgeIndexAtEnd = -3;
 
 class FxTableItem;
 class QQuickTableSectionSizeProviderPrivate;
 
-class Q_QUICK_PRIVATE_EXPORT QQuickTableSectionSizeProvider : public QObject {
+/*! \internal
+ *  TableView uses QQuickTableViewHoverHandler to track where the pointer is
+ *  on top of the table, and change the cursor at the places where a drag
+ *  would start a resize of a row or a column.
+ */
+class QQuickTableViewHoverHandler : public QQuickHoverHandler
+{
     Q_OBJECT
 
 public:
-    QQuickTableSectionSizeProvider(QObject *parent=nullptr);
-    void setSize(int section, qreal size);
-    qreal size(int section);
-    bool resetSize(int section);
-    void resetAll();
+    QQuickTableViewHoverHandler(QQuickTableView *view);
+    inline bool isHoveringGrid() const { return m_row != -1 || m_column != -1; };
 
-Q_SIGNALS:
-    void sizeChanged();
+    int m_row = -1;
+    int m_column = -1;
 
-private:
-    Q_DISABLE_COPY(QQuickTableSectionSizeProvider)
-    Q_DECLARE_PRIVATE(QQuickTableSectionSizeProvider)
+    friend class QQuickTableViewPrivate;
+
+protected:
+    void handleEventPoint(QPointerEvent *event, QEventPoint &point) override;
 };
 
-class Q_QUICK_PRIVATE_EXPORT QQuickTableViewPrivate : public QQuickFlickablePrivate, public QQuickSelectable
+class QQuickTableViewPointerHandler : public QQuickSinglePointHandler
 {
-    Q_DECLARE_PUBLIC(QQuickTableView)
+    Q_OBJECT
 
 public:
+    enum State {
+        Listening, // the pointer is not being pressed between the cells
+        Tracking, // the pointer is being pressed between the cells
+        DraggingStarted, // dragging started
+        Dragging, // a drag is ongoing
+        DraggingFinished // dragging was finished
+    };
+
+    QQuickTableViewPointerHandler(QQuickTableView *view);
+
+    State m_state = Listening;
+    State state() { return m_state; }
+
+protected:
+    bool wantsEventPoint(const QPointerEvent *event, const QEventPoint &point) override;
+};
+
+/*! \internal
+ *  TableView uses QQuickTableViewResizeHandler to enable the user to resize
+ *  rows and columns. By using a custom pointer handler, we can get away with
+ *  using a single pointer handler for the whole content item, rather than
+ *  e.g having to split it up into multiple items with drag handlers placed
+ *  between the cells.
+ */
+class QQuickTableViewResizeHandler : public QQuickTableViewPointerHandler
+{
+    Q_OBJECT
+
+public:
+    QQuickTableViewResizeHandler(QQuickTableView *view);
+
+    int m_row = -1;
+    qreal m_rowStartY = -1;
+    qreal m_rowStartHeight = -1;
+
+    int m_column = -1;
+    qreal m_columnStartX = -1;
+    qreal m_columnStartWidth = -1;
+
+    void updateState(QEventPoint &point);
+    void updateDrag(QPointerEvent *event, QEventPoint &point);
+
+    friend class QQuickTableViewPrivate;
+
+protected:
+    void handleEventPoint(QPointerEvent *event, QEventPoint &point) override;
+    void onGrabChanged(QQuickPointerHandler *grabber, QPointingDevice::GrabTransition transition,
+                       QPointerEvent *ev, QEventPoint &point) override;
+};
+
+#if QT_CONFIG(quick_draganddrop)
+class QQuickTableViewSectionDragHandler : public QQuickTableViewPointerHandler
+{
+    Q_OBJECT
+
+public:
+    QQuickTableViewSectionDragHandler(QQuickTableView *view);
+    ~QQuickTableViewSectionDragHandler();
+
+    void grabSection();
+
+    void handleDrag(QQuickDragEvent *event);
+    void handleDrop(QQuickDragEvent *event);
+    void handleDragDropAction(Qt::DropAction action);
+
+    void setSectionOrientation(Qt::Orientation orientation) { m_sectionOrientation = orientation; }
+
+    friend class QQuickTableViewPrivate;
+
+protected:
+    void handleEventPoint(QPointerEvent *event, QEventPoint &point) override;
+
+private:
+    void resetDragData();
+    void resetSectionOverlay();
+
+    QSharedPointer<QQuickItemGrabResult> m_grabResult;
+    QPointer<QDrag> m_drag;
+    int m_source = -1;
+    int m_destination = -1;
+    QPointer<QQuickDropArea> m_dropArea;
+    Qt::Orientation m_sectionOrientation;
+
+    QPointF m_dragPoint;
+    QSizeF m_step = QSizeF(1, 1);
+    QTimer m_scrollTimer;
+};
+#endif // quick_draganddrop
+
+/*! \internal
+ *  QQuickTableViewTapHandler used to handle tap events explicitly for table view
+ */
+class QQuickTableViewTapHandler : public QQuickTapHandler
+{
+    Q_OBJECT
+
+public:
+    explicit QQuickTableViewTapHandler(QQuickTableView *view);
+    bool wantsEventPoint(const QPointerEvent *event, const QEventPoint &point) override;
+
+    friend class QQuickTableViewPrivate;
+};
+
+class Q_QUICK_EXPORT QQuickTableViewPrivate : public QQuickFlickablePrivate, public QQuickSelectable
+{
+public:
+    Q_DECLARE_PUBLIC(QQuickTableView)
+
     class TableEdgeLoadRequest
     {
         // Whenever we need to load new rows or columns in the
@@ -138,7 +226,7 @@ public:
         inline bool isActive() const { return m_active; }
 
         inline QPoint currentCell() const { return cellAt(m_currentIndex); }
-        inline bool hasCurrentCell() const { return m_currentIndex < m_visibleCellsInEdge.count(); }
+        inline bool hasCurrentCell() const { return m_currentIndex < m_visibleCellsInEdge.size(); }
         inline void moveToNextCell() { ++m_currentIndex; }
 
         inline Qt::Edge edge() const { return m_edge; }
@@ -201,15 +289,19 @@ public:
         LoadInitalTable,
         VerifyTable,
         LayoutTable,
-        LoadAndUnloadAfterLayout,
-        CancelOvershootBottomRight,
-        CancelOvershootTopLeft,
+        CancelOvershoot,
+        UpdateContentSize,
         PreloadColumns,
         PreloadRows,
         MovePreloadedItemsToPool,
         Done
     };
 
+    enum class SectionState {
+        Idle = 0,
+        Moving
+    };
+
     enum class RebuildOption {
         None = 0,
         All = 0x1,
@@ -247,7 +339,7 @@ public:
     // When the applications assignes a new model or delegate to the view, we keep them
     // around until we're ready to take them into use (syncWithPendingChanges).
     QVariant assignedModel = QVariant(int(0));
-    QQmlComponent *assignedDelegate = nullptr;
+    QQmlGuard<QQmlComponent> assignedDelegate;
 
     // loadedRows/Columns describes the rows and columns that are currently loaded (from top left
     // row/column to bottom right row/column). loadedTableOuterRect describes the actual
@@ -255,8 +347,8 @@ public:
     // we need to fill up with more rows/columns. loadedTableInnerRect describes the pixels
     // that the loaded table covers if you remove one row/column on each side of the table, and
     // is used to determine rows/columns that are no longer visible and can be unloaded.
-    QFlatMap<int, int> loadedColumns;
-    QFlatMap<int, int> loadedRows;
+    QMinimalFlatSet<int> loadedColumns;
+    QMinimalFlatSet<int> loadedRows;
     QRectF loadedTableOuterRect;
     QRectF loadedTableInnerRect;
 
@@ -278,13 +370,22 @@ public:
     QQmlTableInstanceModel::ReusableFlag reusableFlag = QQmlTableInstanceModel::Reusable;
 
     bool blockItemCreatedCallback = false;
-    bool layoutWarningIssued = false;
+    mutable bool layoutWarningIssued = false;
     bool polishing = false;
     bool syncVertically = false;
     bool syncHorizontally = false;
     bool inSetLocalViewportPos = false;
     bool inSyncViewportPosRecursive = false;
     bool inUpdateContentSize = false;
+    bool animate = true;
+    bool keyNavigationEnabled = true;
+    bool pointerNavigationEnabled = true;
+    bool alternatingRows = true;
+    bool resizableColumns = false;
+    bool resizableRows = false;
+#if QT_CONFIG(cursor)
+    bool m_cursorSet = false;
+#endif
 
     // isTransposed is currently only used by HeaderView.
     // Consider making it public.
@@ -294,12 +395,10 @@ public:
 
     QJSValue rowHeightProvider;
     QJSValue columnWidthProvider;
-    QQuickTableSectionSizeProvider rowHeights;
-    QQuickTableSectionSizeProvider columnWidths;
 
-    EdgeRange cachedNextVisibleEdgeIndex[4];
-    EdgeRange cachedColumnWidth;
-    EdgeRange cachedRowHeight;
+    mutable EdgeRange cachedNextVisibleEdgeIndex[4];
+    mutable EdgeRange cachedColumnWidth;
+    mutable EdgeRange cachedRowHeight;
 
     // TableView uses contentWidth/height to report the size of the table (this
     // will e.g make scrollbars written for Flickable work out of the box). This
@@ -321,38 +420,73 @@ public:
     Qt::Orientations assignedSyncDirection = Qt::Horizontal | Qt::Vertical;
 
     QPointer<QItemSelectionModel> selectionModel;
+    QQuickTableView::SelectionBehavior selectionBehavior = QQuickTableView::SelectCells;
+    QQuickTableView::SelectionMode selectionMode = QQuickTableView::ExtendedSelection;
+    QItemSelectionModel::SelectionFlag selectionFlag = QItemSelectionModel::NoUpdate;
+    std::function<void(CallBackFlag)> selectableCallbackFunction;
+    bool inSelectionModelUpdate = false;
 
-    int assignedPositionViewAtRow = 0;
-    int assignedPositionViewAtColumn = 0;
-    int positionViewAtRow = 0;
-    int positionViewAtColumn = 0;
+    int assignedPositionViewAtRowAfterRebuild = 0;
+    int assignedPositionViewAtColumnAfterRebuild = 0;
+    int positionViewAtRowAfterRebuild = 0;
+    int positionViewAtColumnAfterRebuild = 0;
     qreal positionViewAtRowOffset = 0;
     qreal positionViewAtColumnOffset = 0;
+    QRectF positionViewAtRowSubRect;
+    QRectF positionViewAtColumnSubRect;
     Qt::Alignment positionViewAtRowAlignment = Qt::AlignTop;
     Qt::Alignment positionViewAtColumnAlignment = Qt::AlignLeft;
 
-    QPoint selectionStartCell;
-    QPoint selectionEndCell;
-    QRectF selectionStartCellRect;
-    QRectF selectionEndCellRect;
+    QQuickPropertyAnimation positionXAnimation;
+    QQuickPropertyAnimation positionYAnimation;
+
+    QPoint selectionStartCell = {-1, -1};
+    QPoint selectionEndCell = {-1, -1};
+    QItemSelection existingSelection;
 
     QMargins edgesBeforeRebuild;
+    QSize tableSizeBeforeRebuild;
 
-    const static QPoint kLeft;
-    const static QPoint kRight;
-    const static QPoint kUp;
-    const static QPoint kDown;
+    int currentRow = -1;
+    int currentColumn = -1;
+
+    QHash<int, qreal> explicitColumnWidths;
+    QHash<int, qreal> explicitRowHeights;
+
+    QQuickTableViewHoverHandler *hoverHandler = nullptr;
+    QQuickTableViewResizeHandler *resizeHandler = nullptr;
+#if QT_CONFIG(quick_draganddrop)
+    QQuickTableViewSectionDragHandler *sectionDragHandler = nullptr;
+#endif
+    QQuickTableViewPointerHandler *activePtrHandler = nullptr;
+
+    QQmlTableInstanceModel *editModel = nullptr;
+    QQuickItem *editItem = nullptr;
+    QPersistentModelIndex editIndex;
+    QQuickTableView::EditTriggers editTriggers = QQuickTableView::DoubleTapped | QQuickTableView::EditKeyPressed;
 
 #ifdef QT_DEBUG
     QString forcedIncubationMode = qEnvironmentVariable("QT_TABLEVIEW_INCUBATION_MODE");
 #endif
 
+    struct SectionData {
+        int index = -1;
+        int prevIndex = -1;
+    };
+
+    QList<SectionData> visualIndices[Qt::Vertical];
+    QList<SectionData> logicalIndices[Qt::Vertical];
+
+    SectionState m_sectionState = SectionState::Idle;
+
 public:
+    void init();
+
     QQuickTableViewAttached *getAttachedObject(const QObject *object) const;
 
     int modelIndexAtCell(const QPoint &cell) const;
     QPoint cellAtModelIndex(int modelIndex) const;
-    int modelIndexToCellIndex(const QModelIndex &modelIndex) const;
+    int modelIndexToCellIndex(const QModelIndex &modelIndex, bool visualIndex = true) const;
     inline bool cellIsValid(const QPoint &cell) const { return cell.x() != -1 && cell.y() != -1; }
 
     qreal sizeHintForColumn(int column) const;
@@ -360,20 +494,24 @@ public:
     QSize calculateTableSize();
     void updateTableSize();
 
-    inline bool isColumnHidden(int column);
-    inline bool isRowHidden(int row);
+    inline bool isColumnHidden(int column) const;
+    inline bool isRowHidden(int row) const;
 
     qreal getColumnLayoutWidth(int column);
     qreal getRowLayoutHeight(int row);
-    qreal getColumnWidth(int column);
-    qreal getRowHeight(int row);
+    qreal getColumnWidth(int column) const;
+    qreal getRowHeight(int row) const;
+    qreal getEffectiveRowY(int row) const;
     qreal getEffectiveRowHeight(int row) const;
+    qreal getEffectiveColumnX(int column) const;
     qreal getEffectiveColumnWidth(int column) const;
+    qreal getAlignmentContentX(int column, Qt::Alignment alignment, const qreal offset, const QRectF &subRect);
+    qreal getAlignmentContentY(int row, Qt::Alignment alignment, const qreal offset, const QRectF &subRect);
 
-    inline int topRow() const { return loadedRows.cbegin().key(); }
-    inline int bottomRow() const { return (--loadedRows.cend()).key(); }
-    inline int leftColumn() const { return loadedColumns.cbegin().key(); }
-    inline int rightColumn() const { return (--loadedColumns.cend()).key(); }
+    int topRow() const { return *loadedRows.cbegin(); }
+    int bottomRow() const { return *loadedRows.crbegin(); }
+    int leftColumn() const { return *loadedColumns.cbegin(); }
+    int rightColumn() const { return *loadedColumns.crbegin(); }
 
     QQuickTableView *rootSyncView() const;
 
@@ -391,18 +529,18 @@ public:
     void updateAverageColumnWidth();
     void updateAverageRowHeight();
     RebuildOptions checkForVisibilityChanges();
-    void forceLayout();
+    void forceLayout(bool immediate);
 
     void updateExtents();
     void syncLoadedTableRectFromLoadedTable();
     void syncLoadedTableFromLoadRequest();
     void shiftLoadedTableRect(const QPointF newPosition);
 
-    int nextVisibleEdgeIndex(Qt::Edge edge, int startIndex);
-    int nextVisibleEdgeIndexAroundLoadedTable(Qt::Edge edge);
-    bool allColumnsLoaded();
-    bool allRowsLoaded();
-    inline int edgeToArrayIndex(Qt::Edge edge);
+    int nextVisibleEdgeIndex(Qt::Edge edge, int startIndex) const;
+    int nextVisibleEdgeIndexAroundLoadedTable(Qt::Edge edge) const;
+    inline bool atTableEnd(Qt::Edge edge) const { return nextVisibleEdgeIndexAroundLoadedTable(edge) == kEdgeIndexAtEnd; }
+    inline bool atTableEnd(Qt::Edge edge, int startIndex) const { return nextVisibleEdgeIndex(edge, startIndex) == kEdgeIndexAtEnd; }
+    inline int edgeToArrayIndex(Qt::Edge edge) const;
     void clearEdgeSizeCache();
 
     bool canLoadTableEdge(Qt::Edge tableEdge, const QRectF fillRect) const;
@@ -423,7 +561,7 @@ public:
     void unloadItem(const QPoint &cell);
     void loadEdge(Qt::Edge edge, QQmlIncubator::IncubationMode incubationMode);
     void unloadEdge(Qt::Edge edge);
-    void loadAndUnloadVisibleEdges();
+    void loadAndUnloadVisibleEdges(QQmlIncubator::IncubationMode incubationMode = QQmlIncubator::AsynchronousIfNested);
     void drainReusePoolAfterLoadRequest();
     void processLoadRequest();
 
@@ -435,20 +573,25 @@ public:
     void layoutAfterLoadingInitialTable();
     void adjustViewportXAccordingToAlignment();
     void adjustViewportYAccordingToAlignment();
-
-    void cancelOvershootBottomRight();
-    void cancelOvershootTopLeft();
+    void cancelOvershootAfterLayout();
 
     void scheduleRebuildTable(QQuickTableViewPrivate::RebuildOptions options);
 
+#if QT_CONFIG(cursor)
+    void updateCursor();
+#endif
+    void updateEditItem();
+    void updateContentSize();
+
     QTypeRevision resolveImportVersion();
     void createWrapperModel();
+    QAbstractItemModel *qaim(QVariant modelAsVariant) const;
 
-    void initItemCallback(int modelIndex, QObject *item);
-    void itemCreatedCallback(int modelIndex, QObject *object);
-    void itemPooledCallback(int modelIndex, QObject *object);
-    void itemReusedCallback(int modelIndex, QObject *object);
-    void modelUpdated(const QQmlChangeSet &changeSet, bool reset);
+    virtual void initItemCallback(int modelIndex, QObject *item);
+    virtual void itemCreatedCallback(int modelIndex, QObject *object);
+    virtual void itemPooledCallback(int modelIndex, QObject *object);
+    virtual void itemReusedCallback(int modelIndex, QObject *object);
+    virtual void modelUpdated(const QQmlChangeSet &changeSet, bool reset);
 
     virtual void syncWithPendingChanges();
     virtual void syncDelegate();
@@ -457,6 +600,7 @@ public:
     virtual void syncModel();
     virtual void syncSyncView();
     virtual void syncPositionView();
+    virtual QAbstractItemModel *selectionSourceModel();
     inline void syncRebuildOptions();
 
     void connectToModel();
@@ -470,6 +614,12 @@ public:
     void columnsRemovedCallback(const QModelIndex &parent, int begin, int end);
     void layoutChangedCallback(const QList<QPersistentModelIndex> &parents, QAbstractItemModel::LayoutChangeHint hint);
     void modelResetCallback();
+    bool compareModel(const QVariant& model1, const QVariant& model2) const;
+
+    void positionViewAtRow(int row, Qt::Alignment alignment, qreal offset, const QRectF subRect = QRectF());
+    void positionViewAtColumn(int column, Qt::Alignment alignment, qreal offset, const QRectF subRect = QRectF());
+    bool scrollToRow(int row, Qt::Alignment alignment, qreal offset, const QRectF subRect = QRectF());
+    bool scrollToColumn(int column, Qt::Alignment alignment, qreal offset, const QRectF subRect = QRectF());
 
     void scheduleRebuildIfFastFlick();
     void setLocalViewportX(qreal contentX);
@@ -478,10 +628,14 @@ public:
     void syncViewportPosRecursive();
 
     bool selectedInSelectionModel(const QPoint &cell) const;
-    void selectionChangedInSelectionModel(const QItemSelection &selected, const QItemSelection &deselected) const;
-    void updateSelectedOnAllDelegateItems() const;
-    void setSelectedOnDelegateItem(const QModelIndex &modelIndex, bool select) const;
-    void setSelectedOnDelegateItem(QQuickItem *delegateItem, bool select) const;
+    void selectionChangedInSelectionModel(const QItemSelection &selected, const QItemSelection &deselected);
+    void updateSelectedOnAllDelegateItems();
+    void setSelectedOnDelegateItem(const QModelIndex &modelIndex, bool select);
+
+    bool currentInSelectionModel(const QPoint &cell) const;
+    void currentChangedInSelectionModel(const QModelIndex &current, const QModelIndex &previous);
+    void setCurrentOnDelegateItem(const QModelIndex &index, bool isCurrent);
+    void updateCurrentRowAndColumn();
 
     void fetchMoreData();
 
@@ -491,19 +645,54 @@ public:
     inline QString tableLayoutToString() const;
     void dumpTable() const;
 
+    void setRequiredProperty(const char *property,
+                             const QVariant &value,
+                             int serializedModelIndex,
+                             QObject *object, bool init);
+
+    void handleTap(const QQuickHandlerPoint &point);
+    void setCurrentIndexFromTap(const QPointF &pos);
+    void setCurrentIndex(const QPoint &cell);
+    bool setCurrentIndexFromKeyEvent(QKeyEvent *e);
+    bool canEdit(const QModelIndex tappedIndex, bool warn);
+    bool editFromKeyEvent(QKeyEvent *e);
+
     // QQuickSelectable
     QQuickItem *selectionPointerHandlerTarget() const override;
+    bool hasSelection() const override;
+    bool startSelection(const QPointF &pos, Qt::KeyboardModifiers modifiers) override;
     void setSelectionStartPos(const QPointF &pos) override;
     void setSelectionEndPos(const QPointF &pos) override;
     void clearSelection() override;
     void normalizeSelection() override;
     QRectF selectionRectangle() const override;
-    QSizeF scrollTowardsSelectionPoint(const QPointF &pos, const QSizeF &step) override;
+    QSizeF scrollTowardsPoint(const QPointF &pos, const QSizeF &step) override;
+    void setCallback(std::function<void(CallBackFlag)> func) override;
+    void cancelSelectionTracking();
 
     QPoint clampedCellAtPos(const QPointF &pos) const;
-    void updateSelection(const QRect &oldSelection, const QRect &newSelection);
+    virtual void updateSelection(const QRect &oldSelection, const QRect &newSelection);
     QRect selection() const;
     // ----------------
+
+    // Section drag handler
+#if QT_CONFIG(quick_draganddrop)
+    void initSectionDragHandler(Qt::Orientation orientation);
+    void destroySectionDragHandler();
+#endif
+    inline void setActivePointerHandler(QQuickTableViewPointerHandler *handler) { activePtrHandler = handler; }
+    inline QQuickTableViewPointerHandler* activePointerHandler() const { return activePtrHandler; }
+    // Row/Column reordering
+    void moveSection(int source , int destination, Qt::Orientations orientation);
+    void initializeIndexMapping();
+    void clearIndexMapping();
+    void clearSection(Qt::Orientations orientation);
+    virtual int logicalRowIndex(const int visualIndex) const;
+    virtual int logicalColumnIndex(const int visualIndex) const;
+    virtual int visualRowIndex(const int logicalIndex) const;
+    virtual int visualColumnIndex(const int logicalIndex) const;
+    void setContainsDragOnDelegateItem(const QModelIndex &modelIndex, bool overlay);
+    int getEditCellIndex(const QModelIndex &index) const;
 };
 
 class FxTableItem : public QQuickItemViewFxItem
