@@ -9,6 +9,7 @@
 #include <QtQuick/private/qquickitemchangelistener_p.h>
 #include <QtQuickTemplates2/private/qquickpopup_p.h>
 #include <QtQuickTemplates2/private/qquickpopupitem_p_p.h>
+#include <QtQuickTemplates2/private/qquickpopupwindow_p_p.h>
 
 QT_BEGIN_NAMESPACE
 
@@ -44,6 +45,9 @@ Q_LOGGING_CATEGORY(lcAttached, "qt.quick.controls.attachedpropertypropagator")
     \li Call \l initialize() in the constructor
     \li Define set/inherit/propagate/reset functions for each property as needed
     \li Reimplement \l attachedParentChange() to handle property inheritance
+    \li Implement a static \c qmlAttachedProperties function and declare the
+        type as an attached QML type with \l QML_ELEMENT and \l QML_ATTACHED,
+        as detailed in \l {Providing Attached Properties}
     \endlist
 
     For an example that demonstrates this in depth, see
@@ -73,6 +77,7 @@ static QQuickAttachedPropertyPropagator *attachedObject(const QMetaObject *type,
 */
 static QQuickAttachedPropertyPropagator *findAttachedParent(const QMetaObject *ourAttachedType, QObject *objectWeAreAttachedTo)
 {
+    qCDebug(lcAttached).noquote() << "findAttachedParent called with" << ourAttachedType->className() << objectWeAreAttachedTo;
     /*
         In the Material ComboBox.qml, we have code like this:
 
@@ -94,37 +99,61 @@ static QQuickAttachedPropertyPropagator *findAttachedParent(const QMetaObject *o
     */
     auto popupItem = qobject_cast<QQuickPopupItem *>(objectWeAreAttachedTo);
     if (popupItem) {
+        qCDebug(lcAttached).noquote() << "- attachee belongs to popup item" << popupItem << "- checking if it has an attached object";
         auto popupItemPrivate = QQuickPopupItemPrivate::get(popupItem);
         QQuickAttachedPropertyPropagator *popupAttached = attachedObject(ourAttachedType, popupItemPrivate->popup);
-        if (popupAttached)
+        if (popupAttached) {
+            qCDebug(lcAttached).noquote() << "- popup item has attached object" << popupAttached << "- returning";
             return popupAttached;
+        } else {
+            qCDebug(lcAttached).noquote() << "- popup item does not have attached object";
+        }
+    } else {
+        qCDebug(lcAttached).noquote() << "- attachee does not belong to a popup";
     }
 
     QQuickItem *item = qobject_cast<QQuickItem *>(objectWeAreAttachedTo);
     if (item) {
+        qCDebug(lcAttached).noquote() << "- attachee is an item; checking its parent items and popups";
         // lookup parent items and popups
         QQuickItem *parent = item->parentItem();
         while (parent) {
+            qCDebug(lcAttached).noquote() << "  - checking parent item" << parent;
             QQuickAttachedPropertyPropagator *attached = attachedObject(ourAttachedType, parent);
-            if (attached)
+            if (attached) {
+                qCDebug(lcAttached).noquote() << "  - parent item has attached object" << attached << "- returning";
                 return attached;
+            }
 
             QQuickPopup *popup = qobject_cast<QQuickPopup *>(parent->parent());
-            if (popup)
+            if (popup) {
+                qCDebug(lcAttached).noquote() << "  - parent popup has attached object" << attached << "- returning";
                 return attachedObject(ourAttachedType, popup);
+            }
 
             parent = parent->parentItem();
         }
 
         // fallback to item's window
+        qCDebug(lcAttached).noquote() << "- checking parent window" << item->window();
         QQuickAttachedPropertyPropagator *attached = attachedObject(ourAttachedType, item->window());
-        if (attached)
+        if (attached) {
+            qCDebug(lcAttached).noquote() << "- parent window has attached object" << attached << "- returning";
             return attached;
+        }
     } else {
         // lookup popup's window
         QQuickPopup *popup = qobject_cast<QQuickPopup *>(objectWeAreAttachedTo);
-        if (popup)
-            return attachedObject(ourAttachedType, popup->popupItem()->window());
+        if (popup) {
+            qCDebug(lcAttached).noquote() << "- attachee is a popup; checking its window";
+            auto* popupWindow = popup->popupItem()->window();
+            auto *object = attachedObject(ourAttachedType, popupWindow);
+            // Check if the attached object exists for the popup window. If it
+            // doesn't, use transient parent attached object
+            if (!object && qobject_cast<QQuickPopupWindow *>(popupWindow))
+                return attachedObject(ourAttachedType, popupWindow->transientParent());
+            return object;
+        }
     }
 
     // lookup parent window
@@ -132,16 +161,20 @@ static QQuickAttachedPropertyPropagator *findAttachedParent(const QMetaObject *o
     if (window) {
         // It doesn't seem like a parent window can be anything but transient in Qt Quick.
         QQuickWindow *parentWindow = qobject_cast<QQuickWindow *>(window->transientParent());
+        qCDebug(lcAttached).noquote() << "- attachee is a window; checking its parent window" << parentWindow;
         if (parentWindow) {
             QQuickAttachedPropertyPropagator *attached = attachedObject(ourAttachedType, parentWindow);
-            if (attached)
+            if (attached) {
+                qCDebug(lcAttached).noquote() << "- parent window has attached object" << attached << "- returning";
                 return attached;
+            }
         }
     }
 
     // fallback to engine (global)
     if (objectWeAreAttachedTo) {
         QQmlEngine *engine = qmlEngine(objectWeAreAttachedTo);
+        qCDebug(lcAttached).noquote() << "- falling back to engine" << engine;
         if (engine) {
             QByteArray name = QByteArray("_q_") + ourAttachedType->className();
             QQuickAttachedPropertyPropagator *attached = engine->property(name).value<QQuickAttachedPropertyPropagator *>();
@@ -227,6 +260,7 @@ public:
     void setAttachedParent(QQuickAttachedPropertyPropagator *parent);
 
     void itemWindowChanged(QQuickWindow *window);
+    void transientParentWindowChanged(QWindow *newTransientParent);
     void itemParentChanged(QQuickItem *item, QQuickItem *parent) override;
 
     QList<QQuickAttachedPropertyPropagator *> attachedChildren;
@@ -235,19 +269,23 @@ public:
 
 void QQuickAttachedPropertyPropagatorPrivate::attachTo(QObject *object)
 {
-    QQuickItem *item = findAttachedItem(object);
-    if (item) {
+    if (QQuickItem *item = findAttachedItem(object)) {
         connect(item, &QQuickItem::windowChanged, this, &QQuickAttachedPropertyPropagatorPrivate::itemWindowChanged);
         QQuickItemPrivate::get(item)->addItemChangeListener(this, QQuickItemPrivate::Parent);
+    } else if (auto *window = qobject_cast<QQuickWindow *>(object)) {
+        QObjectPrivate::connect(window, &QWindow::transientParentChanged, this,
+            &QQuickAttachedPropertyPropagatorPrivate::transientParentWindowChanged);
     }
 }
 
 void QQuickAttachedPropertyPropagatorPrivate::detachFrom(QObject *object)
 {
-    QQuickItem *item = findAttachedItem(object);
-    if (item) {
+    if (QQuickItem *item = findAttachedItem(object)) {
         disconnect(item, &QQuickItem::windowChanged, this, &QQuickAttachedPropertyPropagatorPrivate::itemWindowChanged);
         QQuickItemPrivate::get(item)->removeItemChangeListener(this, QQuickItemPrivate::Parent);
+    } else if (auto *window = qobject_cast<QQuickWindow *>(object)) {
+        QObjectPrivate::disconnect(window, &QWindow::transientParentChanged,
+            this, &QQuickAttachedPropertyPropagatorPrivate::transientParentWindowChanged);
     }
 }
 
@@ -276,27 +314,55 @@ void QQuickAttachedPropertyPropagatorPrivate::setAttachedParent(QQuickAttachedPr
         return;
 
     QQuickAttachedPropertyPropagator *oldParent = attachedParent;
-    qCDebug(lcAttached) << "setAttachedParent called on" << q->parent();
+    qCDebug(lcAttached).noquote() << "setAttachedParent called on" << q << "with parent" << parent;
     if (attachedParent) {
-        qCDebug(lcAttached) << "- removing ourselves as an attached child of" << attachedParent->parent() << attachedParent;
+        qCDebug(lcAttached).noquote() << "- removing ourselves as an attached child of" << attachedParent;
         QQuickAttachedPropertyPropagatorPrivate::get(attachedParent)->attachedChildren.removeOne(q);
     }
     attachedParent = parent;
     if (parent) {
-        qCDebug(lcAttached) << "- adding ourselves as an attached child of" << parent->parent() << parent;
+        qCDebug(lcAttached).noquote() << "- adding ourselves as an attached child of" << parent;
         QQuickAttachedPropertyPropagatorPrivate::get(parent)->attachedChildren.append(q);
     }
     q->attachedParentChange(parent, oldParent);
 }
 
+/*
+    If there's e.g. code like this:
+
+        Behavior on Material.elevation {}
+
+    The meta type will be something like QQuickMaterialStyle_QML_125,
+    whereas QQmlMetaType::attachedPropertiesFunc only has attached
+    property data for QQuickMaterialStyle (i.e. attached property types
+    created from C++). We work around this by finding the first C++
+    meta object, which works even for attached types created in QML.
+*/
+const QMetaObject *firstCppMetaObject(QQuickAttachedPropertyPropagator *propagator)
+{
+    return QQmlData::ensurePropertyCache(propagator)->firstCppMetaObject();
+}
+
 void QQuickAttachedPropertyPropagatorPrivate::itemWindowChanged(QQuickWindow *window)
 {
     Q_Q(QQuickAttachedPropertyPropagator);
     QQuickAttachedPropertyPropagator *attachedParent = nullptr;
-    qCDebug(lcAttached) << "window of" << q->parent() << "changed to" << window;
-    attachedParent = findAttachedParent(q->metaObject(), q->parent());
+    qCDebug(lcAttached).noquote() << "window of" << q << "changed to" << window;
+
+    attachedParent = findAttachedParent(firstCppMetaObject(q), q->parent());
     if (!attachedParent)
-        attachedParent = attachedObject(q->metaObject(), window);
+        attachedParent = attachedObject(firstCppMetaObject(q), window);
+    setAttachedParent(attachedParent);
+}
+
+void QQuickAttachedPropertyPropagatorPrivate::transientParentWindowChanged(QWindow *newTransientParent)
+{
+    Q_Q(QQuickAttachedPropertyPropagator);
+    QQuickAttachedPropertyPropagator *attachedParent = nullptr;
+    qCDebug(lcAttached).noquote() << "transient parent window of" << q << "changed to" << newTransientParent;
+    attachedParent = findAttachedParent(firstCppMetaObject(q), q->parent());
+    if (!attachedParent)
+        attachedParent = attachedObject(firstCppMetaObject(q), newTransientParent);
     setAttachedParent(attachedParent);
 }
 
@@ -305,7 +371,7 @@ void QQuickAttachedPropertyPropagatorPrivate::itemParentChanged(QQuickItem *item
     Q_Q(QQuickAttachedPropertyPropagator);
     Q_UNUSED(item);
     Q_UNUSED(parent);
-    setAttachedParent(findAttachedParent(q->metaObject(), q->parent()));
+    setAttachedParent(findAttachedParent(firstCppMetaObject(q), q->parent()));
 }
 
 /*!
@@ -383,16 +449,19 @@ QQuickAttachedPropertyPropagator *QQuickAttachedPropertyPropagator::attachedPare
 void QQuickAttachedPropertyPropagator::initialize()
 {
     Q_D(QQuickAttachedPropertyPropagator);
+    qCDebug(lcAttached) << "initialize called for" << parent() << "- looking for attached parent...";
     QQuickAttachedPropertyPropagator *attachedParent = findAttachedParent(metaObject(), parent());
     if (attachedParent)
         d->setAttachedParent(attachedParent);
 
     const QList<QQuickAttachedPropertyPropagator *> attachedChildren = findAttachedChildren(metaObject(), parent());
-    qCDebug(lcAttached) << "initialize called for" << parent() << "- found" << attachedChildren.size() << "attached children";
+    qCDebug(lcAttached) << "- found" << attachedChildren.size() << "attached children:";
     for (QQuickAttachedPropertyPropagator *child : attachedChildren) {
-        qCDebug(lcAttached) << "-" << child->parent();
+        qCDebug(lcAttached) << "  -" << child->parent();
         QQuickAttachedPropertyPropagatorPrivate::get(child)->setAttachedParent(this);
     }
+
+    qCDebug(lcAttached) << "... finished initializing";
 }
 
 /*!
@@ -413,6 +482,22 @@ void QQuickAttachedPropertyPropagator::attachedParentChange(QQuickAttachedProper
     Q_UNUSED(oldParent);
 }
 
+#ifndef QT_NO_DEBUG_STREAM
+QDebug operator<<(QDebug debug, const QQuickAttachedPropertyPropagator *propagator)
+{
+    QDebugStateSaver saver(debug);
+    debug.nospace().noquote();
+    if (!propagator) {
+        debug << "QQuickAttachedPropertyPropagator(nullptr)";
+        return debug;
+    }
+
+    // Cast to QObject to avoid recursion.
+    debug << static_cast<const QObject *>(propagator) << " (which is attached to " << propagator->parent() << ')';
+    return debug;
+}
+#endif
+
 QT_END_NAMESPACE
 
 #include "moc_qquickattachedpropertypropagator.cpp"
