@@ -13,7 +13,7 @@ QT_BEGIN_NAMESPACE
 /*!
     \qmltype RangeSlider
     \inherits Control
-//!     \instantiates QQuickRangeSlider
+//!     \nativetype QQuickRangeSlider
     \inqmlmodule QtQuick.Controls
     \since 5.7
     \ingroup qtquickcontrols-input
@@ -225,6 +225,8 @@ void QQuickRangeSliderNode::setHandle(QQuickItem *handle)
     if (d->handle == handle)
         return;
 
+    QQuickControlPrivate::warnIfCustomizationNotSupported(d->slider, handle, QStringLiteral("handle"));
+
     if (!d->handle.isExecuting())
         d->cancelHandle();
 
@@ -484,8 +486,14 @@ bool QQuickRangeSliderPrivate::handlePress(const QPointF &point, ulong timestamp
 
     if (hitNode) {
         hitNode->setPressed(true);
-        if (QQuickItem *handle = hitNode->handle())
+        if (QQuickItem *handle = hitNode->handle()) {
             handle->setZ(1);
+
+            // A specific handle was hit, so it should get focus, rather than the default
+            // (first handle) that gets focus whenever the RangeSlider itself does - see focusInEvent().
+            if (focusPolicy & Qt::ClickFocus)
+                handle->forceActiveFocus(Qt::MouseFocusReason);
+        }
         QQuickRangeSliderNodePrivate::get(hitNode)->touchId = touchId;
     }
     if (otherNode) {
@@ -610,8 +618,14 @@ QQuickRangeSlider::QQuickRangeSlider(QQuickItem *parent)
     Q_D(QQuickRangeSlider);
     d->first = new QQuickRangeSliderNode(0.0, this);
     d->second = new QQuickRangeSliderNode(1.0, this);
+    d->setSizePolicy(QLayoutPolicy::Expanding, QLayoutPolicy::Fixed);
 
     setFlag(QQuickItem::ItemIsFocusScope);
+#ifdef Q_OS_MACOS
+    setFocusPolicy(Qt::TabFocus);
+#else
+    setFocusPolicy(Qt::StrongFocus);
+#endif
     setAcceptedMouseButtons(Qt::LeftButton);
 #if QT_CONFIG(quicktemplates2_multitouch)
     setAcceptTouchEvents(true);
@@ -694,11 +708,11 @@ void QQuickRangeSlider::setTo(qreal to)
 
 /*!
     \since QtQuick.Controls 2.5 (Qt 5.12)
-    \qmlproperty qreal QtQuick.Controls::RangeSlider::touchDragThreshold
+    \qmlproperty real QtQuick.Controls::RangeSlider::touchDragThreshold
 
     This property holds the threshold (in logical pixels) at which a touch drag event will be initiated.
     The mouse drag threshold won't be affected.
-    The default value is \c Qt.styleHints.startDragDistance.
+    The default value is \c Application.styleHints.startDragDistance.
 
     \sa QStyleHints
 
@@ -966,6 +980,11 @@ void QQuickRangeSlider::setOrientation(Qt::Orientation orientation)
     if (d->orientation == orientation)
         return;
 
+    if (orientation == Qt::Horizontal)
+        d->setSizePolicy(QLayoutPolicy::Expanding, QLayoutPolicy::Fixed);
+    else
+        d->setSizePolicy(QLayoutPolicy::Fixed, QLayoutPolicy::Expanding);
+
     d->orientation = orientation;
     emit orientationChanged();
 }
