@@ -3,14 +3,18 @@
 
 /*---
 esid: sec-array.prototype.map
-es5id: 15.4.4.19-5-7
-description: Array.prototype.map - built-in functions can be used as thisArg
+es5id: 15.4.4.19-1-14
+description: Array.prototype.map - applied to Error object
 ---*/
 
 function callbackfn(val, idx, obj) {
-  return this === eval;
+  return obj instanceof Error;
 }
 
-var testResult = [11].map(callbackfn, eval);
+var obj = new Error();
+obj.length = 1;
+obj[0] = 1;
+
+var testResult = Array.prototype.map.call(obj, callbackfn);
 
 assert.sameValue(testResult[0], true, 'testResult[0]');
