@@ -3,21 +3,21 @@
 
 /*---
 esid: sec-array.prototype.reduce
-es5id: 15.4.4.21-3-9
-description: >
-    Array.prototype.reduce - value of 'length' is a number (value is
-    -Infinity)
+es5id: 15.4.4.21-3-3
+description: Array.prototype.reduce - value of 'length' is a number (value is 0)
 ---*/
 
 var accessed = false;
 
 function callbackfn(prevVal, curVal, idx, obj) {
   accessed = true;
+  return 2;
 }
 
 var obj = {
-  0: 9,
-  length: -Infinity
+  0: 1,
+  1: 1,
+  length: 0
 };
 
 assert.sameValue(Array.prototype.reduce.call(obj, callbackfn, 1), 1, 'Array.prototype.reduce.call(obj, callbackfn, 1)');
