@@ -2,10 +2,10 @@
 // This code is governed by the BSD license found in the LICENSE file.
 
 /*---
-es5id: 15.2.3.13-2-11
-description: Object.isExtensible returns true for all built-in objects (Error)
+es5id: 15.2.3.13-2-5
+description: Object.isExtensible returns true for all built-in objects (String)
 ---*/
 
-var e = Object.isExtensible(Error);
+var e = Object.isExtensible(String);
 
 assert.sameValue(e, true, 'e');
