@@ -2,10 +2,10 @@
 // This code is governed by the BSD license found in the LICENSE file.
 
 /*---
-es5id: 15.5.4.20-4-36
+es5id: 15.5.4.20-4-60
 description: >
     String.prototype.trim handles whitepace and lineterminators
-    (ab\u000Bc)
+    (string with just blanks)
 ---*/
 
-assert.sameValue("ab\u000Bc".trim(), "ab\u000Bc", '"ab\u000Bc".trim()');
+assert.sameValue("    ".trim(), "", '"    ".trim()');
