@@ -2,15 +2,12 @@
 // This code is governed by the BSD license found in the LICENSE file.
 
 /*---
-es5id: 10.6-6-3
-description: >
-    'length' property of arguments object for 0 argument function
-    exists
+es5id: 10.6-12-1
+description: Accessing callee property of Arguments object is allowed
 flags: [noStrict]
 ---*/
 
 function testcase() {
-      var arguments= undefined;
-	(function () { assert.sameValue(arguments.length, 0); })();
+    arguments.callee;
  }
 testcase();
