@@ -5,9 +5,9 @@
 info: |
     Check For Statement for automatic semicolon insertion.
     If automatic insertion semicolon would become one of the two semicolons in the header of a For Statement.
-    Use one semicolon
-es5id: 7.9_A6.2_T4
-description: For header is (\n \n semicolon)
+    Don`t use semicolons
+es5id: 7.9_A6.3_T4
+description: For header is (\n false \n)
 negative:
   phase: parse
   type: SyntaxError
@@ -17,7 +17,7 @@ throw "Test262: This statement should not be evaluated.";
 
 //CHECK#1
 for(
-    
-;) {
+    false
+) {
   break;
 }
