@@ -2,14 +2,12 @@
 // This code is governed by the BSD license found in the LICENSE file.
 
 /*---
-es5id: 12.2.1-7-s
-description: >
-    eval - a direct eval declaring a var named 'eval' throws
-    SyntaxError in strict mode
+es5id: 12.2.1-35-s
+description: "'for(var eval = 42 in ...) {...}' throws SyntaxError in strict mode"
 flags: [onlyStrict]
 ---*/
 
 
 assert.throws(SyntaxError, function() {
-    eval('var eval;');
+    eval('for (var eval = 42 in null) {};');
 });
