@@ -1,5 +1,5 @@
 // Copyright (C) 2016 The Qt Company Ltd.
-// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 
 #include <QQmlEngine>
 #include <QQmlContext>
@@ -16,12 +16,17 @@
 #include <QQmlExpression>
 #include <QQmlIncubationController>
 #include <QTemporaryDir>
+ #include <QQmlEngineExtensionPlugin>
 #include <private/qqmlengine_p.h>
 #include <private/qqmltypedata_p.h>
 #include <private/qqmlcomponentattached_p.h>
 #include <QQmlAbstractUrlInterceptor>
 #include <QtQuickTestUtils/private/qmlutils_p.h>
 
+#include "declarativelyregistered.h"
+
+Q_IMPORT_QML_PLUGIN(OnlyDeclarativePlugin)
+
 class tst_qqmlengine : public QQmlDataTest
 {
     Q_OBJECT
@@ -31,8 +36,10 @@ public:
 private slots:
     void initTestCase() override;
     void rootContext();
+#if QT_CONFIG(qml_network)
     void networkAccessManager();
     void synchronousNetworkAccessManager();
+#endif
     void baseUrl();
     void contextForObject();
     void offlineStoragePath();
@@ -62,6 +69,7 @@ private slots:
     void cachedGetterLookup_qtbug_75335();
     void createComponentOnSingletonDestruction();
     void uiLanguage();
+    void markCurrentFunctionAsTranslationBinding();
     void executeRuntimeFunction();
     void captureQProperty();
     void listWrapperAsListReference();
@@ -70,6 +78,13 @@ private slots:
     void stringToColor();
     void qobjectToString();
     void qtNamespaceInQtObject();
+    void nativeModuleImport();
+    void lockedRootObject();
+    void crossReferencingSingletonsDeletion();
+    void bindingInstallUseAfterFree();
+    void objectListArgumentMethod();
+    void variantListQJsonConversion();
+    void attachedObjectOfUnregistered();
 
 public slots:
     QObject *createAQObjectForOwnershipTest ()
@@ -142,6 +157,7 @@ void tst_qqmlengine::rootContext()
     QVERIFY(!engine.rootContext()->parentContext());
 }
 
+#if QT_CONFIG(qml_network)
 class NetworkAccessManagerFactory : public QQmlNetworkAccessManagerFactory
 {
 public:
@@ -157,21 +173,19 @@ public:
 
 void tst_qqmlengine::networkAccessManager()
 {
-    QQmlEngine *engine = new QQmlEngine;
+    std::unique_ptr<QQmlEngine> engine = std::make_unique<QQmlEngine>();
 
     // Test QQmlEngine created manager
     QPointer<QNetworkAccessManager> manager = engine->networkAccessManager();
     QVERIFY(manager != nullptr);
-    delete engine;
 
     // Test factory created manager
-    engine = new QQmlEngine;
+    engine.reset(new QQmlEngine);
     NetworkAccessManagerFactory factory;
     engine->setNetworkAccessManagerFactory(&factory);
     QCOMPARE(engine->networkAccessManagerFactory(), &factory);
     QNetworkAccessManager *engineNam = engine->networkAccessManager(); // calls NetworkAccessManagerFactory::create()
     QCOMPARE(engineNam, factory.manager);
-    delete engine;
 }
 
 class ImmediateReply : public QNetworkReply {
@@ -219,7 +233,7 @@ void tst_qqmlengine::synchronousNetworkAccessManager()
     // reply is finished, so should not be in loading state.
     QVERIFY(!c.isLoading());
 }
-
+#endif
 
 void tst_qqmlengine::baseUrl()
 {
@@ -260,7 +274,7 @@ void tst_qqmlengine::baseUrl()
 
 void tst_qqmlengine::contextForObject()
 {
-    QQmlEngine *engine = new QQmlEngine;
+    std::unique_ptr<QQmlEngine> engine = std::make_unique<QQmlEngine>();
 
     // Test null-object
     QVERIFY(!QQmlEngine::contextForObject(nullptr));
@@ -287,7 +301,7 @@ void tst_qqmlengine::contextForObject()
     QCOMPARE(QQmlEngine::contextForObject(&object), engine->rootContext());
 
     // Delete context
-    delete engine; engine = nullptr;
+    engine.reset();
     QVERIFY(!QQmlEngine::contextForObject(&object));
 }
 
@@ -313,7 +327,9 @@ void tst_qqmlengine::offlineStoragePath()
 
     QCOMPARE(QDir::fromNativeSeparators(engine.offlineStoragePath()), dir.path());
 
+    QSignalSpy offlineStoragePathSpy(&engine, &QQmlEngine::offlineStoragePathChanged);
     engine.setOfflineStoragePath(QDir::homePath());
+    QCOMPARE(offlineStoragePathSpy.size(), 1);
     QCOMPARE(engine.offlineStoragePath(), QDir::homePath());
 }
 
@@ -363,10 +379,9 @@ void tst_qqmlengine::clearComponentCache()
     // Test "test" property
     {
         QQmlComponent component(&engine, fileUrl);
-        QObject *obj = component.create();
-        QVERIFY(obj != nullptr);
+        std::unique_ptr<QObject> obj { component.create() };
+        QVERIFY(obj.get() != nullptr);
         QCOMPARE(obj->property("test").toInt(), 10);
-        delete obj;
     }
 
     // Modify qml file
@@ -386,22 +401,29 @@ void tst_qqmlengine::clearComponentCache()
     // Test cache hit
     {
         QQmlComponent component(&engine, fileUrl);
-        QObject *obj = component.create();
-        QVERIFY(obj != nullptr);
+        std::unique_ptr<QObject> obj { component.create() };
+        QVERIFY(obj.get() != nullptr);
         QCOMPARE(obj->property("test").toInt(), 10);
-        delete obj;
     }
 
     // Clear cache
     engine.clearComponentCache();
 
+    // Nothing holds on to any CU anymore. They should all be gone.
+    QVERIFY(QQmlEnginePrivate::get(&engine)->v4engine()->compilationUnits().isEmpty());
+
     // Test cache refresh
     {
         QQmlComponent component(&engine, fileUrl);
-        QObject *obj = component.create();
-        QVERIFY(obj != nullptr);
+        std::unique_ptr<QObject> obj { component.create() };
+        QVERIFY(obj.get() != nullptr);
         QCOMPARE(obj->property("test").toInt(), 11);
-        delete obj;
+
+        engine.clearComponentCache();
+
+        // The CU we are holding on to is still alive.
+        // Otherwise we cannot mark its objects for GC anymore.
+        QVERIFY(!QQmlEnginePrivate::get(&engine)->v4engine()->compilationUnits().isEmpty());
     }
 
     // Regular Synchronous loading will leave us with an event posted
@@ -428,7 +450,7 @@ public:
         // There might be JS function objects around that hold a last ref to the compilation unit that's
         // keeping the type compilation data (CompilationUnit) around. Let's collect them as well so that
         // trim works well.
-        engine->collectGarbage();
+        gc(*engine);
 
         engine->trimComponentCache();
     }
@@ -487,22 +509,25 @@ void tst_qqmlengine::trimComponentCache_data()
     // empty apart from their inherited elements, and those that define new properties.
     // For each there are five types of composition: extension, aggregation,
     // aggregation via component, property and object-created-via-transient-component.
-    foreach (const QString &test, (QStringList() << "EmptyComponent"
-                                                 << "VMEComponent"
-                                                 << "EmptyExtendEmptyComponent"
-                                                 << "VMEExtendEmptyComponent"
-                                                 << "EmptyExtendVMEComponent"
-                                                 << "VMEExtendVMEComponent"
-                                                 << "EmptyAggregateEmptyComponent"
-                                                 << "VMEAggregateEmptyComponent"
-                                                 << "EmptyAggregateVMEComponent"
-                                                 << "VMEAggregateVMEComponent"
-                                                 << "EmptyPropertyEmptyComponent"
-                                                 << "VMEPropertyEmptyComponent"
-                                                 << "EmptyPropertyVMEComponent"
-                                                 << "VMEPropertyVMEComponent"
-                                                 << "VMETransientEmptyComponent"
-                                                 << "VMETransientVMEComponent")) {
+    const QStringList components = {
+        "EmptyComponent",
+        "VMEComponent",
+        "EmptyExtendEmptyComponent",
+        "VMEExtendEmptyComponent",
+        "EmptyExtendVMEComponent",
+        "VMEExtendVMEComponent",
+        "EmptyAggregateEmptyComponent",
+        "VMEAggregateEmptyComponent",
+        "EmptyAggregateVMEComponent",
+        "VMEAggregateVMEComponent",
+        "EmptyPropertyEmptyComponent",
+        "VMEPropertyEmptyComponent",
+        "EmptyPropertyVMEComponent",
+        "VMEPropertyVMEComponent",
+        "VMETransientEmptyComponent",
+        "VMETransientVMEComponent",
+    };
+    for (const QString &test : components) {
         // For these cases, we first test that the component instance keeps the components
         // referenced, and then that the instantiated object keeps the components referenced
         for (int i = 1; i <= 2; ++i) {
@@ -542,10 +567,14 @@ void tst_qqmlengine::clearSingletons()
     const int cppInstance = qmlRegisterSingletonInstance(
                 "ClearSingletons", 1, 0, "CppInstance", &objectCaller1);
 
+#if QT_DEPRECATED_SINCE(6, 3)
+QT_WARNING_PUSH QT_WARNING_DISABLE_DEPRECATED
     QQmlPrivate::SingletonFunctor deprecatedSingletonFunctor;
     deprecatedSingletonFunctor.m_object = &objectCaller2;
     const int deprecatedCppInstance = qmlRegisterSingletonType<ObjectCaller>(
                 "ClearSingletons", 1, 0, "DeprecatedCppInstance", deprecatedSingletonFunctor);
+QT_WARNING_POP
+#endif // QT_DEPRECATED_SINCE(6, 3)
 
     const int cppFactory = qmlRegisterSingletonType<CppSingleton>(
                 "ClearSingletons", 1, 0, "CppFactory", &CppSingleton::create);
@@ -559,7 +588,9 @@ void tst_qqmlengine::clearSingletons()
     QQmlEngine engine;
 
     QCOMPARE(engine.singletonInstance<ObjectCaller *>(cppInstance), &objectCaller1);
+#if QT_DEPRECATED_SINCE(6, 3)
     QCOMPARE(engine.singletonInstance<ObjectCaller *>(deprecatedCppInstance), &objectCaller2);
+#endif
     const CppSingleton *oldCppSingleton = engine.singletonInstance<CppSingleton *>(cppFactory);
     QVERIFY(oldCppSingleton != nullptr);
     const uint oldCppSingletonId = oldCppSingleton->id;
@@ -577,7 +608,9 @@ void tst_qqmlengine::clearSingletons()
               "import QtQml\n"
               "QtObject {\n"
               "    property QtObject a: CppInstance\n"
+#if QT_DEPRECATED_SINCE(6, 3)
               "    property QtObject f: DeprecatedCppInstance\n"
+#endif
               "    property QtObject b: CppFactory\n"
               "    property int c: JsValue\n"
               "    property QtObject d: JsObject\n"
@@ -586,7 +619,9 @@ void tst_qqmlengine::clearSingletons()
     QVERIFY2(c.isReady(), qPrintable(c.errorString()));
     QScopedPointer<QObject> singletonUser(c.create());
     QCOMPARE(qvariant_cast<QObject *>(singletonUser->property("a")), &objectCaller1);
+#if QT_DEPRECATED_SINCE(6, 3)
     QCOMPARE(qvariant_cast<QObject *>(singletonUser->property("f")), &objectCaller2);
+#endif
     QCOMPARE(qvariant_cast<QObject *>(singletonUser->property("b")), oldCppSingleton);
     QCOMPARE(singletonUser->property("c").toUInt(), 13u);
     QCOMPARE(qvariant_cast<QObject *>(singletonUser->property("d")), oldJsSingleton);
@@ -597,6 +632,7 @@ void tst_qqmlengine::clearSingletons()
 
     QCOMPARE(engine.singletonInstance<ObjectCaller *>(cppInstance), &objectCaller1);
 
+#if QT_DEPRECATED_SINCE(6, 3)
     // Singleton instances created with previous versions of qmlRegisterSingletonInstance()
     // are lost and cannot be accessed anymore. We can only call their synthesized factory
     // functions once. This is not a big problem as you have to recompile in order to use
@@ -610,6 +646,7 @@ void tst_qqmlengine::clearSingletons()
                 "<Unknown File>: qmlRegisterSingletonType(): \"DeprecatedCppInstance\" is not "
                 "available because the callback function returns a null pointer.");
     QCOMPARE(engine.singletonInstance<ObjectCaller *>(deprecatedCppInstance), nullptr);
+#endif
 
     QCOMPARE(CppSingleton::instantiations, oldCppSingletonId);
     const CppSingleton *newCppSingleton = engine.singletonInstance<CppSingleton *>(cppFactory);
@@ -626,7 +663,9 @@ void tst_qqmlengine::clearSingletons()
 
     // Holding on to an old singleton instance is OK. We don't delete those.
     QCOMPARE(qvariant_cast<QObject *>(singletonUser->property("a")), &objectCaller1);
+#if QT_DEPRECATED_SINCE(6, 3)
     QCOMPARE(qvariant_cast<QObject *>(singletonUser->property("f")), &objectCaller2);
+#endif
 
     // Deleting the singletons created by factories has cleared their references in QML.
     // We don't renew them as the point of clearing the singletons is not having any
@@ -698,7 +737,7 @@ void tst_qqmlengine::outputWarningsToStandardError()
     QVERIFY(o != nullptr);
     o.reset();
 
-    QCOMPARE(messageHandler.messages().count(), 1);
+    QCOMPARE(messageHandler.messages().size(), 1);
     QCOMPARE(messageHandler.messages().at(0), QLatin1String("<Unknown File>:1:32: Unable to assign [undefined] to int"));
     messageHandler.clear();
 
@@ -760,7 +799,7 @@ void tst_qqmlengine::objectOwnership()
                         c.createWithInitialProperties({{"test", QVariant::fromValue(this)}}));
             QVERIFY(o != nullptr);
         }
-        QTRY_VERIFY(spy.count());
+        QTRY_VERIFY(spy.size());
     }
     {
         QObject *ptr = new QObject();
@@ -777,7 +816,7 @@ void tst_qqmlengine::objectOwnership()
             QQmlProperty testProp(o.data(), "test");
             testProp.write(QVariant::fromValue<QObject*>(nullptr));
         }
-        QTRY_VERIFY(spy.count());
+        QTRY_VERIFY(spy.size());
     }
 }
 
@@ -880,12 +919,13 @@ void tst_qqmlengine::qtqmlModule()
     QFETCH(QString, expectedError);
     QFETCH(QStringList, expectedWarnings);
 
-    foreach (const QString &w, expectedWarnings)
+    for (const QString &w : std::as_const(expectedWarnings))
         QTest::ignoreMessage(QtWarningMsg, qPrintable(w));
 
     QQmlEngine e;
     QQmlComponent c(&e, testFile);
     if (expectedError.isEmpty()) {
+        QVERIFY2(c.isReady(), qPrintable(c.errorString()));
         QScopedPointer<QObject> o(c.create());
         QVERIFY(o);
     } else {
@@ -1129,6 +1169,8 @@ void tst_qqmlengine::singletonInstance()
         QObject *instance = value.toQObject();
         QVERIFY(instance);
         QCOMPARE(instance->metaObject()->className(), "CppSingleton");
+
+        QCOMPARE(engine.singletonInstance<CppSingleton *>("Test", "CppSingleton"), instance);
     }
 
     {
@@ -1200,6 +1242,20 @@ void tst_qqmlengine::singletonInstance()
         QVERIFY(!noSinglePtr);
     }
 
+    // test the case where we haven't loaded the module yet
+    {
+        auto singleton = engine.singletonInstance<PurelyDeclarativeSingleton *>("OnlyDeclarative", "PurelyDeclarativeSingleton");
+        QVERIFY(singleton);
+        // requesting the singleton twice yields the same result
+        auto again = engine.singletonInstance<PurelyDeclarativeSingleton *>("OnlyDeclarative", "PurelyDeclarativeSingleton");
+        QCOMPARE(again, singleton);
+
+        // different engines -> different singletons
+        QQmlEngine engine2;
+        auto differentEngine = engine2.singletonInstance<PurelyDeclarativeSingleton *>("OnlyDeclarative", "PurelyDeclarativeSingleton");
+        QCOMPARE_NE(differentEngine, singleton);
+    }
+
     {
         // Invalid types
         QJSValue value;
@@ -1232,9 +1288,9 @@ void tst_qqmlengine::singletonInstance()
 
     {
         // deleted object
-        auto dayfly = new Dayfly{};
-        auto id = qmlRegisterSingletonInstance("Vanity", 1, 0, "Dayfly", dayfly);
-        delete dayfly;
+        auto dayfly = std::make_unique<Dayfly>();
+        auto id = qmlRegisterSingletonInstance("Vanity", 1, 0, "Dayfly", dayfly.get());
+        dayfly.reset();
         QTest::ignoreMessage(QtMsgType::QtWarningMsg, "<Unknown File>: The registered singleton has already been deleted. Ensure that it outlives the engine.");
         QObject *instance = engine.singletonInstance<QObject*>(id);
         QVERIFY(!instance);
@@ -1295,6 +1351,8 @@ void tst_qqmlengine::createComponentOnSingletonDestruction()
 
 void tst_qqmlengine::uiLanguage()
 {
+    const QRegularExpression bindingLoopWarningRegex(".*QML QtObject: Binding loop detected for property \"textToTranslate\".*");
+
     {
         QQmlEngine engine;
 
@@ -1306,19 +1364,19 @@ void tst_qqmlengine::uiLanguage()
 
         QQmlComponent component(&engine, testFileUrl("uiLanguage.qml"));
 
-        QTest::ignoreMessage(QtMsgType::QtWarningMsg, (component.url().toString() + ":2:1: QML QtObject: Binding loop detected for property \"textToTranslate\"").toLatin1());
+        QTest::ignoreMessage(QtMsgType::QtWarningMsg, bindingLoopWarningRegex);
         QScopedPointer<QObject> object(component.create());
         QVERIFY(!object.isNull());
 
         QVERIFY(engine.uiLanguage().isEmpty());
         QCOMPARE(object->property("numberOfTranslationBindingEvaluations").toInt(), 1);
 
-        QTest::ignoreMessage(QtMsgType::QtWarningMsg, (component.url().toString() + ":2:1: QML QtObject: Binding loop detected for property \"textToTranslate\"").toLatin1());
+        QTest::ignoreMessage(QtMsgType::QtWarningMsg, bindingLoopWarningRegex);
         engine.setUiLanguage("TestLanguage");
         QCOMPARE(object->property("numberOfTranslationBindingEvaluations").toInt(), 2);
         QCOMPARE(object->property("chosenLanguage").toString(), "TestLanguage");
 
-        QTest::ignoreMessage(QtMsgType::QtWarningMsg, (component.url().toString() + ":2:1: QML QtObject: Binding loop detected for property \"textToTranslate\"").toLatin1());
+        QTest::ignoreMessage(QtMsgType::QtWarningMsg, bindingLoopWarningRegex);
         engine.evaluate("Qt.uiLanguage = \"anotherLanguage\"");
         QCOMPARE(engine.uiLanguage(), QString("anotherLanguage"));
         QCOMPARE(object->property("numberOfTranslationBindingEvaluations").toInt(), 3);
@@ -1329,7 +1387,7 @@ void tst_qqmlengine::uiLanguage()
         QQmlEngine engine;
         QQmlComponent component(&engine, testFileUrl("uiLanguage.qml"));
 
-        QTest::ignoreMessage(QtMsgType::QtWarningMsg, (component.url().toString() + ":2:1: QML QtObject: Binding loop detected for property \"textToTranslate\"").toLatin1());
+        QTest::ignoreMessage(QtMsgType::QtWarningMsg, bindingLoopWarningRegex);
         QScopedPointer<QObject> object(component.create());
         QVERIFY(!object.isNull());
 
@@ -1338,6 +1396,35 @@ void tst_qqmlengine::uiLanguage()
     }
 }
 
+class I18nAwareClass : public QObject {
+    Q_OBJECT
+    QML_NAMED_ELEMENT(I18nAware)
+
+    Q_PROPERTY(QString text READ text NOTIFY textChanged)
+signals:
+    void textChanged();
+public:
+    int counter = 0;
+
+    QString text()
+    {
+        if (auto engine = qmlEngine(this))
+            engine->markCurrentFunctionAsTranslationBinding();
+        return QLatin1String("Hello, %1").arg(QString::number(counter++));
+    }
+};
+
+void tst_qqmlengine::markCurrentFunctionAsTranslationBinding()
+{
+    QQmlEngine engine;
+    qmlRegisterTypesAndRevisions<I18nAwareClass>("i18ntest", 1);
+    QQmlComponent comp(&engine, testFileUrl("markCurrentFunctionAsTranslationBinding.qml"));
+    std::unique_ptr<QObject> root { comp.create() };
+    QCOMPARE(root->property("result"), "Hello, 0");
+    engine.retranslate();
+    QCOMPARE(root->property("result"), "Hello, 1");
+}
+
 void tst_qqmlengine::executeRuntimeFunction()
 {
     QQmlEngine engine;
@@ -1536,8 +1623,10 @@ void tst_qqmlengine::stringToColor()
 
     const QMetaType metaType(QMetaType::QColor);
     QVariant color(metaType);
-    QVERIFY(engine.handle()->metaTypeFromJS(
-                engine.handle()->newString(QStringLiteral("#abcdef"))->asReturnedValue(),
+    QV4::Scope scope(engine.handle());
+    QV4::ScopedValue colorString(scope, engine.handle()->newString(QStringLiteral("#abcdef")));
+    QVERIFY(QV4::ExecutionEngine::metaTypeFromJS(
+                colorString,
                 metaType, color.data()));
     QVERIFY(color.isValid());
     QCOMPARE(color.metaType(), metaType);
@@ -1601,6 +1690,131 @@ void tst_qqmlengine::qtNamespaceInQtObject()
     QVERIFY(qtObject.hasProperty(QStringLiteral("objectName")));
 }
 
+void tst_qqmlengine::nativeModuleImport()
+{
+    QQmlEngine engine;
+
+    QJSValue name("TheName");
+    QJSValue obj = engine.newObject();
+    obj.setProperty("name", name);
+    engine.registerModule("info.mjs", obj);
+
+    QQmlComponent c(&engine, testFileUrl("nativeModuleImport.qml"));
+    QVERIFY2(c.isReady(), qPrintable(c.errorString()));
+
+    QScopedPointer<QObject> o(c.create());
+
+    QCOMPARE(o->property("a").toString(), QStringLiteral("Hello World"));
+    QCOMPARE(o->property("b").toString(), QStringLiteral("TheName"));
+    QCOMPARE(o->property("c").toString(), QStringLiteral("TheName"));
+    QCOMPARE(o->property("d").toString(), QStringLiteral("TheName"));
+    QCOMPARE(o->property("e").toString(), QStringLiteral("TheName"));
+}
+
+void tst_qqmlengine::lockedRootObject()
+{
+    QQmlEngine engine;
+    QQmlComponent c(&engine, testFileUrl("lockedRootObject.qml"));
+    QVERIFY2(c.isReady(), qPrintable(c.errorString()));
+    QTest::ignoreMessage(
+                QtWarningMsg, "You cannot shadow the locked property 'hasOwnProperty' in QML.");
+    QScopedPointer<QObject> o(c.create());
+    QCOMPARE(o->property("myErrorName").toString(), QStringLiteral("MyError1"));
+    QCOMPARE(o->property("errorName").toString(), QStringLiteral("MyError2"));
+    QCOMPARE(o->property("mathMax").toInt(), 4);
+    QCOMPARE(o->property("extendGlobal").toInt(), 32);
+    QCOMPARE(o->property("prototypeTrick").toString(), QStringLiteral("SyntaxError"));
+    QCOMPARE(o->property("shadowMethod1").toString(), QStringLiteral("not a TypeError"));
+    QCOMPARE(o->property("shadowMethod2").toBool(), false);
+    QCOMPARE(o->property("changeObjectProto1").toString(), QStringLiteral("not an Object"));
+    QCOMPARE(o->property("changeObjectProto2").toBool(), false);
+    QCOMPARE(o->property("defineProperty1").toString(), QStringLiteral("not a URIError"));
+    QCOMPARE(o->property("defineProperty2").toBool(), false);
+}
+
+void tst_qqmlengine::crossReferencingSingletonsDeletion()
+{
+    QQmlEngine engine;
+    engine.addImportPath(testFileUrl("crossReferencingSingletonsDeletion").url());
+    QQmlComponent c(&engine, testFileUrl("crossReferencingSingletonsDeletion/Module/Main.qml"));
+    QVERIFY2(c.isReady(), qPrintable(c.errorString()));
+
+    std::unique_ptr<QObject> o{ c.create() };
+    QVERIFY(o);
+    QCOMPARE(o->property("s").toString(), "SingletonA");
+}
+
+void tst_qqmlengine::bindingInstallUseAfterFree()
+{
+    QQmlEngine engine;
+    QQmlComponent c(&engine, testFileUrl("bindingInstallUseAfterFree.qml"));
+    QVERIFY2(c.isReady(), qPrintable(c.errorString()));
+    std::unique_ptr<QObject> o{ c.create() };
+    QVERIFY(o);
+}
+
+void tst_qqmlengine::objectListArgumentMethod()
+{
+    QQmlEngine engine;
+    QQmlComponent c(&engine, testFileUrl("objectListArgumentMethod.qml"));
+    QVERIFY2(c.isReady(), qPrintable(c.errorString()));
+
+    QTest::ignoreMessage(QtMsgType::QtDebugMsg, "5");
+    std::unique_ptr<QObject> o{ c.create() };
+    QVERIFY(o);
+}
+
+void tst_qqmlengine::variantListQJsonConversion()
+{
+    QQmlEngine engine;
+    QQmlComponent c(&engine, testFileUrl("variantListQJsonConversion.qml"));
+    QVERIFY2(c.isReady(), qPrintable(c.errorString()));
+
+    QTest::ignoreMessage(QtMsgType::QtDebugMsg, R"(["cpp","variant","list"])");
+    QTest::ignoreMessage(QtMsgType::QtDebugMsg, R"({"test":["cpp","variant","list"]})");
+    QTest::ignoreMessage(QtMsgType::QtDebugMsg,
+                         R"([{"objectName":"o0"},{"objectName":"o1"},{"objectName":"o2"}])");
+
+    QScopedPointer<QObject> o(c.create());
+    QVERIFY(o);
+}
+
+class UnregisteredAttached : public QObject
+{
+    Q_OBJECT
+public:
+    UnregisteredAttached(QObject *parent = nullptr) : QObject(parent) {}
+};
+
+class Unregistered : public QObject
+{
+    Q_OBJECT
+    QML_ATTACHED(UnregisteredAttached)
+public:
+    static UnregisteredAttached *qmlAttachedProperties(QObject *obj)
+    {
+        return new UnregisteredAttached(obj);
+    }
+};
+
+void tst_qqmlengine::attachedObjectOfUnregistered()
+{
+    QObject o;
+
+    QObject *a = qmlAttachedPropertiesObject<Unregistered>(&o);
+    QVERIFY(a);
+    QVERIFY(qobject_cast<UnregisteredAttached *>(a));
+
+    QObject *b = qmlAttachedPropertiesObject<Unregistered>(&o);
+    QCOMPARE(a, b);
+
+    QObject o2;
+    QObject *c = qmlAttachedPropertiesObject<Unregistered>(&o2);
+    QVERIFY(c);
+    QVERIFY(qobject_cast<UnregisteredAttached *>(c));
+    QVERIFY(c != a);
+}
+
 QTEST_MAIN(tst_qqmlengine)
 
 #include "tst_qqmlengine.moc"
