@@ -1,27 +1,34 @@
 // Copyright (C) 2016 The Qt Company Ltd.
-// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 
 #include <private/qqmlengine_p.h>
 
 #include <qtest.h>
-#include <QDebug>
+
 #include <QQmlEngine>
-#include <QFontDatabase>
-#include <QFileInfo>
 #include <QQmlComponent>
+#include <QtQuick/QQuickItem>
+
+#include <QCryptographicHash>
+#include <QDateTime>
+#include <QDebug>
+#include <QDateTime>
 #include <QDesktopServices>
 #include <QDir>
-#include <QCryptographicHash>
-#include <QtQuick/QQuickItem>
+#include <QFileInfo>
+#include <QFont>
+#include <QFontDatabase>
+#include <QMatrix4x4>
+#include <QQuaternion>
 #include <QSignalSpy>
 #include <QVector2D>
 #include <QVector3D>
 #include <QVector4D>
-#include <QQuaternion>
-#include <QMatrix4x4>
-#include <QFont>
+#include <QTimeZone>
+
 #include <QtQuickTestUtils/private/qmlutils_p.h>
-#include <private/qglobal_p.h>
+#include <private/qtenvironmentvariables_p.h> // for qTzSet()
+#include <private/qqmlengine_p.h>
 
 class tst_qqmlqt : public QQmlDataTest
 {
@@ -50,8 +57,10 @@ private slots:
     void alpha();
     void tint();
     void color();
+#if QT_CONFIG(desktopservices)
     void openUrlExternally();
     void openUrlExternally_pragmaLibrary();
+#endif
     void md5();
     void createComponent();
     void createComponent_pragmaLibrary();
@@ -76,6 +85,8 @@ private slots:
     void timeRoundtrip_data();
     void timeRoundtrip();
 
+    void fontSetsProperties();
+
 private:
     QQmlEngine engine;
 };
@@ -186,6 +197,9 @@ void tst_qqmlqt::hsla()
     QCOMPARE(qvariant_cast<QColor>(object->property("test4")), QColor());
     QCOMPARE(qvariant_cast<QColor>(object->property("test5")), QColor::fromHslF(1, 1, 1, 1));
     QCOMPARE(qvariant_cast<QColor>(object->property("test6")), QColor::fromHslF(0, 0, 0, 0));
+    QColor test7 = qvariant_cast<QColor>(object->property("test7"));
+    QCOMPARE(test7, QColor::fromHslF(-1, 0, 0.5, 1));
+    QCOMPARE(test7.hslHue(), -1.0f);
 }
 
 void tst_qqmlqt::hsva()
@@ -206,6 +220,9 @@ void tst_qqmlqt::hsva()
     QCOMPARE(qvariant_cast<QColor>(object->property("test4")), QColor());
     QCOMPARE(qvariant_cast<QColor>(object->property("test5")), QColor::fromHsvF(1, 1, 1, 1));
     QCOMPARE(qvariant_cast<QColor>(object->property("test6")), QColor::fromHsvF(0, 0, 0, 0));
+    QColor test7 = qvariant_cast<QColor>(object->property("test7"));
+    QCOMPARE(test7, QColor::fromHsvF(-1, 0, 0.5, 1));
+    QCOMPARE(test7.hsvHue(), -1.0f);
 }
 
 void tst_qqmlqt::colorEqual()
@@ -606,6 +623,7 @@ public slots:
     void noteCall(const QUrl &url) { called++; last = url; }
 };
 
+#if QT_CONFIG(desktopservices)
 void tst_qqmlqt::openUrlExternally()
 {
     MyUrlHandler handler;
@@ -652,6 +670,7 @@ void tst_qqmlqt::openUrlExternally_pragmaLibrary()
     QCOMPARE(handler.called,2);
     QCOMPARE(handler.last, htmlTestFile);
 }
+#endif
 
 void tst_qqmlqt::md5()
 {
@@ -743,18 +762,18 @@ void tst_qqmlqt::dateTimeConversion()
     QTime time(14,15,38,200);
     QDateTime dateTime(date, time);
     QDateTime dateTime0(QDate(2021, 7, 1).startOfDay());
-    QDateTime dateTime0utc(QDate(2021, 7, 1).startOfDay(Qt::UTC));
+    QDateTime dateTime0utc(QDate(2021, 7, 1).startOfDay(QTimeZone::UTC));
     QDateTime dateTime1(QDate(2021, 7, 31).startOfDay());
-    QDateTime dateTime1utc(QDate(2021, 7, 31).startOfDay(Qt::UTC));
+    QDateTime dateTime1utc(QDate(2021, 7, 31).startOfDay(QTimeZone::UTC));
     QDateTime dateTime2(QDate(2852,12,31), QTime(23,59,59,500));
     QDateTime dateTime3(QDate(2000,1,1), QTime(0,0,0,0));
     QDateTime dateTime4(QDate(2001,2,2), QTime(0,0,0,0));
     QDateTime dateTime5(QDate(1999,1,1), QTime(2,3,4,0));
     QDateTime dateTime6(QDate(2008,2,24), QTime(14,15,38,200));
-    QDateTime dateTime7(QDate(1970,1,1), QTime(0,0,0,0), Qt::UTC);
-    QDateTime dateTime8(QDate(1586,2,2), QTime(0,0,0,0), Qt::UTC);
-    QDateTime dateTime9(QDate(955,1,1), QTime(0,0,0,0), Qt::UTC);
-    QDateTime dateTime10(QDate(113,2,24), QTime(14,15,38,200), Qt::UTC);
+    QDateTime dateTime7(QDate(1970, 1, 1), QTime(0, 0), QTimeZone::UTC);
+    QDateTime dateTime8(QDate(1586, 2, 2), QTime(0, 0), QTimeZone::UTC);
+    QDateTime dateTime9(QDate(955, 1, 1), QTime(0, 0), QTimeZone::UTC);
+    QDateTime dateTime10(QDate(113, 2, 24), QTime(14, 15, 38, 200), QTimeZone::UTC);
 
     QQmlEngine eng;
     QQmlComponent component(&eng, testFileUrl("dateTimeConversion.qml"));
@@ -799,7 +818,7 @@ void tst_qqmlqt::dateTimeFormatting()
         << component.url().toString() + ":40: TypeError: Passing incompatible arguments to C++ functions from JavaScript is not allowed."
         << component.url().toString() + ":43: TypeError: Passing incompatible arguments to C++ functions from JavaScript is not allowed.";
 
-    foreach (const QString &warning, warnings)
+    for (const QString &warning : std::as_const(warnings))
         QTest::ignoreMessage(QtWarningMsg, qPrintable(warning));
 
     warnings.clear();
@@ -814,7 +833,7 @@ void tst_qqmlqt::dateTimeFormatting()
         << "Could not convert argument 1 at"
         << "expression for err_dateTime2@";
 
-    foreach (const QString &warning, warnings)
+    for (const QString &warning : std::as_const(warnings))
         QTest::ignoreMessage(QtWarningMsg, QRegularExpression(warning));
 
     QScopedPointer<QObject> object(component.createWithInitialProperties({
@@ -827,7 +846,7 @@ void tst_qqmlqt::dateTimeFormatting()
 
     QVERIFY(inputProperties.size() > 0);
     QVariant result;
-    foreach(const QString &prop, inputProperties) {
+    for (const QString &prop : std::as_const(inputProperties)) {
         QVERIFY(QMetaObject::invokeMethod(object.data(), method.toUtf8().constData(),
                 Q_RETURN_ARG(QVariant, result),
                 Q_ARG(QVariant, prop)));
@@ -903,7 +922,10 @@ void tst_qqmlqt::dateTimeFormattingVariants()
         QTest::ignoreMessage(QtWarningMsg, QRegularExpression(warning));
 
     warnings.clear();
-    if (method == QStringLiteral("formatTime") && variant.typeId() == QMetaType::QString) {
+
+    if (method == QStringLiteral("formatTime")
+            && variant.typeId() == QMetaType::QString
+            && QByteArrayView(QTest::currentDataTag()).endsWith("ISO")) {
         for (int i = 0; i < 4; ++i) {
             QTest::ignoreMessage(QtWarningMsg,
                                  "\"2011/05/31 11:16:39.755\" is a "
@@ -919,13 +941,18 @@ void tst_qqmlqt::dateTimeFormattingVariants()
                     QtWarningMsg,
                     QRegularExpression("formatting.qml:19: Error: Invalid argument passed to "
                                        "formatTime"));
-        } else {
-            QTest::ignoreMessage(QtWarningMsg,
-                                 QRegularExpression("Could not convert argument 0 at"));
-            QTest::ignoreMessage(QtWarningMsg, QRegularExpression(method + "@"));
-            QTest::ignoreMessage(QtWarningMsg, QRegularExpression(
-                                     "TypeError: Passing incompatible arguments to "
-                                     "C.. functions from JavaScript is not allowed."));
+        } else if (method == "formatDate") {
+            // formatDate has special error handling as it parses the strings itself.
+            QTest::ignoreMessage(
+                    QtWarningMsg,
+                    QRegularExpression("formatting.qml:10: Error: Invalid argument passed to "
+                                       "formatDate"));
+        } else if (method == "formatDateTime") {
+            // formatDateTime has special error handling as it parses the strings itself.
+            QTest::ignoreMessage(
+                    QtWarningMsg,
+                    QRegularExpression("formatting.qml:29: Error: Invalid argument passed to "
+                                       "formatDateTime"));
         }
     }
 
@@ -961,7 +988,7 @@ void tst_qqmlqt::dateTimeFormattingVariants_data()
 
     QDate date(2011,5,31);
     // V4 reads the date in UTC but DateObject::toQDateTime() gives it back in local time:
-    temporary = QDateTime(date, QTime(0, 0, 0), Qt::UTC).toLocalTime();
+    temporary = date.startOfDay(QTimeZone::UTC).toLocalTime();
     QTest::newRow("formatDate, qdate")
         << "formatDate" << QVariant::fromValue(date)
         << (QStringList()
@@ -1004,27 +1031,47 @@ void tst_qqmlqt::dateTimeFormattingVariants_data()
             << temporary.time().toString("H:m:s a")
             << temporary.time().toString("hh:mm:ss.zzz"));
 
-    QString string(QLatin1String("2011/05/31 11:16:39.755"));
-    temporary = QDateTime::fromString(string, "yyyy/MM/dd HH:mm:ss.zzz");
-    QTest::newRow("formatDate, qstring")
-        << "formatDate" << QVariant::fromValue(string)
+    const QString isoString(QLatin1String("2011/05/31 11:16:39.755"));
+    temporary = QDateTime::fromString(isoString, "yyyy/MM/dd HH:mm:ss.zzz");
+    const QString jsString = engine.coerceValue<QDateTime, QString>(temporary);
+    QTest::newRow("formatDate, qstring, ISO")
+        << "formatDate" << QVariant::fromValue(isoString)
         << (QStringList()
             << QLocale().toString(temporary.date(), QLocale::ShortFormat)
             << QLocale().toString(temporary.date(), QLocale::LongFormat)
             << temporary.date().toString("ddd MMMM d yy"));
-    QTest::newRow("formatDateTime, qstring")
-        << "formatDateTime" << QVariant::fromValue(string)
+    QTest::newRow("formatDate, qstring, JS")
+        << "formatDate" << QVariant::fromValue(jsString)
+        << (QStringList()
+            << QLocale().toString(temporary.date(), QLocale::ShortFormat)
+            << QLocale().toString(temporary.date(), QLocale::LongFormat)
+            << temporary.date().toString("ddd MMMM d yy"));
+    QTest::newRow("formatDateTime, qstring, ISO")
+        << "formatDateTime" << QVariant::fromValue(isoString)
         << (QStringList()
             << QLocale().toString(temporary, QLocale::ShortFormat)
             << QLocale().toString(temporary, QLocale::LongFormat)
             << temporary.toString("M/d/yy H:m:s a"));
-    QTest::newRow("formatTime, qstring")
-        << "formatTime" << QVariant::fromValue(string)
+    QTest::newRow("formatDateTime, qstring, JS")
+        << "formatDateTime" << QVariant::fromValue(jsString)
+        << (QStringList()
+            << QLocale().toString(temporary, QLocale::ShortFormat)
+            << QLocale().toString(temporary, QLocale::LongFormat)
+            << temporary.toString("M/d/yy H:m:s a"));
+    QTest::newRow("formatTime, qstring, ISO")
+        << "formatTime" << QVariant::fromValue(isoString)
         << (QStringList()
             << QLocale().toString(temporary.time(), QLocale::ShortFormat)
             << QLocale().toString(temporary.time(), QLocale::LongFormat)
             << temporary.time().toString("H:m:s a")
             << temporary.time().toString("hh:mm:ss.zzz"));
+    QTest::newRow("formatTime, qstring, JS")
+        << "formatTime" << QVariant::fromValue(jsString)
+        << (QStringList()
+            << QLocale().toString(temporary.time(), QLocale::ShortFormat)
+            << QLocale().toString(temporary.time(), QLocale::LongFormat)
+            << temporary.time().toString("H:m:s a")
+            << temporary.time().toString("hh:mm:ss.000")); // JS Date to string coercion drops milliseconds
 
     QColor color(Qt::red);
     temporary = QVariant::fromValue(color).toDateTime();
@@ -1246,12 +1293,13 @@ void tst_qqmlqt::later_data()
 
 void tst_qqmlqt::later()
 {
+    QQmlEngine engine;
     QFETCH(QString, function);
     QFETCH(QStringList, expectedWarnings);
     QFETCH(QStringList, propNames);
     QFETCH(QVariantList, values);
 
-    foreach (const QString &w, expectedWarnings)
+    for (const QString &w : std::as_const(expectedWarnings))
         QTest::ignoreMessage(QtWarningMsg, qPrintable(w));
 
     QQmlComponent component(&engine, testFileUrl("later.qml"));
@@ -1266,7 +1314,7 @@ void tst_qqmlqt::later()
             QCoreApplication::sendPostedEvents(nullptr, QEvent::DeferredDelete);
             QCoreApplication::processEvents();
         } else if (propNames.at(i) == QLatin1String("collectGarbage")) {
-            engine.collectGarbage();
+            gc(engine, GCFlags::DontSendPostedEvents);
         } else {
             QCOMPARE(root->property(qPrintable(propNames.at(i))), values.at(i));
         }
@@ -1331,6 +1379,7 @@ class TimeProvider: public QObject
 {
     Q_OBJECT
     QML_NAMED_ELEMENT(TimeProvider)
+    QML_UNCREATABLE("")
     Q_PROPERTY(QTime time READ time WRITE setTime NOTIFY timeChanged)
 
 public:
@@ -1428,6 +1477,20 @@ void tst_qqmlqt::timeRoundtrip()
     QCOMPARE(tp.m_getTime, tp.m_putTime);
 }
 
+void tst_qqmlqt::fontSetsProperties() {
+    QQmlComponent component(&engine, testFileUrl("fontProperties.qml"));
+
+    QScopedPointer<QObject> object(component.create());
+    QVERIFY(object != nullptr);
+
+    QFont fontProperty = qvariant_cast<QFont>(object->property("fontProperty"));
+    QVERIFY(fontProperty.styleStrategy() & QFont::ContextFontMerging);
+    QCOMPARE(fontProperty.variableAxisTags().size(), 1);
+    QCOMPARE(fontProperty.variableAxisValue("abcd"), 23.0625);
+    QCOMPARE(fontProperty.featureTags().size(), 1);
+    QCOMPARE(fontProperty.featureValue("abcd"), 23);
+}
+
 QTEST_MAIN(tst_qqmlqt)
 
 #include "tst_qqmlqt.moc"
