@@ -1,30 +1,5 @@
-/****************************************************************************
-**
-** Copyright (C) 2020 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the test suite of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:GPL-EXCEPT$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 as published by the Free Software
-** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2020 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 
 #include <qtest.h>
 
@@ -44,8 +19,7 @@
 #include <QtGui/private/qguiapplication_p.h>
 #include <QtGui/qpa/qplatformintegration.h>
 
-#include <QtGui/private/qrhi_p.h>
-#include <QtQuick/private/qquickrendercontrol_p.h>
+#include <rhi/qrhi.h>
 
 #if QT_CONFIG(vulkan)
 #include <QVulkanInstance>
@@ -88,6 +62,7 @@ private slots:
     void renderAndReadBackWithRhi_data();
     void renderAndReadBackWithRhi();
     void renderAndReadBackWithVulkanNative();
+    void renderAndReadBackWithVulkanAndCustomDepthTexture();
 
 private:
 #if QT_CONFIG(vulkan)
@@ -130,27 +105,35 @@ void tst_RenderControl::renderAndReadBackWithRhi_data()
     QTest::addColumn<QSGRendererInterface::GraphicsApi>("api");
 
 #if QT_CONFIG(opengl)
-    QTest::newRow("OpenGL") << QSGRendererInterface::OpenGLRhi;
+    QTest::newRow("OpenGL") << QSGRendererInterface::OpenGL;
 #endif
 #if QT_CONFIG(vulkan)
-    QTest::newRow("Vulkan") << QSGRendererInterface::VulkanRhi;
+    QTest::newRow("Vulkan") << QSGRendererInterface::Vulkan;
 #endif
 #ifdef Q_OS_WIN
-    QTest::newRow("D3D11") << QSGRendererInterface::Direct3D11Rhi;
+    QTest::newRow("D3D11") << QSGRendererInterface::Direct3D11;
+    QTest::newRow("D3D12") << QSGRendererInterface::Direct3D12;
 #endif
-#if defined(Q_OS_MACOS) || defined(Q_OS_IOS)
-    QTest::newRow("Metal") << QSGRendererInterface::MetalRhi;
+#if QT_CONFIG(metal)
+    QTest::newRow("Metal") << QSGRendererInterface::Metal;
 #endif
 }
 
 void tst_RenderControl::renderAndReadBackWithRhi()
 {
     QFETCH(QSGRendererInterface::GraphicsApi, api);
+
 #if QT_CONFIG(vulkan)
-    if (api == QSGRendererInterface::VulkanRhi && !vulkanInstance.isValid())
+    if (api == QSGRendererInterface::Vulkan && !vulkanInstance.isValid())
         QSKIP("Skipping Vulkan-based QRhi readback test due to failing to create a VkInstance");
 #endif
 
+#ifdef Q_OS_ANDROID
+    // QTBUG-102780
+    if (api == QSGRendererInterface::Vulkan)
+        QSKIP("Vulkan-based rendering tests on Android are flaky.");
+#endif
+
     // Changing the graphics api is not possible once a QQuickWindow et al is
     // created, however we do support changing it once all QQuickWindow,
     // QQuickRenderControl, etc. instances are destroyed, before creating new
@@ -161,7 +144,7 @@ void tst_RenderControl::renderAndReadBackWithRhi()
     QScopedPointer<QQuickRenderControl> renderControl(new QQuickRenderControl);
     QScopedPointer<QQuickWindow> quickWindow(new QQuickWindow(renderControl.data()));
 #if QT_CONFIG(vulkan)
-    if (api == QSGRendererInterface::VulkanRhi)
+    if (api == QSGRendererInterface::Vulkan)
         quickWindow->setVulkanInstance(&vulkanInstance);
 #endif
 
@@ -184,7 +167,8 @@ void tst_RenderControl::renderAndReadBackWithRhi()
 
     QQuickItem *rootItem = qobject_cast<QQuickItem *>(rootObject);
     QVERIFY(rootItem);
-    QCOMPARE(rootItem->size(), QSize(200, 200));
+    static const QSize ITEM_SIZE = QSize(200, 200);
+    QCOMPARE(rootItem->size(), ITEM_SIZE);
 
     quickWindow->contentItem()->setSize(rootItem->size());
     quickWindow->setGeometry(0, 0, rootItem->width(), rootItem->height());
@@ -199,7 +183,7 @@ void tst_RenderControl::renderAndReadBackWithRhi()
     // failed. The exception for now is OpenGL - that should (usually) work.
     if (!initSuccess) {
 #if QT_CONFIG(opengl)
-        if (api != QSGRendererInterface::OpenGLRhi
+        if (api != QSGRendererInterface::OpenGL
                 || !QGuiApplicationPrivate::platformIntegration()->hasCapability(QPlatformIntegration::OpenGL))
 #endif
         {
@@ -211,13 +195,7 @@ void tst_RenderControl::renderAndReadBackWithRhi()
 
     QCOMPARE(quickWindow->rendererInterface()->graphicsApi(), api);
 
-    // What comes now is technically cheating - as long as QRhi is not a public
-    // API this is not something applications can follow doing. However, it
-    // allows us to test out the pipeline without having to write 4 different
-    // native (Vulkan, Metal, D3D11, OpenGL) implementations of all what's below.
-
-    QQuickRenderControlPrivate *rd = QQuickRenderControlPrivate::get(renderControl.data());
-    QRhi *rhi = rd->rhi;
+    QRhi *rhi = renderControl->rhi();
     Q_ASSERT(rhi);
 
     const QSize size = rootItem->size().toSize();
@@ -225,9 +203,6 @@ void tst_RenderControl::renderAndReadBackWithRhi()
                                                     QRhiTexture::RenderTarget | QRhiTexture::UsedAsTransferSource));
     QVERIFY(tex->create());
 
-    // depth-stencil is mandatory with RHI, although strictly speaking the
-    // scenegraph could operate without one, but it has no means to figure out
-    // the lack of a ds buffer, so just be nice and provide one.
     QScopedPointer<QRhiRenderBuffer> ds(rhi->newRenderBuffer(QRhiRenderBuffer::DepthStencil, size, 1));
     QVERIFY(ds->create());
 
@@ -251,9 +226,27 @@ void tst_RenderControl::renderAndReadBackWithRhi()
             QVERIFY(tex->create()); // internally we now have a whole new native texture object
             ds->setPixelSize(currentSize);
             QVERIFY(ds->create());
-            QVERIFY(texRt->create());
-            // setRenderTarget is mandatory upon changing something, even if the texRt pointer itself is the same
-            quickWindow->setRenderTarget(QQuickRenderTarget::fromRhiRenderTarget(texRt.data()));
+
+            // Starting from Qt 6.3 we need neither a texRt->create() nor a
+            // quickWindow->setRenderTarget() here. Just recreating the
+            // internal native objects behing tex and ds should (must!) be
+            // functional on its own.
+
+        } else if (frame == 85) {
+            // like the above but now change the size radically so that we can
+            // test that rendering (viewports etc.) is corect.
+            currentSize = QSize(100, 100);
+            tex->setPixelSize(currentSize);
+            QVERIFY(tex->create());
+            ds->setPixelSize(currentSize);
+            QVERIFY(ds->create());
+        } else if (frame == 86) {
+            // reset to the default size
+            currentSize = ITEM_SIZE;
+            tex->setPixelSize(currentSize);
+            QVERIFY(tex->create());
+            ds->setPixelSize(currentSize);
+            QVERIFY(ds->create());
         } else if (frame == 90) {
             // Go berserk, destroy and recreate the texture and related stuff
             // (the QRhi objects themselves, not just the native stuff
@@ -305,7 +298,7 @@ void tst_RenderControl::renderAndReadBackWithRhi()
         };
         QRhiResourceUpdateBatch *readbackBatch = rhi->nextResourceUpdateBatch();
         readbackBatch->readBackTexture(tex.data(), &readResult);
-        rd->cb->resourceUpdate(readbackBatch);
+        renderControl->commandBuffer()->resourceUpdate(readbackBatch);
 
         // our frame is done, submit
         renderControl->endFrame();
@@ -328,24 +321,31 @@ void tst_RenderControl::renderAndReadBackWithRhi()
         QVERIFY(qAbs(qGreen(background) - 130) < maxFuzz);
         QVERIFY(qAbs(qBlue(background) - 180) < maxFuzz);
 
-        background = img.pixel(195, 195);
-        QVERIFY(qAbs(qRed(background) - 70) < maxFuzz);
-        QVERIFY(qAbs(qGreen(background) - 130) < maxFuzz);
-        QVERIFY(qAbs(qBlue(background) - 180) < maxFuzz);
+        // Frame 85 is where we resize to (100, 100), skip for that but from 86
+        // we will back to the proper (200, 200). If failures occur from frame
+        // 85 or - more likely - 86, that is likely because the
+        // scenegraph/QQuickWindow does not correctly pick up the render target
+        // size changes.
+        if (frame != 85) {
+            background = img.pixel(195, 195);
+            QVERIFY(qAbs(qRed(background) - 70) < maxFuzz);
+            QVERIFY(qAbs(qGreen(background) - 130) < maxFuzz);
+            QVERIFY(qAbs(qBlue(background) - 180) < maxFuzz);
 
-        // after about 1.25 seconds (animation time, one iteration is 16 ms
-        // thanks to our custom animation driver) the rectangle reaches a 90
-        // degree rotation, that should be frame 76
-        if (frame <= 2 || (frame >= 76 && frame <= 80)) {
-            QRgb c = img.pixel(28, 28); // rectangle
-            QVERIFY(qAbs(qRed(c) - 152) < maxFuzz);
-            QVERIFY(qAbs(qGreen(c) - 251) < maxFuzz);
-            QVERIFY(qAbs(qBlue(c) - 152) < maxFuzz);
-        } else {
-            QRgb c = img.pixel(28, 28); // background because rectangle got rotated so this pixel is not covered by it
-            QVERIFY(qAbs(qRed(c) - 70) < maxFuzz);
-            QVERIFY(qAbs(qGreen(c) - 130) < maxFuzz);
-            QVERIFY(qAbs(qBlue(c) - 180) < maxFuzz);
+            // after about 1.25 seconds (animation time, one iteration is 16 ms
+            // thanks to our custom animation driver) the rectangle reaches a 90
+            // degree rotation, that should be frame 76
+            if (frame <= 2 || (frame >= 76 && frame <= 80)) {
+                QRgb c = img.pixel(28, 28); // rectangle
+                QVERIFY(qAbs(qRed(c) - 152) < maxFuzz);
+                QVERIFY(qAbs(qGreen(c) - 251) < maxFuzz);
+                QVERIFY(qAbs(qBlue(c) - 152) < maxFuzz);
+            } else {
+                QRgb c = img.pixel(28, 28); // background because rectangle got rotated so this pixel is not covered by it
+                QVERIFY(qAbs(qRed(c) - 70) < maxFuzz);
+                QVERIFY(qAbs(qGreen(c) - 130) < maxFuzz);
+                QVERIFY(qAbs(qBlue(c) - 180) < maxFuzz);
+            }
         }
     }
 }
@@ -356,7 +356,7 @@ void tst_RenderControl::renderAndReadBackWithVulkanNative()
     if (!vulkanInstance.isValid())
         QSKIP("Skipping native Vulkan test due to failing to create a VkInstance");
 
-    QQuickWindow::setGraphicsApi(QSGRendererInterface::VulkanRhi);
+    QQuickWindow::setGraphicsApi(QSGRendererInterface::Vulkan);
 
     // We will create our own VkDevice and friends, which will then get used by
     // Qt Quick as well (instead of creating its own objects), so this is a test
@@ -384,7 +384,7 @@ void tst_RenderControl::renderAndReadBackWithVulkanNative()
     f->vkGetPhysicalDeviceQueueFamilyProperties(physDev, &queueCount, queueFamilyProps.data());
 
     int gfxQueueFamilyIdx = -1;
-    for (int i = 0; i < queueFamilyProps.count(); ++i) {
+    for (int i = 0; i < queueFamilyProps.size(); ++i) {
         if (queueFamilyProps[i].queueFlags & VK_QUEUE_GRAPHICS_BIT) {
             gfxQueueFamilyIdx = i;
             break;
@@ -662,21 +662,6 @@ void tst_RenderControl::renderAndReadBackWithVulkanNative()
             QVERIFY(qAbs(qGreen(background) - 130) < maxFuzz);
             QVERIFY(qAbs(qBlue(background) - 180) < maxFuzz);
 
-            // after about 1.25 seconds (animation time, one iteration is 16 ms
-            // thanks to our custom animation driver) the rectangle reaches a 90
-            // degree rotation, that should be frame 76
-            if (frame <= 2 || (frame >= 76 && frame <= 80)) {
-                QRgb c = img.pixel(28, 28); // rectangle
-                QVERIFY(qAbs(qRed(c) - 152) < maxFuzz);
-                QVERIFY(qAbs(qGreen(c) - 251) < maxFuzz);
-                QVERIFY(qAbs(qBlue(c) - 152) < maxFuzz);
-            } else {
-                QRgb c = img.pixel(28, 28); // background because rectangle got rotated so this pixel is not covered by it
-                QVERIFY(qAbs(qRed(c) - 70) < maxFuzz);
-                QVERIFY(qAbs(qGreen(c) - 130) < maxFuzz);
-                QVERIFY(qAbs(qBlue(c) - 180) < maxFuzz);
-            }
-
             img = QImage();
             df->vkUnmapMemory(dev, bufMem);
         }
@@ -698,6 +683,151 @@ void tst_RenderControl::renderAndReadBackWithVulkanNative()
 #endif
 }
 
+void tst_RenderControl::renderAndReadBackWithVulkanAndCustomDepthTexture()
+{
+#if QT_CONFIG(vulkan)
+    if (!vulkanInstance.isValid())
+        QSKIP("Skipping native Vulkan test due to failing to create a VkInstance");
+
+    QQuickWindow::setGraphicsApi(QSGRendererInterface::Vulkan);
+
+#ifdef Q_OS_ANDROID
+    if (QQuickWindow::graphicsApi() == QSGRendererInterface::Vulkan)
+        QSKIP("Vulkan support is broken in Android emulator, skipping");
+#endif
+
+    QScopedPointer<QQuickRenderControl> renderControl(new QQuickRenderControl);
+    QScopedPointer<QQuickWindow> quickWindow(new QQuickWindow(renderControl.data()));
+
+    quickWindow->setVulkanInstance(&vulkanInstance);
+
+    QScopedPointer<QQmlEngine> qmlEngine(new QQmlEngine);
+    QScopedPointer<QQmlComponent> qmlComponent(new QQmlComponent(qmlEngine.data(),
+                                                                 testFileUrl(QLatin1String("rect_depth.qml"))));
+    QVERIFY(!qmlComponent->isLoading());
+    if (qmlComponent->isError()) {
+        for (const QQmlError &error : qmlComponent->errors())
+            qWarning() << error.url() << error.line() << error;
+    }
+    QVERIFY(!qmlComponent->isError());
+
+    QObject *rootObject = qmlComponent->create();
+    if (qmlComponent->isError()) {
+        for (const QQmlError &error : qmlComponent->errors())
+            qWarning() << error.url() << error.line() << error;
+    }
+    QVERIFY(!qmlComponent->isError());
+
+    QQuickItem *rootItem = qobject_cast<QQuickItem *>(rootObject);
+    QVERIFY(rootItem);
+    static const QSize ITEM_SIZE = QSize(200, 200);
+    QCOMPARE(rootItem->size(), ITEM_SIZE);
+
+    quickWindow->contentItem()->setSize(rootItem->size());
+    quickWindow->setGeometry(0, 0, rootItem->width(), rootItem->height());
+
+    rootItem->setParentItem(quickWindow->contentItem());
+
+    const bool initSuccess = renderControl->initialize();
+
+    if (!initSuccess)
+        QSKIP("Could not initialize graphics, perhaps unsupported graphics API, skipping");
+
+    QRhi *rhi = renderControl->rhi();
+    Q_ASSERT(rhi);
+
+    const QSize size = rootItem->size().toSize();
+    QScopedPointer<QRhiTexture> tex(rhi->newTexture(QRhiTexture::RGBA8, size, 1,
+                                                    QRhiTexture::RenderTarget | QRhiTexture::UsedAsTransferSource));
+    QVERIFY(tex->create());
+
+    // In this test we use QRhiTexture to create resources but pull out the
+    // native VkImage and pass that in via fromVulkanImage(). This is to
+    // exercise additional features, such as setting a custom depth texture,
+    // which is not available (not relevant) when using fromRhiRenderTarget().
+    // In particular, this setup (fromVulkanImage + setDepthTexture) exercises
+    // what we get in Qt Quick 3D with OpenXR, where rendering happens into a
+    // XrSwapchain-provided color and depth texture. (granted, here we only
+    // exercise the non-multiview, non-MSAA case).
+
+    QScopedPointer<QRhiTexture> depthTex(rhi->newTexture(QRhiTexture::D24S8, size, 1, QRhiTexture::RenderTarget));
+    QVERIFY(depthTex->create());
+
+    QQuickRenderTarget rt = QQuickRenderTarget::fromVulkanImage(VkImage(tex->nativeTexture().object),
+                                                                        VkImageLayout(tex->nativeTexture().layout),
+                                                                        VK_FORMAT_R8G8B8A8_UNORM,
+                                                                        VK_FORMAT_R8G8B8A8_UNORM,
+                                                                        size,
+                                                                        1,
+                                                                        0,
+                                                                        {});
+    rt.setDepthTexture(depthTex.data());
+    quickWindow->setRenderTarget(rt);
+
+    QSize currentSize = size;
+
+    for (int frame = 0; frame < 100; ++frame) {
+        QCoreApplication::processEvents();
+
+        if (frame > 0)
+            animDriver->advance();
+
+        renderControl->polishItems();
+
+        renderControl->beginFrame();
+
+        renderControl->sync();
+        renderControl->render();
+
+        bool readCompleted = false;
+        QRhiReadbackResult readResult;
+        QImage result;
+        readResult.completed = [&readCompleted, &readResult, &result, &rhi] {
+            readCompleted = true;
+            QImage wrapperImage(reinterpret_cast<const uchar *>(readResult.data.constData()),
+                                readResult.pixelSize.width(), readResult.pixelSize.height(),
+                                QImage::Format_RGBA8888_Premultiplied);
+            if (rhi->isYUpInFramebuffer())
+                result = wrapperImage.mirrored();
+            else
+                result = wrapperImage.copy();
+        };
+        QRhiResourceUpdateBatch *readbackBatch = rhi->nextResourceUpdateBatch();
+        readbackBatch->readBackTexture(tex.data(), &readResult);
+        renderControl->commandBuffer()->resourceUpdate(readbackBatch);
+
+        renderControl->endFrame();
+
+        QVERIFY(readCompleted);
+
+        QImage img = result;
+        QVERIFY(!img.isNull());
+        QCOMPARE(img.size(), currentSize);
+
+        const int maxFuzz = 2;
+
+        // The scene is: background, rectangle, text
+        // where rectangle rotates
+
+        QRgb background = img.pixel(5, 5);
+
+        // with rect_depth.qml the following would not pass if the depth buffering
+        // was not functional (because the red rectangle with z: -1 would still
+        // go on top, not below)
+        QVERIFY(qAbs(qRed(background) - 70) < maxFuzz);
+        QVERIFY(qAbs(qGreen(background) - 130) < maxFuzz);
+        QVERIFY(qAbs(qBlue(background) - 180) < maxFuzz);
+
+        background = img.pixel(195, 195);
+        QVERIFY(qAbs(qRed(background) - 70) < maxFuzz);
+        QVERIFY(qAbs(qGreen(background) - 130) < maxFuzz);
+        QVERIFY(qAbs(qBlue(background) - 180) < maxFuzz);
+    }
+#else
+    QSKIP("No Vulkan support in Qt build, skipping native Vulkan test");
+#endif
+}
+
 #include "tst_qquickrendercontrol.moc"
 
 QTEST_MAIN(tst_RenderControl)
