@@ -1,5 +1,5 @@
 // Copyright (C) 2016 The Qt Company Ltd.
-// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 
 #include <qtest.h>
 #include <qtesttouch.h>
@@ -10,7 +10,13 @@
 #include <QtQuick/qquickview.h>
 #include <QtQuick/qquickitem.h>
 #include <QtQuick/private/qquickitem_p.h>
+#include <QtQuick/private/qquickmousearea_p.h>
+#include <QtQuick/private/qquicktaphandler_p.h>
+#include <QtQuickTemplates2/private/qquickbutton_p.h>
+#include <QtQuickTemplates2/private/qquickoverlay_p.h>
+#include <QtQuickTemplates2/private/qquickpopup_p.h>
 #include <QtQuickTestUtils/private/qmlutils_p.h>
+#include <QtQuickTestUtils/private/visualtestutils_p.h>
 #include <QtGui/QWindow>
 #include <QtGui/QScreen>
 #include <QtGui/QImage>
@@ -21,6 +27,7 @@
 #include <QtGui/qstylehints.h>
 #include <QtWidgets/QBoxLayout>
 #include <QtWidgets/QLabel>
+#include <QtWidgets/private/qapplication_p.h>
 
 #include <QtQuickWidgets/QQuickWidget>
 
@@ -123,13 +130,26 @@ private slots:
     void mouseEventWindowPos();
     void synthMouseFromTouch_data();
     void synthMouseFromTouch();
+    void touchTapMouseArea();
+    void touchTapButton();
+    void touchTapHandler_data();
+    void touchTapHandler();
+    void touchMultipleWidgets();
     void tabKey();
+    void focusChain_data();
+    void focusChain();
     void resizeOverlay();
+    void overlayGeometry_data();
+    void overlayGeometry();
     void controls();
     void focusOnClick();
 #if QT_CONFIG(graphicsview)
     void focusOnClickInProxyWidget();
 #endif
+    void focusPreserved();
+    void accessibilityHandlesViewChange();
+    void cleanupRhi();
+    void dontRecreateRootElementOnWindowChange();
 
 private:
     QPointingDevice *device = QTest::createTouchDevice();
@@ -208,6 +228,7 @@ void tst_qquickwidget::resizemodeitem()
     QCOMPARE(view->size(), QSize(200, 200));
     QCOMPARE(view->size(), view->sizeHint());
     QCOMPARE(view->size(), view->initialSize());
+    QCOMPARE(view->quickWindow()->contentItem()->size(), view->size());
 
     // size update from view
     view->resize(QSize(80,100));
@@ -216,6 +237,8 @@ void tst_qquickwidget::resizemodeitem()
     QCOMPARE(item->height(), 100.0);
     QCOMPARE(view->size(), QSize(80, 100));
     QCOMPARE(view->size(), view->sizeHint());
+    QCOMPARE(view->quickWindow()->contentItem()->size(), view->size());
+
 
     view->setResizeMode(QQuickWidget::SizeViewToRootObject);
 
@@ -224,6 +247,8 @@ void tst_qquickwidget::resizemodeitem()
     QCOMPARE(item->width(), 80.0);
     QCOMPARE(item->height(), 100.0);
     QTRY_COMPARE(view->size(), QSize(60, 80));
+    QCOMPARE(view->quickWindow()->contentItem()->size(), view->size());
+
 
     // size update from root object
     item->setWidth(250);
@@ -233,6 +258,7 @@ void tst_qquickwidget::resizemodeitem()
     QTRY_COMPARE(view->size(), QSize(250, 350));
     QCOMPARE(view->size(), QSize(250, 350));
     QCOMPARE(view->size(), view->sizeHint());
+    QCOMPARE(view->quickWindow()->contentItem()->size(), view->size());
 
     // reset window
     window.hide();
@@ -250,6 +276,7 @@ void tst_qquickwidget::resizemodeitem()
     QCOMPARE(item->height(), 200.0);
     QCOMPARE(view->size(), view->sizeHint());
     QCOMPARE(view->size(), view->initialSize());
+    QCOMPARE(view->quickWindow()->contentItem()->size(), view->size());
 
     // size update from root object
     item->setWidth(80);
@@ -258,6 +285,7 @@ void tst_qquickwidget::resizemodeitem()
     QCOMPARE(item->height(), 100.0);
     QTRY_COMPARE(view->size(), QSize(80, 100));
     QCOMPARE(view->size(), view->sizeHint());
+    QCOMPARE(view->quickWindow()->contentItem()->size(), view->size());
 
     // size update from root object disabled
     view->setResizeMode(QQuickWidget::SizeRootObjectToView);
@@ -266,6 +294,7 @@ void tst_qquickwidget::resizemodeitem()
     QCOMPARE(view->width(), 80);
     QCOMPARE(view->height(), 100);
     QCOMPARE(QSize(item->width(), item->height()), view->sizeHint());
+    QCOMPARE(view->quickWindow()->contentItem()->size(), view->size());
 
     // size update from view
     view->resize(QSize(200,300));
@@ -273,6 +302,7 @@ void tst_qquickwidget::resizemodeitem()
     QCOMPARE(item->height(), 300.0);
     QCOMPARE(view->size(), QSize(200, 300));
     QCOMPARE(view->size(), view->sizeHint());
+    QCOMPARE(view->quickWindow()->contentItem()->size(), view->size());
 
     window.hide();
 
@@ -296,6 +326,8 @@ void tst_qquickwidget::resizemodeitem()
     QTRY_COMPARE(view->size(), QSize(300, 300));
     QCOMPARE(view->size(), view->sizeHint());
     QCOMPARE(view->initialSize(), QSize(200, 200)); // initial object size
+    QCOMPARE(view->quickWindow()->contentItem()->size(), view->size());
+
 }
 
 void tst_qquickwidget::layoutSizeChange()
@@ -613,6 +645,7 @@ void tst_qquickwidget::synthMouseFromTouch()
     QTest::touchEvent(&window, device).move(0, p2, &window);
     QTest::touchEvent(&window, device).release(0, p2, &window);
 
+    qCDebug(lcTests) << item->m_touchEvents << item->m_mouseEvents;
     QCOMPARE(item->m_touchEvents.size(), synthMouse ? 0 : (acceptTouch ? 3 : 1));
     QCOMPARE(item->m_mouseEvents.size(), synthMouse ? 3 : 0);
     QCOMPARE(childView->m_mouseEvents.size(), 0);
@@ -620,6 +653,141 @@ void tst_qquickwidget::synthMouseFromTouch()
         QCOMPARE(ev, Qt::MouseEventSynthesizedByQt);
 }
 
+void tst_qquickwidget::touchTapMouseArea()
+{
+    QWidget window;
+    window.resize(100, 100);
+    window.setObjectName("window widget");
+    window.setAttribute(Qt::WA_AcceptTouchEvents);
+    QVERIFY(QCoreApplication::testAttribute(Qt::AA_SynthesizeMouseForUnhandledTouchEvents));
+    QQuickWidget *quick = new QQuickWidget(&window);
+    quick->setSource(testFileUrl("mouse.qml"));
+    quick->move(50, 50);
+    quick->setObjectName("quick widget");
+    window.show();
+    QVERIFY(QTest::qWaitForWindowExposed(&window));
+    QQuickItem *rootItem = quick->rootObject();
+    QVERIFY(rootItem);
+    QQuickMouseArea *ma = rootItem->findChild<QQuickMouseArea *>();
+    QVERIFY(ma);
+
+    QPoint p1 = QPoint(70, 70);
+    QTest::touchEvent(&window, device).press(0, p1, &window);
+    QTRY_COMPARE(ma->isPressed(), true);
+    QTest::touchEvent(&window, device).move(0, p1, &window);
+    QTest::touchEvent(&window, device).release(0, p1, &window);
+    QTRY_COMPARE(ma->isPressed(), false);
+    QVERIFY(rootItem->property("wasClicked").toBool());
+}
+
+void tst_qquickwidget::touchTapButton()
+{
+    QWidget window;
+    QQuickWidget *quick = new QQuickWidget;
+    quick->setSource(testFileUrl("button.qml"));
+
+    QHBoxLayout hbox;
+    hbox.addWidget(quick);
+    window.setLayout(&hbox);
+    window.show();
+    QVERIFY(QTest::qWaitForWindowExposed(&window));
+
+    QQuickItem *rootItem = quick->rootObject();
+    QVERIFY(rootItem);
+    QQuickButton *button = rootItem->findChild<QQuickButton *>("button");
+    QVERIFY(button);
+
+    const QPoint point = quick->mapTo(&window, button->mapToScene(button->boundingRect().center()).toPoint());
+    QTest::touchEvent(&window, device).press(0, point, &window).commit();
+    QTRY_VERIFY(rootItem->property("wasPressed").toBool());
+    QTest::touchEvent(&window, device).release(0, point, &window).commit();
+    QTRY_VERIFY(rootItem->property("wasReleased").toBool());
+    QTRY_VERIFY(rootItem->property("wasClicked").toBool());
+}
+
+void tst_qquickwidget::touchTapHandler_data()
+{
+    QTest::addColumn<bool>("guiSynthMouse"); // AA_SynthesizeMouseForUnhandledTouchEvents
+    QTest::addColumn<QQuickTapHandler::GesturePolicy>("gesturePolicy");
+
+    // QTest::newRow("nosynth: passive grab") << false << QQuickTapHandler::DragThreshold; // still failing
+    QTest::newRow("nosynth: exclusive grab") << false << QQuickTapHandler::ReleaseWithinBounds;
+    QTest::newRow("allowsynth: passive grab") << true << QQuickTapHandler::DragThreshold; // QTBUG-113558
+    QTest::newRow("allowsynth: exclusive grab") << true << QQuickTapHandler::ReleaseWithinBounds;
+}
+
+void tst_qquickwidget::touchTapHandler()
+{
+    QFETCH(bool, guiSynthMouse);
+    QFETCH(QQuickTapHandler::GesturePolicy, gesturePolicy);
+
+    QCoreApplication::setAttribute(Qt::AA_SynthesizeMouseForUnhandledTouchEvents, guiSynthMouse);
+    QQuickWidget quick;
+    if (!quick.testAttribute(Qt::WA_AcceptTouchEvents))
+        QSKIP("irrelevant on non-touch platforms");
+
+    quick.setSource(testFileUrl("tapHandler.qml"));
+    quick.show();
+    QVERIFY(QTest::qWaitForWindowExposed(&quick));
+
+    QQuickItem *rootItem = quick.rootObject();
+    QVERIFY(rootItem);
+    QQuickTapHandler *th = rootItem->findChild<QQuickTapHandler *>();
+    QVERIFY(th);
+    th->setGesturePolicy(gesturePolicy);
+    QSignalSpy tappedSpy(th, &QQuickTapHandler::tapped);
+
+    const QPoint p(50, 50);
+    QTest::touchEvent(&quick, device).press(0, p, &quick);
+    QTRY_COMPARE(th->isPressed(), true);
+    QTest::touchEvent(&quick, device).release(0, p, &quick);
+    QTRY_COMPARE(tappedSpy.size(), 1);
+    QCOMPARE(th->isPressed(), false);
+}
+
+void tst_qquickwidget::touchMultipleWidgets()
+{
+    QWidget window;
+    QQuickWidget *leftQuick = new QQuickWidget;
+    leftQuick->setSource(testFileUrl("button.qml"));
+    if (!leftQuick->testAttribute(Qt::WA_AcceptTouchEvents))
+        QSKIP("irrelevant on non-touch platforms");
+
+    QQuickWidget *rightQuick = new QQuickWidget;
+    rightQuick->setSource(testFileUrl("button.qml"));
+
+    QHBoxLayout hbox;
+    hbox.addWidget(leftQuick);
+    hbox.addWidget(rightQuick);
+    window.setLayout(&hbox);
+    window.show();
+    QVERIFY(QTest::qWaitForWindowExposed(&window));
+
+    QQuickItem *leftRootItem = leftQuick->rootObject();
+    QQuickItem *rightRootItem = rightQuick->rootObject();
+    QVERIFY(leftRootItem);
+    QVERIFY(rightRootItem);
+    QQuickButton *leftButton = leftRootItem->findChild<QQuickButton *>("button");
+    QQuickButton *rightButton = rightRootItem->findChild<QQuickButton *>("button");
+    QVERIFY(leftButton);
+    QVERIFY(rightButton);
+
+    const QPoint leftPoint = leftQuick->mapTo(&window, leftButton->mapToScene(
+                                              leftButton->boundingRect().center()).toPoint());
+    const QPoint rightPoint = rightQuick->mapTo(&window, rightButton->mapToScene(
+                                                rightButton->boundingRect().center()).toPoint());
+    QTest::touchEvent(&window, device).press(0, leftPoint, &window).commit();
+    QTRY_VERIFY(leftRootItem->property("wasPressed").toBool());
+    QTest::touchEvent(&window, device).press(1, rightPoint, &window).commit();
+    QTRY_VERIFY(rightRootItem->property("wasPressed").toBool());
+    QTest::touchEvent(&window, device).release(1, rightPoint, &window).commit();
+    QTRY_VERIFY(rightRootItem->property("wasReleased").toBool());
+    QVERIFY(rightRootItem->property("wasClicked").toBool());
+    QTest::touchEvent(&window, device).release(0, leftPoint, &window).commit();
+    QTRY_VERIFY(leftRootItem->property("wasReleased").toBool());
+    QVERIFY(leftRootItem->property("wasClicked").toBool());
+}
+
 void tst_qquickwidget::tabKey()
 {
     if (QGuiApplication::styleHints()->tabFocusBehavior() != Qt::TabFocusAllControls)
@@ -631,9 +799,9 @@ void tst_qquickwidget::tabKey()
     qqw2->setSource(testFileUrl("noActiveFocusOnTab.qml"));
     qqw2->move(100, 0);
     window1.show();
-    qqw->setFocus();
     QVERIFY(QTest::qWaitForWindowExposed(&window1, 5000));
-    QVERIFY(qqw->hasFocus());
+    qqw->setFocus();
+    QTRY_VERIFY(qqw->hasFocus());
     QQuickItem *item = qobject_cast<QQuickItem *>(qqw->rootObject());
     QQuickItem *topItem = item->findChild<QQuickItem *>("topRect");
     QQuickItem *middleItem = item->findChild<QQuickItem *>("middleRect");
@@ -654,7 +822,98 @@ void tst_qquickwidget::tabKey()
     QVERIFY(topItem2->property("activeFocus").toBool());
     QTest::keyClick(qqw2, Qt::Key_Tab);
     QTRY_VERIFY(qqw->hasFocus());
-    QVERIFY(middleItem->property("activeFocus").toBool());
+    // Focus-in by tab/backtab events makes the first/last focusable item to be focused
+    QVERIFY(topItem->property("activeFocus").toBool());
+}
+
+void tst_qquickwidget::focusChain_data()
+{
+    if (QGuiApplication::styleHints()->tabFocusBehavior() != Qt::TabFocusAllControls)
+        QSKIP("This function doesn't support NOT iterating all.");
+
+    QTest::addColumn<bool>("forward");
+
+    QTest::addRow("Forward") << true;
+    QTest::addRow("Backward") << false;
+}
+
+void tst_qquickwidget::focusChain()
+{
+    QFETCH(bool, forward);
+
+    QWidget window;
+    window.resize(300, 300);
+    QHBoxLayout layout(&window);
+    QQuickWidget qqw1;
+    qqw1.setObjectName("qqw1");
+    qqw1.setResizeMode(QQuickWidget::SizeRootObjectToView);
+    qqw1.setSource(testFileUrl("activeFocusOnTab.qml"));
+    QWidget middleWidget;
+    middleWidget.setObjectName("middleWidget");
+    middleWidget.setFocusPolicy(Qt::FocusPolicy::TabFocus);
+    QQuickWidget qqw2;
+    qqw2.setObjectName("qqw2");
+    qqw1.setResizeMode(QQuickWidget::SizeRootObjectToView);
+    qqw2.setSource(testFileUrl("activeFocusOnTab.qml"));
+    layout.addWidget(&qqw1);
+    layout.addWidget(&middleWidget);
+    layout.addWidget(&qqw2);
+    window.show();
+    window.windowHandle()->requestActivate();
+    QVERIFY(QTest::qWaitForWindowExposed(&window));
+
+    QQuickItem *root1 = qqw1.rootObject();
+    QVERIFY(root1);
+    QQuickItem *topItem1 = root1->findChild<QQuickItem *>("topRect");
+    topItem1->setObjectName("topRect1");
+    QVERIFY(topItem1);
+    QQuickItem *middleItem1 = root1->findChild<QQuickItem *>("middleRect");
+    middleItem1->setObjectName("middleRect1");
+    QVERIFY(middleItem1);
+    QQuickItem *bottomItem1 = root1->findChild<QQuickItem *>("bottomRect");
+    bottomItem1->setObjectName("bottomRect1");
+    QVERIFY(bottomItem1);
+
+    QQuickItem *root2 = qqw2.rootObject();
+    QVERIFY(root2);
+    QQuickItem *topItem2 = root2->findChild<QQuickItem *>("topRect");
+    topItem2->setObjectName("topRect2");
+    QVERIFY(topItem2);
+    QQuickItem *middleItem2 = root2->findChild<QQuickItem *>("middleRect");
+    middleItem2->setObjectName("middleRect2");
+    QVERIFY(middleItem2);
+    QQuickItem *bottomItem2 = root2->findChild<QQuickItem *>("bottomRect");
+    bottomItem2->setObjectName("bottomRect2");
+    QVERIFY(bottomItem2);
+
+    qqw1.setFocus();
+    QTRY_VERIFY(qqw1.hasFocus());
+
+    auto hasActiveFocus = [](const QObject *targetObject) -> bool
+    {
+        if (const QQuickItem *targetItem = qobject_cast<const QQuickItem *>(targetObject))
+            return targetItem->hasActiveFocus();
+        else if (const QWidget *targetWidget = qobject_cast<const QWidget *>(targetObject))
+            return targetWidget->hasFocus();
+        return false;
+    };
+
+    QList<QObject *> expectedFocusChain;
+    if (forward) {
+        expectedFocusChain << topItem1 << middleItem1 << bottomItem1 << &middleWidget << topItem2
+                           << middleItem2 << bottomItem2 << topItem1;
+    } else {
+        expectedFocusChain << topItem1 << bottomItem2 << middleItem2 << topItem2 << &middleWidget
+                           << bottomItem1 << middleItem1 << topItem1;
+    }
+
+    const Qt::Key tabKey = forward ? Qt::Key_Tab : Qt::Key_Backtab;
+
+    for (QObject *expectedFocusTarget : expectedFocusChain) {
+        const QByteArray objectName = expectedFocusTarget->objectName().toLatin1();
+        QTRY_VERIFY2(hasActiveFocus(expectedFocusTarget), "expectedFocusTarget: " + objectName);
+        QTest::keyClick(QGuiApplication::focusWindow(), tabKey);
+    }
 }
 
 class Overlay : public QQuickItem, public QQuickItemChangeListener
@@ -725,6 +984,55 @@ void tst_qquickwidget::resizeOverlay()
     QCOMPARE(overlay->height(), rootItem->height());
 }
 
+/*
+    Overlaps with resizeOverlay, but uses a proper
+    Qt Quick Controls Overlay.
+*/
+void tst_qquickwidget::overlayGeometry_data()
+{
+    QTest::addColumn<QQuickWidget::ResizeMode>("resizeMode");
+
+    QTest::addRow("SizeRootObjectToView") << QQuickWidget::SizeRootObjectToView;
+    QTest::addRow("SizeViewToRootObject") << QQuickWidget::SizeViewToRootObject;
+}
+
+void tst_qquickwidget::overlayGeometry()
+{
+    QFETCH(const QQuickWidget::ResizeMode, resizeMode);
+
+    QWidget window;
+
+    QQuickWidget widget(testFileUrl("overlayGeometry.qml"), &window);
+    widget.setResizeMode(resizeMode);
+
+    QCOMPARE(widget.status(), QQuickWidget::Ready);
+    const auto *rootItem = qobject_cast<QQuickItem*>(widget.rootObject());
+    QVERIFY(rootItem);
+    const QSizeF preferredSize = rootItem->size();
+
+    auto *popup = rootItem->findChild<QQuickPopup *>("popup");
+    QVERIFY(popup);
+    popup->open();
+    QTRY_VERIFY(popup->isOpened());
+
+    const auto *overlay = QQuickOverlay::overlay(widget.quickWindow());
+    QVERIFY(overlay);
+
+    QTestPrivate::androidCompatibleShow(&window);
+    QVERIFY(QTest::qWaitForWindowExposed(&window));
+    QCOMPARE(widget.size(), preferredSize.toSize());
+    QCOMPARE(overlay->position(), QPointF(0, 0));
+    QCOMPARE(overlay->size(), widget.size());
+
+    widget.resize((preferredSize * 2).toSize());
+    QCOMPARE(overlay->position(), QPointF(0, 0));
+    QCOMPARE(overlay->size(), widget.size());
+
+    // click outside the popup, make sure it closes
+    QTest::mouseClick(&widget, Qt::LeftButton, {}, QPoint(10, 10));
+    QTRY_VERIFY(!popup->isOpened());
+}
+
 void tst_qquickwidget::controls()
 {
     // Smoke test for having some basic Quick Controls in a scene in a QQuickWidget.
@@ -841,6 +1149,138 @@ void tst_qquickwidget::focusOnClickInProxyWidget()
 }
 #endif
 
+void tst_qquickwidget::focusPreserved()
+{
+    SKIP_IF_NO_WINDOW_ACTIVATION
+    if (QGuiApplication::platformName() == "android")
+        QSKIP("Test doesn't exit cleanly on Android and generates many warnings - QTBUG-112696");
+
+    QScopedPointer<QWidget> widget(new QWidget());
+    QScopedPointer<QQuickWidget> quick(new QQuickWidget());
+    QQuickItem *root = new QQuickItem(); // will be owned by quick after setContent
+    QScopedPointer<QQuickItem> content(new QQuickItem());
+    content->setActiveFocusOnTab(true);
+    content->setFocus(true);
+    quick->setFocusPolicy(Qt::StrongFocus);
+    quick->setContent(QUrl(), nullptr, root);
+    root->setFlag(QQuickItem::ItemHasContents);
+    content->setParentItem(root);
+
+    // added content2 and content3 to test more complicated case
+    QScopedPointer<QQuickItem> content2(new QQuickItem());
+    content2->setActiveFocusOnTab(true);
+    content2->setFocus(true);
+    content2->setParentItem(root);
+    QScopedPointer<QQuickItem> content3(new QQuickItem());
+    content3->setActiveFocusOnTab(true);
+    content3->setFocus(true);
+    content3->setParentItem(root);
+
+    quick->setGeometry(0, 0, 200, 200);
+    quick->show();
+    quick->setFocus();
+    quick->activateWindow();
+    QVERIFY(QTest::qWaitForWindowExposed(quick.get()));
+    QTRY_VERIFY(quick->hasFocus());
+    QTRY_VERIFY(content->hasFocus());
+    QTRY_VERIFY(content->hasActiveFocus());
+
+    content2->forceActiveFocus();
+    QVERIFY(content2->hasFocus());
+    QVERIFY(content2->hasActiveFocus());
+
+    widget->show();
+    widget->setFocus();
+    widget->activateWindow();
+    QVERIFY(QTest::qWaitForWindowExposed(widget.get()));
+    QTRY_VERIFY(widget->hasFocus());
+
+    quick->setParent(widget.get());
+
+    quick->show();
+    quick->setFocus();
+    quick->activateWindow();
+    QTRY_VERIFY(quick->hasFocus());
+    QTRY_VERIFY(content2->hasFocus());
+    QTRY_VERIFY(content2->hasActiveFocus());
+}
+
+/*
+    Reparenting the QQuickWidget recreates the offscreen QQuickWindow.
+    Since the accessible interface that is cached for the QQuickWidget dispatches
+    all calls to the offscreen QQuickWindow, it must fix itself when the offscreen
+    view changes. QTBUG-108226
+*/
+void tst_qquickwidget::accessibilityHandlesViewChange()
+{
+    if (QGuiApplication::platformName() == "offscreen")
+        QSKIP("Doesn't test anything on offscreen platform.");
+    if (QGuiApplication::platformName() == "android")
+        QSKIP("Test doesn't exit cleanly on Android and generates many warnings - QTBUG-112696");
+
+    QWidget window;
+
+    QPointer<QQuickWindow> backingScene;
+
+    QQuickWidget *childView = new QQuickWidget(&window);
+    childView->setSource(testFileUrl("rectangle.qml"));
+    window.show();
+    QVERIFY(QTest::qWaitForWindowExposed(&window));
+    backingScene = childView->quickWindow();
+    QVERIFY(backingScene);
+
+    QAccessibleInterface *iface = QAccessible::queryAccessibleInterface(childView);
+    QVERIFY(iface);
+    (void)iface->child(0);
+
+    std::unique_ptr<QQuickWidget> quickWidget(childView);
+    childView->setParent(nullptr);
+    childView->show();
+    QVERIFY(QTest::qWaitForWindowExposed(childView));
+    QVERIFY(!backingScene); // the old QQuickWindow should be gone now
+    QVERIFY(childView->quickWindow()); // long live the new QQuickWindow
+
+    iface = QAccessible::queryAccessibleInterface(childView);
+    QVERIFY(iface);
+    // this would crash if QAccessibleQuickWidget hadn't repaired itself to
+    // delegate calls to the new (or at least not the old, destroyed) QQuickWindow.
+    (void)iface->child(0);
+}
+
+class CreateDestroyWidget : public QWidget
+{
+public:
+    using QWidget::create;
+    using QWidget::destroy;
+};
+
+void tst_qquickwidget::cleanupRhi()
+{
+    CreateDestroyWidget topLevel;
+    QQuickWidget quickWidget(&topLevel);
+    quickWidget.setSource(testFileUrl("rectangle.qml"));
+    topLevel.show();
+    QVERIFY(QTest::qWaitForWindowExposed(&topLevel));
+
+    topLevel.destroy();
+    topLevel.create();
+}
+
+void tst_qquickwidget::dontRecreateRootElementOnWindowChange()
+{
+    auto *quickWidget = new QQuickWidget();
+    quickWidget->setSource(testFileUrl("rectangle.qml"));
+    QObject *item = quickWidget->rootObject();
+
+    bool wasDestroyed = false;
+    QObject::connect(item, &QObject::destroyed, this, [&] { wasDestroyed = true; });
+
+    QEvent event(QEvent::WindowChangeInternal);
+    QCoreApplication::sendEvent(quickWidget, &event);
+
+    QVERIFY(!wasDestroyed);
+}
+
 QTEST_MAIN(tst_qquickwidget)
 
 #include "tst_qquickwidget.moc"
