@@ -8,6 +8,7 @@
 #include <QtCore/qstack.h>
 #include <QtCore/qdir.h>
 #include <QtCore/qloggingcategory.h>
+#include <QtQml/private/qqmlsignalnames_p.h>
 
 #include <private/qqmljsutils_p.h>
 
@@ -61,8 +62,9 @@ static bool isOrUnderComponent(QQmlJSScope::ConstPtr type)
 {
     Q_ASSERT(type->isComposite()); // we're dealing with composite types here
     for (; type; type = type->parentScope()) {
-        if (isExplicitComponent(type) || isImplicitComponent(type))
+        if (isExplicitComponent(type) || isImplicitComponent(type)) {
             return true;
+        }
     }
     return false;
 }
@@ -72,7 +74,7 @@ QmltcVisitor::QmltcVisitor(const QQmlJSScope::Ptr &target, QQmlJSImporter *impor
                            const QStringList &qmldirFiles)
     : QQmlJSImportVisitor(target, importer, logger, implicitImportDirectory, qmldirFiles)
 {
-    m_qmlTypeNames.append(QFileInfo(logger->fileName()).baseName()); // put document root
+    m_qmlTypeNames.append(QFileInfo(logger->filePath()).baseName()); // put document root
 }
 
 void QmltcVisitor::findCppIncludes()
@@ -87,7 +89,7 @@ void QmltcVisitor::findCppIncludes()
         return false;
     };
     const auto addCppInclude = [this](const QQmlJSScope::ConstPtr &type) {
-        if (QString includeFile = type->filePath(); includeFile.endsWith(u".h"))
+        if (QString includeFile = filePath(type); !includeFile.isEmpty())
             m_cppIncludes.insert(std::move(includeFile));
     };
 
@@ -100,9 +102,15 @@ void QmltcVisitor::findCppIncludes()
         // look in type
         addCppInclude(type);
 
+        if (type->isListProperty())
+            addCppInclude(type->valueType());
+
         // look in type's base type
         auto base = type->baseType();
-        Q_ASSERT(base || !type->isComposite());
+        if (!base && type->isComposite())
+            // in this case, qqmljsimportvisitor would have already print an error message
+            // about the missing type, so just return silently without crashing
+            return;
         if (!base || visitType(base))
             return;
         addCppInclude(base);
@@ -129,8 +137,9 @@ void QmltcVisitor::findCppIncludes()
         Q_ASSERT(type);
 
         const auto scopeType = type->scopeType();
-        if (scopeType != QQmlJSScope::QMLScope && scopeType != QQmlJSScope::GroupedPropertyScope
-            && scopeType != QQmlJSScope::AttachedPropertyScope) {
+        if (scopeType != QQmlSA::ScopeType::QMLScope
+            && scopeType != QQmlSA::ScopeType::GroupedPropertyScope
+            && scopeType != QQmlSA::ScopeType::AttachedPropertyScope) {
             continue;
         }
 
@@ -142,8 +151,8 @@ void QmltcVisitor::findCppIncludes()
             for (const QQmlJSMetaProperty &p : properties) {
                 findInType(p.type());
 
-                if (p.isPrivate() && t->filePath().endsWith(u".h")) {
-                    const QString ownersInclude = t->filePath();
+                if (p.isPrivate()) {
+                    const QString ownersInclude = filePath(t);
                     QString privateInclude = constructPrivateInclude(ownersInclude);
                     if (!privateInclude.isEmpty())
                         m_cppIncludes.insert(std::move(privateInclude));
@@ -155,9 +164,9 @@ void QmltcVisitor::findCppIncludes()
             for (const QQmlJSMetaMethod &m : methods) {
                 findInType(m.returnType());
 
-                const auto parameters = m.parameterTypes();
+                const auto parameters = m.parameters();
                 for (const auto &param : parameters)
-                    findInType(param);
+                    findInType(param.type());
             }
         }
 
@@ -165,47 +174,47 @@ void QmltcVisitor::findCppIncludes()
     }
 
     // remove own include
-    m_cppIncludes.remove(m_exportedRootScope->filePath());
+    m_cppIncludes.remove(filePath(m_exportedRootScope));
 }
 
 static void addCleanQmlTypeName(QStringList *names, const QQmlJSScope::ConstPtr &scope)
 {
-    Q_ASSERT(scope->scopeType() == QQmlJSScope::QMLScope);
+    Q_ASSERT(scope->scopeType() == QQmlSA::ScopeType::QMLScope);
     Q_ASSERT(!scope->isArrayScope());
     Q_ASSERT(!scope->baseTypeName().isEmpty());
     // the scope is guaranteed to be a new QML type, so any prefixes (separated
     // by dots) should be import namespaces
-    names->append(scope->baseTypeName().replace(u'.', u'_'));
+    const std::optional<QString> &inlineComponentName = scope->inlineComponentName();
+    QString name = inlineComponentName ? *inlineComponentName : scope->baseTypeName();
+    names->append(name.replace(u'.', u'_'));
 }
 
 bool QmltcVisitor::visit(QQmlJS::AST::UiObjectDefinition *object)
 {
-    const bool processingRoot = !rootScopeIsValid();
-
     if (!QQmlJSImportVisitor::visit(object))
         return false;
 
-    if (processingRoot) {
-        Q_ASSERT(rootScopeIsValid());
-        setRootFilePath();
-    }
-
     // we're not interested in non-QML scopes
-    if (m_currentScope->scopeType() != QQmlJSScope::QMLScope)
+    if (m_currentScope->scopeType() != QQmlSA::ScopeType::QMLScope)
         return true;
 
+    if (m_currentScope->isInlineComponent()) {
+        m_inlineComponentNames.append(m_currentRootName);
+        m_inlineComponents[m_currentRootName] = m_currentScope;
+    }
+
     if (m_currentScope != m_exportedRootScope) // not document root
         addCleanQmlTypeName(&m_qmlTypeNames, m_currentScope);
     // give C++-relevant internal names to QMLScopes, we can use them later in compiler
     m_currentScope->setInternalName(uniqueNameFromPieces(m_qmlTypeNames, m_qmlTypeNameCounts));
-    m_qmlTypesWithQmlBases.append(m_currentScope);
+    m_qmlTypesWithQmlBases[m_currentRootName].append(m_currentScope);
 
     return true;
 }
 
 void QmltcVisitor::endVisit(QQmlJS::AST::UiObjectDefinition *object)
 {
-    if (m_currentScope->scopeType() == QQmlJSScope::QMLScope)
+    if (m_currentScope->scopeType() == QQmlSA::ScopeType::QMLScope)
         m_qmlTypeNames.removeLast();
     QQmlJSImportVisitor::endVisit(object);
 }
@@ -220,7 +229,7 @@ bool QmltcVisitor::visit(QQmlJS::AST::UiObjectBinding *uiob)
     // give C++-relevant internal names to QMLScopes, we can use them later in compiler
     m_currentScope->setInternalName(uniqueNameFromPieces(m_qmlTypeNames, m_qmlTypeNameCounts));
 
-    m_qmlTypesWithQmlBases.append(m_currentScope);
+    m_qmlTypesWithQmlBases[m_currentRootName].append(m_currentScope);
     return true;
 }
 
@@ -257,7 +266,7 @@ bool QmltcVisitor::visit(QQmlJS::AST::UiPublicMember *publicMember)
             owner->addOwnProperty(property);
         }
 
-        const QString notifyName = name + u"Changed"_s;
+        const QString notifyName = QQmlSignalNames::propertyNameToChangedSignalName(name);
         // also check that notify is already a method of the scope
         {
             auto owningScope = m_savedBindingOuterScope ? m_savedBindingOuterScope : m_currentScope;
@@ -267,12 +276,12 @@ bool QmltcVisitor::visit(QQmlJS::AST::UiPublicMember *publicMember)
                         methods.isEmpty() ? u"no signal"_s : u"too many signals"_s;
                 m_logger->log(
                         u"internal error: %1 found for property '%2'"_s.arg(errorString, name),
-                        Log_Compiler, publicMember->identifierToken);
+                        qmlCompiler, publicMember->identifierToken);
                 return false;
-            } else if (methods[0].methodType() != QQmlJSMetaMethod::Signal) {
+            } else if (methods[0].methodType() != QQmlJSMetaMethodType::Signal) {
                 m_logger->log(u"internal error: method %1 of property %2 must be a signal"_s.arg(
                                       notifyName, name),
-                              Log_Compiler, publicMember->identifierToken);
+                              qmlCompiler, publicMember->identifierToken);
                 return false;
             }
         }
@@ -299,10 +308,6 @@ bool QmltcVisitor::visit(QQmlJS::AST::UiInlineComponent *component)
 {
     if (!QQmlJSImportVisitor::visit(component))
         return false;
-    m_logger->log(u"Inline components are not supported"_s, Log_Compiler,
-                  component->firstSourceLocation());
-    // despite the failure, return true here so that we do not assert in
-    // QQmlJSImportVisitor::endVisit(UiInlineComponent)
     return true;
 }
 
@@ -313,11 +318,11 @@ void QmltcVisitor::endVisit(QQmlJS::AST::UiProgram *program)
         return;
 
     QHash<QQmlJSScope::ConstPtr, QList<QQmlJSMetaPropertyBinding>> bindings;
-    for (const QQmlJSScope::ConstPtr &type : std::as_const(m_qmlTypes)) {
-        if (isOrUnderComponent(type))
-            continue;
+
+    // Yes, we want absolutely all bindings in the document.
+    // Not only the ones immediately assigned to QML types.
+    for (const QQmlJSScope::ConstPtr &type : std::as_const(m_scopesByIrLocation))
         bindings.insert(type, type->ownPropertyBindingsInQmlIROrder());
-    }
 
     postVisitResolve(bindings);
     setupAliases();
@@ -327,28 +332,28 @@ void QmltcVisitor::endVisit(QQmlJS::AST::UiProgram *program)
 
     findCppIncludes();
 
-    for (const QQmlJSScope::ConstPtr &type : m_pureQmlTypes)
-        checkForNamingCollisionsWithCpp(type);
+    for (const QList<QQmlJSScope::ConstPtr> &qmlTypes : m_pureQmlTypes)
+        for (const QQmlJSScope::ConstPtr &type : qmlTypes)
+            checkNamesAndTypes(type);
 }
 
 QQmlJSScope::ConstPtr fetchType(const QQmlJSMetaPropertyBinding &binding)
 {
     switch (binding.bindingType()) {
-    case QQmlJSMetaPropertyBinding::Object:
+    case QQmlSA::BindingType::Object:
         return binding.objectType();
-    case QQmlJSMetaPropertyBinding::Interceptor:
+    case QQmlSA::BindingType::Interceptor:
         return binding.interceptorType();
-    case QQmlJSMetaPropertyBinding::ValueSource:
+    case QQmlSA::BindingType::ValueSource:
         return binding.valueSourceType();
-    // TODO: AttachedProperty and GroupProperty are not supported yet,
-    // but have to also be acknowledged here
+    case QQmlSA::BindingType::AttachedProperty:
+        return binding.attachingType();
+    case QQmlSA::BindingType::GroupProperty:
+        return binding.groupType();
     default:
         return {};
     }
-    // coverity complains about the unreachable return below,
-    // some compilers; some compilers complain without it
-    Q_UNREACHABLE();
-    return {};
+    Q_UNREACHABLE_RETURN({});
 }
 
 template<typename Predicate>
@@ -422,10 +427,15 @@ void QmltcVisitor::postVisitResolve(
         const QHash<QQmlJSScope::ConstPtr, QList<QQmlJSMetaPropertyBinding>> &qmlIrOrderedBindings)
 {
 
+    // add the document root (that is not an inline component), as we usually
+    // want to iterate on all inline components, followed by the document root
+    m_inlineComponentNames.append(RootDocumentNameType());
+    m_inlineComponents[RootDocumentNameType()] = m_exportedRootScope;
+
     // match scopes to indices of QmlIR::Object from QmlIR::Document
     qsizetype count = 0;
     const auto setIndex = [&](const QQmlJSScope::Ptr &current) {
-        if (current->scopeType() != QQmlJSScope::QMLScope || current->isArrayScope())
+        if (current->scopeType() != QQmlSA::ScopeType::QMLScope || current->isArrayScope())
             return;
         Q_ASSERT(!m_qmlIrObjectIndices.contains(current));
         m_qmlIrObjectIndices[current] = count;
@@ -450,7 +460,10 @@ void QmltcVisitor::postVisitResolve(
         }
         return false;
     };
-    iterateBindings(m_exportedRootScope, qmlIrOrderedBindings, findDeferred);
+    for (const auto &inlineComponentName : m_inlineComponentNames) {
+        iterateBindings(m_inlineComponents[inlineComponentName], qmlIrOrderedBindings,
+                        findDeferred);
+    }
 
     const auto isOrUnderDeferred = [&deferredTypes](QQmlJSScope::ConstPtr type) {
         for (; type; type = type->parentScope()) {
@@ -461,7 +474,6 @@ void QmltcVisitor::postVisitResolve(
     };
 
     // find all "pure" QML types
-    m_pureQmlTypes.reserve(m_qmlTypes.size());
     QList<QQmlJSScope::ConstPtr> explicitComponents;
     for (qsizetype i = 0; i < m_qmlTypes.size(); ++i) {
         const QQmlJSScope::ConstPtr &type = m_qmlTypes.at(i);
@@ -479,15 +491,28 @@ void QmltcVisitor::postVisitResolve(
             }
         }
 
-        m_creationIndices[type] = m_pureQmlTypes.size();
-        m_pureQmlTypes.append(type);
+        const InlineComponentOrDocumentRootName inlineComponent =
+                type->enclosingInlineComponentName();
+        QList<QQmlJSScope::ConstPtr> &pureQmlTypes = m_pureQmlTypes[inlineComponent];
+        m_creationIndices[type] = pureQmlTypes.size();
+        pureQmlTypes.append(type);
+    }
+
+    // update the typeCounts
+    for (const auto &inlineComponent : m_inlineComponentNames) {
+        m_inlineComponentTypeCount[inlineComponent] = m_pureQmlTypes[inlineComponent].size();
     }
 
     // add explicit components to the object creation indices
     {
-        qsizetype index = 0;
-        for (const QQmlJSScope::ConstPtr &c : std::as_const(explicitComponents))
-            m_creationIndices[c] = m_pureQmlTypes.size() + index++;
+        QHash<InlineComponentOrDocumentRootName, qsizetype> index;
+        for (const QQmlJSScope::ConstPtr &c : std::as_const(explicitComponents)) {
+            const InlineComponentOrDocumentRootName inlineComponent =
+                    c->enclosingInlineComponentName();
+            m_creationIndices[c] =
+                    m_pureQmlTypes[inlineComponent].size() + index[inlineComponent]++;
+            m_inlineComponentTypeCount[inlineComponent]++;
+        }
     }
 
     // m_qmlTypesWithQmlBases should contain the types to be compiled.
@@ -500,28 +525,32 @@ void QmltcVisitor::postVisitResolve(
     // only known after visitation is over!
 
     // filter step:
-    QList<QQmlJSScope::ConstPtr> filteredQmlTypesWithQmlBases;
-    filteredQmlTypesWithQmlBases.reserve(m_qmlTypesWithQmlBases.size());
-    std::copy_if(m_qmlTypesWithQmlBases.cbegin(), m_qmlTypesWithQmlBases.cend(),
-                 std::back_inserter(filteredQmlTypesWithQmlBases),
-                 [&](const QQmlJSScope::ConstPtr &type) {
-                     auto base = type->baseType();
-                     return base && base->isComposite() && !isOrUnderComponent(type)
-                             && !isOrUnderDeferred(type);
-                 });
-    qSwap(m_qmlTypesWithQmlBases, filteredQmlTypesWithQmlBases);
+    for (const auto &inlineComponent : m_inlineComponentNames) {
+        QList<QQmlJSScope::ConstPtr> filteredQmlTypesWithQmlBases;
+        QList<QQmlJSScope::ConstPtr> &unfilteredQmlTypesWithQmlBases =
+                m_qmlTypesWithQmlBases[inlineComponent];
+        filteredQmlTypesWithQmlBases.reserve(unfilteredQmlTypesWithQmlBases.size());
+        std::copy_if(unfilteredQmlTypesWithQmlBases.cbegin(), unfilteredQmlTypesWithQmlBases.cend(),
+                     std::back_inserter(filteredQmlTypesWithQmlBases),
+                     [&](const QQmlJSScope::ConstPtr &type) {
+                         auto base = type->baseType();
+                         return base && base->isComposite() && !isOrUnderComponent(type)
+                                 && !isOrUnderDeferred(type);
+                     });
+        qSwap(unfilteredQmlTypesWithQmlBases, filteredQmlTypesWithQmlBases);
+    }
 
     // count QmlIR::Objects in the document - the amount is used to calculate
     // object indices of implicit components
-    int qmlScopeCount = 0;
+    QHash<InlineComponentOrDocumentRootName, qsizetype> qmlScopeCount;
     const auto countQmlScopes = [&](const QQmlJSScope::ConstPtr &scope) {
         if (scope->isArrayScope()) // special kind of QQmlJSScope::QMLScope
             return;
         switch (scope->scopeType()) {
-        case QQmlJSScope::QMLScope:
-        case QQmlJSScope::GroupedPropertyScope:
-        case QQmlJSScope::AttachedPropertyScope: {
-            ++qmlScopeCount;
+        case QQmlSA::ScopeType::QMLScope:
+        case QQmlSA::ScopeType::GroupedPropertyScope:
+        case QQmlSA::ScopeType::AttachedPropertyScope: {
+            ++qmlScopeCount[scope->enclosingInlineComponentName()];
             break;
         }
         default:
@@ -531,10 +560,9 @@ void QmltcVisitor::postVisitResolve(
     };
     // order doesn't matter (so re-use QQmlJSUtils)
     QQmlJSUtils::traverseFollowingQmlIrObjectStructure(m_exportedRootScope, countQmlScopes);
-    Q_ASSERT(qmlScopeCount >= int(m_qmlTypes.size()));
 
     // figure synthetic indices of QQmlComponent-wrapped types
-    int syntheticCreationIndex = 0;
+    int syntheticCreationIndex;
     const auto addSyntheticIndex = [&](const QQmlJSScope::ConstPtr &type) {
         // explicit component
         if (isExplicitComponent(type)) {
@@ -543,30 +571,40 @@ void QmltcVisitor::postVisitResolve(
         }
         // implicit component
         if (isImplicitComponent(type)) {
-            const int index = int(qmlScopeCount) + syntheticCreationIndex++;
+            const int index =
+                    qmlScopeCount[type->enclosingInlineComponentName()] + syntheticCreationIndex++;
             m_syntheticTypeIndices[type] = index;
             return true;
         }
         return false;
     };
-    iterateTypes(m_exportedRootScope, qmlIrOrderedBindings, addSyntheticIndex);
+
+    for (const auto &inlineComponentName : m_inlineComponentNames) {
+        syntheticCreationIndex = 0; // reset for each inline component
+        iterateTypes(m_inlineComponents[inlineComponentName], qmlIrOrderedBindings,
+                     addSyntheticIndex);
+    }
 
     // figure runtime object ids for non-component wrapped types
-    int currentId = 0;
+    int currentId;
     const auto setRuntimeId = [&](const QQmlJSScope::ConstPtr &type) {
         // any type wrapped in an implicit component shouldn't be processed
         // here. even if it has id, it doesn't need to be set by qmltc
-        if (type->isComponentRootElement())
+        if (type->isComponentRootElement()) {
             return true;
+        }
 
-        if (m_typesWithId.contains(type))
+        if (m_typesWithId.contains(type)) {
             m_typesWithId[type] = currentId++;
+        }
 
-        // otherwise we need to `return true` here
-        Q_ASSERT(!type->isInlineComponent());
         return false;
     };
-    iterateTypes(m_exportedRootScope, qmlIrOrderedBindings, setRuntimeId);
+
+    for (const auto &inlineComponentName : m_inlineComponentNames) {
+        currentId = 0; // reset for each inline component
+        iterateTypes(m_inlineComponents[inlineComponentName], qmlIrOrderedBindings, setRuntimeId);
+    }
 }
 
 static void setAliasData(QQmlJSMetaProperty *alias, const QQmlJSUtils::ResolvedAlias &origin)
@@ -579,8 +617,12 @@ static void setAliasData(QQmlJSMetaProperty *alias, const QQmlJSUtils::ResolvedA
         return;
     if (origin.property.isWritable() && alias->write().isEmpty())
         alias->setWrite(compiledData.write);
-    if (!origin.property.notify().isEmpty() && alias->notify().isEmpty())
+
+    // the engine always compiles a notify for properties/aliases defined in qml code
+    // Yes, this generated notify will never be emitted.
+    if (alias->notify().isEmpty())
         alias->setNotify(compiledData.notify);
+
     if (!origin.property.bindable().isEmpty() && alias->bindable().isEmpty())
         alias->setBindable(compiledData.bindable);
 }
@@ -602,7 +644,7 @@ void QmltcVisitor::setupAliases()
                                                     QQmlJSUtils::AliasResolutionVisitor {});
             if (result.kind == QQmlJSUtils::AliasTarget_Invalid) {
                 m_logger->log(QStringLiteral("Cannot resolve alias \"%1\"").arg(p.propertyName()),
-                              Log_Alias, current->sourceLocation());
+                              qmlUnresolvedAlias, current->sourceLocation());
                 continue;
             }
             setAliasData(&p, result);
@@ -611,7 +653,7 @@ void QmltcVisitor::setupAliases()
     }
 }
 
-void QmltcVisitor::checkForNamingCollisionsWithCpp(const QQmlJSScope::ConstPtr &type)
+void QmltcVisitor::checkNamesAndTypes(const QQmlJSScope::ConstPtr &type)
 {
     static const QString cppKeywords[] = {
         u"alignas"_s,
@@ -630,20 +672,21 @@ void QmltcVisitor::checkForNamingCollisionsWithCpp(const QQmlJSScope::ConstPtr &
         u"case"_s,
         u"catch"_s,
         u"char"_s,
-        u"char8_t"_s,
         u"char16_t"_s,
         u"char32_t"_s,
+        u"char8_t"_s,
         u"class"_s,
-        u"compl"_s,
-        u"concept"_s,
-        u"const"_s,
-        u"consteval"_s,
-        u"constexpr"_s,
-        u"const_cast"_s,
-        u"continue"_s,
         u"co_await"_s,
         u"co_return"_s,
         u"co_yield"_s,
+        u"compl"_s,
+        u"concept"_s,
+        u"const"_s,
+        u"const_cast"_s,
+        u"consteval"_s,
+        u"constexpr"_s,
+        u"constinit"_s,
+        u"continue"_s,
         u"decltype"_s,
         u"default"_s,
         u"delete"_s,
@@ -711,6 +754,7 @@ void QmltcVisitor::checkForNamingCollisionsWithCpp(const QQmlJSScope::ConstPtr &
         u"xor"_s,
         u"xor_eq"_s,
     };
+    Q_ASSERT(std::is_sorted(std::begin(cppKeywords), std::end(cppKeywords)));
 
     const auto isReserved = [&](QStringView word) {
         if (word.startsWith(QChar(u'_')) && word.size() >= 2
@@ -724,9 +768,26 @@ void QmltcVisitor::checkForNamingCollisionsWithCpp(const QQmlJSScope::ConstPtr &
         if (!isReserved(name))
             return;
         m_logger->log(errorPrefix + u" '" + name + u"' is a reserved C++ word, consider renaming",
-                      Log_Compiler, type->sourceLocation());
+                      qmlCompiler, type->sourceLocation());
     };
 
+    const auto validateType = [&type, this](const QQmlJSScope::ConstPtr &typeToCheck,
+                                            QStringView name, QStringView errorPrefix) {
+        if (type->moduleName().isEmpty() || typeToCheck.isNull())
+            return;
+
+        if (typeToCheck->isComposite() && typeToCheck->moduleName() != type->moduleName()) {
+            m_logger->log(
+                    QStringLiteral(
+                            "Can't compile the %1 type \"%2\" to C++ because it "
+                            "lives in \"%3\" instead of the current file's \"%4\" QML module.")
+                            .arg(errorPrefix, name, typeToCheck->moduleName(), type->moduleName()),
+                    qmlCompiler, type->sourceLocation());
+        }
+    };
+
+    validateType(type->baseType(), type->baseTypeName(), u"QML base");
+
     const auto enums = type->ownEnumerations();
     for (auto it = enums.cbegin(); it != enums.cend(); ++it) {
         const QQmlJSMetaEnum e = it.value();
@@ -741,28 +802,38 @@ void QmltcVisitor::checkForNamingCollisionsWithCpp(const QQmlJSScope::ConstPtr &
     for (auto it = properties.cbegin(); it != properties.cend(); ++it) {
         const QQmlJSMetaProperty &p = it.value();
         validate(p.propertyName(), u"Property");
+
+        if (!p.isAlias() && !p.typeName().isEmpty())
+            validateType(p.type(), p.typeName(), u"QML property");
     }
 
     const auto methods = type->ownMethods();
     for (auto it = methods.cbegin(); it != methods.cend(); ++it) {
         const QQmlJSMetaMethod &m = it.value();
         validate(m.methodName(), u"Method");
+        if (!m.returnTypeName().isEmpty())
+            validateType(m.returnType(), m.returnTypeName(), u"QML method return");
 
-        const auto parameterNames = m.parameterNames();
-        for (const auto &name : parameterNames)
-            validate(name, u"Method '%1' parameter"_s.arg(m.methodName()));
+        for (const auto &parameter : m.parameters()) {
+            validate(parameter.name(), u"Method '%1' parameter"_s.arg(m.methodName()));
+            if (!parameter.typeName().isEmpty())
+                validateType(parameter.type(), parameter.typeName(), u"QML parameter");
+        }
     }
 
     // TODO: one could also test signal handlers' parameters but we do not store
     // this information in QQmlJSMetaPropertyBinding currently
 }
 
-void QmltcVisitor::setRootFilePath()
+/*! \internal
+ *  Returns the file path for the C++ header of \a scope or the header created
+ *  by qmltc for it and its inline components.
+ */
+QString QmltcVisitor::filePath(const QQmlJSScope::ConstPtr &scope) const
 {
-    const QString filePath = m_exportedRootScope->filePath();
-    if (filePath.endsWith(u".h")) // assume the correct path is set
-        return;
-    Q_ASSERT(filePath.endsWith(u".qml"_s));
+    const QString filePath = scope->filePath();
+    if (!filePath.endsWith(u".qml")) // assume the correct path is set
+        return scope->filePath();
 
     const QString correctedFilePath = sourceDirectoryPath(filePath);
     const QStringList paths = m_importer->resourceFileMapper()->resourcePaths(
@@ -774,38 +845,23 @@ void QmltcVisitor::setRootFilePath()
         qCDebug(lcQmltcCompiler,
                 "Failed to find a header file name for path %s. Paths checked:\n%s",
                 correctedFilePath.toUtf8().constData(), matchedPaths.toUtf8().constData());
-        return;
+        return QString();
     }
     // NB: get the file name to avoid prefixes
-    m_exportedRootScope->setFilePath(QFileInfo(*firstHeader).fileName());
+    return QFileInfo(*firstHeader).fileName();
 }
 
-/*! \internal
-
-    Returns a source directory path for a corresponding \a path which is either
-    a source directory path (rarely - when processing currently-compiled file)
-    or a build directory path (mostly - when importing a module/file).
-*/
-QString QmltcVisitor::sourceDirectoryPath(const QString &path)
+QString QmltcVisitor::sourceDirectoryPath(const QString &path) const
 {
-    const QStringList resourcePaths = m_importer->resourceFileMapper()->resourcePaths(
-            QQmlJSResourceFileMapper::localFileFilter(path));
-    // path could be pointing to a source directory already
-    const QString uniquePrefix = u"/qt_qml_module_dir_mapping/"_s;
-    if (resourcePaths.size() != 1 || !resourcePaths[0].startsWith(uniquePrefix)) {
-        // found nothing or path is a source directory path already. either way
-        // return the input here and let the caller take care of it
-        const QString matchedPaths =
-                resourcePaths.isEmpty() ? u"<none>"_s : resourcePaths.join(u", ");
-        qCDebug(lcQmltcCompiler,
-                "Path %s is not a build directory path. Resource paths that matched:\n%s",
-                path.toUtf8().constData(), matchedPaths.toUtf8().constData());
-        return path;
-    }
-    QString sourceDirPath = resourcePaths[0];
-    sourceDirPath.remove(0, uniquePrefix.size());
-    Q_ASSERT(QFileInfo(sourceDirPath).exists());
-    return sourceDirPath;
+    auto result = QQmlJSUtils::sourceDirectoryPath(m_importer, path);
+    if (const QString *srcDirPath = std::get_if<QString>(&result))
+        return *srcDirPath;
+
+    const QQmlJS::DiagnosticMessage *error = std::get_if<QQmlJS::DiagnosticMessage>(&result);
+    Q_ASSERT(error);
+    qCDebug(lcQmltcCompiler, "%s", error->message.toUtf8().constData());
+    // return input as a fallback
+    return path;
 }
 
 QT_END_NAMESPACE
